local TOC, CCD = ...

CCD.Modules = CCD.Modules or {}
CCD.Modules.RCT = CCD.Modules.RCT or {}
CCD.Modules.RBT = CCD.Modules.RBT or {}
CCD.Modules.CPT = CCD.Modules.CPT or {}
CCD.Modules.CPUT = CCD.Modules.CPUT or {}
CCD.Modules.ErrorManager = CCD.Modules.ErrorManager or {}
CCD.Buff = CCD.Buff or {}
CCD.Ally = CCD.Ally or {}
CCD.Target = CCD.Target or {
	Id,
	IsHostile,			-- everything that is not friendly is considered hostile including neutrals
	RaidDebuffs,		-- indexed by Buff Type; debuffs that affect everyone in the raid i.e all support debuffs, lethal poison, tenebrious distortion, etc but not debuffs such as Accelerant nor Storm Locus
	RaidDebuffsById,
}
CCD.Allies = CCD.Allies or {}
CCD.Debug = CCD.Debug or {}
CCD.UI = CCD.UI or {}
CCD.UI.MenuButton = CCD.UI.MenuButton or {}
CCD.Importer = CCD.Importer or {}
CCD.Messaging = CCD.Messaging or {}
--
CCD.Version = "3.4.4"
CCD.IsFirstTime = nil -- do not set
CCD.EnableEditModeOnMainWindow = true
CCD.AutoCancelWildGrowthOn = true
local MechanicFrame
local FirstUpdate = true
CCD.PartialAvailabilityTimersByAlly = {}
local PartialAvailabilityTimersByAlly = CCD.PartialAvailabilityTimersByAlly -- Key: Ally; Value: LibCron Timer Tag
CCD.SlashCommands = {
	["reset"] = { function() CCD.ResetAllFrames() end, "Resets all frame positions" },
	["locate"] = { function() CCD.UI.MenuButton:Locate() end, "Highlights the CCD button" },
	["clear"] = { function() CCD.RemoveAllCooldowns() end, "Clears the cooldown list" },
	["tracked"] = { function() CCD.PrintAllCooldowns() end, "Prints a list of cooldowns that are being tracked" },
	["edit"] = { function() CCD.UI.SetEditMode() end, "Toggles Edit Mode on or off" },
	["wild"] = { function() SetAutoCancelWildGrowth() end, "Toggles the automatic cancellation of Wild Growth on you" },
	["rct"] = { function() RCT.Set() end, "Toggles the Cooldown Tracker on or off" },
	["version"] = { function() CCD.PrintVersion() end, "Prints the addon version" },
	["print"] = { function() CCD.PrintCommands() end, "Prints a list of commands" },
	["help"] = { function() CCD.Help() end, "Prints a list of commands" },
	["rbt"] = { function() RaidBuffTracker:Set() end, "Toggles the Buff/Debuff Tracker on or off" },
	["test"] = { function() CCD.Test() end, "You probably shouldn't use this command", true },
	["printlog"] = { function() CCD.MainWindow.ChangeLogTab:PrintToDebug() end, "Prints the change log to the debug window", true },
	["printforum"] = { function() CCD.MainWindow.ChangeLogTab:PrintForum() end, "", true },
	["debug"] = { function() CCD.Debug.SetDeveloperMode() end, "Toggles Developer Mode on or off" },
	["clients"] = { function() CCD.PrintClients() end, "Prints a list of players in your group with CCD" },
	["printerrors"] = { function() CCD.Modules.ErrorManager:PrintAll() end, "Prints a list of CCD errors" },
	["clearerrors"] = { function() CCD.Modules.ErrorManager:ClearAll() end, "Clears the CCD error list" },
	["find"] = { function(BuffName) CCD.FindBuff(BuffName) end, "", true},
}
CCD.BuffLibrary = {}
CCD.DisableInWarfronts = false

-- Aliases
local RCT = CCD.Modules.RCT
local RBT = CCD.Modules.RBT
local CPT = CCD.Modules.CPT
local CPUT = CCD.Modules.CPUT
local EM = CCD.Modules.ErrorManager
local Importer = CCD.Importer
local Buff = CCD.Buff
local Allies = CCD.Allies
local Target = CCD.Target
local Debug = CCD.Debug
local MenuButton = CCD.UI.MenuButton
local Version = CCD.Version

-- Change Log
CCD.ChangeLog = {} -- Key: Version; Value: Table of Strings
local ChangeLog = CCD.ChangeLog
table.insert(ChangeLog, {"3.4.4", {
	"@New Features & Changes",
	"Crit Pot Tracker (CPT)",
		"#CPT alert can now be hidden when there are no crit pots in your bag.",
		"#New crit pots added: Illustrious Bottle of Critical Strikes and Illustrious Bottle of Spellstrikes.",
	"Cooldown Tracker",
		"#Removed Focused Fire as it is no longer a raid buff.",
	"Buff Tracker",
		"#New vials added: Illustrious Brightsurge Vial, Illustrious Enduring Vial, Illustrious Powersurged Vial.",
		"#New debuffs added: Slagged - 5% non-physical damage (Vulcanist/Conflagrate), Squall - 5% crit chance (Dervish).",
	"$Bug Fixes",
	"Another fix for Error: CCD/Modules/CritPotTracker.lua:84: attempt to index field \"MyPot\" (a nil value).",
	"Cooldown Bars and Alerts should no longer disappear while in Edit Mode.",
}})
table.insert(ChangeLog, {"3.4.3", {
	"@New Features",
	"Crit Pot Tracker alert now appears on login even if you do not have any crit pots.",
	"The /ccd and /ccd help commands have been revamped to: print a list of commands, show the CCD button and open the CCD window.",
	"/ccd locate now also forces the CCD button to be shown.",
	"$Bug Fixes",
	"Fixed Error: CCD/Modules/CritPotTracker.lua:84: attempt to index field \"MyPot\" (a nil value).",
	"Heal cooldown selections should now be saved. - Reported by Chickie",
	"Crit Pot Tracker alert should now show the appropriate icon on login for casters.",
}})
table.insert(ChangeLog, {"3.4.2", {
	"$Bug Fixes",
	"Background opacity settings for cooldown bars should now work properly. - Reported by Loqque@Typhiria",
}})
table.insert(ChangeLog, {"3.4.1", {
	"@New Features",
	"Crit Pot Tracker",
		"#The CPT alert can now be set to only be shown while you are in a spec flagged as damage.",
		"#The CPT alert header now displays the number of crit pots in your bag.",
		"#The CPT alert is now displayed on login.",
	"$Bug Fixes",
	"Added a fix for Error: CCD/Modules/RaidCooldownTracker.lua:3250: attempt to index local 'abilityDetails' (a nil value). - Reported by Kabloom",	
	"Frames linked to a cooldown bar anchor should no longer overlap. Cooldown bar dimensions may have been reduced slightly.",
	"Cooldown alerts should now transition more smoothly (particularly from Active to Ready).",
	"Hidden cooldown bars are no longer included when resizing their container. This means you shouldn't see any undefined space between a frame linked to a bar. ",
}})
table.insert(ChangeLog, {"3.4.0", {
	"@New Features",
	"CCD Messenger Revamp - The new messaging system is NOT compatible with older versions so encourage members of your raid to update.",
	"All opacity settings now only apply to backgrounds and icons (text will always be opaque).",
	"Improved Frame Customization - Includes more precise icon sizing, optional header and footers and more. Frame sizes may have changed slightly. - Suggested by Nehira, Niarb, Alfiexx",
	"Healing Cooldowns - Now completely separate from damage-related cooldowns. New additions include: Beastmaster Commands, Gathering of Flames (Puri), Intensive Care (Phys), Life Insurance (Phys), Life's Rapture (Pally), Mass Rescue (Lib) and Natural Splendor (Chloro) among others. There are still quite a few abilities missing mostly due to the lack of a buff to track e.g. Warden cooldowns, Full Heal abilities, etc.",
	"Healing Cooldowns (cont'd) - There is now only one alert that acts as an anchor for all active cds (similar to the anchor used for bars). Healing cds also now have their own bar group.",
	"Improved Cooldown Detection - CCD users will now broadcast their cooldowns from any map, world or instance as long as they are in your group.",
	"Improved Role Change Detection - As with cooldowns, role changes will also be reported from anywhere.",
	"The CCD button can now be hidden via its menu. Use /ccd to make it visible again. - Suggested by Nehira@Zaviel",
	"Cogwheel menu positions are now static and saved per character.",
	"The Edit Mode window now has a Frame Spacing slider for linking. - Inspired by Kabloom@Wolfsbane",
	"Improved Buff Tracker Tooltips - The list of buffs tracked is now included in each alert's tooltip.",
	"The CCD button no longer flashes. - Suggested by Caldorian",
	"Any frames linked to the bottom left of a cooldown bar anchor will now always be linked to the last bar of that group rather than the first.",
	"Added the option of disabling alerts and bars in warfronts. This option is account-wide and located in the Misc tab of the main window.",
	"$Bug Fixes",
	"Fixed a bug that removed cooldowns on log in.",
	"Added proper error handling for CCD/Modules/RaidCooldownTracker.lua:1447: in function 'GetAllCooldownInfo'.",
	"Old cooldowns should no longer persist after that player changes role.",
	"Buff Tracker buffs should now show the correct timer colour with Alert Mode: Missing.",
	"^Other Changes",
	"Dragging the Show Disabled button now moves the Edit Mode window.",
	"Changed all references of \"Inert\" cooldowns to \"Down\".",
	}})
table.insert(ChangeLog, {"3.3.5", {
	"$Bug Fixes",
	"Fixed a bug that occurred on load when grouped with an Archon or Chloromancer.",
	"Cooldown Bar length should no longer reset on load when set below 125. - Reported by Loqque@Typhiria",
	}})
table.insert(ChangeLog, {"3.3.4", {
	"@New Features",
	"The Importer now remembers the last character selected for import. - Suggested by Chickie",
	"Improved Archon Detection - Characters with the ability Vitality of Stone (36 points) are now instantly flagged as Archons. This only works for yourself and CCD 3.3.4+ users.",
	"Added Flame of Life (Purifier), Healer's Covenant (Sentinel), Spiritual Conflagration (Purifier) and Biofeedback (Physician) to the Cooldown Tracker. - Requested by Chickie and Ekhstaz@Gelidra",
	"Improved role swap detection from users of CCD 3.3.4+ via the CCD Messenger. This also mean that CCD's features should be fully compatible with any language if cooldown users have 3.3.4+. Role swaps from non-CCD users are detected through castbars which can be unreliable due to lag. ",
	"Archons and Chloromancers are now properly unflagged if their soul tree is reset.",
	"Group members that reset their soul trees now have their cooldowns removed.",
	"Disabled frames (alerts, bar, cpu tracker) are now hidden by default in Edit Mode. There is a movable \"Show Disabled\" button in the middle of the screen to show these alerts.",
	"Readded the option for adjusting cooldown bar background opacity. - Requested by Loqque@Typhiria",	
	"Reduced the minimum length of cooldown bars by half. - Requested by Loqque@Typhiria",
	"The opacity setting for cooldown bars no longer affects text (i.e. text will remain fully opaque).",
	"$Bug Fixes",
	"Debuffs alerts should now work properly for target of target. - Reported by Kabloom",
	"If there are multiple Archons and one of them changes role, CCD should still recognize that there is an Archon. Same applies for Chloromancers.",
	"^Upcoming Features",
	"Heal and Tank Cooldowns will be separated from Damage Cooldowns. The current system was created with only damage in mind and this leads to the UI being less than desirable for cooldowns of other types. The new system will aim to make the Cooldown Tracker as appealing to Healers and Tanks as it currently is to damage dealers.",
	--"Cooldowns will be as customizable as Buffs are in terms of when you want to see them. This will probably be similar to the Alert Show Modes that are used in the Buff Tracker.",
	}})
table.insert(ChangeLog, {"3.3.3", {
	"@New Features",
	"Instant Cooldown Detection - Cooldown information will now be sent to users with CCD 3.3.3+ as soon as they are loaded into your list of abilities. The only restriction is that users must be in the same \"world\" or instance i.e. Mathosia, Storm Legion Zones, the Plane of Water or Conquest.",
	"Each Buff Tracker alert now has an Alert Show Mode option of either Never, Always, Missing or Active. NEVER is equivalent to disabling the alert. ALWAYS means the alert will always be shown. MISSING means it will only be shown when all associated buffs are missing. ACTIVE means it will only be shown while an associated buff is active. Always, Missing and Active modes have additional options once selected. You can access these options from an alert's cogwheel menu or in the Buffs tab. - Suggested by Megron",
	"Command to Attack (Beastmaster) now has its own alert, separate from Flaring Power. For the most part you can probably just disable it since very few encounters utilize a BM instead of an Archon.",
	"New alerts now default to the middle of the screen rather than the top left.",
	"$Bug Fixes",	
	"The option for showing/hiding alerts while cooldowns are Inert (i.e. on cooldown) should actually work now. It may be changed in either the cogwheel menu or in the Cooldowns tab. - Inspired by Maatang",
	"Fixed a possible error with detecting Archons and Chloros after a Role Changing cast.",
	"^Changes",
	"Optimized the formula for draining cooldown bars.",
	"All alert and bar timers should be in sync with the client's timers (if they weren't already).",
}})
table.insert(ChangeLog, {"3.3.2", {
	"$Bug Fixes",
	"Fixed a bug that showed an incorrect cooldown timer for someone else's cooldown. You would only have noticed it when someone with CCD 3.3.X used a cooldown. - Reported by Hankee",
}})
table.insert(ChangeLog, {"3.3.1", {
	"@New Features",
	"In an attempt to help track down certain bugs that do not trigger addon errors, I have added an Error Manager. There's nothing flashy about it - just use /ccd printerrors to print the list of errors and /ccd clearerrors to clear that list.",
	"Buff alerts may now be set to show while a buff is active (rather than missing), similar to Karuul Alert's \"Active\" feature.",
	"Buff alert timer colours may now be changed automatically when a buff timer reaches a certain value.",
	"Tooltips can now be changed to: never be shown, always be shown, only be shown in combat or only be shown outside of combat. They will always be shown in Edit Mode. See the Buffs tab for options. - Suggested by Kabloom",
	"$Bug Fixes",
	"Cooldowns should now properly be removed when the caster leaves your map. This does not apply if both users have CCD 3.3.0+.",
	"Cooldowns should now properly be removed when the caster changes their role/spec.",
}})
table.insert(ChangeLog, {"3.3.0", {
	"@New Features",
	"Added tooltips to raid buffs (not cooldowns). Just hover over the icons.",
	"Casting a tracked cooldown on any friendly target will now trigger the appropriate alert and bar.",
	"CCD now tracks the Battle Weary debuff to get more accurate timers for the cooldowns of Flaring Power and Command to Attack. If LESS than half of the raid has Battle Weary then CCD will assume a cooldown of 2 minutes. Therefore, if half or more of the raid are affected by Battle Weary then their cooldowns are assumed to be 5 minutes.",
	"Added a new module called the CCD Messenger. This module enables your addon to communicate with other users of CCD. Going forward you will see improved performance with more users of CCD in your raid, especially those with support buffs. There's not much to it right now but over the new few updates I should be able to add some interesting features that were previously impossible such as gathering information before cooldowns are used.",
	"^Changes",
	"The formula used to fill the cooldown bars was changed slightly. Please report if you notice any inconsistencies.",
	"Updated the Known Issues section.",
	"$Bug Fixes",
	"Using the mouse wheel to resize the CCD button should no longer produce an error.",
	"The Living Energy alert will now properly show if there is a Mage with a Heal flag in your group.",
	"Casting your own Wild Growth with 10 or less charge will now trigger an alert. Anyone using CCD 3.3.0 or later versions should also be notified.",
	"Casting your own Lava Field will always trigger an alert regardless of whether anyone received the buff or not. Like Wild Growth, other users will be notified if they have a new version of CCD.",
	"Cooldown alerts should no longer appear to freeze.",
	
}})
table.insert(ChangeLog, {"3.2.4", {
	"@New Features",
	"Added more options for customizing raid cooldown alerts including changing of the font size for headers and footers. These options may be applied individually by right clicking a cogwheel or globally through the Raid Cooldowns tab. -- Inspired by Xonk",
	"Added an alert for the Armour Debuff provided by: Lightning Fury (Marksman), Piercing Shot (Ranger), Splinter Bomb (Saboteur) and Mark of Inevitability (Champion). - Suggested by Iamneb",
	"You can now adjust the opacity of all alerts and bars. - Suggested by someone in an HK pug run",
	"You can now use the mouse wheel to adjust different properties on various frames including scale, font sizes and length. This works for the CCD button as well but does not apply to the CPU Tracker's frame yet. - Suggested by Kabloom",
	"You can now lock the position and size of the CCD button from its menu. - Suggested by Kabloom",
	"$Bug Fixes",
	"Leaving Edit Mode with the Raid Cooldown Tracker disabled should no longer cause alerts to stay on the screen. - Reported by Iamneb",
	"Removed some redundant code that might have been increasing CPU usage.",
	"Middle clicking the CCD button should no longer give the error: CCD/UI/MenuButton.lua:100: attempt to index global \"RBT\" (a nil value).",
	"Possible fix for Error: CCD/RaidCooldown.lua:44: attempt to index a nil value.",
	"You should no longer receive error: CCD/Modules/RaidBuffTracker.lua:452: attempt to index a nil value.",
	"^Quick Tips",
	"The amount of space between two frames being linked is set by the slider in the Frame Linking tab",
	"You can use the Importer to transfer your UI and settings from one toon to another. Simply right click the CCD button and select \"Importer\" from there.",
	"You need to press Enter after typing new values in any of the text boxes related to CCD.",
	"All menus brought up by right clicking can be moved around the screen.",
}})
table.insert(ChangeLog, {"3.2.3", {
	"@New Features",
	"Raid Cooldown alerts and bars are now completely independent of each other i.e. if you don't like the alerts you can disable them and just use the bars or vice versa. This change also brings a few more options in various menus such as \"Enable Cooldown Bars for this group\". If alerts/bars are not being shown as you would expect, double check that you don't have them globally disabled under \"Global Options\". - Suggested by Fathermonster",
	"All menus are now draggable.",
	"You can now adjust the height and font size for all Cooldown Bars (in addition to their length). - Suggested by Chickie",
	"$Bug Fixes",
	"All visible timers should be more in sync with Rift's timers.",
	"Chloromancer's Resurgence will no longer trigger the Crit Pot Tracker's Alert.",
	"^Did You Know:",
	"The amount of space between two frames being linked is set by the slider in the Frame Linking tab",
	"You can use the Importer to transfer your UI and settings from one toon to another. Simply right click the CCD button and select \"Importer\" from there.",
	"You need to press Enter after typing new values in any of the text boxes related to CCD.",
}})
table.insert(ChangeLog, {"3.2.2", {
	"@New Features",
	"You can now change the opacity of the Cooldown Bars' background.",
	"The max number of Cooldown Bars is no longer limited to 20.",
	"$Bug Fixes",
	"You should no longer see multiple Symbiosis alerts from the same toon. - Reported by Kabloom",
	"Possible fix for a rare bug where multiple cooldown timers were updating a single cooldown bar. - Reported by Chickie",
	"Fixed Error: CCD/Modules/RaidBuffTracker.lua:477: attempt to call field \"DebugPrint\" (a nil value).",
	"You should no longer see Error: CCD/Modules/RaidBuffTracker.lua:517 attempt to index a nil value. It hasn't been fixed but it's fine to ignore it until I find the problem.",
	"^Highlights",
	"You can use the Importer to transfer your UI and settings from one toon to another. Simply right click the CCD button and select \"Importer\" from there.",
	"You need to press Enter after typing new values in any of the text boxes related to CCD.",
}})
table.insert(ChangeLog, {"3.2.1", {
	"@New Features",
	"The alert for weapon stones will now display the image for the crafted, level 61 stone appropriate for your class. Warriors with two-handers will see the Coral Whetstone icon.",
	"The 20 Cooldown Bars in Edit Mode have been replaced with a single bar that acts as an anchor for others.",
	"Options for Cooldown Bars have been added to the Raid Cooldowns tab for convenience.",
	"Cooldown Bars now have a minimum length of 250 (up from 50). Any bars under 250 will be resized automatically.",
	"$Bug Fixes",
	"The Symbiosis alert now triggers correctly for the your Synthesis target.",
	"Cooldown Bars of a particular group will no longer show when that group's Alert has been turned off.",
	"Cooldown Bars should now be shown or hidden based on settings more reliably.",
	--"^Changes",
}})
table.insert(ChangeLog, {"3.2.0", {
	"NEW: Added tracking for the Enduring Tank Vial. - Suggested by Chickie",
	"NEW: Added Beastmaster's Spotter's Call to the list of debuffs tracked.",
	"NEW: All raid cooldowns now have the option to be individually: hidden, only shown in combat and only shown in groups. - Suggested by Kabloom",
	"NEW: Cooldown Bars! A lot of you requested the option to be able to split raid cooldown alerts per person but that wasn't really doable so these are a compromise. You now have 20 cooldowns bars that will display the status of each cooldown being tracked (up to 20). If an alert is disabled, it's bar is also disabled. Bars have the option to be collectively: shown/hidden, only shown in combat and only shown in groups of a certain size like every other alert. - Suggested by The People",
	"NEW: Added a frame for the 5% non-physical damage provided by Clinging Spirit (Inquisitor), Sigil of Spirits (Cabalist), Tenebrious Distortion (Defiler) and Embers Bomb(Sab).",
	"NEW: Added a frame for the 5% critical chance debuff provided by Lethal Poison (Assassin), Dauntless Strike (Bladedancer) and Mark of Extermination (Champion).",
	"NEW: Added 4 new buffs to the cooldown tracker: Unstable Transformation, Shield Defense, Symbiosis and Fiery Will. These are just experimental but if you find them useful and would like another buff tracked just send me a tell.",
	"NEW: Added an alert to track the L61 PA consumable for gloves, Weapon Aegis.",
	"NEW: Included the L61 main stat, spell power and attack power PA weapon stones to the list of recognized consumables. - Suggested by Sherwood",
	"See the Changes page on Curse or RiftUI for the full list."}})
table.insert(ChangeLog, {"3.1.7", {
	"FIX: Possible fix for a small bug related to tracking raid buffs. - Reported by Topiland",
}})
table.insert(ChangeLog, {"3.1.6", {
	"NEW: Added an alert for tracking Phenomenal Crit Bottles. Currently you can only adjust settings for the alert by right clicking the cog in Edit mode. It shares the footers (Active, Ready, etc) from the Raid Cooldown Tracker.  - Suggested by Lachtate@Wolfsbane",
	"NEW: There is now an option to set the number of people needed in your group before the raid support message and raid buffs are shown. - Suggested by Egm@Wolfsbane",
	"FIX: The Clear Text button in the Debug tab now works. - Reported by Topiland",
	"FIX: Cleaned up some of the code regarding debuffs. This may have fixed some inconsistencies with the debuff alerts.",
	"FIX: Oracle's Boon of Resurgence not overwriting Bard's Fanfare of Power (RIFT bug) was not being handled well by CCD but it should work fine now.",
	"CHANGE: Using default alert settings for Raid Cooldowns should now show text as \"Ready\" rather than \"Read\". - Reported by Kajara",
	"CHANGE: Cleaned up some code related to the Raid Buff and Raid Cooldown tracking. If you were experiencing fps lag before then try this release and let CCD.Me know if the problem persists.",
}})
table.insert(ChangeLog, {"3.1.5", {
	"NEW: Added Focused Fire to the Raid Cooldown Tracker. There may be a few bugs related to Wavelength resetting.",
	"NEW: You can now use /ccd locate to help find the menu button. Use /ccd print for a list of commands.",
	"NEW: Frames that are only to be shown during combat will have a light red tint in Edit Mode.",
	"CHANGE: Revamped the Frame Linking system such that any two frames may be linked by drag-and-drop. See the Frame Linking tab for more details on how the new system works. A side effect of the revamp is that importing old Link settings will not work. You will need to re-save settings under the new version by logging on to the toon you are importing from.",
	"FIX: The Combat Only option for Raid Buffs should work more reliably. - Reported by Coffeejelly",
}})
table.insert(ChangeLog, {"3.1.4", {
	"NEW: Each Raid Buff frame now has an \"Only show in combat\" option. There is also an option in the Raid Buffs tab enable/disable this for all frames.",
	"FIX: The Sigil of Concentration frame should now disappear if you are using the percentage-based buff rather than a flat value. Credit to Notsherwoodd@Greybriar and Coffeejelly@Greybriar for reporting.",
	"CHANGE: Closing the main window will no longer disable Edit Mode.",
}})
table.insert(ChangeLog, {"3.1.3", {
	"FIX: The weapon enchant frame for rogues and warriors will now only show if there are no enchants on either weapon. This may change to show when an enchant is missing on just one weapon in the future.",
	"FIX: Feast of the Ghar alert should now work properly.",
	"NEW: The appropriate vial and stone icons (based on class) will now be displayed.",
}})
table.insert(ChangeLog, {"3.1.2", {
	"FIX: Importing should actually work now.",
}})
table.insert(ChangeLog, {"3.1.1", {
	"FIX: Linking Raid Cooldowns now works properly.",
}})
table.insert(ChangeLog, {"3.1.1", {
	"NEW: There is a now a tab to track addon changes.",
	"FIX: Some textual errors.",
	"FIX: Bug that caused Orchestra of the Planes' active timer to refresh every time you got within range of its aura.",
	"NEW: You can now change the Raid Support Message's background.",
	"NEW: You can now change the CPU Tracker's font and background.",
	"NEW: Added the following consumables to the Raid Buff Tracker: crafted NT weapon stones, phenomenal main stat and endurance vials, AP and CP banners, AP and CP sigils, and Feast of the Ghar.",
	"NEW: You can now link Raid Buff frames to each other by simple drag-and-drop. Spacing and Link Point are configured in the \"Frame Spacing\" section of the Raid Buffs tab. Drag-and-drop linking for all frames will be available in a later version.",
}})

-- Target
function CCD.ReadMyTarget(UnitId)
	if not UnitId then
		CCD.Me.Target = nil
		CCD.Me.TargetOfTarget = nil
		return
	end

	CCD.Me.Target = CCD.Target:Create(UnitId)
	
	if not CCD.Me.Target then return end -- add error here

	if CCD.Me.Target.IsHostile then
		CCD.Me.Target:ReadAllBuffs()
	end
	
	local targetOfTargetDetails = Inspect.Unit.Detail("player.target.target")
	
	if targetOfTargetDetails then
		CCD.Me.TargetOfTarget = CCD.Target:CreateFromDetails(targetOfTargetDetails)
		-- don't do anything more
	end
end
function CCD.ReadMyTargetOfTarget(UnitId)
	if UnitId then
		CCD.Me.TargetOfTarget = CCD.Target:Create(UnitId)
	
		if not CCD.Me.TargetOfTarget then return end
		
		if not CCD.Me.Target.IsHostile and CCD.Me.TargetOfTarget.IsHostile then -- only read buffs of hostile tots, and only if target is friendly
			CCD.Me.TargetOfTarget:ReadAllBuffs()
		end	
	else
		CCD.Me.TargetOfTarget = nil
	end	
end

-- Table Manipulation
function CCD.RemoveAllCooldowns()
	local count = 0

	-- remove cooldowns from Allies.ValuesById
	for unitId, ally in pairs(Allies.ValuesById) do
		ally.RaidCooldowns = {}
	end

	-- remove cooldowns from RCT.DamageCooldownsByType
	for buffType, cooldownTable in pairs(RCT.DamageCooldownsByType) do
		for unitId, cooldown in pairs(cooldownTable) do
			RCT.DamageCooldownsByType[buffType][unitId] = nil
			count = count + 1
		end
	end

	-- clear all DisplayedCooldowns
	for i in ipairs(RCT.Frames) do
		local frame = RCT.Frames[i]
		frame.RCD = nil
		frame:SetVisible(false)
	end

	for group, cooldowns in pairs(RCT.DamageCooldownsByGroup) do
		for i in ipairs(cooldowns) do
			table.remove(cooldowns, i)
		end
	end

	print(string.format("All cooldowns have been removed. Total count: %d.", count))
end

-- Modes
function CCD.SetAutoCancelWildGrowth(Switch)
	if Switch == nil then
		CCD.AutoCancelWildGrowthOn = not CCD.AutoCancelWildGrowthOn
	else
		CCD.AutoCancelWildGrowthOn = Switch
	end

	if CCD.AutoCancelWildGrowthOn then
		print("Will now automatically cancel Wild Growth's charge draining buff.")
	else
		print("No longer cancelling Wild Growth's charge draining buff.")
	end
end

-- Frame Coordinates
function CCD.ResetAllFrames()
	RCT.ResetFramePositions()
	RBT:ResetFramePositions()
	CPUT:ResetFramePositions()
	CCD.UI.MenuButton:ResetFramePositions()
end

-- Misc
function CCD.PrintAllCooldowns()
	print(string.format("Group Count: %d", Allies.Count))
	print("----")
	print("DAMAGE:")
	for buffType, cooldowns in pairs(RCT.DamageCooldownsByType) do
		for allyId, rcd in pairs(cooldowns) do
			print(string.format("%s's %s", tostring(CCD.Allies:Get(allyId).Name), tostring(rcd.Name)))
		end
	end
	print("HEAL:")
	for buffType, cooldowns in pairs(RCT.HealCooldownsByType) do
		for allyId, rcd in pairs(cooldowns) do
			print(string.format("%s's %s", tostring(CCD.Allies:Get(allyId).Name), tostring(rcd.Name)))
		end
	end
	print("----")
end
function CCD.PrintCommands()
	local keys = {}
	
	for key, values in pairs(CCD.SlashCommands) do
		if not values[3] then
			table.insert(keys, key)
		end
	end
	
	table.sort(keys, function (a, b) return a < b end)
	
	for _, key in ipairs(keys) do
		print(string.format("/ccd %s - %s", key, CCD.SlashCommands[key][2]))
	end
end
function CCD.PrintVersion()
	print("Version: " .. Version)
end
function CCD.Test()
	--[[for k, v in pairs(CCD.Me:GetInventory()) do
		if v then
			local details = Inspect.Item.Detail(k)
			
			if details.type == "I295F0C974C8478AC,3AABAA06DD2E68B0,,,,,," then
				print(details.name .. " found! " .. details.stack)
			else 
				print(details.name .. " found! " .. details.type)
			end
		
		end
	end--]]
end
function CCD.PrintClients()
	if CCD.Messaging.ClientCount == 0 then
		print("No one in your group has CCD 3.4+.")
		return
	else
		print("CCD 3.4+ Users:")
		for id, version in pairs(CCD.Messaging.Clients) do
			print(string.format("%s (%s)",CCD.Allies:Get(id).Name, version))
		end
		print("----------------------------")
	end
end
function CCD.FindBuff(BuffName)
	--for name in pairs(CCD.BuffLibrary) do
	--	print(name)
	--end
	if CCD.BuffLibrary[BuffName] then
		local entry = CCD.BuffLibrary[BuffName]
		local s = string.format("%s - Buff = %s; Ability = %s", BuffName, CCD.BuffLibrary[BuffName].BuffTypeId, CCD.BuffLibrary[BuffName].AbilityId)
		local nameWithoutSpaces = table.concat(CCD.Messaging.SplitData(BuffName, " "))
		Debug:Record(string.format("CooldownData:Create(\"%s\", \"%s\", %d, %d, <GROUP>, nil, \"%s\", CCD.Modules.RCT.Role.Heal)", nameWithoutSpaces, entry.BuffTypeId, entry.Cooldown or -1, entry.Duration or -1, entry.AbilityId))
	end
end
function CCD.Help()
	CCD.PrintCommands()
	CCD.UI.MenuButton:Show()
	CCD.MainWindow:Show()
end

-- Event Handlers
function CCD.OnBuffAdd(Handle, UnitId, BuffTable)
	local ally = Allies.ValuesById[UnitId]
	
	if not ally then		
		local checkForCooldown = false
		
		if CCD.Me and CCD.Me.Target then
			if UnitId == CCD.Me.Target.Id then
				if CCD.Me.Target.IsHostile then -- currently ignore targets that are NOT hostile
					CCD.OnTargetBuff(CCD.Me.Target, BuffTable)
				else
					checkForCooldown = true
				end
			elseif CCD.Me.TargetOfTarget and UnitId == CCD.Me.TargetOfTarget.Id then
				if CCD.Me.TargetOfTarget.IsHostile then -- currently ignore targets of your target that are NOT hostile
					CCD.OnTargetBuff(CCD.Me.TargetOfTarget, BuffTable)
				else
					checkForCooldown = true
				end
			else -- something or someone outside the group this isn't my target nor the target of my target received a buff
				checkForCooldown = true				
			end
		else
			checkForCooldown = true	
		end
				
		if checkForCooldown then
			for buffId, buffType in pairs(BuffTable) do
				RCT:OnBuffAdd(UnitId, buffId, buffType)
			end
		end
		
		return -- ignore all player units outside of my group
	end

	for buffId, buffType in pairs(BuffTable) do -- buffType "true" for rune buffs but inspecting.details gives nil
		local buff = nil
		
		if CCD.Debug.DeveloperModeOn then -- for debugging only
			buff = Buff:Create(ally.Id, buffId)
			Debug:Record(string.format("%s = %s = %s = %s", buff.Name, buff.Type, buff.Icon, buff.Description))
			--Debug:Record(string.format("[\"%s\"] = {0, \"CLASS\", \"%s\", \"LONG\"}, -- %s", buff.Type or buff.Rune, buff.Icon, buff.Name))
			--print(string.format("%s's StartTime = %d vs Inspect.Time.Frame's %d", buff.Name, buff.StartTime or -1, Inspect.Time.Frame()))
			local name = string.lower(buff.Name)
			CCD.BuffLibrary[name] = CCD.BuffLibrary[name] or {}
			CCD.BuffLibrary[name].BuffTypeId = buff.Type
			CCD.BuffLibrary[name].Duration = buff.Duration
		end

		-- check for rune buff
		if buffType == true then
			buffType = Inspect.Buff.Detail(UnitId, buffId).rune
		end

		if buffType ~= nil then
			RBT:OnBuffAdd(ally, buffId, buffType, buff) -- RBT is first because it adds important buffs to Ally.Buffs
			RCT:OnBuffAdd(ally.Id, buffId, buffType)
			CPT:OnBuffAdd(ally, buffId, buffType)
		end
	end
end
function CCD.OnTargetBuff(Target, BuffTable)
	Target:ReadBuffs(BuffTable)

	if CCD.Debug.DeveloperModeOn then
		for buffId, buffType in pairs(BuffTable) do
			local buff = Buff:Create(Target.Id, buffId)
			Debug:Record(string.format("[\"%s\"] = {0, \"CLASS\", \"%s\"}, -- %s", buff.Type, buff.Icon, buff.Name))
		end
	end
end
function CCD.OnBuffRemove(Handle, UnitId, BuffTable)
	local ally = Allies.ValuesById[UnitId]

	if ally then
		for buffId, discarded in pairs(BuffTable) do
			local buff = ally:RemoveBuff(buffId)
			if buff then RBT:OnBuffRemove(ally, buff) end
		end
	else
		if CCD.Me.Target then
			if UnitId == CCD.Me.Target.Id and CCD.Me.Target.IsHostile then
				CCD.Me.Target:RemoveBuffs(BuffTable)
			elseif CCD.Me.TargetOfTarget and UnitId == CCD.Me.TargetOfTarget.Id and CCD.Me.TargetOfTarget.IsHostile then
				CCD.Me.TargetOfTarget:RemoveBuffs(BuffTable)
			end
		end
	end
end

function CCD.OnUpdate(Handle)
	--[[CCD.Five = CCD.Five or Inspect.Time.Frame()
	if CCD.Five and Inspect.Time.Frame() >= CCD.Five + 1 then
		CPUT:Update()
		CCD.Five = nil
	end
	if 1 == 1 then return end--]]
	if FirstUpdate then
		local details = Inspect.Unit.Detail("player.target") -- causes Unit Details to be read twice; fix this

		if details then
			CCD.OnUnitChanged(nil, details.id)
		end

		CCD.Me.RoleId = Inspect.TEMPORARY.Role()
		CPT:OnFirstUpdate()
		FirstUpdate = false
	end
	
	RCT:OnUpdate()
	CPUT:Update()
	CCD.Messaging.OnUpdate()	
end

function CCD.OnSlashCCD(Handle, Parameter)
	local params = CCD.Messaging.SplitData(string.lower(Parameter), " ")
	
	if #params == 0 then
		CCD.Help()
		return
	end
	
	if CCD.SlashCommands[params[1]] ~= nil then
		CCD.SlashCommands[params[1]][1](table.concat(params, " ", 2, #params))
	else
		CCD.UI.MenuButton.BaseFrame:SetVisible(true)
		CCD.UI.MenuButton.ContextMenu.MenuButtonHiddenCheckbox:SetChecked(false)
		CCD.MainWindow:Show()
	end
end

function CCD.OnGroupJoin(Handle, UnitId)
	local newAlly = Allies:Get(UnitId)

	if newAlly then	
	else
		local newAlly = CCD.Ally:Create(UnitId)

		if not newAlly then  -- should never happen
			Debug:Record("Could not find details for new group member with ID %s", UnitId)
			return
		end
		
		Allies:Add(newAlly)		
		
		if newAlly.Availability == "full" then
			-- FFU: may need to initialise buffs here
			RBT:CheckRoles()
		end
		
		RCT:OnNewAlly(newAlly)		
	end
end
function CCD.OnGroupLeave(Handle, UnitId)
	if UnitId == CCD.Me.Id then return end
	
	local ally = Allies:Get(UnitId)
	
	if ally then
		Allies:Remove(UnitId)

		RCT:RemoveAllCooldowns(ally)
		RBT:OnGroupLeave(ally)		
		CCD.Messaging.OnLeftGroup(UnitId)
	end
end
function CCD.OnPlayerJoin()
	CCD.Messaging.OnJoinedGroup()
end

function CCD.OnCast(Handle, UnitTable)
	for unitId, status in pairs(UnitTable) do -- status: true = cast active; false = cast ended
		local ally = Allies:Get(unitId)
		
		if ally then
			if status then
				local details = Inspect.Unit.Castbar(unitId)
				
				if details then
					if details.abilityName ==  "Changing Role" then -- possible localization issue
						CCD.OnRoleCastBegin(ally, details)
					end
				end
			else
				CCD.OnCastEnded(ally)
			end
		end
	end
end
function CCD.OnCastEnded(Ally)
	if not Ally.RoleCastBegin then return end
	
	local successfulTime = Ally.RoleCastBegin + Ally.RoleCastDuration - 0.5 -- as long as cast is within 0.5s of completion time it is considered sucessful
	local currentTime = Inspect.Time.Frame()
	--print(successfulTime .. " vs " .. currentTime)
	
	if currentTime >= successfulTime then
		CCD.OnRoleCastCompleted(Ally)
	else
		--print(Ally.Name .. " broke their role cast. Current Time: " .. currentTime .. " vs Successful Time: " .. successfulTime)
	end
	
	Ally.RoleCastBegin = nil
	Ally.RoleCastDuration = nil
end
function CCD.OnRoleCastBegin(Ally, CastbarInfo)
	if Ally == CCD.Me then return end -- use Messaging for role swap dectection when possible
	
	Ally.RoleCastBegin = CastbarInfo.begin
	Ally.RoleCastDuration = CastbarInfo.duration
end
function CCD.OnRoleCastCompleted(Ally)
	RBT:OnRoleCastCompleted(Ally)
	RCT:OnRoleCastCompleted(Ally)
	-- At this point role flags have not been updated yet so wait half a second before checking roles
	Library.LibCron.new(TOC.identifier, 0.5, 1, false, CCD.OnRoleCastCompleted_Task, Ally)
end
function CCD.OnRoleCastCompleted_Task(Ally)
	RBT:CheckRoles()
end

function CCD.OnDeath(Handle, Info)
	local ally = Allies.ValuesById[Info.target]
	if ally or not ally then return end

	local dUnitInfo = Inspect.Unit.Detail(Info.target)
	Debug:Record("%s is dying.", dUnitInfo.name)

	local details = Inspect.Unit.Detail(Allies.Ids) -- get info for every group member
	local count = 0

	for unitSpecifier, unitInfo in pairs(details) do -- don't know what unit specifier is
		-- check if this unit is dying, is out of los or has been dead
		if unitInfo.id == Info.target or unitInfo.availability ~= "full" or unitInfo.health == 0 then
			count = count + 1
			--Debug:Record("%s has been flagged as dead. Dead count: %d.", unitInfo.name, count)
		end
	end

	--[[ effectively checks for a raid wipe
	if count >= LibSRM.GroupCount() - 3 then -- (-3) added in 3.3.0 for accuracy
		RCT:ResetFlaring()
		print("Raid wipe detected.")
	end--]]
end

function CCD.OnOfflineStatusChanging(Handle, StatusesByUnitId)
	for unitId, isOffline in pairs(StatusesByUnitId) do
		local ally = Allies.ValuesById[unitId]

		if ally then
			ally.IsOffline = isOffline
			ally:ClearAllBuffs()
			RCT:RemoveAllCooldowns(ally)
			RBT:OnOffline(ally)
			Debug:Record("%s's offline status changed to %s", ally.Name, ally.IsOffline)
		end
	end
end

function CCD.OnVariablesLoaded(Event, AddonId)
	if AddonId ~= TOC.identifier then return end

	RCT.Initialize()
	RBT:Initialize()
	CPUT:Initialize()
	CPT:Initialize()
	EM:Initialize()
	
	Allies:AddSelf()
	RCT:OnNewAlly(CCD.Me)
	
	-- load save table
	CCDAccountSettings = CCDAccountSettings or {}
	if CCDAccountSettings.EnableEditModeOnMainWindow ~= nil then
		CCD.EnableEditModeOnMainWindow = CCDAccountSettings.EnableEditModeOnMainWindow
	end
	if CCDAccountSettings.Version then CCD.SavedVersion = CCDAccountSettings.Version end
	if CCDAccountSettings.ErrorManager then EM:Load(CCDAccountSettings.ErrorManager) end
	if CCDAccountSettings.DisableInWarfronts ~= nil then CCD.DisableInWarfronts = CCDAccountSettings.DisableInWarfronts end
	if CCDCharacterSettings then
		local cst = CCDCharacterSettings

		MenuButton:Load(cst.MenuButton)
		if cst.MainWindow ~= nil then
			CCD.MainWindow.X = cst.MainWindow.X or CCD.MainWindow.X
			CCD.MainWindow.Y = cst.MainWindow.Y or CCD.MainWindow.Y
		end
		if cst.AutoCancelWildGrowthOn ~= nil then CCD.AutoCancelWildGrowthOn = cst.AutoCancelWildGrowthOn end

		RCT.Load(cst.RCT)
		RBT:Load(cst.RBT)
		CPT:Load(cst.CPT)
		CPUT:Load(cst.CPUTracker)	
		Importer:Load(cst.Importer)
		CCD.UI.ContextMenu:Load(cst.ContextMenu)
		CCD.UI.EditModeWindow:Load(cst.EditModeWindow)
	else
		CCDCharacterSettings = {}
		CCD.IsFirstTime = true
		RCT.OnFirstRun()
		RBT:OnFirstRun()
		CPT:OnFirstRun()
		CPUT:OnFirstRun()
	end

	CCD.Importer:Initialize()
	CCD.UI.MenuButton:Create()
	CCD.MainWindow:Initialize()
	--CCD.MainWindow:Show()
	--CCD.IsFirstTime = true
	if CCD.IsFirstTime then
		CCD.ResetAllFrames()
		--CCD.UI.SetEditMode(true)
	end

	Library.LibCron.new(TOC.identifier, 1, math.huge, false, CCD.OnUpdate) -- start Update Hook
end
function CCD.OnVariablesSaving(Event, AddonId)
	if AddonId ~= TOC.identifier then return end

	CCDAccountSettings = CCDAccountSettings or {} -- do not overwrite Account settings
	CCDAccountSettings.EnableEditModeOnMainWindow = CCD.EnableEditModeOnMainWindow
	CCDAccountSettings.DisableInWarfronts = CCD.DisableInWarfronts
	CCDAccountSettings.Version = Version
	CCDAccountSettings.ErrorManager = EM:CreateSaveTable()

	local name = CCD.Me.Name .. "@" .. Inspect.Shard().name -- Name@Shard
	CCDAccountSettings[name] = {}
	CCDCharacterSettings = {}

	local saveTables = { CCDAccountSettings[name], CCDCharacterSettings }

	for _, svTable in pairs(saveTables) do
		svTable.AutoCancelWildGrowthOn = CCD.AutoCancelWildGrowthOn
		
		svTable.MenuButton = MenuButton:CreateSaveTable()
		svTable.MainWindow = { X = CCD.MainWindow.X, Y = CCD.MainWindow.Y }

		-- Modules		
		svTable.RCT = RCT:CreateSaveTable()
		svTable.RBT = RBT:CreateSaveTable()
		svTable.CPT = CPT:CreateSaveTable()
		svTable.CPUTracker = CPUT:CreateSaveTable()
		
		-- Importer
		svTable.Importer = Importer:CreateSaveTable()
		
		-- Context Menu
		svTable.ContextMenu = CCD.UI.ContextMenu:CreateSaveTable()
		
		-- Edit Mode Window
		svTable.EditModeWindow = CCD.UI.EditModeWindow:CreateSaveTable()
	end
end
function CCD.OnShutDownEnd(Handle)
	
end

function CCD.OnEnterFullAvailability(Handle, UnitIds)
	local shouldCheckRoles = false -- used to delay Role check until all Units have been processed
	
	for unitId, _ in pairs(UnitIds) do		
		local ally = Allies:Get(unitId)

		if ally then			
			local prevAvailability = ally.Availability
			
			ally:ReadDetails() -- recheck this positioning and function
			ally.Availability = "full" -- this should not be necessary
			
			if ally == CCD.Me and ally.Class then
				RBT:OnPlayerFullyAvailable()
				CCD.Messaging.OnPlayerFullAvailability()
			end			
						
			CCD.Messaging.OnAllyFullAvailability(ally)

			local partialTimerTag = PartialAvailabilityTimersByAlly[ally]
			
			if partialTimerTag then
				Library.LibCron.delete(partialTimerTag)				
				PartialAvailabilityTimersByAlly[ally] = nil
			elseif ally == CCD.Me then
				-- FFU: may need to remove the above check for other allies
				ally:InitializeBuffs()
			end

			shouldCheckRoles = true
			Debug:Record("%s's availability changed from %s to %s", ally.Name, prevAvailability, ally.Availability)
		else
			--Debug:Record("%s's availability changed to FULL", id)
		end
		
		if shouldCheckRoles then RBT:CheckRoles() end
	end
end
function CCD.OnEnterNoneAvailability(Handle, UnitIds)
	for id, discarded in pairs(UnitIds) do
	
		local ally = Allies.ValuesById[id]
		
		if ally ~= nil then			
			--print(string.format("%s's availability changed to NONE", ally.Name))
			local prevAvailability = ally.Availability
			ally.Availability = "none"
			Debug:Record("%s's availability changed from %s to %s", ally.Name, prevAvailability, ally.Availability)
			
			RCT:RemoveAllCooldowns(ally)
			ally:ClearAllBuffs()
			
			RBT:OnEnterNoneAvailability(ally)		
		else
			--Debug:Record("%s's availability changed to NONE", id)
		end
	end
end
function CCD.OnEnterPartialAvailability(Handle, UnitIds)
	for id in pairs(UnitIds) do
		local ally = Allies.ValuesById[id]
		
		if ally then
			--print(string.format("%s's availability changed to PARTIAL", ally.Name))
			local prevAvailability = ally.Availability
			ally.Availability = "partial"
			Debug:Record("%s's availability changed from %s to %s", ally.Name, prevAvailability, ally.Availability)
			ally:ClearAllBuffs()
			if PartialAvailabilityTimersByAlly[ally] then Library.LibCron.delete(PartialAvailabilityTimersByAlly[ally]) end
			PartialAvailabilityTimersByAlly[ally] = Library.LibCron.new(TOC.identifier, 5, 1, false, CCD.OnEnterPartialAvailability_Task, ally)
			--print("Timer " .. PartialAvailabilityTimersByAlly[ally] .. " started for " .. ally.Name)
			
		else
			--Debug:Record("%s's availability changed to PARTIAL", id)
			--local details = Inspect.Unit.Detail(id)
			--if details then
			--	print(string.format("%s's availability changed to PARTIAL", details.name))
			--end
		end
	end
end
function CCD.OnEnterPartialAvailability_Task(Ally)
	Debug:Record("%s's partial availability timer expired", Ally.Name)
	--print("Timer " .. PartialAvailabilityTimersByAlly[Ally] .. " running for " .. Ally.Name)
	if not Allies:Get(Ally.Id) then return end -- verify that the Ally is still in group
	
	--if Ally.IsOffline then--or not CCD.Messaging.Clients[Ally.Id] then
		RCT:RemoveAllCooldowns(Ally)
	--end
	
	RBT:OnEnterPartialAvailability(Ally)	
	
	PartialAvailabilityTimersByAlly[Ally] = nil
end

function CCD.OnHealthChange(Handle, UnitIds)
	for id, health in pairs(UnitIds) do
		if Allies.ValuesById[id] ~= nil then
			Allies.ValuesById[id].Health = health
		end
	end
end

function CCD.OnRoleChange(Handle, UnitIds) -- only reliable for role flag changes
	for id, role in pairs(UnitIds) do
		local ally = Allies.ValuesById[id]
		if ally then			
			ally.Role = role
			RBT:OnRoleFlagChanged(ally)
			RCT:OnRoleFlagChanged(ally)
			CPT:OnRoleFlagChanged(ally)
			Debug:Record("%s's role changed to %s", ally.Name, role)
		end
	end

	RBT:CheckRoles()
end
function CCD.OnMyRoleIdChanged(Handle, RoleId)
	-- the redundacy is needed because tracking cast bars is unreliable
	local previousRoleId = CCD.Me.RoleId
	CCD.Me.RoleId = RoleId
	
	if previousRoleId and previousRoleId ~= CCD.Me.RoleId then -- RoleId changes on full availability
		RBT:OnRoleCastCompleted(CCD.Me)
		RCT:OnRoleCastCompleted(CCD.Me)
		CCD.Messaging.BroadcastRoleChanged()
	end
	--print("New role ID: " .. RoleId)	
end

function CCD.OnUnitChanged(Handle, UnitId)
	-- store old target info
	CCD.Me.OldTarget = CCD.Me.Target
	CCD.Me.OldTargetOfTarget = CCD.Me.TargetOfTarget
	
	local targetInfo = Inspect.Unit.Detail("player.target")
	
	-- no target; just return
	if not targetInfo then
		CCD.Me.Target = nil
		CCD.Me.TargetOfTarget = nil
		return
	end
		
	CCD.Me.Target = CCD.Target:CreateFromDetails(targetInfo)
	
	if CCD.Me.Target.IsHostile then
		CCD.Me.Target:ReadAllBuffs()
		CCD.Me.TargetOfTarget = nil
	else
		local targetOfTargetInfo = Inspect.Unit.Detail("player.target.target")
		
		if targetOfTargetInfo then
			CCD.Me.TargetOfTarget = CCD.Target:CreateFromDetails(targetOfTargetInfo)
			
			if CCD.Me.TargetOfTarget.IsHostile then
				CCD.Me.TargetOfTarget:ReadAllBuffs()
			end
		else	
			CCD.Me.TargetOfTarget = nil			
		end
	end
	
	RBT:OnUnitChanged()
end

function CCD.OnCombatStatusChanged(Handle, Units)
	for id, status in pairs(Units) do
		local ally = Allies.ValuesById[id]

		if ally then
			ally.InCombat = status

			if ally == CCD.Me then
				RCT:OnCombatStatusChanged()
				RBT:UpdateAllFrames()
				CPT:OnCombatStatusChanged()
			end
		end
	end
end

function CCD.OnItemSlotUpdated(Handle, ItemIdsBySlotId)
	for slotId, itemId in pairs(ItemIdsBySlotId) do
		CPT:OnItemSlotUpdate(slotId, itemId)
		--[[local name
		local details
		if itemId then
			name = Inspect.Item.Detail(itemId).name
			details = Inspect.Item.Detail(itemId)
			print(string.format("Updated SLOT %s with ITEM %s type: %s", slotId, name or "nil", details.type))
		end --]]
		
	end
end
function CCD.OnItemUpdated(Handle, Updates)
	for slotId, itemId in pairs(Updates) do -- itemId: false if the slot is now empty, or the string "nil" if the slot no longer exists
		CPT:OnItemUpdated(slotId, itemId)
	end
end

function CCD.OnAbilityNewCooldownBegin(Handle, Cooldowns)	
	for abilityId, cooldownValue in pairs(Cooldowns) do
		--if cooldownValue > 1.5 then print(Inspect.Ability.New.Detail(abilityId).name) end
		RCT:OnAbilityNewCooldownBegin(CCD.Me.Id, abilityId, cooldownValue)		
	end
end
function CCD.OnAbilityCooldownEnd(Handle, Cooldowns)	
	for abilityId, cooldown in pairs(Cooldowns) do
		RCT:OnAbilityNewCooldownEnd(CCD.Me.Id, abilityId, cooldown)
	end
end
function CCD.OnAbilitiesAdded(Handle, Abilities)
	for abilityId in pairs(Abilities) do	
		RCT:OnAbilityAdded(abilityId, nil, CCD.Me.Id)
		RBT:OnAbilityAdded(abilityId)
		
		--
		if Debug.DeveloperModeOn then
			local details = Inspect.Ability.New.Detail(abilityId)
			if details then
				local name = string.lower(details.name)
				CCD.BuffLibrary[name] = CCD.BuffLibrary[name] or {}
				CCD.BuffLibrary[name].AbilityId = abilityId
				CCD.BuffLibrary[name].Cooldown = details.cooldown
			end
		end--]]
	end
end

function CCD.OnZoneChange(Handle, ZonesByUnitId)
	--print("triggered")
	for unitId, zoneId in pairs(ZonesByUnitId) do
		if unitId == CCD.Me.Id and zoneId then
			local details = Inspect.Zone.Detail(zoneId)
			print(details.name)
			if details.type == "warfront" then
			end
			
			
		end
	end
end

-- Event Handlers
Command.Event.Attach(Event.Buff.Add, CCD.OnBuffAdd, "BuffAddEvent")
Command.Event.Attach(Event.Buff.Remove, CCD.OnBuffRemove, "BuffRemoveEvent")
Command.Event.Attach(Command.Slash.Register("ccd"), CCD.OnSlashCCD, "SlashRegisterCommand")
Command.Event.Attach(Event.SafesRaidManager.Group.Join, CCD.OnGroupJoin, "Event.SafesRaidManager.Group.Join")
Command.Event.Attach(Event.SafesRaidManager.Group.Leave, CCD.OnGroupLeave, "Event.SafesRaidManager.Group.Leave")
Command.Event.Attach(Event.SafesRaidManager.Player.Join, CCD.OnPlayerJoin, "Event.SafesRaidManager.Player.Join")
Command.Event.Attach(Event.Unit.Castbar, CCD.OnCast, "Event.Unit.Castbar")
Command.Event.Attach(Event.Combat.Death, CCD.OnDeath, "Event.Combat.Death")
Command.Event.Attach(Event.Unit.Detail.Offline, CCD.OnOfflineStatusChanging, "Event.Unit.Detail.Offline")
Command.Event.Attach(Event.Addon.SavedVariables.Load.End, CCD.OnVariablesLoaded, "Event.Addon.SavedVariables.Load.End")
Command.Event.Attach(Event.Addon.SavedVariables.Save.Begin, CCD.OnVariablesSaving, "Event.Addon.SavedVariables.Save.Begin")
Command.Event.Attach(Event.Unit.Availability.Full, CCD.OnEnterFullAvailability, "Event.Unit.Availability.Full")
Command.Event.Attach(Event.Unit.Availability.None, CCD.OnEnterNoneAvailability, "Event.Unit.Availability.None")
Command.Event.Attach(Event.Unit.Availability.Partial, CCD.OnEnterPartialAvailability, "Event.Unit.Availability.Partial")
Command.Event.Attach(Event.Unit.Detail.Health, CCD.OnHealthChange, "Event.Unit.Detail.Health")
Command.Event.Attach(Event.Unit.Detail.Role, CCD.OnRoleChange, "Event.Unit.Detail.Role")
Command.Event.Attach(Library.LibUnitChange.Register("player.target"), CCD.OnUnitChanged, "OnTargetChanged")
Command.Event.Attach(Library.LibUnitChange.Register("player.target.target"), CCD.OnUnitChanged, "OnTargetOfTargetChanged")
Command.Event.Attach(Event.Unit.Detail.Combat, CCD.OnCombatStatusChanged, "OnCombatStatusChanged")
Command.Event.Attach(Event.Item.Slot, CCD.OnItemSlotUpdated, "CCD.OnItemSlotUpdated")
Command.Event.Attach(Event.Item.Update, CCD.OnItemUpdated, "CCD.OnItemUpdated")
Command.Event.Attach(Event.Ability.New.Cooldown.Begin, CCD.OnAbilityNewCooldownBegin, "CCD.OnAbilityNewCooldownBegin")
Command.Event.Attach(Event.Ability.New.Cooldown.End, CCD.OnAbilityCooldownEnd, "CCD.OnAbilityCooldownEnd")
Command.Event.Attach(Event.Ability.New.Add, CCD.OnAbilitiesAdded, "CCD.OnAbilitiesAdded")
Command.Event.Attach(Event.TEMPORARY.Role, CCD.OnMyRoleIdChanged, "CCD.OnMyRoleIdChanged")
Command.Event.Attach(Event.Addon.Shutdown.End, CCD.OnShutDownEnd, "CCD.OnShutDownEnd")
--Command.Event.Attach(Event.Unit.Detail.Zone, CCD.OnZoneChange, "Event.Unit.Detail.Zone")

print(string.format("CCD v%s loaded.", Version))