local TOC, CCD = ...
CCD.Modules = CCD.Modules or {}
CCD.Modules.RCT = CCD.Modules.RCT or {}
CCD.Modules.RBT = CCD.Modules.RBT or {}
CCD.Modules.CPT = CCD.Modules.CPT or {}
local RCT = CCD.Modules.RCT
local RBT = CCD.Modules.RBT
local CPT = CCD.Modules.CPT
local CPUT = CCD.Modules.CPUT
CCD.Importer = CCD.Importer or {}
local Importer = CCD.Importer
local MainWindow = CCD.MainWindow
local MenuButton = CCD.UI.MenuButton

Importer.X = UIParent:GetWidth() / 4
Importer.Y = UIParent:GetHeight() / 4

local SecureContext = UI.CreateContext("SecureContext")
SecureContext:SetSecureMode("restricted")

function Importer:Initialize()
	Importer.FrameDescription = "Importer"
	Importer.Frame = UI.CreateFrame("Mask", Importer.FrameDescription, SecureContext)
	Importer.Frame:SetSecureMode("restricted")
	Importer.Frame:SetWidth(180)
	Importer.Frame:SetHeight(210)
	Importer.Frame:SetBackgroundColor(0, 0, 0, 0.5) -- medium black tint
	Importer.Frame:SetLayer(4)
	Library.LibSimpleWidgets.SetBorder("rounded", Importer.Frame)
	Importer.Frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", self.X, self.Y)
	Importer.Frame:SetVisible(false)
	function Importer.Frame.Event:LeftDown()
		if Inspect.System.Secure() then	return end
		self.MouseDown = true
		mouseData = Inspect.Mouse()
		self.MyStartX = self:GetLeft()
		self.MyStartY = self:GetTop()
		self.StartX = mouseData.x - self.MyStartX
		self.StartY = mouseData.y - self.MyStartY
	end
	function Importer.Frame.Event:MouseMove(mouseX, mouseY)
		if Inspect.System.Secure() then	return end
		if self.MouseDown then
			self:ClearPoint("TOPLEFT")
			self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", (mouseX - self.StartX), (mouseY - self.StartY))
		end
	end
	function Importer.Frame.Event:LeftUp()
			if Inspect.System.Secure() then	return end
			if self.MouseDown then
				self.MouseDown = false
			end
		end
	function Importer.Frame.Event:Move()
		Importer.X = self:GetLeft()
		Importer.Y = self:GetTop()
	end

	-- Title
	Importer.Frame.Title = UI.CreateFrame("Text", Importer.FrameDescription .. "_Title", Importer.Frame)
	Importer.Frame.Title:SetText("CCD Importer")
	Importer.Frame.Title:SetFontSize(20)
	Importer.Frame.Title:SetFontColor(1, 1, 1, 1)
	Importer.Frame.Title:SetPoint("TOPCENTER", Importer.Frame, "TOPCENTER", 0, 1)
	Importer.Frame.Title:SetVisible(true)

	-- Dropdown
	Importer.Frame.CharacterListDropdown = UI.CreateFrame("SimpleSelect", Importer.FrameDescription .. "CharacterListDropdown", Importer.Frame)
	local items	= {}
	local values = {}
	local name = CCD.Me.Name .. "@" .. Inspect.Shard().name
	for k in pairs(CCDAccountSettings) do
		if k ~= name and string.find(k, "@") then
			table.insert(items, k)
			table.insert(values, k)
		end
	end
	Importer.Frame.CharacterListDropdown:SetItems(items, values)
	
	Importer.Frame.CharacterListDropdown:ResizeToFit()
	if self.LastCharacterSelected and CCDAccountSettings[self.LastCharacterSelected] then
		Importer.Frame.CharacterListDropdown:SetSelectedValue(self.LastCharacterSelected)
	else
		Importer.Frame.CharacterListDropdown:SetSelectedIndex(1)
	end	
	Importer.Frame.CharacterListDropdown:SetPoint("TOPCENTER", Importer.Frame.Title, "BOTTOMCENTER", 0, 5)
	Importer.Frame.CharacterListDropdown:SetVisible(true)

	-- Import Layout Button
	Importer.Frame.ImportLayoutButton = UI.CreateFrame("RiftButton",Importer.FrameDescription .. "_Button", Importer.Frame)
	Importer.Frame.ImportLayoutButton:SetText("Import Layout")
	Importer.Frame.ImportLayoutButton:SetPoint("TOPCENTER", Importer.Frame.CharacterListDropdown, "BOTTOMCENTER", 0, 5)
	Importer.Frame.ImportLayoutButton:SetVisible(true)
	function Importer.Frame.ImportLayoutButton.Event:LeftPress()
		if Inspect.System.Secure() then	return end
		Importer:ImportLayout(Importer.Frame.CharacterListDropdown:GetSelectedValue())
	end

	-- Reload UI Button
	Importer.Frame.ReloadButton = UI.CreateFrame("RiftButton", Importer.FrameDescription .. "_ReloadButton", Importer.Frame)
	Importer.Frame.ReloadButton:SetSecureMode("restricted")
	Importer.Frame.ReloadButton:SetText("Reload UI")
	Importer.Frame.ReloadButton:EventMacroSet(Event.UI.Input.Mouse.Left.Click, "reloadui")
	Importer.Frame.ReloadButton:SetPoint("TOPCENTER", Importer.Frame.ImportLayoutButton, "BOTTOMCENTER", 0, 5)
	Importer.Frame.ReloadButton:SetVisible(true)

	-- Close Button
	Importer.Frame.CloseButton = UI.CreateFrame("RiftButton",Importer.FrameDescription .. "_Button", Importer.Frame)
	Importer.Frame.CloseButton:SetText("Close")
	Importer.Frame.CloseButton:SetPoint("TOPCENTER", Importer.Frame.ReloadButton, "BOTTOMCENTER", 0, 5)
	Importer.Frame.CloseButton:SetVisible(true)
	function Importer.Frame.CloseButton.Event:LeftPress()
		if Inspect.System.Secure() then	return end
		Importer.Frame:SetVisible(false)
	end

	-- Status Text
	Importer.Frame.StatusText = UI.CreateFrame("Text", Importer.FrameDescription .. "_StatusText", Importer.Frame)
	Importer.Frame.StatusText:SetWidth(Importer.Frame:GetWidth())
	Importer.Frame.StatusText:SetText("* You must either reload your UI or relog to finish importing.")
	Importer.Frame.StatusText:SetFontSize(12)
	Importer.Frame.StatusText:SetWordwrap(true)
	Importer.Frame.StatusText:SetPoint("TOPLEFT", Importer.Frame, "BOTTOMLEFT", 0, -40)
	Importer.Frame.StatusText:SetVisible(true)
end
function Importer:ImportLayout(Character)
	local settings = CCDAccountSettings[Character]

	if not settings then return end
	print(string.format("Importing settings from %s...", Character))
	
	-- Make sure all linkable frames have their points cleared to prevent layout errors; this should be move to LinkableFrame.lua
	for _, frame in pairs(CCD.UI.LinkableFramesById) do
		frame.EditBolt:ClearAll()
	end

	-- Modules (Linkable Frames)
	RCT.Load(settings.RCT)
	RBT:Load(settings.RBT)
	CPT:Load(settings.CPT)
	CPUT:Load(settings.CPUTracker)

	-- Main Window
	if settings.MainWindow then
		MainWindow.X = settings.MainWindow.X or MainWindow.X
		MainWindow.Y = settings.MainWindow.Y or MainWindow.Y
	end
	
	-- Menu Button
	MenuButton:Load(settings.MenuButton)
	
	-- Importer
	if settings.Importer then
		Importer.X = settings.Importer.X or Importer.X
		Importer.Y = settings.Importer.Y or Importer.Y
	end

	-- wild growth
	if settings.AutoCancelWildGrowthOn then AutoCancelWildGrowthOn = settings.AutoCancelWildGrowthOn end

	print("Completed. Reload UI to get new settings.")
end
function Importer:CreateSaveTable()
	local data = {
		X = Importer.Frame:GetLeft(),
		Y = Importer.Frame:GetTop(),
		LastCharacterSelected = Importer.Frame.CharacterListDropdown:GetSelectedValue(),
	}
	
	return data
end
function Importer:Load(Settings)
	if not Settings then return end
	
	Importer.X = Settings.X
	Importer.Y = Settings.Y
	Importer.LastCharacterSelected = Settings.LastCharacterSelected	
end