local addon, CCD = ...

CCD.MainWindow = CCD.MainWindow or UI.CreateFrame("SimpleWindow", "CCDMainWindow", CCD.UI.Context)
local ChangeLog = CCD.ChangeLog
CCD.Debug = CCD.Debug or {}
local Debug = CCD.Debug

function string.startswith(String, Start1, Start2, Start3)
   if string.sub(String,1,string.len(Start1))==Start1 then return true
   elseif Start2 and string.sub(String,1,string.len(Start2))==Start2 then return true
   elseif Start3 and string.sub(String,1,string.len(Start3))==Start3 then return true
   else return false end
end

local MainWindow = CCD.MainWindow
MainWindow.X = 0
MainWindow.Y = 0
function MainWindow.Event:Move()
	self.X = self:GetLeft()
	self.Y = self:GetTop()
end
function MainWindow:Initialize()
	local windowWidth = 1000
	local windowHeight = 750

	-- Window
	self:SetWidth(windowWidth)
	self:SetHeight(windowHeight)
	self:SetLayer(4)
	if (CCD.IsFirstTime) then
		self:SetPoint("CENTER", UIParent, "CENTER", self.X, self.Y) -- 0, 0 but for consistency
	else
		self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", self.X, self.Y)
	end
	self:SetCloseButtonVisible(true)
	self:SetTitle("Choco's Cooldowns v" .. CCD.Version)
	self:SetVisible(CCD.Version ~= CCD.SavedVersion) -- always show on first run

	-- Get border info
	local border = self:GetBorder()
	local borderBottom = {X = border:GetRight(), Y = border:GetBottom()}
	local tabWidth = borderBottom.X - 52 - self:GetLeft()
	local tabHeight = borderBottom.Y - 100 - self:GetTop()

	-- Tab View
	self.TabView = UI.CreateFrame("SimpleTabView", self:GetName() .. "_TabView", self)
	self.TabView:SetWidth(tabWidth)
	self.TabView:SetHeight(tabHeight)
	self.TabView:SetPoint("CENTER", self, "CENTER", 0, 10)
	self.TabView:SetTabPosition("top")
	self.TabView:SetVisible(true)

	self.TabView:AddTab("Cooldowns", CCD.Modules.RCT:CreateTab(tabWidth))
	self.TabView:AddTab("Buffs", CCD.Modules.RBT:CreateTab(tabWidth))
	self.FrameLinkingTab = self:CreateFrameLinkingTab()
	self.TabView:AddTab("Frame Linking", self.FrameLinkingTab)
	local clTab = self:CreateChangeLogTab(tabWidth)
	self.TabView:AddTab("Change Log", clTab)

	
	self:CreateMiscTab()
	self:CreateDebugTab()

	--[[
	self.CreditText = CCD.UI.CreateText(self:GetName() .. ".CreditText", self)
	self.CreditText:SetEffectGlow({
		colorA = 1, colorB = 0, colorG = 0, colorR = 0,
		offsetX = 0, offsetY = 0,
		blurX = 2, blurY = 2,
		knockout = false,
		replace = false,
		strength = 10
	})
	self.CreditText:SetText("Credit to Kabloom@Wolfsbane!", 15, CCD.UI.Colors.Yellow)
	self.CreditText:SetPoint("TOPRIGHT", self, "TOPRIGHT", -60, 18)--]]

	if CCD.IsFirstTime or CCD.Version ~= CCD.SavedVersion then self.TabView:SetActiveTab(5) end
end
function MainWindow:CreateFrameLinkingTab()
	local name = self:GetName() .. ".FrameLinkingTab"

	-- Base
	local tab = UI.CreateFrame("Mask", name, self)
	tab:SetBackgroundColor(0, 0, 0, 1) -- black
	
	-- Link 2 Frames
	tab.Link2FramesTitle = CCD.UI.CreateTitle(name .. ".Link2FramesTitle", tab, "How to Link 2 Frames", "YELLOWRED")
	tab.Link2FramesTitle:SetPoint("TOPLEFT", tab, "TOPLEFT", 5, 10)
	--
	tab.Link2FramesGuide = CCD.UI.CreateText(name .. ".Link2FramesGuide", tab)
	local text = {
	"1. Set the Edit Mode Frame Spacing slider.",
	"2. Left click and hold the cogwheel in the top left corner of the frame you want to be linked. We'll call this Frame A; Frame B is the frame you are anchoring Frame A to.",
	"3. Drag A to one of the highlighted corners of B and release. A yellow cogwheel indicates that the frame was linked successfully.",
	}
	tab.Link2FramesGuide:SetText(table.concat(text, "\n"), 15, CCD.UI.Colors.White)
	tab.Link2FramesGuide:SetPoint("TOPRIGHT", tab, "TOPRIGHT")
	tab.Link2FramesGuide:SetPoint("TOPLEFT", tab.Link2FramesTitle, "BOTTOMLEFT", 0, 2)
			
	-- Link Module Frames
	tab.LinkModuleFramesTitle = CCD.UI.CreateTitle(name .. ".LinkModuleFramesTitle", tab, "How to Link All Frames of a Particular Module", "YELLOWRED")
	tab.LinkModuleFramesTitle:SetPoint("TOPLEFT", tab.Link2FramesGuide, "BOTTOMLEFT", 0, 10)
	--
	tab.LinkModuleFramesGuide = CCD.UI.CreateText(name .. ".LinkModuleFramesGuide", tab)
	local text = {
	"1. Select the tab for the module you want.",
	"2. Set the \"Frame Spacing\" slider and select a value from the Horizontal/Vertical dropdown menu.",
	"3. If frames are already linked and you wish to change their order then press \"Unlink Frames\".",
	"4. Arrange frames from left to right or top to bottom; don't worry about their alignment.",
	"6. Press \"Link Frames\".",
	}
	tab.LinkModuleFramesGuide:SetText(table.concat(text, "\n"), 15, CCD.UI.Colors.White)
	tab.LinkModuleFramesGuide:SetPoint("TOPRIGHT", tab, "TOPRIGHT")
	tab.LinkModuleFramesGuide:SetPoint("TOPLEFT", tab.LinkModuleFramesTitle, "BOTTOMLEFT", 0, 2)
	--
	
	return tab
end
function MainWindow:CreateMiscTab()
	local name = self:GetName()

	-- Base
	self.MiscTab = UI.CreateFrame("Mask", self:GetName() .. "_MiscTab", self)
	self.MiscTab:SetBackgroundColor(0, 0, 0, 1) -- black
	self.TabView:AddTab("Misc", self.MiscTab)

	-- Quick Start Guide Label
	self.MiscTab.QuickStartGuideLabel = CCD.UI.CreateTitle(name .. ".QuickStartGuideLabel", self.MiscTab, "Quick Start Guide", "YELLOWRED")
	self.MiscTab.QuickStartGuideLabel:SetPoint("TOPLEFT", self.MiscTab, "TOPLEFT", 5, 10)
	self.MiscTab.QuickStartGuideLabel:SetVisible(true)

	local guidelines = {
		"1. Find the         button on your screen. Left click the button to unlock frames and begin editing your layout. You can right click the cog in the top-left corner of any frame to bring up that frame's menu.",
		"2. Right click the CCD button to bring up the main menu.",
		"3. After adjusting your layout for one character, use the Importer to copy that layout to your other characters quickly. You will need to relog or reload ui to complete the import.",
		"4. You can bring up the main window through either middle clicking the CCD button, selecting it from the menu or entering /ccd. More detailed information can be found in the main window.",
			
	}
	local lastLine = self.MiscTab.QuickStartGuideLabel
	self.MiscTab.QuickStartGuideTextLines = {}
	for i, text in ipairs(guidelines) do
		self.MiscTab.QuickStartGuideTextLines[i] = CCD.UI.CreateText(string.format("%s.QuickStartGuideTextLines[%d]", name, i), self.MiscTab)
		self.MiscTab.QuickStartGuideTextLines[i]:SetWidth(self.MiscTab:GetWidth() - 5)
		self.MiscTab.QuickStartGuideTextLines[i]:SetText(text, 15, CCD.UI.Colors.White)
		self.MiscTab.QuickStartGuideTextLines[i]:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT")
		lastLine = self.MiscTab.QuickStartGuideTextLines[i]
	end

	-- Menu Icon within text
	self.MiscTab.MenuButtonTexture = UI.CreateFrame("Texture", self:GetName() .. "_Texure", self.MiscTab)
	self.MiscTab.MenuButtonTexture:SetTexture("CCD", "Images\\Button1.png")
	self.MiscTab.MenuButtonTexture:SetWidth(24)
	self.MiscTab.MenuButtonTexture:SetHeight(24)
	self.MiscTab.MenuButtonTexture:SetLayer(1) -- on top of GuideText
	self.MiscTab.MenuButtonTexture:SetPoint("TOPLEFT", self.MiscTab.QuickStartGuideTextLines[1], "TOPLEFT", 80, 0)
	self.MiscTab.MenuButtonTexture:SetVisible(true)

	-- Auto Cancel Wild Growth Checkbox
	self.MiscTab.WildGrowthCheckbox = UI.CreateFrame("SimpleCheckbox", self:GetName() .. "_WildGrowthCheckbox", self.MiscTab)
	self.MiscTab.WildGrowthCheckbox:SetFontSize(15)
	self.MiscTab.WildGrowthCheckbox:SetText("Enable Automatic Cancelling of Wild Growth's Charge Draining Buff")
	self.MiscTab.WildGrowthCheckbox:SetChecked(CCD.AutoCancelWildGrowthOn)
	self.MiscTab.WildGrowthCheckbox:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", 0, 5)
	self.MiscTab.WildGrowthCheckbox:SetVisible(true)
	function self.MiscTab.WildGrowthCheckbox.Event:CheckboxChange()
		CCD.SetAutoCancelWildGrowth(self:GetChecked())
	end

	-- CPU Tracker Checkbox
	self.MiscTab.CPUTrackerCheckBox = UI.CreateFrame("SimpleCheckbox", self:GetName() .. "_CPUTrackerCheckBox", self.MiscTab)
	self.MiscTab.CPUTrackerCheckBox:SetFontSize(15)
	self.MiscTab.CPUTrackerCheckBox:SetText("Enable CPU Tracker")
	self.MiscTab.CPUTrackerCheckBox:SetChecked(CCD.Modules.CPUT.On)
	self.MiscTab.CPUTrackerCheckBox:SetPoint("TOPLEFT", self.MiscTab.WildGrowthCheckbox, "BOTTOMLEFT", 0, 5)
	self.MiscTab.CPUTrackerCheckBox:SetVisible(true)
	function self.MiscTab.CPUTrackerCheckBox.Event:CheckboxChange()
		local isChecked = self:GetChecked()
		CCD.Modules.CPUT:Set(isChecked)
		CCD.Modules.CPUT.Frame.ContextMenu.ShowGroupCheckbox:SetChecked(isChecked)
	end

	-- Enable EditMode On Main Window Checkbox
	self.MiscTab.EMMSCheckbox = UI.CreateFrame("SimpleCheckbox", self:GetName() .. "_EMMSCheckbox", self.MiscTab)
	self.MiscTab.EMMSCheckbox:SetFontSize(15)
	self.MiscTab.EMMSCheckbox:SetText("Enable Edit Mode When Main Window Opens")
	self.MiscTab.EMMSCheckbox:SetChecked(CCD.EnableEditModeOnMainWindow)
	self.MiscTab.EMMSCheckbox:SetPoint("TOPLEFT", self.MiscTab.CPUTrackerCheckBox, "BOTTOMLEFT", 0, 5)
	self.MiscTab.EMMSCheckbox:SetVisible(true)
	function self.MiscTab.EMMSCheckbox.Event:CheckboxChange()
		CCD.EnableEditModeOnMainWindow = self:GetChecked()
	end

	-- Disable In Warfronts Checkbox
	self.MiscTab.DisableInWfsCheckbox = CCD.UI.CreateCheckbox(self:GetName() .. ".DisableInWfsCheckbox", self.MiscTab, CCD.DisableInWarfronts)
	self.MiscTab.DisableInWfsCheckbox:SetText("Disable All Alerts and Bars in Warfronts (Experimental)", 15, CCD.UI.Color.White())
	self.MiscTab.DisableInWfsCheckbox:SetPoint("TOPLEFT", self.MiscTab.EMMSCheckbox, "BOTTOMLEFT", 0, 5)
	function self.MiscTab.DisableInWfsCheckbox.Event:CheckboxChange()
		CCD.DisableInWarfronts = self:GetChecked()
	end
	
	-- Instructions Label
	self.MiscTab.KnownIssuesLabel = CCD.UI.CreateText(self:GetName() .. ".KnownIssuesLabel", self.MiscTab)
	self.MiscTab.KnownIssuesLabel:SetText("Known Issues", 18, CCD.UI.Colors.White)
	self.MiscTab.KnownIssuesLabel:SetPoint("TOPLEFT", self.MiscTab.DisableInWfsCheckbox, "BOTTOMLEFT", 0, 10)

	lastLine = self.MiscTab.KnownIssuesLabel
	local issues = {
	"1. If Wild Growth is cast with only 10 charge such that it cancelled immediately, CCD cannot track that it has been cast. This is not an issue if the caster is using CCD 3.0+.",
	"2. If Lava Field is dropped at the same time as Orchestra, the Lava Field alert and bar may not show that it has been used. This is not an issue if the caster is using CCD 3.0+.",
	"3. CCD is currently only guaranteed to work with the NA client. The only issue may be cooldowns that aren't removed on role swap due to the Changing Role cast not being translated. If the caster of those cooldowns has CCD 3.3.4+ then this is not an issue.",
	}

	for i, text in ipairs(issues) do
		local lineText = UI.CreateFrame("Text", self:GetName() .. "_Instruction" .. i, self.MiscTab)
		lineText:SetWidth(self.MiscTab:GetWidth() - 10)
		--Library.LibSimpleWidgets.SetBorder("plain", lineText, 1, 1, 1, 1, 1, "tblr")
		lineText:SetWordwrap(true)
		lineText:SetText(text)
		lineText:SetEffectGlow({
			colorA = 0.8, colorB = 0, colorG = 0, colorR = 0,
			offsetX = 0, offsetY = 0,
			blurX = 1, blurY = 1,
			knockout = false,
			replace = false,
			strength = 3
		})
		lineText:SetBackgroundColor(0, 0, 0, 0) -- transparent background
		lineText:SetFontSize(15)
		lineText:SetFontColor(1, 1, 1) -- white
		lineText:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT")
		lineText:SetVisible(true)

		--table.insert(lineFrames, lineText)

		lastLine = lineText
	end
end
function MainWindow:CreateChangeLogTab(TabWidth)
	local parentName = self:GetName()
	
	-- Scroll View
	local scrollView = UI.CreateFrame("SimpleScrollView", ".ChangeLogTab.ScrollView", self)
	
	
	-- Base
	self.ChangeLogTab = UI.CreateFrame("Mask", parentName .. ".ChangeLogTab", self)
	self.ChangeLogTab:SetWidth(TabWidth)
	self.ChangeLogTab:SetHeight(750)
	self.ChangeLogTab:SetBackgroundColor(0, 0, 0, 1) -- black
	local tab = self.ChangeLogTab
	parentName = parentName .. ".ChangeLogTab"	

	-- just an anchor now
	tab.Title = CCD.UI.CreateFrame(parentName .. ".Title", tab)
	tab.Title:SetSize(0, 0)
	tab.Title:SetPoint("TOPLEFT", tab, "TOPLEFT", 5, 00)

	local lastLine = tab.Title

	-- lines
	local tbl = {}
	for j, versionLinesTable in ipairs(ChangeLog) do
		if j <= 1 then
			local version = versionLinesTable[1]
			local lineTable = versionLinesTable[2]

			tbl[version] = CCD.UI.CreateTitle(parentName .. ".tbl." .. version, tab, string.format("v%s", version), "YELLOWRED")
			tbl[version]:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", 0, 10)

			tbl[version].Lines = {}
			parentName = parentName .. ".tbl." .. version
			lastLine = tbl[version]

			for i, line in ipairs(lineTable) do
				tbl[version].Lines[i] = CCD.UI.CreateText(parentName .. string.format(".Lines[%d]", i), tab)
				if string.startswith(line, "@") then
					tbl[version].Lines[i]:SetText(string.sub(line, 2), 15, CCD.UI.Colors.Turquoise)
				elseif string.startswith(line, "$") then
					tbl[version].Lines[i]:SetText(string.sub(line, 2), 15, CCD.UI.Color:Create(1, 0.4, 0, 1))
				elseif string.startswith(line, "^") then
					tbl[version].Lines[i]:SetText(string.sub(line, 2), 15, CCD.UI.Color:Create(1, 0, 1, 1))
				elseif string.startswith(line, "#") then
					tbl[version].Lines[i]:SetText("\t- " .. string.sub(line, 2), 15, CCD.UI.Colors.White)
				else
					tbl[version].Lines[i]:SetText("* " .. line, 15, CCD.UI.Colors.White)
				end
				tbl[version].Lines[i]:SetWidth(TabWidth - 25)
				if i == 1 then tbl[version].Lines[i]:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", 0, 5)
				else tbl[version].Lines[i]:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", 0, 0) end
				lastLine = tbl[version].Lines[i]
			end
		end
	end

	function tab:PrintToDebug()
		local allText = ""

		for _, versionLinesTable in ipairs(ChangeLog) do
			local version = versionLinesTable[1]
			local lineTable = versionLinesTable[2]
			local editedTable = {}

			for i, line in ipairs(lineTable) do
				if string.startswith(line, "@", "$", "^") then
					local editedLine = "===" .. string.sub(line, 2) .. "==="
					table.insert(editedTable, editedLine)
				else
					local editedLine = "* " .. line
					table.insert(editedTable, editedLine)
				end
			end

			local text = string.format("==v%s==\n", version) .. table.concat(editedTable, "\n")
			allText = allText .. "\n\n" .. text
		end

		MainWindow.DebugTextArea:SetText(allText)
	end
	function tab:PrintForum()
		local allText = ""

		for _, versionLinesTable in ipairs(ChangeLog) do
			local version = versionLinesTable[1]
			local lineTable = versionLinesTable[2]
			local editedTable = {}

			for i, line in ipairs(lineTable) do
				if string.startswith(line, "@") then
					local editedLine = "[COLOR=\"#00ffff\"]" .. string.sub(line, 2) .. "[/COLOR]"
					table.insert(editedTable, editedLine)
				elseif string.startswith(line, "$") then
					local editedLine = "[COLOR=\"#ff8c00\"]" .. string.sub(line, 2) .. "[/COLOR]"
					table.insert(editedTable, editedLine)
				elseif string.startswith(line, "^") then
					local editedLine = "[COLOR=\"#ff00ff\"]" .. string.sub(line, 2) .. "[/COLOR]"
					table.insert(editedTable, editedLine)
				else
					table.insert(editedTable, line)
				end
			end

			local text = string.format("[CENTER][SIZE=\"3\"]Updated to %s[/SIZE][/CENTER]\n", version) .. table.concat(editedTable, "\n")
			allText = allText .. "\n\n" .. text
		end

		MainWindow.DebugTextArea:SetText(allText)
	end

	scrollView:SetContent(self.ChangeLogTab)
	return scrollView
end
function MainWindow:CreateDebugTab()
	-- Base Frame for Instructions Tab
	self.DebugTabBackground = UI.CreateFrame("Mask", self:GetName() .. "_DebugTabBackground", self)
	self.DebugTabBackground:SetBackgroundColor(0, 0, 0, 1)
	self.TabView:AddTab("Debug", self.DebugTabBackground)

	-- Debug Text Box
	self.DebugTextArea = UI.CreateFrame("SimpleTextArea", self:GetName() .. "_DebugTextArea", self.DebugTabBackground)
	self.DebugTextArea:SetPoint("TOPCENTER", self.DebugTabBackground, "TOPCENTER", 0, 0)
	self.DebugTextArea:SetWidth(self.DebugTabBackground:GetWidth())
	self.DebugTextArea:SetHeight(self.DebugTabBackground:GetHeight() - 50)
	self.DebugTextArea:SetBackgroundColor(0, 0, 0, 1) -- black
	self.DebugTextArea:SetVisible(true)

	-- Debug Button
	self.DebugButton = UI.CreateFrame("RiftButton", self:GetName() .. "_DebugButton", self.DebugTabBackground)
	self.DebugButton:SetPoint("TOPLEFT", self.DebugTextArea, "BOTTOMLEFT", 5, 20)
	self.DebugButton:SetText("Print Log") -- black
	self.DebugButton:SetEnabled(true)
	self.DebugButton:SetVisible(true)
	function self.DebugButton.Event:LeftPress()
		CCD.Debug:Print()
	end

	-- Clear Text Button
	self.ClearTextButton = UI.CreateFrame("RiftButton", self:GetName() .. "_ClearTextButton", self.DebugTabBackground)
	self.ClearTextButton:SetPoint("TOPLEFT", self.DebugButton, "TOPRIGHT", 10, 0)
	self.ClearTextButton:SetText("Clear Text") -- black
	self.ClearTextButton:SetEnabled(true) -- set to true in Debug:Record()
	self.ClearTextButton:SetVisible(true)
	function self.ClearTextButton.Event:LeftPress()
		MainWindow.DebugTextArea:SetText("")
	end

end
function MainWindow:ToggleVisibility()
	local visibility = MainWindow:GetVisible()

	MainWindow:SetVisible(not visibility)

	if not visibility and EnableEditModeOnMainWindow then
		CCD.UI.SetEditMode(true)
	end
end
function MainWindow:Show()
	self:SetVisible(true)

	if CCD.EnableEditModeOnMainWindow then
		CCD.UI.SetEditMode(true)
	end
end
function MainWindow:Hide()
	self:SetVisible(false)
end
