local TOC, CCD = ...
CCD.Modules = CCD.Modules or {}
CCD.Modules.CPUT = CCD.Modules.CPUT or {}
local CPUT = CCD.Modules.CPUT
CCD.UI.Context = CCD.UI.Context or UI.CreateContext("CCDContext")
local MyUIContext = CCD.UI.Context
CCD.UI = CCD.UI or {}
CCD.UI.Color = CCD.UI.Color or {}
local Color = CCD.UI.Color
CCD.UI.Colors = CCD.UI.Colors or {}
local Colors = CCD.UI.Colors
local RCT = CCD.Modules.RCT

function CPUT:Initialize()
	self.Id = "CPUT"
	self.On = true
	self.FontColor = CCD.UI.Colors.White
	self.BackgroundColor = CCD.UI.Colors.HeavyBlackTint	

	-- Base Frame
	self.Frame = CCD.UI.CreateLinkableFrame(self.Id .. ".Frame", MyUIContext)
	local frame = self.Frame
	frame:SetWidth(386)
	frame:SetHeight(72)
	frame:SetMouseMasking("limited")
	function frame:OnEditModeEnd()
		if not CPUT.On then
			frame:Hide()
		end
	end
	
	local description = self.Frame:GetName()

	-- First Line
	frame.Line1 = CCD.UI.CreateText(description .. ".Line1", frame)
	frame.Line1:SetEffectGlow(CCD.UI.FrameTextEffectGlow)
	frame.Line1:SetPoint("TOPCENTER", frame, "TOPCENTER")

	-- Second Line
	frame.Line2 = CCD.UI.CreateText(description .. ".Line2", frame)
	frame.Line2:SetEffectGlow(CCD.UI.FrameTextEffectGlow)
	frame.Line2:SetPoint("TOPCENTER", frame.Line1, "BOTTOMCENTER")

	-- 3rd Line
	frame.Line3 = CCD.UI.CreateText(description .. ".Line3", frame)
	frame.Line3:SetEffectGlow(CCD.UI.FrameTextEffectGlow)
	frame.Line3:SetPoint("TOPCENTER", frame.Line2, "BOTTOMCENTER")

	function frame.EditBolt.Event:RightClick()
		frame.ContextMenu:Toggle(CPUT.Frame)
	end
	
	self.Frame.ContextMenu = self:CreateContextMenu()
end
function CPUT:Load(Settings)
	if Settings ~= nil then
		if Settings.On ~= nil then self.On = Settings.On end
		self.FontColor = Settings.FontColor or self.FontColor
		self.BackgroundColor = Settings.BackgroundColor or self.BackgroundColor

		if Settings.Linked ~= nil then
			self.Frame.X = Settings.X or self.Frame.X
			self.Frame.Y = Settings.Y or self.Frame.Y
			-- Linking
			self.Frame.Linked = Settings.Linked
			self.Frame.AnchorId = Settings.AnchorId
			self.Frame.LinkSpace = Settings.LinkSpace
			self.Frame.LinkPoint = Settings.LinkPoint
		else
			-- old data
			self.Frame.X = Settings.X or self.Frame.X
			self.Frame.Y = Settings.Y or self.Frame.Y
		end
	end

	CPUT:Start()
end
function CPUT:OnFirstRun()
	self:Start()
	self:ResetFramePositions()
end
function CPUT:Start()
	self.Frame.Line1:SetText(self.Frame.Line1:GetName(), 15, self.FontColor)
	self.Frame.Line2:SetText(self.Frame.Line2:GetName(), 15, self.FontColor)
	self.Frame.Line3:SetText(self.Frame.Line3:GetName(), 15, self.FontColor)

	self.Frame:SetBackgroundColor(self.BackgroundColor.R, self.BackgroundColor.G, self.BackgroundColor.B, self.BackgroundColor.A)
	
	if self.Frame.Linked then
		self.Frame:Link()
	else
		self.Frame:Unlink(true)
	end
	
	self.Frame:SetVisible(self.On)
	self.Frame.IsMobile = self.On
end
function CPUT:CreateSaveTable()
	local saveTable = {
		X = self.Frame.X,
		Y = self.Frame.Y,
		On = self.On,
		FontColor = self.FontColor,
		BackgroundColor = self.BackgroundColor,
		Linked = self.Frame.Linked,
		AnchorId = self.Frame.AnchorId,
		LinkPoint = self.Frame.LinkPoint,
		LinkSpace = self.Frame.LinkSpace,
	}

	return saveTable
end

function CPUT:Set(NewState)
	if NewState == self.On then return end

	self.On = NewState

	if self.On then
		self.Frame:SetVisible(true)
	elseif not CCD.UI.EditModeOn then
		self.Frame:SetVisible(false)
	end
	
	self.Frame.IsMobile = self.On
end

function CPUT:CreateContextMenu()
	-- Context Menu
	local frame = CCD.UI.CreateContextMenu(CPUT.Id .. ".ContextMenu", 320, 340, "CPU Tracker")

	-- Show Group Checkbox
	frame.ShowGroupCheckbox = frame:AddCheckbox("ShowGroupCheckbox", "Enable CPU Tracker", true)
	function frame.ShowGroupCheckbox.Event:CheckboxChange()
		local isChecked = self:GetChecked()
		CPUT:Set(isChecked)
		CCD.MainWindow.MiscTab.CPUTrackerCheckBox:SetChecked(isChecked)
	end

	frame.FontColorPicker = frame:AddColourPicker("FontColorPicker", "Font Color", false, true)
	frame.FontColorPicker.OnColorChange = function(R, G, B)
		CPUT:SetFontColor(R, G, B)
		CPUT.FontColor = Color:Create(R, G, B)
	end
	
	frame.BackgroundColorPicker = frame:AddColourPicker("BackgroundColorPicker", "Background Color", true, false)
	frame.BackgroundColorPicker.OnColorChange = function(R, G, B, A)
		CPUT.Frame:SetBackgroundColor(R, G, B, A)
		CPUT.BackgroundColor = Color:Create(R, G, B, A)
	end

	-- Load Function
	function frame:Load()
		frame.ShowGroupCheckbox:SetChecked(CPUT.On)
		frame.FontColorPicker:SetColor(CPUT.FontColor)
		frame.BackgroundColorPicker:SetColor(CPUT.BackgroundColor)
	end

	return frame
end
function CPUT:RepositionFrame()
	self.Frame.EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", self.X, self.Y)
end
function CPUT:Update()
	if not self.On then return end

	corout.new(self.Update_Task, "CPUT.Update_Task()", self)
end
function CPUT:Update_Task()
	local cpu = Inspect.Addon.Cpu()
	local subIds = {}

	for k, v in pairs(cpu["CCD"]) do
		table.insert(subIds, {k, v})
	end

	table.sort(subIds, function(a, b) return(a[2] > b[2]) end)

	self.Frame.Line1:SetText(string.format("%s = %.02f%%", subIds[1][1], subIds[1][2] * 100))
	corout.check()
	self.Frame.Line2:SetText(string.format("%s = %.02f%%", subIds[2][1], subIds[2][2] * 100))
	corout.check()
	self.Frame.Line3:SetText(string.format("%s = %.02f%%", subIds[3][1], subIds[3][2] * 100))
end
function CPUT:ResetFramePositions()
	self.Frame:Unlink()
	self.Frame.EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", RCT.HealBarContainer.X, RCT.HealBarContainer:GetBottom() + 30)
end
function CPUT:SetFontColor(R, G, B, A)
	A = A or 1
	self.Frame.Line1:SetFontColor(R, G, B, A)
	self.Frame.Line2:SetFontColor(R, G, B, A)
	self.Frame.Line3:SetFontColor(R, G, B, A)
end
