local TOC, CCD = ...

CCD.Modules = CCD.Modules or {}
CCD.Modules.CPT = CCD.Modules.CPT or {}
CCD.UI = CCD.UI or {}
CCD.UI.Color = CCD.UI.Color or {}

CCD.MainWindow = CCD.MainWindow or UI.CreateFrame("SimpleWindow", "CCDMainWindow", MyUIContext)
local MainWindow = CCD.MainWindow
local Buff = CCD.Buff
local RCT = CCD.Modules.RCT
local RBT = CCD.Modules.RBT
local Status = CCD.RCDStatus

-- Aliases
local CPT = CCD.Modules.CPT
local Color = CCD.UI.Color

function CPT:Initialize()
	self.Id = "CPT"
	self.On = true
	self.UseGreyscale = true
	self.ContextMenu = self:CreateContextMenu()
	self.TimerShowValue = 10
	self.SpellCritPot = {
		Id = "BFF8A4CB1E4A8F3DD",	
		Icon = "legendary_potion_spherical_3_1.dds",
		GreyIcon = "legendary_potion_spherical_3_1_grey.jpg",
		Name = "Phenomenal Bottle of Spellstrikes",
		ItemType = "I295F0C96CEA3BC69,4A564C21737F1E06,,,,,,",
	}
	self.IllustriousSpellStrikes = {
		Id = "B62F500A92CAECB61",
		Icon = "legendary_potion_spherical_2.dds",
		GreyIcon = "legendary_potion_spherical_3_1_grey.jpg",
		Name = "Illustrious Bottle of Spellstrikes",
		ItemType = "I14494141ABD2F48A,D19DA7218F93CD32,,,,,,",
	}
	self.PhysCritPot = {
		Id = "B60179E5916900017",
		Icon = "legendary_potion_spherical_2.dds",
		GreyIcon = "legendary_potion_spherical_2_grey.jpg",
		Name = "Phenomenal Bottle of Critical Strikes",
		ItemType = "I295F0C974C8478AC,3AABAA06DD2E68B0,,,,,,",
	}
	self.IllustriousCritStrikes = {
		Id = "B53C2F529F5DC2DA1",
		Icon = "legendary_potion_spherical_2.dds",
		GreyIcon = "legendary_potion_spherical_2_grey.jpg",
		Name = "Illustrious Bottle of Critical Strikes",
		ItemType = "I149ADF5136F85DE8,D73D488B6D8DC21C,,,,,,",
	}
	
	self.ItemCooldown = 180 -- 3 minutes
	self.PotReady = true
	self.ShowInertTimer = true
	CPT.Inventory = CPT.Inventory or {}
	self.HideAlertWhenPotCountIsZero = false
	
	local id = string.format("%s.Frame", self.Id)
	local frame = CCD.UI.AlertFrame.Create(id, CCD.UI.Context)
	self.Frame = frame
	CPT.Frame.X = 300
	CPT.Frame.Y = 800
	frame.OnlyShowInDamageRole = false
	--frame.TimerTextFontSizeLocked = true
	
	-- Edit Bolt
	function frame.EditBolt.Event:RightClick()
		CPT.ContextMenu:Toggle(CPT.Frame)
	end
	
	-- frame functions
	function frame:Update()
		if CPT.PotReady then
			CPT.Frame:SetFontColor(CCD.UI.Colors.Yellow)
			CPT.Frame.Texture:SetTexture("Rift", CPT.MyPot[1].Icon) -- this sets it to the appropriate icon regardless of which pot was used
			CPT.Frame.TimerText:SetVisible(false)
			CPT.Frame:SetHeaderText(CPT:GetPotCount())
			CPT.Frame:SetFooterText(RCT.StatusText[Status.Ready])
			
			if self:ShouldBeShown() then
				CPT.Frame:ShowX(CPT.On)
			else
				CPT.Frame:Hide()
			end
		else
			CPT.Frame:Hide()
			-- the timer will update frame automatically
		end
	end
	function frame:OnEditModeBegin()		
		frame:SetFontColor(CCD.UI.Colors.White)
		frame.Texture:SetTexture("Rift", CPT.MyPot[1].Icon)
		self.TimerText:SetText(string.format("%d", CPT.TimerShowValue))
		self.TimerText:SetVisible(true)
		frame:SetFooterText("Edit")
	end
	function frame:OnEditModeEnd()
		frame:Update()
	end
	function frame:OnIconSizeChanged()
		if CPT.ContextMenu.Group == frame.Group then
			CPT.ContextMenu.IconScaleSlider:SetPosition(self.IconSize, true)
		end
	end
	function frame:OnTimerTextFontSizeChanged()
		if CPT.ContextMenu.Group == frame.Group then
			CPT.ContextMenu.TimerTextFontSizeSlider:SetPosition(self.TimerTextFontSize, true)
		end
	end
	function frame:OnGroupShowValueChanged()
		if not CCD.UI.EditModeOn then
			frame:Update()
		end
	end
	function frame:SetCombatShow(Status)
		self.CombatShow = Status
		self:UpdateForeground()
	end
	function frame:UpdateForeground()
		if CPT.On then
			if self.CombatShow then
				self:SetForegroundColor(CCD.UI.Colors.LightRedTint)
			else
				self:SetForegroundColor(CCD.UI.Colors.ForegroundTint)
			end
		else
			self:SetForegroundColor(CCD.UI.Colors.ForegroundBlackTint)
		end
	end
	function frame:ShouldBeShown()
		if CCD.UI.EditModeOn then return false end
		if not CPT.On then return false end
		if self.CombatShow and not CCD.Me.InCombat then return false end
		if self.OnlyShowInDamageRole and CCD.Me:GetRole() ~= "dps" then return false end
		if self.GroupShowValue >= 2 and CCD.Allies.Count < self.GroupShowValue then return false end
		if CPT.HideAlertWhenPotCountIsZero and CPT.GetPotCount() == 0 and CPT.PotReady then return false end
		return true
	end
end
function CPT:OnFirstRun()
	self:ResetFramePositions()
	self:Start()
end
function CPT:Load(Settings)
	if Settings ~= nil then
		if Settings.On ~= nil then CPT.On = Settings.On end
		if Settings.ShowInertTimer ~= nil then CPT.ShowInertTimer = Settings.ShowInertTimer end
		if Settings.UseGreyscale ~= nil then CPT.UseGreyscale = Settings.UseGreyscale end
		CPT.TimerShowValue = Settings.TimerShowValue or CPT.TimerShowValue
		if Settings.HideAlertWhenPotCountIsZero ~= nil then CPT.HideAlertWhenPotCountIsZero = Settings.HideAlertWhenPotCountIsZero end
		
		if Settings.Frame then
			CPT.Frame.X = Settings.Frame.X or CPT.Frame.X
			CPT.Frame.Y = Settings.Frame.Y or CPT.Frame.Y
			if Settings.Frame.CombatShow ~= nil then CPT.Frame.CombatShow = Settings.Frame.CombatShow end
			
			if Settings.Frame.Linked ~= nil then CPT.Frame.Linked = Settings.Frame.Linked end
			CPT.Frame.LinkPoint = Settings.Frame.LinkPoint or CPT.Frame.LinkPoint
			CPT.Frame.AnchorId = Settings.Frame.AnchorId or CPT.Frame.AnchorId
			CPT.Frame.LinkSpace = Settings.Frame.LinkSpace or CPT.Frame.LinkSpace
			
			CPT.Frame.IconSize = Settings.Frame.IconSize or CPT.Frame.IconSize
			if CPT.Frame.IconSize <= 10 then CPT.Frame.IconSize = 39 + ((CPT.Frame.IconSize - 1) * 11) end
			CPT.Frame.GroupShowValue = Settings.Frame.GroupShowValue or CPT.Frame.GroupShowValue
			if Settings.Frame.OnlyShowInDamageRole ~= nil then CPT.Frame.OnlyShowInDamageRole = Settings.Frame.OnlyShowInDamageRole end
			CPT.Frame.Alpha = Settings.Frame.Alpha or CPT.Frame:GetAlpha()
			CPT.Frame.TimerTextFontSize = Settings.TimerTextFontSize or Settings.Frame.TimerTextFontSize or CPT.Frame.TimerTextFontSize
			if Settings.Frame.TimerTextFontSizeLocked ~= nil then CPT.Frame.TimerTextFontSizeLocked = Settings.Frame.TimerTextFontSizeLocked end

			if Settings.Frame.HeaderOn ~= nil then CPT.Frame.HeaderOn = Settings.Frame.HeaderOn end
			CPT.Frame.HeaderTextLength = Settings.Frame.HeaderTextLength or CPT.Frame.HeaderTextLength
			CPT.Frame.HeaderTextFontSize = Settings.Frame.HeaderTextFontSize or CPT.Frame.HeaderTextFontSize
			if Settings.Frame.FooterOn ~= nil then CPT.Frame.FooterOn = Settings.Frame.FooterOn end
			CPT.Frame.FooterTextFontSize = Settings.Frame.FooterTextFontSize or CPT.Frame.FooterTextFontSize
		end
	end
	
	CPT:Start()
end
function CPT:Start()
	if CPT.Frame.Linked then
		CPT.Frame:Link(CCD.UI.LinkableFramesById[CPT.Frame.AnchorId], CPT.Frame.LinkPoint, CPT.Frame.LinkSpace)
	else
		CPT.Frame.EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", CPT.Frame.X, CPT.Frame.Y)
	end
	
	CPT.Frame:UpdateForeground()	
	CPT.Frame.IsMobile = CPT.On
	CPT.Frame:Initialize()
	
	CCD.Me.OnReadDetailsEnd = function()
		if CCD.Me.Class ~= nil then
			CPT:InitializeMyPot()
			
			for slotId, itemId in pairs(CCD.Me:GetInventory()) do
				if itemId then
					local details = Inspect.Item.Detail(slotId)
					
					if details.type == CPT.MyPot[1].ItemType or details.type == CPT.MyPot[2].ItemType then
						CPT.Inventory[slotId] = details.stack or 1
					end
				end
			end
			
			CPT.Frame:Update()
		end
	end
end
function CPT:CreateSaveTable()
	local st = {}
	
	st.On = CPT.On
	st.ShowInertTimer = CPT.ShowInertTimer
	st.UseGreyscale = CPT.UseGreyscale
	st.TimerShowValue = CPT.TimerShowValue
	st.HideAlertWhenPotCountIsZero = CPT.HideAlertWhenPotCountIsZero
	
	st.Frame = {}	
	st.Frame.X = CPT.Frame.X
	st.Frame.Y = CPT.Frame.Y
	st.Frame.Linked = CPT.Frame.Linked
	st.Frame.LinkPoint = CPT.Frame.LinkPoint
	st.Frame.AnchorId = CPT.Frame.AnchorId
	st.Frame.LinkSpace = CPT.Frame.LinkSpace
	st.Frame.IconSize = CPT.Frame.IconSize
	st.Frame.CombatShow = CPT.Frame.CombatShow
	st.Frame.GroupShowValue = CPT.Frame.GroupShowValue
	st.Frame.OnlyShowInDamageRole = CPT.Frame.OnlyShowInDamageRole
	st.Frame.Alpha = CPT.Frame.Alpha
	st.Frame.TimerTextFontSize = CPT.Frame.TimerTextFontSize
	st.Frame.TimerTextFontSizeLocked = CPT.Frame.TimerTextFontSizeLocked
	
	st.Frame.HeaderOn = CPT.Frame.HeaderOn
	st.Frame.HeaderTextLength = CPT.Frame.HeaderTextLength
	st.Frame.HeaderTextFontSize = CPT.Frame.HeaderTextFontSize
	st.Frame.FooterOn = CPT.Frame.FooterOn
	st.Frame.FooterTextFontSize = CPT.Frame.FooterTextFontSize
		
	return st
end
function CPT:ResetFramePositions()
	CPT.Frame:Unlink()
	--[[CPT.Frame.EditBolt:ClearAll()
	CPT.Frame.EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", 100, 660)
	CPT.Frame.X = 100
	CPT.Frame.Y = 660--]]
	CPT.Frame:Link(RCT.DamageAlertsByGroup[11], "TOPRIGHT", 5)
end

function CPT:OnFirstUpdate()
	local buff = {
		StartTime = nil,
		Duration = 0,
		Remaining = 0,
	}
	local found = false
	
	for slotId, itemId in pairs(CCD.Me:GetInventory()) do
		if itemId then
			local details = Inspect.Item.Detail(slotId)
			
			if details.type == CPT.MyPot[1].ItemType or details.type == CPT.MyPot[2].ItemType then
				if not found then
					if details.cooldownRemaining then
						buff.StartTime = details.cooldownBegin
					end					
					
					found = true
				end
				
				CPT.Inventory[slotId] = details.stack or 1
			end
		end	
	end
	
	if found then
		if buff.StartTime then
			CPT:StartActiveTimer(buff, CPT.MyPot)
		else
			CPT.PotReady = true
			CPT.Frame:Update()
		end
	end	
end
function CPT:OnBuffAdd(Ally, BuffId, BuffType)
	if not CPT.MyPot then return end
	
	if Ally ~= CCD.Me then return end
	
	if BuffType == CPT.MyPot[1].Id or BuffType == CPT.MyPot[2].Id then
		buff = Ally.Buffs[BuffId] or Buff:Create(Ally.Id, BuffId)
				
		self.PotReady = false
		self:StartActiveTimer(buff)
	end
end
function CPT:OnCombatStatusChanged()
	CPT.Frame:Update()
end
function CPT:OnRoleFlagChanged(Ally)
	if Ally == CCD.Me then
		CPT.Frame:Update()
	end
end
function CPT:OnItemSlotUpdate(SlotId, ItemId)
	if not CPT.MyPot then return end
	
	local slotType = Utility.Item.Slot.Parse(SlotId) -- inventory, vault, equipment, etc
	
	if slotType ~= "inventory" then return end
	
	if CPT.Inventory[SlotId] then
		if ItemId then
			CPT.Inventory[SlotId] = Inspect.Item.Detail(ItemId).stack or 1			
		else
			CPT.Inventory[SlotId] = nil
		end
	elseif ItemId and ItemId ~= "nil" then
		local details = Inspect.Item.Detail(ItemId)
		
		if details.type == CPT.MyPot[1].ItemType or details.type == CPT.MyPot[2].ItemType then
			CPT.Inventory[SlotId] = details.stack or 1
		end		
	end
	
	CPT.Frame:SetHeaderText(CPT:GetPotCount())
	
	if CPT.Frame:ShouldBeShown() then
		CPT.Frame:ShowX(CPT.On)
	else
		CPT.Frame:Hide()
	end
end
function CPT:OnItemUpdated(SlotId, ItemId)
	if not CPT.MyPot then return end
	
	local slotType = Utility.Item.Slot.Parse(SlotId) -- inventory, vault, equipment, etc
	
	if slotType ~= "inventory" then return end
	
	if CPT.Inventory[SlotId] then
		if ItemId and ItemId ~= "nil" then
			CPT.Inventory[SlotId] = Inspect.Item.Detail(ItemId).stack or 1			
		else
			CPT.Inventory[SlotId] = nil
		end	
	end
	
	CPT.Frame:SetHeaderText(CPT:GetPotCount())
	
	if CPT.Frame:ShouldBeShown() then
		CPT.Frame:ShowX(CPT.On)
	else
		CPT.Frame:Hide()
	end
end

function CPT:Switch(NewState)
	if NewState == self.On then return end

	self.On = NewState

	if self.On then
		
	elseif not CCD.UI.EditModeOn then
		CPT.Frame:Hide()
	end
	CPT.Frame:UpdateForeground()
	
	CPT.Frame.IsMobile = CPT.On
end
function CPT:InitializeMyPot()	
	local resource = CCD.Me:GetResource()

	if resource == "MANA" then
		CPT.MyPot = { self.SpellCritPot, self.IllustriousSpellStrikes }
	else
		CPT.MyPot = { self.PhysCritPot, self.IllustriousCritStrikes} -- for any classes other than mage and cleric
	end
end
function CPT:GetPotCount()
	local count = 0
	
	for _, stack in pairs(CPT.Inventory) do
		count = count + stack
	end
	
	return count
end

function CPT:StartActiveTimer(Buff)
	if self.LastTag then Library.LibCron.delete(self.LastTag) end
	local repetitions = self.ItemCooldown - (Inspect.Time.Frame() - Buff.StartTime) + 1 -- time left on item cooldown + 1s for setting frame to "Ready"
	
	self.LastTag = Library.LibCron.new(TOC.identifier, 1, repetitions, false, self.StartActiveTimer_Task, self, Buff, CritPot)
	if CCD.UI.EditModeOn then return end
	
	if Buff.Remaining and Buff.Remaining > 0 then
		CPT.Frame:SetFontColor(CCD.UI.Colors.Green)
		self.Frame.Texture:SetTexture("Rift", CPT.MyPot[1].Icon)		
		self.Frame.TimerText:SetText(string.format("%d", Buff.Remaining - 1))
		self.Frame.TimerText:SetVisible(true)
		self.Frame:SetHeaderText(CPT:GetPotCount())
		self.Frame:SetFooterText(RCT.StatusText[Status.Active])
	end
	
	if CPT.Frame:ShouldBeShown() then		
		CPT.Frame:ShowX(CPT.On)
	else
		CPT.Frame:Hide()
	end
end
function CPT:StartActiveTimer_Task(Buff)
	if not self.On or CCD.UI.EditModeOn then return end
	
	local activeTime = Buff.Duration - (Inspect.Time.Frame() - Buff.StartTime + 1)
	local inertTime = self.ItemCooldown - (Inspect.Time.Frame() - Buff.StartTime)
	local visible = true

	if activeTime > -1 then
		self.Frame:SetFontColor(Color.Green())
		self.Frame.Texture:SetTexture("Rift", CPT.MyPot[1].Icon)	
		self.Frame.TimerText:SetText(string.format("%d", activeTime))
		self.Frame.TimerText:Show()
		self.Frame:SetHeaderText(CPT:GetPotCount())
		self.Frame:SetFooterText(RCT.StatusText[Status.Active])
	elseif inertTime > 0 then
		if CPT.UseGreyscale then
			self.Frame.Texture:SetTexture("CCD", "\\Images\\" .. CPT.MyPot[1].GreyIcon)
		else
			self.Frame.Texture:SetTexture("Rift", CPT.MyPot[1].Icon)
		end
		
		self.Frame.TimerText:Show()
		
		if inertTime <= self.TimerShowValue then
			CPT.Frame:SetFontColor(CCD.UI.Colors.Turquoise)
			self.Frame.TimerText:SetText(string.format("%d", inertTime))
			self.Frame:SetHeaderText(CPT:GetPotCount())
			self.Frame:SetFooterText(RCT.StatusText[Status.Soon])
		elseif CPT.ShowInertTimer then
			CPT.Frame:SetFontColor(CCD.UI.Colors.Red)
			self.Frame.TimerText:SetText(string.format("%d", inertTime))
			self.Frame:SetHeaderText(CPT:GetPotCount())
			self.Frame:SetFooterText(RCT.StatusText[Status.Inert])
		else
			visible = false
		end
	else
		self.PotReady = true
		CPT.Frame:SetFontColor(CCD.UI.Colors.Yellow)
		self.Frame.Texture:SetTexture("Rift", CPT.MyPot[1].Icon)
		self.Frame.TimerText:Hide()
		self.Frame:SetHeaderText(CPT:GetPotCount())
		self.Frame:SetFooterText(RCT.StatusText[Status.Ready])
		Library.LibCron.delete(CPT.LastTag)
	end	
	
	if CPT.Frame:ShouldBeShown() then
		CPT.Frame:ShowX(visible)
	else
		CPT.Frame:Hide()
	end
end

function CPT:CreateContextMenu()
	-- Context Menu
	local frame = CCD.UI.CreateContextMenu(CPT.Id .. ".ContextMenu", 325, 548, "Crit Pot Tracker")
	--[[function frame.Event:WheelBack()
		self:SetHeight(self:GetHeight() - 1)
		print(self:GetHeight())
	end	--]]

	-- Show Group Checkbox
	frame.ShowGroupCheckbox = frame:AddCheckbox("ShowGroupCheckbox", "Enable")
	function frame.ShowGroupCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		CPT:Switch(checked)
	end

	-- Combat Show Checkbox
	frame.CombatShowCheckbox = frame:AddCheckbox("CombatShowCheckbox", "Only show while in combat")
	function frame.CombatShowCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		CPT.Frame:SetCombatShow(checked)
	end

	-- GSV Text Field
	frame.GSVTextField = frame:AddTextField(frame:GetName() .. ".GSVTextField", "Only show in groups of", "or more")
	function frame.GSVTextField.OnTextFieldChanged()
		local value = frame.GSVTextField:GetConvertedText()

		if value then			
			CPT.Frame:SetGroupShowValue(value)
		end
	end
	frame.GSVTextField:SetTextFieldChangedEvent(frame.GSVTextField.OnTextFieldChanged, frame.GSVTextField:GetName() .. ".OnTextFieldChanged")
	
	-- Show icon at... Text Box
	frame.TSVTextField = frame:AddTextField(frame:GetName() .. ".TSVTextField", "Show at", "seconds left")
	function frame.TSVTextField.OnTextFieldChanged()
		local value = frame.TSVTextField:GetConvertedText()

		if value then
			CPT.TimerShowValue = value
			CPT.Frame.TimerText:SetText(string.format("%0d", CPT.TimerShowValue))
		end
	end
	frame.TSVTextField:SetTextFieldChangedEvent(frame.TSVTextField.OnTextFieldChanged, frame.TSVTextField:GetName() .. ".OnTextFieldChanged")
	
	-- Show Inert Timers Checkbox
	frame.ShowInertTimerCheckbox = frame:AddCheckbox(frame:GetName() .. ".ShowInertTimerCheckbox", "Show while Crit Pot is DOWN")
	function frame.ShowInertTimerCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		CPT.ShowInertTimer = checked
		-- frame status will be updated by timer
	end
	
	-- Greyscale Checkbox
	frame.GreyscaleCheckbox = frame:AddCheckbox(frame:GetName() .. ".GreyscaleCheckbox", "Use greyscale while Crit Pot is DOWN")
	function frame.GreyscaleCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		CPT.UseGreyscale = checked
		-- texture will be updated by timer
	end
	
	-- DPS Role Only Checkbox
	frame.DamageRoleOnlyCheckbox = frame:AddCheckbox(frame:GetName() .. ".DamageRoleOnlyCheckbox", "Only show in damage roles")
	function frame.DamageRoleOnlyCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		CPT.Frame.OnlyShowInDamageRole = checked
	end
	
	-- Hide When Pot Count is Zero Checkbox
	frame.HideWhenPotCountIsZeroCheckbox = frame:AddCheckbox(frame:GetName() .. ".HideWhenPotCountIsZeroCheckbox", "Hide when there are no Crit Pots left", true)
	function frame.HideWhenPotCountIsZeroCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		CPT.HideAlertWhenPotCountIsZero = checked
	end
	
	-- Frame Alpha Slider
	frame.AlphaSlider = frame:AddSlider(frame:GetName() .. ".ContextMenu.AlphaSlider", "Opacity:", 0, 100)
	function frame.AlphaSlider.Slider.Event:SliderChange()
		local sliderValue = self:GetPosition()		
		CPT.Frame:SetAlphaFromSlider(sliderValue)
	end
	
	-- Icon SCALE Slider
	frame.IconScaleSlider = frame:AddSlider(frame:GetName() .. ".IconScaleSlider", "Icon Size:", CCD.UI.AlertFrame.IconSizeRange.Min, CCD.UI.AlertFrame.IconSizeRange.Max)
	function frame.IconScaleSlider.Slider.Event:SliderChange()
		local scale = self:GetPosition()
		CPT.Frame:SetIconSize(scale)
	end
	
	-- Timer Text Size Slider
	frame.TimerTextFontSizeSlider = frame:AddSlider(frame:GetName() .. ".TimerTextFontSizeSlider", "Timer Size:", CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max)
	function frame.TimerTextFontSizeSlider.Slider.Event:SliderChange()
		local size = self:GetPosition()
		CPT.Frame:SetTimerTextFontSize(size)
	end

	-- Timer Font Size Locked Check Box
	frame.TimerTextFontSizeLockedCheckbox = frame:AddCheckbox(frame:GetName() .. ".TimerTextFontSizeLockedCheckbox", "Lock Timer Font Size", true)
	function frame.TimerTextFontSizeLockedCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		CPT.Frame.TimerTextFontSizeLocked = checked
	end
		
	-- Show Header Check Box
	frame.ShowHeaderCheckbox = frame:AddCheckbox(frame:GetName() .. ".ShowHeaderCheckbox", "Show Header", false)
	function frame.ShowHeaderCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		CPT.Frame:SetHeaderVisibility(checked)
		frame.HeaderTextLengthSlider:SetVisible(checked)
		frame.HeaderTextFontSizeSlider:SetVisible(checked)
		frame.HeaderTextFontSizeSlider.Line:SetVisible(checked)
		
		if checked then
			frame.ShowFooterCheckbox:AnchorTo("ORIGINAL")
		else
			frame.ShowFooterCheckbox:AnchorTo(frame.ShowHeaderCheckbox, 0)
		end		
	end
	
	-- Header Text Length Slider
	frame.HeaderTextLengthSlider = frame:AddSlider(frame:GetName() .. ".HeaderTextLengthSlider", "Header Length:", 3, 19)
	function frame.HeaderTextLengthSlider.Slider.Event:SliderChange()
		CPT.Frame:SetHeaderTextLength(self:GetPosition())
	end

	-- Header Font Size Slider
	frame.HeaderTextFontSizeSlider = frame:AddSlider(frame:GetName() .. ".HeaderTextFontSizeSlider", "Header Font Size:", CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max, true)
	function frame.HeaderTextFontSizeSlider.Slider.Event:SliderChange()
		local size = self:GetPosition()		
		CPT.Frame:SetHeaderTextFontSize(size)		
	end
	
	-- Show Footer Check Box
	frame.ShowFooterCheckbox = frame:AddCheckbox(frame:GetName() .. ".ShowFooterCheckbox", "Show Footer", false)
	function frame.ShowFooterCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		CPT.Frame:SetFooterVisibility(checked)
		
		frame.FooterTextFontSizeSlider:SetVisible(checked)
		
		if checked then
			--frame.FooterTextFontSizeSlider.Line:AnchorTo("ORIGINAL")
		else
			--frame.FooterTextFontSizeSlider.Line:AnchorTo(frame.ShowFooterCheckbox, 0)
		end		
	end
	
	-- Footer Font Size Slider
	frame.FooterTextFontSizeSlider = frame:AddSlider(frame:GetName() .. ".FooterTextFontSizeSlider", "Footer Font Size:", CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max, false)
	function frame.FooterTextFontSizeSlider.Slider.Event:SliderChange()
		local size = self:GetPosition()		
		CPT.Frame:SetFooterTextFontSize(size)		
	end		
	
	-- Load Function
	function frame:Load()
		frame.ShowGroupCheckbox:SetChecked(CPT.On)
		frame.CombatShowCheckbox:SetChecked(CPT.Frame.CombatShow)
		frame.GreyscaleCheckbox:SetChecked(CPT.UseGreyscale)
		frame.DamageRoleOnlyCheckbox:SetChecked(CPT.Frame.OnlyShowInDamageRole)
		frame.HideWhenPotCountIsZeroCheckbox:SetChecked(CPT.HideAlertWhenPotCountIsZero)
		
		frame.ShowInertTimerCheckbox:SetChecked(CPT.ShowInertTimer)
		frame.IconScaleSlider:SetPosition(CPT.Frame.IconSize, true)
		frame.TSVTextField:SetTextField(CPT.TimerShowValue)
		frame.TimerTextFontSizeSlider:SetPosition(CPT.Frame.TimerTextFontSize, true)
		frame.TimerTextFontSizeLockedCheckbox:SetChecked(CPT.Frame.TimerTextFontSizeLocked)
		frame.GSVTextField:SetTextField(CPT.Frame.GroupShowValue)
		frame.AlphaSlider:SetPosition(CCD.Math.Round(CPT.Frame.Alpha * 100), true)
		
		self.ShowHeaderCheckbox:SetChecked(CPT.Frame.HeaderOn)
		self.HeaderTextLengthSlider:SetPosition(CPT.Frame.HeaderTextLength, true)
		self.HeaderTextFontSizeSlider:SetPosition(CPT.Frame.HeaderTextFontSize, true)
		self.ShowFooterCheckbox:SetChecked(CPT.Frame.FooterOn)
		self.FooterTextFontSizeSlider:SetPosition(CPT.Frame.FooterTextFontSize, true)
	end

	return frame
end
