-- Error Manager.lua
local TOC, CCD = ...

CCD.UI = CCD.UI or {}
CCD.Modules = CCD.Modules or {}
CCD.Modules.ErrorManager = CCD.Modules.ErrorManager or {}

-- Aliases
local EM = CCD.Modules.ErrorManager

-- Classes
EM.CCDError = {
	Type,
	DateTime,
	Message,
	CharacterName,
}
function EM.CCDError:Create(Type, Message)
	local ccdError = {}
	setmetatable(ccdError, self)
	self.__index = self
	
	ccdError.Type = Type or "None"
	ccdError.DateTime = os.time()
	ccdError.Message = Message or "None"
	ccdError.CharacterName = CCD.Me.Name .. "@" .. Inspect.Shard().name
	
	return ccdError
end

-- Public Functions
function EM:Initialize()
	EM.Errors = {}
end
function EM:OnFirstRun()
	EM:Start()
end
function EM:Load(Settings)
	if Settings then
		if Settings.Errors then
			for _, ccdError in ipairs(Settings.Errors) do
				table.insert(EM.Errors, ccdError)
			end
		end
		EM:Start()
	else
		EM:OnFirstRun()
	end
end
function EM:Start()
	
end
function EM:CreateSaveTable()
	local saveTable = {	
		Errors = {},
	}
	
	for _, ccdError in ipairs(EM.Errors) do
		table.insert(saveTable.Errors, ccdError)
	end
	
	return saveTable
end

function EM:AddError(Type, Message)
	local ccdError = EM.CCDError:Create(Type, Message)
	table.insert(EM.Errors, ccdError)
end

function EM:PrintAll()
	for _, ccdError in ipairs(EM.Errors) do
		print(string.format("[%s] [%s]: %s @ %s", ccdError.CharacterName, ccdError.Type, ccdError.Message, os.date("%c", ccdError.DateTime)))
	end
	if #EM.Errors == 0 then print("No errors recorded.") end
end
function EM:ClearAll()
	for i = #EM.Errors, 1, -1 do
		table.remove(EM.Errors, i)
	end
	print("Errors have been cleared.")
end