local TOC, CCD = ...

-- Globals
CCD.Modules = CCD.Modules or {}
CCD.Modules.RBT = CCD.Modules.RBT or {}
CCD.Buff = CCD.Buff or {}
CCD.Allies = CCD.Allies or {}
CCD.UI = CCD.UI or {}
CCD.UI.Color = CCD.UI.Color or {}
CCD.UI.Colors = CCD.UI.Colors or {}

-- Aliases
local RBT = CCD.Modules.RBT
local Buff = CCD.Buff
local Allies = CCD.Allies
local Color = CCD.UI.Color
local Colors = CCD.UI.Colors

-- Locals
local function Compare(a, b) -- Used for sorting frames by their X or Y position
	local xDistance = math.abs(a[2] - b[2])
	local yDistance = math.abs(a[3] - b[3])

	if xDistance >= yDistance then
		return a[2] < b[2]
	else
		return a[3] < b[3]
	end
end
local TooltipViewMode = {
	Off = 1,
	Always = 2,
	CombatOnly = 3,
	PassiveOnly = 4,
}
local AlertShowMode = {
	Never = 1,
	Always = 2,
	Missing = 3,	-- only when buff is missing
	Active = 4,		-- only when buff is active
}

-- Global Functions
function RBT:Initialize()
	RBT.Id = "RBT"
	RBT.On = true
	RBT.Buffs = {
		--	[BuffType]		  =	{Buff Group, Class, Icon}
		["B62F881BBF0FCCD2C"] = {1, "BARD", "fanfare_of_vigor.dds"},		-- Fanfare of Vigor
		["B4F0EF3B2483B48F7"] = {1, "ORACLE", "boon_of_vitality_b.dds"},	-- Boon of Vitality
		["B5F75E0061F5806B1"] = {2, "BARD", "fanfare_of_power.dds"},		-- Fanfare of Power
		["B3007E29CF12D03AC"] = {2, "ORACLE", "boon_of_resurgence_b.dds"},	-- Boon of Resurgence
		["B2B02AF47B53C21FC"] = {3, "BARD", "anthem_of_glory.dds"},			-- Anthem of Glory (RBT and Aura)
		["B652E7F8D09C92616"] = {3, "ORACLE", "defensive_favor.dds"},		-- Defensive Favor
		["B65677DACA87066AD"] = {3, "ARCHON", "invigoration2a.dds"},		-- Arcane Aegis (RBT)
		["B65677DAAA1A41F86"] = {3, "ARCHON", "invigoration2a.dds"},		-- Arcane Aegis (Aura)
		["BFBB53A78D37A5FC9"] = {3, "BEASTMASTER", "beastmaster_comp_intervention.dds"}, -- Bond of Shelter (Both)
		["B06C97A8D745AC7D4"] = {4, "ARCHON", "rally1x.dds"},				-- Shared Vigor (RBT)
		["B023CD8D851AD339D"] = {4, "ARCHON", "rally1x.dds"},				-- Shared Vigor (Aura)
		["B51B584CD2CC55E53"] = {5, "ARCHON", "mend1f.dds"},				-- Vitality of Stone (RBT)
		["B51B584CB7C4213D2"] = {5, "ARCHON", "mend1f.dds"},				-- Vitality of Stone (Aura)
		["B72AED2881E9C8CD7"] = {5, "BEASTMASTER", "beastmaster_bond_of_might_a.dds"}, -- Bond of Power (Both)
		["B66916BFA6D43F95A"] = {6, "ARCHON", "ardent_doctrine_01.dds"},	-- Burning Purpose (RBT)
		["B66916BF834330732"] = {6, "ARCHON", "ardent_doctrine_01.dds"},	-- Burning Purpose (Aura)
		["B72C5A8B3ED13ACC3"] = {6, "BEASTMASTER", "spiritoftree1.dds"},	-- Call of the Forest
		["B3789B303AAC4EED9"] = {7, "ARCHON", "archon-earthen_barrage.dds"}, -- Earthen Barrage
		["BFBA778B2D2CEE7B1"] = {7, "BEASTMASTER", "nysyrs_rebuke.dds"},	-- Call of Blood
		["B0E9C7498E1524C2B"] = {8, "BARD", "resonance.dds"},					-- Resonance
		["B798CF51711D54B71"] = {8, "ORACLE", "sentinel_wrathful_exuberance_b.dds"}, -- Vital Inspiration
		["B60F47F310B4793F6"] = {9, "BARD", "motif_of_tenacity.dds"},			-- Motif of Tenacity
		["B13FEC95AAB509ADD"] = {9, "ORACLE", "voidknight_ravenous_defence_01_a.dds"}, -- Inspiration of the Keep (%Dmg Redeuc)
		["BFA15035328F46502"] = {9, "BEASTMASTER", "archon_rock_slide.dds"},	-- Call of Stone
		["B75CF79A3A7B75E32"] = {10, "BARD", "motif_of_bravery.dds"},			-- Motif of Bravery
		["B44E09C3BAC84FAF8"] = {10, "ORACLE", "generic_ability_020.dds"},		-- Inspiration of Battle
		["B1452CFB0E72FA883"] = {11, "CHLOROMANCER", "opportunity_01_a.dds"},	-- Living Energy
		["B59056DA97EA511E1"] = {11, "BARD", "anthem_of_fervor.dds"},			-- Anthem of Fervor
		["BFBB4F54187B25A02"] = {11, "ORACLE", "rally3.dds"},					-- Marshaled Favor
		["B3A65A7E265FF4A76"] = {12, "BARD", "anthem_of_competence.dds"},		-- Anthem of Competence
		["B30F7C39941447343"] = {12, "ORACLE", "favor_of_the_current.dds"},		-- Favor of the Current
		["BFA3B0FEE8837F793"] = {12, "ARCHON", "surging_flames.dds"},			-- Spark Shower
		["B133FEAF4C172CAD6"] = {12, "BEASTMASTER", "swiftstrike2.dds"},		-- Bond of Alacrity (Both)
		["B40B9609C0DF37AF1"] = {13, "ARCHON", "slam4.dds"},					-- Volcanic Bomb
		["BFEFB07C231CDA112"] = {13, "BEASTMASTER", "beastmaster_enraged_companion_aura_01_a.dds"}, -- Call of Savagery
		["B0FA41E09E7A89F58"] = {14, "BARD", "motif_of_regeneration_a.dds"},	-- Motif of Regeneration
		["BFC309907EA2F0FBB"] = {14, "ORACLE", "healed.dds"},					-- Inspiration of Survival (Absorb)
		["B24561799878AD6DB"] = {14, "ORACLE", "healed.dds"},					-- Inspiration of Survival (%5 Healing)
		["B206F8F2D48B2721B"] = {18, "SELF", "food5.dds"},						-- Feast of the Ghar (SP)
		["B206F8F2F21AE709B"] = {18, "SELF", "food5.dds"},						-- Feast of the Ghar (AP)
		["B2A1357781A34EE07"] = {19, "SELF", "medium_potion_spherical_3.dds", "MANA"}, 		-- Phenomenal Brightsurge Vial
		["B5FB47C7B1CE019F8"] = {19, "SELF", "medium_potion_spherical_3.dds", "MANA"}, 		-- Illustrious Brightsurge Vial
		["B4E57169B18162520"] = {19, "SELF", "medium_potion_spherical_1.dds", "ENERGY"},	-- Phenomenal Powersurge Vial
		["B45441A66942B2875"] = {19, "SELF", "medium_potion_spherical_1.dds", "ENERGY"},	-- Illustrious Powersurge Vial
		["B00FB053AEBE90DDB"] = {19, "SELF", "medium_potion_spherical_5.dds"},		-- Phenomenal Enduring Vial
		["B57196D010E66419C"] = {19, "SELF", "medium_potion_spherical_5.dds"},		-- Illustrious Enduring Vial
		["r12AC495C6CA32B19"] = {20, "SELF", "insoles_03.dds"},						-- Spongy Insoles
		["r54A72A7D7486A939"] = {21, "SELF", "fatestone_02_blue.dds", "MANA"},		-- Pelagic Powerstone
		["r7273BFB75132FBD6"] = {21, "SELF", "ench_material_30_a.dds", "MANA"},		-- Wind Emblem
		["r61C14C434DDF615D"] = {21, "SELF", "spellstone_03_a.dds", "MANA"},		-- Flaring Sigil
		["r3BDE026769180CFF"] = {21, "SELF", "spellstone_03.dds", "MANA"},			-- Flaring Emblem
		["r12FE7D95752512F0"] = {21, "SELF", "ench_material_15_b.dds", "MANA"}, 	-- Thunder Emblem
		["r0C2445386997ECE6"] = {21, "SELF", "fatestone_03_violet.dds", "ENERGY"},	-- Coral Whetstone
		["r1B3A3BDB5E50AAC7"] = {21, "SELF", "whetstone2.dds", "ENERGY"},			-- Searrated Edge
		["r082D24455D03DDD8"] = {21, "SELF", "harbinger_vengeful_shock.dds", "ENERGY"}, -- Lightning Emblem
		["r762BD1423747FA3F"] = {21, "SELF", "fatestone_02_violet.dds", "ENERGY"},	-- Coral Oilstone
		["r6B788FB70D2FEA29"] = {21, "SELF", "whetstone3_a.dds", "ENERGY"},			-- Primal Edge
		["r6F97576F54C8E782"] = {21, "SELF", "harbinger_storm_nova_c.dds", "ENERGY"}, -- Storm Emblem
		["B51A854DC90D23C59"] = {22, "SELF", "banner_port_scion2a.dds"}, -- Summon Banner of Concentration
		["B769AA268C04A0FAC"] = {22, "SELF", "banner_port_scion2a.dds"}, -- Sigil of Concentration (Flat)
		["B77DA5A1FC46AD0DC"] = {22, "SELF", "banner_port_scion2a.dds"}, -- Sigil of Concentration (3%)
		["B61FBAEB54B9E553A"] = {23, "SELF", "flag_iron_circle.dds"}, -- Summon Banner of Rage
		["B7A0BCB33C0776758"] = {23, "SELF", "flag_iron_circle.dds"}, -- Sigil of Rage		
		["r64A1B77D51F16669"] = {26, "SELF", "grips_01_a.dds"}, -- Weapon Aegis
	}
	RBT.Debuffs = {
		["B7E3F5A42867CB671"] = {15, "ARCHON", "archon_crumbling_resistance_01.dds"}, -- Crumbling Resistance
		["B28FB2E2C8B594765"] = {15, "BARD", "coda_of_distress.dds"}, -- Coda of Distress
		["B6E2FAAA28AC69397"] = {15, "ORACLE", "curse_of_consumption.dds"}, -- Curse of Consumption
		["B4EB172833919BDE4"] = {15, "ORACLE", "curse_of_consumption.dds"}, -- Seeping Consumption
		["B3B6BDCA1CFEA464C"] = {15, "INQUISITOR", "menacingroar4b.dds"}, -- Spiritual Deficiency
		["B5085DB9C5B1F666A"] = {16, "ARCHON", "castigation1a.dds"}, -- Ashen Defense
		["B6BA1ADFD5A93F825"] = {16, "BARD", "coda_of_cowardice.dds"}, -- Coda of Cowardice
		["B6586D0E97741AF9E"] = {16, "ORACLE", "curse_of_frailty.dds"}, -- Curse of Frailty
		["BFECB7D5207EC913F"] = {16, "ORACLE", "curse_of_frailty.dds"}, -- Seeping Frailty
		["B1F112B8C62C155A2"] = {17, "BARD", "coda_of_distress_a.dds"}, -- Coda of Jeopardy
		["B0E4796AC7FD8BC72"] = {17, "ORACLE", "corroded_defense.dds"}, -- Corroded Defense
		["B11A16BF84A35E31E"] = {17, "ARCHON", "riftblade-surging_energy.dds"}, -- Illuminate
		["BFF1245B32DA8F020"] = {17, "BEASTMASTER", "awareness_01.dds"}, -- Spotter's Call
		["B58C8F3A95CAE1CFA"] = {24, "INQUISITOR", "shroud_of_agony.dds"}, -- Clinging Spirit
		["BFE71296F8F1AAA68"] = {24, "CABALIST", "shroud_of_agony_a.dds"}, -- Sigil of Spirits
		["B491E02360D5FB16B"] = {24, "DEFILER", "dark_containment_01.dds"}, -- Tenebrious Distortion
		["B1C94A46AE6BD5813"] = {24, "VULCANIST", "elemward2.dds"}, -- Slagged
		["BFF388C670F95E5A8"] = {24, "SABOTEUR", "charge_ember.dds"}, -- Embers Bomb
		["B65A5EB24B4A0128C"] = {25, "ASSASSIN", "poisonone2a.dds"}, --	Lethal Poison
		["B0AD5DE5D3D0E580C"] = {25, "BLADEDANCER", "finalstrike2.dds"}, -- Dauntless Strike
		["B73F07014EC337C95"] = {25, "CHAMPION", "finalstrike2.dds"}, -- Mark of Extermination
		["B139F74A8C28D3F73"] = {25, "DERVISH", "harbinger_essence_of_wind_b.dds"}, -- Squall
		["B205BF07C7A6EA89B"] = {27, "MARKSMAN", "lighting_fury.dds"}, -- Lightning Fury
		["B439A050AC305E060"] = {27, "SABOTEUR", "charge_splinter.dds"}, -- Splinter Bomb
		["B601C85607B7D6397"] = {27, "RANGER", "piercing_shot_01_a.dds"}, -- Piercing Shot
		["B2D65EC137A0B1B38"] = {27, "CHAMPION", "exposeweakness1.dds"}, -- Mark of Inevitability
	}
	RBT.CombinedBuffs = {}
	RBT.LivingEnergy = "B1452CFB0E72FA883"
	RBT.GroupStrings = {
		{ Title = "Endurance #1", BuffNames = { "Fanfare of Vigor (Bard)", "Boon of Vitality (Oracle)" } },
		{ Title = "Main Stat #1", BuffNames = { "Fanfare of Power (Bard)", "Boon of Resurgence (Oracle)" } },
		{ Title = "Armour + Resist", BuffNames = { "Anthem of Glory (Bard)", "Defensive Favor (Oracle)", "Arcane Aegis (Archon)", "Bond of Shelter (Beastmaster)" } },
		{ Title = "Endurance #2", BuffNames = { "Shared Vigor (Archon)", "Boon of Vitality (Oracle)" } },
		{ Title = "Main Stat #2", BuffNames = { "Vitality of Stone (Archon)", "Bond of Power (Beastmaster)" } },
		{ Title = "7% Cast Time Reduction", BuffNames = { "Burning Purpose (Archon)", "Call of the Forest (Beastmaster)" } },
		{ Title = "5% Crit Buff", BuffNames = { "Earthen Barrage (Archon)", "Call of Blood (Beastmaster)" } },
		{ Title = "5% Main Stat + End.", BuffNames = { "Resonance (Bard)", "Vital Inspiration (Oracle)" } },
		{ Title = "5% Damage Reduction", BuffNames = { "Motif of Tenacity (Bard)", "Inspiration of the Keep (Oracle)", "Call of Stone (Beastmaster)" } },
		{ Title = "Flat AP + SP", BuffNames = { "Motif of Bravery (Bard)", "Inspiration of Battle (Oracle)" } },
		{ Title = "5% Cost Reduction", BuffNames = { "Living Energy (Chloromancer)", "Anthem of Fervor (Bard)", "Marshaled Favor (Oracle)" } },
		{ Title = "15% Speed", BuffNames = { "Anthem of Competence (Bard)", "Favor of the Current (Oracle)", "Spark Shower (Archon)", "Bond of Alacrity (Beastmaster)" } },
		{ Title = "5% Damage", BuffNames = { "Volcanic Bomb (Archon)", "Call of Savagery (Beastmaster)" } },
		{ Title = "5% Healing Increase", BuffNames = { "Motif of Regeneration (Bard)", "Inspiration of Survival (Oracle)" } }, -- 14
		{ Title = "7% Non-Physical Dmg", BuffNames = { "Crumbling Resistance (Archon)", "Coda of Distress (Bard)", "Curse of Consumption (Oracle)", "Spiritual Deficiency (Inquisitor)" } },
		{ Title = "5% Physical Dmg", BuffNames = { "Ashen Defense (Archon)", "Coda of Cowardice (Bard)", "Curse of Frailty (Oracle)" } },
		{ Title = "1900 Dmg (3s ICD)", BuffNames = { "Coda of Jeopardy (Bard)", "Corroded Defense (Oracle)", "Illuminate (Archon)", "Spotter's Call (Beastmaster)" } },
		{ Title = "Feast of the Ghar", BuffNames = { "Feast of the Ghar" } },
		{ Title = "Phenomenal Vial", BuffNames = { "Phenomenal Brightsurge Vial (Wisdom + Intelligence)", "Phenomenal Powersurge Vial (Strength + Dexterity)", "Phenomenal Enduring Vial (Endurance)" } },
		{ Title = "Spongy Insoles", BuffNames = { "Spongy Insoles" } },
		{ Title = "Weapon Stone", BuffNames =
			{	"Pelagic Powerstone (350 Spell Power, L61, Crafted)", "Flaring Emblem (280 Spell Power, L61, PA)", "Flaring Sigil (160 Spell Power, L51, PA)",
				"Coral Oilstone (350 Attack Power, L61, Crafted, 2 Hander)", "Primal Edge (280 Attack Power, L61, PA, 2 Hander", "Coral Whetstone (175 Attack Power, L61, Crafted, 1 Hander)", "Serrated Edge (140 Attack Power, L61, PA, 1 Hander)",
				"Thunder Emblem (120 Intelligence, L61, PA)", "Wind Emblem (120 Wisdom, L61, PA)", "Lightning Emblem (120 Dexterity, L61, PA)", "Storm Emblem (120 Strength Power, L61, PA)",
			},
		},
		{ Title = "CP Banner/Sigil", BuffNames = { "Sigil of Concentration", "Banner of Concentration" } },
		{ Title = "AP/SP Banner/Sigil", BuffNames = { "Sigil of Rage", "Banner of Rage" } },
		{ Title = "5% Non-Phys", BuffNames = { "Clinging Spirit (Inquisitor)", "Sigil of Spirits (Cabalist)", "Tenebrious Distortion (Defiler)", "Embers Bomb (Saboteur)", "Slagged (Vulcanist)" } },
		{ Title = "5% Crit Debuff", BuffNames = { "Lethal Poison (Assassin)", "Dauntless Strike (Bladedancer)", "Mark of Extermination (Champion)", "Squall (Dervish)" } },
		{ Title = "Weapon Aegis", BuffNames = { "Weapon Aegis" } },
		{ Title = "Armour Debuff", BuffNames = { "Lightning Fury (Marksman)", "Splinter Bomb (Saboteur)", "Piercing Shot (Ranger)", "Mark of Inevitability (Champion)" } },
	}
	RBT.SupportPresent = {
		Bard = false,
		Oracle = false,
		Archon = false,
		Beastmaster = false,
		Chloromancer = false,
	}
	RBT.Archon = nil
	RBT.Chloro = nil
	RBT.EditModeTextures = {}
	RBT.DefaultGroupOrder = {1, 2, 5, 4, 3, 11, 12, 6, 7, 13, 8, 9, 10, 14, 15, 16, 17, 24, 25, 27, 18, 19, 20, 21, 22, 23, 26}
	RBT.GroupOrder = {1, 2, 5, 4, 3, 11, 12, 6, 7, 13, 8, 9, 10, 14, 15, 16, 17, 24, 25, 27, 18, 19, 20, 21, 22, 23, 26}
	RBT.ContextMenu = nil
	RBT.CombatShowToggle = true
	RBT.AlwaysShowToggle = true
	RBT.Frames = {}
	RBT.MessageFrame = RBT:CreateMessageFrame()
	RBT.BuffsByGroup = {}	
	RBT.DebuffFrames = {}
	RBT.TooltipViewMode = TooltipViewMode.Always
	
	RBT.Tooltip = RBT:CreateTooltip()
	
	-- create buff frames
	for buffType, values in pairs(RBT.Buffs) do
		local group = values[1]
		local support = values[2]
		local texture = values[3]
		local resource = values[4]

		-- create frame
		if not RBT.Frames[group] then
			RBT.Frames[group] = RBT:CreateFrame(group)
			local frame = RBT.Frames[group]
			
			function frame:Update()
				-- Mobility
				if RBT.On and frame.ShowMode ~= AlertShowMode.Never then
					frame.IsMobile = true
				else
					frame.IsMobile = false
				end
				
				if CCD.UI.EditModeOn then return end -- needed for frame:OnGroupShowValueChanged()
				
				if not RBT.On then return end -- RBT:OnEditModeEnd will hide all frames
				if frame.ShowMode == AlertShowMode.Never then frame:Hide() return end
				
				if frame.TimerTag then -- timer is updating frame
					RBT:UpdateAlertTimer(frame, frame.Buff, Inspect.Time.Frame())					
					return
				end		
				
				-- All buffs related to this alert are guaranteed to be missing from this point on
				
				if frame.ShowMode == AlertShowMode.Active then frame:Hide() return end
				
				-- No need to check for any more AlertShowModes since Always and Missing have* to be shown from now on
				
				local isSupportPresent
				local class
				local texture
				
				for _, buffType in ipairs(RBT.BuffsByGroup[frame.Group]) do
					--[[if CCD.Me.BuffsByType[buffType] then
						frame:Hide() -- timer will draw; eventually fix this to avoid alert "flashing"
						return
					end--]]
					
					--[[	currently cannot determine which texture will be chosen if group is supported by multiple
							support specs that are present; i.e. cannot choose between jeopardy and illuminate]]
					if not isSupportPresent then -- loop until support is found
						local values = RBT.Buffs[buffType]
						class = values[2]
						texture = values[3]
						
						if class == "SELF" then
							local resource = values[4]
							
							if not resource or (resource and resource == CCD.Me:GetResource()) then
								isSupportPresent = true
							end
						elseif RBT:IsClassSupportPresent(class) then
							isSupportPresent = true
						end
					end
				end	
				
				if isSupportPresent then					
					frame.TimerText:SetVisible(false) -- timer will draw back if necessary; eventually fix this to avoid alert "flashing"
					
					if class == "SELF" then
						frame.Texture:SetTexture("Rift", RBT.EditModeTextures[frame.Group])
					else
						frame.Texture:SetTexture("Rift", texture)
					end
					
					frame:ShowX()					
				else
					frame:Hide()
				end
			end			
		end

		-- use bard, archon and living energy for Edit mode textures
		if (RBT.EditModeTextures[group] == nil and (support == "BARD" or support == "ARCHON" or support == "SELF"))
		or buffType == RBT.LivingEnergy then
			RBT.EditModeTextures[group] = texture
		end

		-- populate BuffsByGroup
		if not self.BuffsByGroup[group] then self.BuffsByGroup[group] = {} end
		table.insert(self.BuffsByGroup[group], buffType)	

		self.CombinedBuffs[buffType] = values
	end
	
	-- create debuff frames
	for buffType, values in pairs(RBT.Debuffs) do
		local group = values[1]
		local support = values[2]
		local texture = values[3]
		local resource = values[4]

		-- create frame
		if not RBT.Frames[group] then
			RBT.Frames[group] = RBT:CreateFrame(group)
			local frame = RBT.Frames[group]
			
			function frame:Update()
				-- Mobility
				if RBT.On and frame.ShowMode ~= AlertShowMode.Never then
					frame.IsMobile = true
				else
					frame.IsMobile = false
				end
				
				if CCD.UI.EditModeOn then return end
				
				if not RBT.On then return end -- RBT:OnEditModeEnd will hide all frames
				if frame.ShowMode == AlertShowMode.Never then frame:Hide() return end
				
				if frame.TimerTag then -- timer is updating frame
					
					RBT:UpdateAlertTimer(frame, frame.Buff, Inspect.Time.Frame())					
					return
				end	
				
				-- All buffs related to this alert are guaranteed to be missing from this point on
				
				if frame.ShowMode == AlertShowMode.Active then frame:Hide() return end
				
				-- No need to check for any more AlertShowModes since Always and Missing have* to be shown from now on
			
				local target = CCD.Me.Target
				local isHostileTarget = false
				
				-- determine if either my target or my target's target is hostile
				if CCD.Me.Target then
					if CCD.Me.Target.IsHostile then
						isHostileTarget = true
					elseif CCD.Me.TargetOfTarget and CCD.Me.TargetOfTarget.IsHostile then
						isHostileTarget = true
						target = CCD.Me.TargetOfTarget
					end
				end
				
				if not isHostileTarget then frame:Hide() return end				
								
				--[[ on reaching this far if there is a hostile target without this debuff
					texture is set to editmodetexture but eventually it should be set based on current support if any are present]]
				frame.Texture:SetTexture("Rift", RBT.EditModeTextures[frame.Group])
				frame.TimerText:SetVisible(false)				
				frame:ShowX()
			end			
			
			table.insert(RBT.DebuffFrames, frame)
		end

		-- use bard, archon and living energy for Edit mode textures
		if (RBT.EditModeTextures[group] == nil and (support == "BARD" or support == "ARCHON" or support == "SELF"))
		or buffType == RBT.LivingEnergy then
			RBT.EditModeTextures[group] = texture
		end

		-- populate BuffsByGroup
		if not self.BuffsByGroup[group] then self.BuffsByGroup[group] = {} end
		table.insert(self.BuffsByGroup[group], buffType)
		
		self.CombinedBuffs[buffType] = values
	end

	RBT.EditModeTextures[15] = RBT.Debuffs["B7E3F5A42867CB671"][3]
	RBT.EditModeTextures[16] = RBT.Debuffs["B5085DB9C5B1F666A"][3]
	RBT.EditModeTextures[17] = RBT.Debuffs["B1F112B8C62C155A2"][3]
	RBT.EditModeTextures[24] = RBT.Debuffs["B58C8F3A95CAE1CFA"][3]
	RBT.EditModeTextures[25] = RBT.Debuffs["B65A5EB24B4A0128C"][3]	
	RBT.EditModeTextures[27] = RBT.Debuffs["B205BF07C7A6EA89B"][3]
	--RBT.EditModeTextures[26] = RBT.Debuffs["r64A1B77D51F16669"][3]
	RBT.MessageFrame.On = true
	RBT.MessageFrame.FontColor = CCD.UI.Colors.White
	RBT.MessageFrame.FontSize = 18
	RBT.MessageFrame.BackgroundColor = CCD.UI.Colors.HeavyBlackTint

	self:CreateContextMenu()
end
function RBT:OnFirstRun()	
	self:ResetFramePositions()
	self:Start()
end
function RBT:Load(Settings)
	if Settings then
		if Settings.On ~= nil then self.On = Settings.On end
		self.GroupOrder = Settings.GroupOrder or self.GroupOrder
		RBT.TooltipViewMode = Settings.TooltipViewMode or RBT.TooltipViewMode

		-- Frames
		for i, frame in ipairs(self.Frames) do
			if Settings.Frames[i] ~= nil then
				frame.X = Settings.Frames[i].X or frame.X
				frame.Y = Settings.Frames[i].Y or frame.Y

				if Settings.Frames[i].Linked ~= nil then frame.Linked = Settings.Frames[i].Linked end
				frame.AnchorId = Settings.Frames[i].AnchorId or frame.AnchorId
				frame.LinkPoint =  Settings.Frames[i].LinkPoint or frame.LinkPoint
				frame.LinkSpace = Settings.Frames[i].LinkSpace or frame.LinkSpace

				if Settings.Frames[i].IconScale then -- pre 3.4.0 setting
					frame.IconSize = 39 + ((Settings.Frames[i].IconScale - 1) * 9)
				end
				frame.IconSize = Settings.Frames[i].Scale or frame.IconSize
				if Settings.Frames[i].TimerScale ~= nil then -- pre 3.4.0 setting
					frame.TimerTextFontSize = 10 + ((Settings.Frames[i].TimerScale  - 1) * 10)
				end
				frame.TimerTextFontSize = Settings.Frames[i].TimerTextFontSize or frame.TimerTextFontSize
				if Settings.Frames[i].TimerTextFontSizeLocked ~= nil then frame.TimerTextFontSizeLocked = Settings.Frames[i].TimerTextFontSizeLocked end
				
				if Settings.Frames[i].TimerColour ~= nil then
					frame.TimerColour = Settings.Frames[i].TimerColour
				end
				if Settings.Frames[i].CombatShow ~= nil then frame.CombatShow = Settings.Frames[i].CombatShow end
				if Settings.Frames[i].GroupShowValue ~= nil then frame.GroupShowValue = Settings.Frames[i].GroupShowValue end
				frame.Alpha = Settings.Frames[i].Alpha or frame:GetAlpha()
				
				if Settings.Frames[i].IsTracked == false then -- Pre 3.3.3 setting
					frame.ShowMode = AlertShowMode.Never
				elseif Settings.Frames[i].AlwaysShow == true then -- Pre 3.3.3 setting
					frame.ShowMode = AlertShowMode.Always
					frame.ShowModeAlwaysTimerValue = Settings.Frames[i].TSV or frame.ShowModeAlwaysTimerValue
					frame.ShowModeActiveTimerValue = Settings.Frames[i].TSV or frame.ShowModeActiveTimerValue
					frame.ShowModeAlwaysTimerColor = Settings.Frames[i].TimerShowValueColor or frame.ShowModeAlwaysTimerColor
				else
					frame.ShowMode = Settings.Frames[i].ShowMode or frame.ShowMode
					frame.ShowModeMissingTimerValue = Settings.Frames[i].TSV or frame.ShowModeMissingTimerValue
				end
				
				frame.ShowModeMissingTimerValue = Settings.Frames[i].ShowModeMissingTimerValue or frame.ShowModeMissingTimerValue
				frame.ShowModeActiveTimerValue = Settings.Frames[i].ShowModeActiveTimerValue or frame.ShowModeActiveTimerValue
				frame.ShowModeAlwaysTimerValue = Settings.Frames[i].ShowModeAlwaysTimerValue or frame.ShowModeAlwaysTimerValue
				frame.ShowModeAlwaysTimerColor = Settings.Frames[i].ShowModeAlwaysTimerColor or frame.ShowModeAlwaysTimerColor
			end
		end

		-- Message Frame
		if Settings.MessageFrame ~= nil then
			local frame = self.MessageFrame
			local settings = Settings.MessageFrame

			if settings.On ~= nil then frame.On = settings.On end

			frame.FontColor = settings.FontColor or frame.FontColor
			frame.FontSize = settings.FontSize or frame.FontSize			
			frame.BackgroundColor = settings.BackgroundColor or frame.BackgroundColor
			frame.GroupShowValue = settings.GroupShowValue or frame.GroupShowValue
			
			-- Linking
			frame.X = settings.X or frame.X
			frame.Y = settings.Y or frame.Y
			if settings.Linked ~= nil then frame.Linked = settings.Linked end
			frame.AnchorId = settings.AnchorId or frame.AnchorId
			frame.LinkSpace = settings.LinkSpace or frame.LinkSpace
			frame.LinkPoint = settings.LinkPoint or frame.LinkPoint				
		end
	else
		self:OnFirstRun()
	end

	self:Start()
end
function RBT:Start()
	-- Initial Frame Updates based on either default values or loaded settings
	for i, frame in ipairs(self.Frames) do
		frame:Initialize()
		frame:SetShowMode() -- foreground
		frame:Update()
		
		if frame.Linked then
			if frame.AnchorId == "NONE" then
				local Settings = CCDCharacterSettings.RBT
				-- this means we're loading settings from a version before 3.1.5
				local oldAnchor = self.Frames[Settings.Frames[i].AnchorGroup]
				local oldLinkPoint = Settings.Frames[i].LinkPoint
				local oldLinkSpacing = Settings.Frames[i].Spacing
				frame:Link(oldAnchor, oldLinkPoint, oldLinkSpacing)
			else
				-- v3.1.5+
				if CCD.UI.LinkableFramesById[frame.AnchorId] then
					frame:Link(CCD.UI.LinkableFramesById[frame.AnchorId], frame.LinkPoint, frame.LinkSpace)
				else -- v3.2.2+ - some frames were linked to the now created-on-demand cooldown bars 2+
					local x, y = frame.X, frame.Y
					frame.EditBolt:ClearAll()
					frame.EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", x, y)
				end
			end
		else
			local x, y = frame.X, frame.Y
			frame.EditBolt:ClearAll()
			frame.EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", x, y)
		end
	end
	
	-- message frame
	self.MessageFrame.Text:SetText("RBT - Support Message", self.MessageFrame.FontSize, self.MessageFrame.FontColor)
	self.MessageFrame:SetBackgroundColor(self.MessageFrame.BackgroundColor.R, self.MessageFrame.BackgroundColor.G, self.MessageFrame.BackgroundColor.B, self.MessageFrame.BackgroundColor.A)
	if self.MessageFrame.Linked then
		self.MessageFrame:Link(CCD.UI.LinkableFramesById[self.MessageFrame.AnchorId], self.MessageFrame.LinkPoint, self.MessageFrame.LinkSpace)
	else
		self.MessageFrame:Unlink(true)
	end
	RBT:UpdateMessageFrame()

	RBT:Set(RBT.On)	
end
function RBT:CreateSaveTable()
	local saveTable = {
		On = RBT.On,
		GroupOrder = RBT.GroupOrder,
		--IsUsingTSV = RBT.IsUsingTSV, -- Removed in 3.3.3
		Frames = {},
		TooltipViewMode = RBT.TooltipViewMode,
	}
	
	for buffGroup, frame in ipairs(RBT.Frames) do
		saveTable.Frames[buffGroup] = {}
		saveTable.Frames[buffGroup].X = frame.X
		saveTable.Frames[buffGroup].Y = frame.Y
		saveTable.Frames[buffGroup].Linked = frame.Linked
		saveTable.Frames[buffGroup].LinkPoint = frame.LinkPoint
		saveTable.Frames[buffGroup].AnchorId = frame.AnchorId
		saveTable.Frames[buffGroup].LinkSpace = frame.LinkSpace
		--saveTable.Frames[buffGroup].IsTracked = frame.IsTracked -- Removed in 3.3.3
		--saveTable.Frames[buffGroup].TSV = frame.TimerShowValue -- Removed in 3.3.3
		--saveTable.Frames[buffGroup].TimerShowValueColor = frame.TimerShowValueColor -- Removed in 3.3.3
		--saveTable.Frames[buffGroup].IconScale = frame.IconScale -- Removed in 3.4.0
		saveTable.Frames[buffGroup].Scale = frame.IconSize
		--saveTable.Frames[buffGroup].TimerScale = frame.TimerScale -- Removed in 3.4.0
		saveTable.Frames[buffGroup].TimerTextFontSize = frame.TimerTextFontSize
		saveTable.Frames[buffGroup].TimerTextFontSizeLocked = frame.TimerTextFontSizeLocked
		saveTable.Frames[buffGroup].TimerColour = frame.TimerColour
		saveTable.Frames[buffGroup].CombatShow = frame.CombatShow
		saveTable.Frames[buffGroup].GroupShowValue = frame.GroupShowValue
		saveTable.Frames[buffGroup].Alpha = frame.Alpha
		--saveTable.Frames[buffGroup].AlwaysShow = frame.AlwaysShow -- Added in 3.3.1; Removed in 3.3.3
		saveTable.Frames[buffGroup].ShowMode = frame.ShowMode
		saveTable.Frames[buffGroup].ShowModeMissingTimerValue = frame.ShowModeMissingTimerValue
		saveTable.Frames[buffGroup].ShowModeActiveTimerValue = frame.ShowModeActiveTimerValue
		saveTable.Frames[buffGroup].ShowModeAlwaysTimerValue = frame.ShowModeAlwaysTimerValue
		saveTable.Frames[buffGroup].ShowModeAlwaysTimerColor = frame.ShowModeAlwaysTimerColor
	end

	saveTable.MessageFrame = {}
	saveTable.MessageFrame.X = RBT.MessageFrame.X
	saveTable.MessageFrame.Y = RBT.MessageFrame.Y
	saveTable.MessageFrame.On = RBT.MessageFrame.On
	saveTable.MessageFrame.FontColor = RBT.MessageFrame.FontColor
	saveTable.MessageFrame.FontSize = RBT.MessageFrame.FontSize
	saveTable.MessageFrame.BackgroundColor = RBT.MessageFrame.BackgroundColor
	saveTable.MessageFrame.GroupShowValue = RBT.MessageFrame.GroupShowValue

	-- Linking
	saveTable.MessageFrame.Linked = RBT.MessageFrame.Linked
	saveTable.MessageFrame.AnchorId = RBT.MessageFrame.AnchorId
	saveTable.MessageFrame.LinkSpace = RBT.MessageFrame.LinkSpace
	saveTable.MessageFrame.LinkPoint = RBT.MessageFrame.LinkPoint
	
	return saveTable
end
function RBT:Set(NewState)
	self.On = NewState

	if self.On then
		RBT:UpdateAllFrames()
	elseif not CCD.UI.EditModeOn then
		for _, frame in pairs(RBT.Frames) do
			frame:Hide()
		end
		self.MessageFrame:Hide()
	end
	
	--[[ Mobility
	for _, frame in pairs(RBT.Frames) do
		if RBT.On then
			if frame.AlertShowMode ~= AlertShowMode.Never then
				frame.IsMobile = true
			end
		else
			frame.IsMobile = false
		end
	end	--]]
end

-- Event Handlers
function RBT:OnBuffAdd(Ally, BuffId, BuffType, NewBuff) -- not sure how this function handles stacking buffs
	if Ally ~= CCD.Me then return end -- Only handling buffs for Player atm
	if not RBT.Buffs[BuffType] then return end
	
	NewBuff = NewBuff or Buff:Create(Ally.Id, BuffId)
	
	-- annoying bug
	if NewBuff.Type ~= BuffType then
		print(string.format("%s has a .Type of %s vs given %s", NewBuff.Name, NewBuff.Type, BuffType))
		CCD.Debug:Record("%s has a .Type of %s vs given %s", NewBuff.Name, NewBuff.Type, BuffType)
		return -- comment this to get error
	end
	
	Ally:AddBuff(NewBuff)
	local buffGroup = RBT.Buffs[NewBuff.Type][1]
	
	-- if buff has been generated by a rune
	if buffGroup == 21 then
		-- search ally's buffs
		for _, buff in pairs(Ally.Buffs) do
			-- for a buff of the same group as the one just added
			if RBT.Buffs[NewBuff.Type][1] == buffGroup then
				-- if another buff exists then remove flag
				self.Frames[buffGroup].GroupJustRemoved = false
			end
		end
	end
	
	local caster = Allies:Get(NewBuff.CasterId)
	if caster then -- needed because buff may be read before ally especially on loading ui/logging in
		-- check for archon aura
		if RBT:IsArchonBuff(NewBuff.Type) then
			-- if new archon then recheck NewBuffs
			RBT:AddArchon(caster)
		-- check for Living Energy
		elseif RBT:IsLivingEnergy(NewBuff) then
			RBT:AddLEProvider(caster)
		end
	end
	RBT:SetBuffStatus(NewBuff, true)
end
function RBT:OnBuffRemove(Ally, Buff) -- Triggered after Ally:RemoveBuff()
	if not Buff.Type then return end -- all relevant buffs will have a type; rune buffs have their type set in CCD.lua:OnBuffAdd()

	local removeBuff = true
	
	--[[check for weapon enhancement because rogues and warriors can have 2 of the same
	if RBT.Buffs[Buff.Type] and RBT.Buffs[Buff.Type][1] == 21 then
		for _, buff in pairs(Ally.Buffs) do
			if removeBuff and buff.Type == Buff.Type then -- check if another buff has the same type i.e. only one weapon enhancement fell off
				removeBuff = false
			end
		end
	end]]

	if removeBuff then
		if Ally == CCD.Me then			
			if RBT.Buffs[Buff.Type] then
				-- if buff is an enchant
				if RBT.Buffs[Buff.Type][1] == 21 then
					-- search buffs
					for _, buff in pairs(Ally.Buffs) do
						-- for a buff belonging to the same group as the one that was removed
						if RBT.Buffs[Buff.Type][1] == RBT.Buffs[Buff.Type][1] then
							-- if found, flag that a buff has just been removed
							self.Frames[RBT.Buffs[Buff.Type][1]].GroupJustRemoved = true
						end
					end
				end
				
				RBT:SetBuffStatus(Buff, false)
			end
		end
	end
end
function RBT:OnAllyCountChanged()
	RBT:UpdateAllFrames()
end
function RBT:OnRoleCastCompleted(Ally)
	if Ally.IsArchon then
		RBT:RemoveArchon(Ally)
	end
	
	if Ally.CanProvideLivingEnergy then
		RBT:RemoveLEProvider(Ally)
	end
end
function RBT:OnPlayerFullyAvailable()
	RBT:SetMyEditModeTextures()
end
function RBT:OnUnitChanged()
	if CCD.Me.Target and CCD.Me.OldTarget and CCD.Me.Target.Id == CCD.Me.OldTarget.Id and CCD.Me.Target.IsHostile then return end
	if CCD.Me.TargetOfTarget  and CCD.Me.OldTargetOfTarget and CCD.Me.TargetOfTarget.Id == CCD.Me.OldTargetOfTarget.Id and CCD.Me.TargetOfTarget.IsHostile and (CCD.Me.Target and not CCD.Me.Target.IsHostile) then return end

	-- delete all current timers
	for _, frame in ipairs(RBT.DebuffFrames) do
		if frame.TimerTag then
			Library.LibCron.delete(frame.TimerTag)
			frame.TimerTag = nil
		end
	end
	
	RBT:UpdateDebuffFrames() -- necessary
	
	local target = CCD.Me.Target
	local isHostileTarget = false
	
	-- determine if either my target or my target's target is hostile
	if CCD.Me.Target then
		if CCD.Me.Target.IsHostile then
			isHostileTarget = true
		elseif CCD.Me.TargetOfTarget and CCD.Me.TargetOfTarget.IsHostile then
			isHostileTarget = true
			target = CCD.Me.TargetOfTarget
		end
	end
	
	if not isHostileTarget then return end
	
	for _, buff in pairs(target.RaidDebuffs) do
		RBT:SetBuffStatus(buff, true)
	end
end
function RBT:OnEditModeEnd()
	if not RBT.On then -- hide all frames
		for _, frame in ipairs(RBT.Frames) do
			frame:Hide()
		end
	end
end
function RBT:OnAbilityAdded(AbilityId)
	if AbilityId == "A6660BA5F8F8CB707" then -- Vitality of Stone
		RBT:AddArchon(CCD.Me)
		CCD.Messaging.BroadcastArchonStatus()
	end
end
function RBT:OnEnterPartialAvailability(Ally)
	if Ally.IsArchon then
		RBT:RemoveArchon(Ally)
	end
	
	if Ally.CanProvideLivingEnergy then
		RBT:RemoveLEProvider(Ally)
	end
	
	RBT:CheckRoles()
end
function RBT:OnEnterNoneAvailability(Ally)
	if Ally.IsArchon then
		RBT:RemoveArchon(Ally)
	end
	
	if Ally.CanProvideLivingEnergy then
		RBT:RemoveLEProvider(Ally)
	end
	
	RBT:CheckRoles()
end
function RBT:OnGroupLeave(Ally)
	if Ally.IsArchon then
		RBT:RemoveArchon(Ally)
	end
	
	if Ally.CanProvideLivingEnergy then
		RBT:RemoveLEProvider(Ally)
	end
	
	RBT:CheckRoles()
end
function RBT:OnRoleFlagChanged(Ally)
	if Ally.Role == false then -- empty role
		if Ally.IsArchon then
			RBT:RemoveArchon(Ally)
		end
		
		if Ally.CanProvideLivingEnergy then
			RBT:RemoveLEProvider(Ally)
		end
	end
end
function RBT:OnOffline(Ally)
	if Ally.IsArchon then
		RBT:RemoveArchon(Ally)
	end
	
	if Ally.CanProvideLivingEnergy then
		RBT:RemoveLEProvider(Ally)
	end
	
	RBT:CheckRoles()
end

-- Frame Creation and Linking
function RBT:CreateFrame(Group)
	local id = string.format("%s.Frame%d", RBT.Id, Group)
	local frame = CCD.UI.AlertFrame.Create(id, CCD.UI.Context)
	
	frame.Group = Group
	frame.GroupName = RBT.GroupStrings[frame.Group].Title
	--frame.IsTracked = true -- Removed in 3.3.3
	--frame.TimerShowValue = 15 -- Removed in 3.3.3
	--frame.TimerShowValueColor = Color.Red() -- Removed in 3.3.3
	frame.TimerColour = Color.White()
	frame.ShowMode = AlertShowMode.Missing
	frame.ShowModeMissingTimerValue = 15 -- in Missing mode, the alert is shown when the buff has X or less number of seconds left; X = 0 disables this feature
	frame.ShowModeActiveTimerValue = 0 -- in Active mode, the alert is ONLY shown when the buff has X or less number of seconds left; X = 0 disables this feature
	frame.ShowModeAlwaysTimerValue = 15 -- in Always mode, sets X for ShowModeAlwaysTimerColor below
	frame.ShowModeAlwaysTimerColor = Color.Red() -- in Always mode, the timer color changes to this when the buff has X seconds or less left; x = 0 disables this feature

	frame:HideHeader()
	frame:HideFooter()
	
	function frame:OnIconSizeChanged()		
		if RBT.ContextMenu.Group == frame.Group then
			RBT.ContextMenu.IconSizeSlider:SetPosition(self.IconSize, true)
		end
	end
	function frame:OnTimerTextFontSizeChanged()
		if RBT.ContextMenu.Group == frame.Group then
			RBT.ContextMenu.TimerTextSizeSlider:SetPosition(self.TimerTextFontSize, true)
		end
	end
	function frame:SetTimerColour(R, G, B, A)
		self.TimerColour.R = R
		self.TimerColour.G = G
		self.TimerColour.B = B
		self.TimerColour.A = A
		self.TimerText:SetFontColor(R, G, B, A)		
	end
	function frame:SetCombatShow(Status)
		self.CombatShow = Status
		self:UpdateForeground()
		self:Update()
	end
	function frame:UpdateForeground()
		if frame.ShowMode ~= AlertShowMode.Never then
			if self.CombatShow then
				self:SetForegroundColor(CCD.UI.Colors.LightRedTint)
			else
				self:SetForegroundColor(CCD.UI.Colors.ForegroundTint)
			end
		else
			self:SetForegroundColor(CCD.UI.Colors.ForegroundBlackTint)
		end
	end
	function frame:SetShowMode(Mode)
		frame.ShowMode = Mode or frame.ShowMode
		frame:UpdateForeground()
		
		if CCD.UI.EditModeOn then
			local timerValue = frame:GetShowModeTimerValue()
		
			if timerValue and timerValue > 0 then
				self.TimerText:SetText(string.format("%d", timerValue), nil, self.TimerColour)			
				self.TimerText:Show()
			else
				self.TimerText:Hide()
			end
		end
		
		--[[if frame.ShowMode == AlertShowMode.Never then
			frame.IsMobile = false
		elseif RBT.On then
			frame.IsMobile = true
		end--]]
	end
	function frame:SetShowModeTimerValue(Value)
		if frame.ShowMode == AlertShowMode.Never then -- nothing
		elseif frame.ShowMode == AlertShowMode.Always then frame.ShowModeAlwaysTimerValue = Value or frame.ShowModeAlwaysTimerValue
		elseif frame.ShowMode == AlertShowMode.Missing then frame.ShowModeMissingTimerValue = Value or frame.ShowModeMissingTimerValue
		elseif frame.ShowMode == AlertShowMode.Active then frame.ShowModeActiveTimerValue = Value or frame.ShowModeActiveTimerValue
		end
		
		if not CCD.UI.EditModeOn then return end
		
		local timerValue = frame:GetShowModeTimerValue()
		
		if timerValue and timerValue > 0 then
			self.TimerText:SetText(string.format("%d", timerValue), nil, self.TimerColour)			
			self.TimerText:Show()
		else
			self.TimerText:Hide()
		end
	end
	function frame:GetShowModeTimerValue()
		if frame.ShowMode == AlertShowMode.Never then return nil
		elseif frame.ShowMode == AlertShowMode.Always then return frame.ShowModeAlwaysTimerValue
		elseif frame.ShowMode == AlertShowMode.Missing then return frame.ShowModeMissingTimerValue
		elseif frame.ShowMode == AlertShowMode.Active then return frame.ShowModeActiveTimerValue
		end
	end
	--
	function frame:OnEditModeBegin()
		self.Texture:SetTexture("Rift", RBT.EditModeTextures[self.Group])
		
		local timerValue = self:GetShowModeTimerValue()
		
		if timerValue and timerValue > 0 then
			self.TimerText:SetText(string.format("%d", timerValue), nil, self.TimerColour)			
			self.TimerText:Show()
		else
			self.TimerText:Hide()
		end
	end
	function frame:OnEditModeEnd()
		frame:Update()
	end
	function frame:OnGroupShowValueChanged()
		frame:Update()
	end
	-- Events
	function frame.Event:MouseIn()
		if not CCD.UI.EditModeOn and
			(RBT.TooltipViewMode == TooltipViewMode.Off
				or (RBT.TooltipViewMode == TooltipViewMode.CombatOnly and not CCD.Me.InCombat)
				or (RBT.TooltipViewMode == TooltipViewMode.PassiveOnly and CCD.Me.InCombat)) then return end
		RBT.Tooltip:ClearAll()		
		RBT.Tooltip:SetText(frame.GroupName, table.concat(RBT.GroupStrings[frame.Group].BuffNames, "\n"))
		
		local tooltipPoint
		local parentPoint
		local yOffset
		
		if self:GetTop() - (RBT.Tooltip:GetHeight() + 5) > UIParent:GetTop() then
			tooltipPoint = "BOTTOM"
			parentPoint = "TOP"
			yOffset = -5
		else
			tooltipPoint = "TOP"
			parentPoint = "BOTTOM"
			yOffset = 0
		end
		
		local centerX = self:GetLeft() + (self:GetWidth() / 2)
		local halfTooltip = (RBT.Tooltip:GetWidth() / 2)
		
		if centerX - halfTooltip < UIParent:GetLeft() then
			tooltipPoint = tooltipPoint .. "LEFT"
			parentPoint = parentPoint .. "LEFT"
		elseif centerX + halfTooltip > UIParent:GetRight() then
			tooltipPoint = tooltipPoint .. "RIGHT"
			parentPoint = parentPoint .. "RIGHT"
		else
			tooltipPoint = tooltipPoint .. "CENTER"
			parentPoint = parentPoint .. "CENTER"
		end
		
		RBT.Tooltip:SetPoint(tooltipPoint, self, parentPoint, 0, yOffset)
		RBT.Tooltip:SetVisible(true)
	end
	function frame.Event:MouseOut()
		if not CCD.UI.EditModeOn and
			(RBT.TooltipViewMode == TooltipViewMode.Off
				or (RBT.TooltipViewMode == TooltipViewMode.Combat and not CCD.Me.InCombat)) then return end
		RBT.Tooltip:SetVisible(false)
	end

	-- Edit Bolt
	frame.EditBolt:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	function frame.EditBolt.Event:RightClick()
		if RBT.ContextMenu.Group == frame.Group then
			RBT.ContextMenu:Hide()
			RBT.ContextMenu.Group = -1
		else
			RBT.ContextMenu.Group = frame.Group
			RBT.ContextMenu:Show(frame)
		end
	end
	
	return frame
end
function RBT:CreateMessageFrame()
	-- base
	local frame = CCD.UI.CreateLinkableFrame(self.Id .. ".MessageFrame", CCD.UI.Context)
	function frame.OnEditModeEnd()
		RBT:UpdateMessageFrame()
	end
	function frame:OnGroupShowValueChanged()
		if not CCD.UI.EditModeOn then
			frame:ShowX(RBT.MessageFrame.On)
		end
	end
	function frame:SetFontSize(Value)
		frame.FontSize = Value or frame.FontSize or frame.Text:GetFontSize()
		frame.FontSize = CCD.Math.MinMax(frame.FontSize, 15, 100)
		frame.Text:SetFontSize(frame.FontSize)
		frame.ContextMenu.FontSizeSlider:SetPosition(frame.FontSize, true)
	end
	function frame.Event:WheelForward()
		if not CCD.UI.EditModeOn then return end
		frame:SetFontSize(self.FontSize + 5)
	end
	function frame.Event:WheelBack()
		if not CCD.UI.EditModeOn then return end
		frame:SetFontSize(self.FontSize - 5)
	end
	
	-- Text
	frame.Text = CCD.UI.CreateText(self.Id .. ".MessageFrame.Text", frame)
	frame.Text:SetEffectGlow(CCD.UI.FrameTextEffectGlow)
	frame.Text:SetPoint("CENTER", frame, "CENTER")
	function frame.Text.Event:Size()
		frame:SetWidth(self:GetWidth())
		frame:SetHeight(self:GetHeight())
	end
	
	-- Mouse Event Handlers
	function frame.EditBolt.Event:RightClick()
		frame.ContextMenu:Toggle(frame)
	end

	-- Context Menu
	local description = self.Id .. ".MessageFrame.ContextMenu"
	frame.ContextMenu = CCD.UI.CreateContextMenu(RBT.Id .. ".MessageFrame.ContextMenu", 320, 420, "Support Message")

	frame.ContextMenu.ShowFrameCheckbox = frame.ContextMenu:AddCheckbox("ShowFrameCheckbox", "Enable Message", false)
	function frame.ContextMenu.ShowFrameCheckbox.Event:CheckboxChange()
		frame.On = self:GetChecked()
	end

	frame.ContextMenu.GSVTextField = frame.ContextMenu:AddTextField(frame.ContextMenu:GetName() .. ".GSVTextField", "Show in group of ", "or more", true)
	frame.ContextMenu.GSVTextField:SetTextFieldChangedEvent(function ()
		local group = frame.Group
		local value = frame.ContextMenu.GSVTextField:GetConvertedText()

		if value then
			frame:SetGroupShowValue(value)
		end
	end, frame.ContextMenu.GSVTextField:GetName() .. ".OnTextFieldChanged")
	
	frame.ContextMenu.FontSizeSlider = frame.ContextMenu:AddSlider("FontSizeSlider", "Font Size:", 15, 100,  true)
	function frame.ContextMenu.FontSizeSlider.Slider.Event:SliderChange()
		local fontSize = self:GetPosition()
		frame:SetFontSize(fontSize)
	end

	frame.ContextMenu.FontColorPicker = frame.ContextMenu:AddColourPicker("FontColorPicker", "Font Color", false, true)
	frame.ContextMenu.FontColorPicker.OnColorChange = function(R, G, B)
		frame.Text:SetFontColor(R, G, B)
		frame.FontColor = Color:Create(R, G, B)
	end

	frame.ContextMenu.BackgroundColorPicker = frame.ContextMenu:AddColourPicker("BackgroundColorPicker", "Background Color", true, false)
	frame.ContextMenu.BackgroundColorPicker.OnColorChange = function(R, G, B, A)
		frame:SetBackgroundColor(R, G, B, A)
		frame.BackgroundColor = Color:Create(R, G, B, A)
	end

	function frame.ContextMenu:Load()
		frame.ContextMenu.ShowFrameCheckbox:SetChecked(frame.On)
		frame.ContextMenu.FontColorPicker:SetColor(frame.FontColor)
		frame.ContextMenu.FontSizeSlider:SetPosition(frame.FontSize)
		frame.ContextMenu.BackgroundColorPicker:SetColor(frame.BackgroundColor)
		frame.ContextMenu.GSVTextField:SetTextField(frame.GroupShowValue)
	end

	return frame
end
function RBT:LinkFrames(Point, Spacing)
	local lastFrame = self.Frames[self.GroupOrder[1]]

	for i, groupIndex in ipairs(self.GroupOrder) do
		if i ~= 1 then
			local frame = self.Frames[groupIndex]

			frame:Link(lastFrame, Point, Spacing)
			lastFrame = frame
		end
	end
end
function RBT:UnlinkFrames()
	for i, frame in ipairs(self.Frames) do
		frame:Unlink(true)
	end
end
function RBT:ResetFramePositions()
	self.Frames[1].EditBolt:ClearAll()
	self.Frames[1].EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", 100, 300)
	RBT:UseDefaultGroupOrder()
	self:LinkFrames("TOPRIGHT", 5)
	self.MessageFrame:Link(self.Frames[1], "BOTTOMLEFT", 5)
end
function RBT:CreateTooltip()
	-- Container
	local container = CCD.UI.CreateFrame("RBT.Tooltip", CCD.UI.Context)
	container:SetLayer(2)
	container:Hide()
	Library.LibSimpleWidgets.SetBorder("plain", container, 1, 0.5, 0.5, 0.5, 1, "tblr")
	function container:SetText(Title, BodyText)
		container.Title:SetText(Title)
		container.BodyText:SetText(BodyText)
		local width = math.max(container.Title:GetWidth(), container.BodyText:GetWidth())
		local height = container.Title:GetHeight() + container.BodyText:GetHeight() + 1
		container:SetWidth(width)
		container:SetHeight(height)
		container.Background:SetWidth(width)
		container.Background:SetHeight(height)
	end
	
	-- Background
	container.Background = CCD.UI.CreateFrame("RBT.Tooltip.Background", container)
	container.Background:SetBackgroundColor(Color.Black(0.75))
	container.Background:SetPoint("TOPLEFT", container, "TOPLEFT")
	
	--Title
	container.Title = CCD.UI.CreateText("RBT.Tooltip.Title", container)
	container.Title:SetText("Title")
	container.Title:SetFontColor(Color.Orange())
	container.Title:SetLayer(1)
	container.Title:SetPoint("TOPCENTER", container, "TOPCENTER")
	
	-- Body Text
	container.BodyText = CCD.UI.CreateText("RBT.Tooltip.BodyText", container)
	container.BodyText:SetText("Line1\nLine2\nLine3")
	container.BodyText:SetFontColor(Color.White())
	container.BodyText:SetLayer(1)
	container.BodyText:SetPoint("TOPLEFT", container, "TOPLEFT", 0, container.Title:GetHeight())
	
	return container
end

-- Frame Timers
function RBT:SetBuffStatus(Buff, IsActive)
	local group = self.CombinedBuffs[Buff.Type][1]	
	local frame = self.Frames[group]
	local class = self.CombinedBuffs[Buff.Type][2] -- unused atm but will be needed when textures are based on class
	
	frame.Texture:SetTexture("Rift", Buff.Icon)
	
	if IsActive then
		frame.Buff = Buff
		if frame.TimerTag then
			Library.LibCron.delete(frame.TimerTag)
		end
		frame.TimerTag = Library.LibCron.new(TOC.identifier, 0.1, math.huge, false, RBT.RaidBuffTimer_Task, self, Buff, Inspect.Time.Frame())
				
		RBT:UpdateAlertTimer(frame, Buff, Inspect.Time.Frame())
	else
		frame.Buff = nil				
		Library.LibCron.delete(frame.TimerTag) -- it may be possible for frame.TimerTag to be nil here since Deframe.TimerTag can be nil; not sure how
		frame.TimerTag = nil
		frame:Update()
	end
end
function RBT:RaidBuffTimer_Task(Buff, MyStartTime)
	--print("Running timer for " .. Buff.Name)
	if not RBT.On or CCD.UI.EditModeOn then return end
	if Buff.Target == CCD.Me.Id and not CCD.Me.Buffs[Buff.Id] then return end -- check if this actually happens
	local supportClass = RBT.CombinedBuffs[Buff.Type][2]
	if Buff.Target == CCD.Me.Id and not RBT:IsClassSupportPresent(supportClass) then return end
	local group = RBT.CombinedBuffs[Buff.Type][1]
	-- may need to add a check for if alert is disabled here

	local frame = RBT.Frames[group]
	
	frame.Texture:SetTexture("Rift", Buff.Icon) -- set texture regardless

	RBT:UpdateAlertTimer(frame, Buff, MyStartTime)
end
function RBT:UpdateAlertTimer(frame, Buff, MyStartTime)
	local asm = frame.ShowMode
	local remaining
	
	if Buff.TargetId == CCD.Me.Id then
		--remaining = Buff.Remaining - (Inspect.Time.Frame() - Buff.StartTime + 1)
		if not Buff.StartTime then
			remaining = math.huge
		else
			local timeElapsed = Inspect.Time.Frame() - Buff.StartTime
			remaining = math.floor(Buff.Duration - timeElapsed)
		end
	else
		if not Buff.StartTime then
			remaining = math.huge
			--CCD.Modules.ErrorManager:AddError("RBT:UpdateAlertTimer()", Buff.Name .. " is a debuff with a nil .StartTime")
		else
			local timeElapsed = Inspect.Time.Frame() - Buff.StartTime
			remaining = math.floor(Buff.Duration - timeElapsed)
		end
		--remaining = math.floor(Inspect.Buff.Detail(Buff.TargetId, Buff.Id).remaining) --Buff.TargBuff.Remaining - (Inspect.Time.Frame() - MyStartTime +1)
	end
	
	if remaining < 0 then return end
	
	if asm == AlertShowMode.Never then
		frame:Hide()
	elseif asm == AlertShowMode.Always then
		if remaining < math.huge then
			if frame.ShowModeAlwaysTimerValue > 0 and remaining <= frame.ShowModeAlwaysTimerValue then
				frame.TimerText:SetFontColor(frame.ShowModeAlwaysTimerColor)
			else
				frame.TimerText:SetFontColor(frame.TimerColour)
			end
			frame.TimerText:SetText(string.format("%d", remaining))			
		else
			frame.TimerText:SetText("A")		
		end
		frame.TimerText:SetVisible(true)
		frame:ShowX()
	elseif asm == AlertShowMode.Missing then
		frame.TimerText:SetFontColor(frame.TimerColour)
		if remaining < math.huge then		
			if frame.ShowModeMissingTimerValue > 0 and remaining <= frame.ShowModeMissingTimerValue then
				frame.TimerText:SetText(string.format("%d", remaining))
				frame.TimerText:SetVisible(true)
				frame:ShowX()
			else
				frame:Hide()
			end
		else
			frame.TimerText:SetText("A")	
			frame.TimerText:Show()
			frame:Hide()
		end
	elseif asm == AlertShowMode.Active then
		if remaining < math.huge then
			if (frame.ShowModeActiveTimerValue > 0 and remaining <= frame.ShowModeActiveTimerValue) or frame.ShowModeActiveTimerValue == 0 then
				frame.TimerText:SetText(string.format("%d", remaining))
				frame.TimerText:SetVisible(true)
				frame:ShowX()
			else
				frame:Hide()
			end
		else
			frame.TimerText:SetText("A")	
			frame.TimerText:Show()
			frame:ShowX()
		end
	end	
end

-- Frame Update
function RBT:UpdateAllFrames() -- Updates both Buff and Debuff frames as well as the Support Message frame
	if not RBT.On then return end
	if CCD.UI.EditModeOn then return end
	
	for _, frame in pairs(RBT.Frames) do
		frame:Update()
	end
	
	RBT:UpdateMessageFrame()
end
function RBT:UpdateDebuffFrames()
	if not RBT.On then return end
	if CCD.UI.EditModeOn then return end
	
	for _, debuffFrame in ipairs(RBT.DebuffFrames) do
		debuffFrame:Update()
	end
end
function RBT:UpdateMessageFrame()	
	-- Mobility
	if RBT.On and RBT.MessageFrame.On then
		RBT.MessageFrame.IsMobile = true
	else
		RBT.MessageFrame.IsMobile = false
	end

	if CCD.UI.EditModeOn then return end
	
	if not RBT.On or not RBT.MessageFrame.On then
		RBT.MessageFrame:Hide()
		return
	end
	
	local currentText = RBT.MessageFrame.Text:GetText()
	local newText

	if not RBT.SupportPresent.Bard and not RBT.SupportPresent.Oracle and not RBT.SupportPresent.Archon and not RBT.SupportPresent.Beastmaster then
		newText = "No Support Found"
	elseif (RBT.SupportPresent.Bard or RBT.SupportPresent.Oracle) and (not RBT.SupportPresent.Archon and not RBT.SupportPresent.Beastmaster) then
		newText = "No Archon/Beastmaster Found"
	elseif (RBT.SupportPresent.Archon or RBT.SupportPresent.Beastmaster) and (not RBT.SupportPresent.Bard and not RBT.SupportPresent.Oracle) then
		newText = "No Bard/Oracle Found"
	end

	if newText then
		if newText ~= currentText then RBT.MessageFrame.Text:SetText(newText) end
		
		RBT.MessageFrame:ShowX()
	else
		RBT.MessageFrame:Hide()
	end
end
function RBT:HideAllDebuffFrames()
	for _, debuffFrame in ipairs(RBT.DebuffFrames) do
		debuffFrame:Hide()
	end
end

function RBT:SetGlobalIconScale(Scale)
	for _, frame in ipairs(self.Frames) do
		frame:SetIconSize(Scale)
	end
end
function RBT:SetGlobalTimerScale(Scale)
	for _, frame in ipairs(self.Frames) do
		frame:SetTimerTextFontSize(Scale)
	end
end
function RBT:SetAllTimerTextColors(R, G, B, A)
	for _, frame in ipairs(self.Frames) do
		frame:SetTimerColour(R, G, B, A)
	end
end
function RBT:UseDefaultGroupOrder()
	for i in ipairs(self.GroupOrder) do
		self.GroupOrder[i] = nil
	end

	for _, groupIndex in ipairs(self.DefaultGroupOrder) do
		table.insert(self.GroupOrder, groupIndex)
	end
end
function RBT:SetGlobalAlertAlpha(Value)
	for _, alert in ipairs(RBT.Frames) do
		alert:SetAlphaFromSlider(Value)
		if RBT.ContextMenu.Group == alert.Group then
			RBT.ContextMenu.AlphaSlider:SetPosition(Value, true)
		end
	end
end
function RBT:SetGlobalShowMode(Mode)
	for _, frame in ipairs(RBT.Frames) do
		frame:SetShowMode(Mode)
		frame:Update()
	end
end
function RBT:SetGlobalASMTimerValue(Value)
	for _, frame in ipairs(RBT.Frames) do
		frame:SetShowModeTimerValue(Value)
	end
end
function RBT:SetGlobalASMAlwaysTimerColor(R, G, B, A)
	for _, frame in ipairs(RBT.Frames) do
		frame.ShowModeAlwaysTimerColor = Color:Create(R, G, B, A)
	end
end

function RBT:CreateContextMenu()
	-- Context Menu
	local frame = CCD.UI.CreateContextMenu(RBT.Id .. ".ContextMenu", 325, 635)
	local menuId = frame:GetName()
	self.ContextMenu = frame
	
	-- Alert Show Mode Radio Buttons
	frame.ShowModeNeverRadioButton = frame:AddRadioButton(menuId .. ".ShowModeNeverRadioButton", "Never", false, false)
	function frame.ShowModeNeverRadioButton.Event:RadioButtonSelect()
		RBT.Frames[frame.Group]:SetShowMode(AlertShowMode.Never)
		frame.ASMTextField:Hide()
		frame.ASMColorPicker:Hide()
	end
	
	frame.ShowModeAlwaysRadioButton = frame:AddRadioButton(menuId .. ".ShowModeAlwaysRadioButton", "Always", true, false)
	function frame.ShowModeAlwaysRadioButton.Event:RadioButtonSelect()
		RBT.Frames[frame.Group]:SetShowMode(AlertShowMode.Always)
		frame.ASMTextField:SetLabel1("Change colour when timer reaches:")
		frame.ASMTextField:SetLabel2("")
		frame.ASMTextField:SetTextField(RBT.Frames[frame.Group]:GetShowModeTimerValue())
		frame.ASMTextField:Show()
		frame.ASMColorPicker:Show()
	end
	
	frame.ShowModeMissingRadioButton = frame:AddRadioButton(menuId .. ".ShowModeMissingRadioButton", "Missing", true, false)
	function frame.ShowModeMissingRadioButton.Event:RadioButtonSelect()
		RBT.Frames[frame.Group]:SetShowMode(AlertShowMode.Missing)
		frame.ASMTextField:SetLabel1("Show when timer reaches:")
		frame.ASMTextField:SetLabel2("")
		frame.ASMTextField:SetTextField(RBT.Frames[frame.Group]:GetShowModeTimerValue())
		frame.ASMTextField:Show()
		frame.ASMColorPicker:Hide()
	end
	
	frame.ShowModeActiveRadioButton = frame:AddRadioButton(menuId .. ".ShowModeActiveRadioButton", "Active", true, false)
	function frame.ShowModeActiveRadioButton.Event:RadioButtonSelect()
		RBT.Frames[frame.Group]:SetShowMode(AlertShowMode.Active)
		frame.ASMTextField:SetLabel1("ONLY show when timer reaches:")
		frame.ASMTextField:SetLabel2("")
		frame.ASMTextField:SetTextField(RBT.Frames[frame.Group]:GetShowModeTimerValue())
		frame.ASMTextField:Show()
		frame.ASMColorPicker:Hide()
	end
	
	-- Alert Show Mode Radio Button Controller
	frame.ShowModeRadioButtonController = Library.LibSimpleWidgets.RadioButtonGroup(menuId .. ".AlertAlwaysShowRadioButtonController")
	frame.ShowModeRadioButtonController:AddRadioButton(frame.ShowModeNeverRadioButton)
	frame.ShowModeRadioButtonController:AddRadioButton(frame.ShowModeAlwaysRadioButton)
	frame.ShowModeRadioButtonController:AddRadioButton(frame.ShowModeMissingRadioButton)
	frame.ShowModeRadioButtonController:AddRadioButton(frame.ShowModeActiveRadioButton)
	
	-- ASM Timer Value Text Field
	frame.ASMTextField = frame:AddTextField(menuId .. ".ASMTextField", "", "", false, frame.ShowModeNeverRadioButton)
	frame.ASMTextField:SetTextFieldChangedEvent(function ()
		local group = frame.Group
		local value = frame.ASMTextField:GetConvertedText()

		if value then
			RBT.Frames[group]:SetShowModeTimerValue(value)
		end
	end, frame.ASMTextField:GetName() .. ".OnTextFieldChanged")
	
	 -- Alert Show Mode Timer Value Colour Picker (Always)
	frame.ASMColorPicker = frame:AddColourPicker("ASMColorPicker", "", false)
	frame.ASMColorPicker.OnColorChange = function(r, g, b, a)
		RBT.Frames[frame.Group].ShowModeAlwaysTimerColor = Color:Create(r, g, b, a)		
	end

	-- Combat Show Checkbox
	frame.CombatShowCheckbox = frame:AddCheckbox("CombatShowCheckbox", "Only show while in combat")
	function frame.CombatShowCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RBT.Frames[frame.Group]:SetCombatShow(checked)
		local radButton = RBT.Tab.AlertCombatShowRadioButtonController:GetSelectedRadioButton()
		if radButton then
			radButton:SetSelected(false, true)
		end
	end

	-- GSV Text Field
	frame.GSVTextField = frame:AddTextField(menuId .. ".GSVTextField", "Only show in groups of", "or more", true)
	function frame.GSVTextField.OnTextFieldChanged()
		local value = frame.GSVTextField:GetConvertedText()

		if value then			
			local selectedFrame = RBT.Frames[frame.Group]
			selectedFrame:SetGroupShowValue(value)
		end
	end
	frame.GSVTextField:SetTextFieldChangedEvent(frame.GSVTextField.OnTextFieldChanged, frame.GSVTextField:GetName() .. ".OnTextFieldChanged")
		
	-- Frame Alpha Slider
	frame.AlphaSlider = frame:AddSlider(RBT.Id .. ".ContextMenu.AlphaSlider", "Opacity:", 0, 100, false)
	function frame.AlphaSlider.Slider.Event:SliderChange()
		local sliderValue = self:GetPosition()		
		RBT.Frames[frame.Group]:SetAlphaFromSlider(sliderValue)
	end

	-- Icon Size Slider
	frame.IconSizeSlider = frame:AddSlider(RBT.Id .. ".ContextMenu.IconSizeSlider", "Icon Size:", CCD.UI.AlertFrame.IconSizeRange.Min, CCD.UI.AlertFrame.IconSizeRange.Max)
	function frame.IconSizeSlider.Slider.Event:SliderChange()
		local scale = self:GetPosition()
		RBT.Frames[frame.Group]:SetIconSize(scale)
	end
	
	-- Timer Text Size Slider
	frame.TimerTextSizeSlider = frame:AddSlider(menuId .. ".TimerTextSizeSlider", "Timer Size:", CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max)
	function frame.TimerTextSizeSlider.Slider.Event:SliderChange()
		local scale = self:GetPosition()
		frame:GetAlert():SetTimerTextFontSize(scale)
	end

	-- Timer Font Size Locked Check Box
	frame.TimerTextFontSizeLockedCheckbox = frame:AddCheckbox(menuId .. ".TimerTextFontSizeLockedCheckbox", "Lock Timer Font Size", true)
	function frame.TimerTextFontSizeLockedCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		frame:GetAlert().TimerTextFontSizeLocked = checked
	end
		
	 -- Colour Picker
	frame.TimerTextColorPicker = frame:AddColourPicker(menuId .. ".TimerTextColorPicker", "Timer Colour:", false, true)
	frame.TimerTextColorPicker.OnColorChange = function(r, g, b, a)
		RBT.Frames[frame.Group]:SetTimerColour(r, g, b, a)
	end

	-- Global Header
	self.ContextMenu.GlobalOptionsLabel = frame:AddHeader("GlobalOptionsLabel", "Global Options")

	frame.OnCheckbox = frame:AddCheckbox("OnCheckbox", "Enable Buff Tracker", true)
	function frame.OnCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		if checked == RBT.On then return end
		RBT:Set(checked)
		RBT.Tab.EnableRBTCheckbox:SetChecked(checked)
	end

	self.ContextMenu.FooterText = frame:AddFooterNote("FooterText", "Click here for more options")
	function self.ContextMenu.FooterText.Event:LeftClick()
		CCD.MainWindow:Show()
		CCD.MainWindow.TabView:SetActiveTab(2)
	end

	-- Load Function
	function frame:Load()
		local alert = frame:GetAlert()
		frame.HeaderText:SetText(alert.GroupName)	
		
		--
		frame.ShowModeRadioButtonController:SetSelectedIndex(alert.ShowMode, false) -- not silent because we need to load UI elements based on selection
		frame.ASMTextField:SetTextField(alert:GetShowModeTimerValue())		
		frame.ASMColorPicker:SetColor(alert.ShowModeAlwaysTimerColor)
		frame.CombatShowCheckbox:SetChecked(alert.CombatShow, true)
		frame.GSVTextField.TextField:SetText(alert.GroupShowValue)
		--
		frame.IconSizeSlider:SetPosition(alert.IconSize, true)		
		frame.TimerTextSizeSlider:SetPosition(alert.TimerTextFontSize, true)
		self.TimerTextFontSizeLockedCheckbox:SetChecked(alert.TimerTextFontSizeLocked)
		frame.AlphaSlider:SetPosition(CCD.Math.Round(alert.Alpha * 100), true)
		--		
		frame.TimerTextColorPicker:SetColor(alert.TimerColour)
		--	
		frame.OnCheckbox:SetChecked(RBT.On)
		
	end
	function frame:GetAlert()
		if frame.Group then return
			RBT.Frames[frame.Group]
		else
			return nil
		end		
	end
end
function RBT:CreateTab(Width)
	-- Base Frame
	local tabName = "RBTTab"
	self.Tab = UI.CreateFrame("Mask", tabName, CCD.MainWindow)
	local tab = self.Tab
	tab:SetBackgroundColor(0, 0, 0, 1) -- black

	-- "Enable RBT" Checkbox
	tab.EnableRBTCheckbox = CCD.UI.CreateCheckbox(tabName .. ".EnableRBTCheckbox", tab, RBT.On)
	tab.EnableRBTCheckbox:SetText("Enable Buff Tracker", 15, CCD.UI.Colors.White)
	tab.EnableRBTCheckbox:SetPoint("TOPLEFT", tab, "TOPLEFT", 10, 10)
	function tab.EnableRBTCheckbox.Event:CheckboxChange()
		RBT:Set(self:GetChecked())
		RBT.ContextMenu.OnCheckbox:SetChecked(self:GetChecked())
	end

	-- "Reset Frames" Button
	tab.ResetFramesButton = UI.CreateFrame("RiftButton", tabName .. ".ResetFramesButton", tab)
	tab.ResetFramesButton:SetText("Reset Frames")
	tab.ResetFramesButton:SetPoint("TOPRIGHT", tab, "TOPRIGHT", 0, 2) -- top-right corner
	tab.ResetFramesButton:SetVisible(true)
	function tab.ResetFramesButton.Event:LeftPress()
		RBT:ResetFramePositions()
	end

	--
	tab.GlobalOptionsNote = CCD.UI.CreateText(tabName .. ".GlobalOptionsNote", tab)
	tab.GlobalOptionsNote:SetText("*Options in this tab will adjust the properties for ALL buff alerts. If you want to edit an individual alert enter Edit Mode and right click that alert's cogwheel.", 15, Color.Yellow())
	tab.GlobalOptionsNote:SetPoint("TOPLEFT", tab.EnableRBTCheckbox, "BOTTOMLEFT", 0, 10)
	tab.GlobalOptionsNote:SetPoint("RIGHT", tab, "RIGHT")
	
	-- ASM Label
	tab.ASMLabel = CCD.UI.CreateText(tabName .. ".ASMLabel", tab)
	tab.ASMLabel:SetText("Alert Show Mode", 15, Color.Green())
	tab.ASMLabel:SetPoint("TOPLEFT", tab.GlobalOptionsNote, "BOTTOMLEFT", 0, 5)
	
	-- Alert Show Mode Radio Buttons
	tab.ShowModeNeverRadioButton = CCD.UI.RadioButton:Create(tabName .. ".ShowModeNeverRadioButton", tab, "Never")
	tab.ShowModeNeverRadioButton:SetSelected(true)
	tab.ShowModeNeverRadioButton:SetPoint("TOPLEFT", tab.ASMLabel, "BOTTOMLEFT", 0, 5)
	function tab.ShowModeNeverRadioButton.Event:RadioButtonSelect()
		RBT:SetGlobalShowMode(AlertShowMode.Never)
		tab.ASMTextField:SetTextField("")
		tab.ASMTextField:Hide()
		tab.ASMColorPicker:Hide()
		tab.ASMColorPickerButton:SetVisible(false)
	end
	
	tab.ShowModeAlwaysRadioButton = CCD.UI.RadioButton:Create(tabName .. ".ShowModeAlwaysRadioButton", tab, "Always")
	tab.ShowModeAlwaysRadioButton:SetPoint("TOPLEFT", tab.ShowModeNeverRadioButton, "TOPRIGHT")
	function tab.ShowModeAlwaysRadioButton.Event:RadioButtonSelect()
		RBT:SetGlobalShowMode(AlertShowMode.Always)
		tab.ASMTextField:SetLabel1("Change colour when timer reaches:")
		tab.ASMTextField:SetLabel2("")
		tab.ASMTextField:SetTextField("")
		tab.ASMTextField:Show()
		tab.ASMColorPicker:Show()
		tab.ASMColorPickerButton:SetVisible(true)
	end
	
	tab.ShowModeMissingRadioButton = CCD.UI.RadioButton:Create(tabName .. ".ShowModeMissingRadioButton", tab, "Missing")
	tab.ShowModeMissingRadioButton:SetPoint("TOPLEFT", tab.ShowModeAlwaysRadioButton, "TOPRIGHT")
	function tab.ShowModeMissingRadioButton.Event:RadioButtonSelect()
		RBT:SetGlobalShowMode(AlertShowMode.Missing)
		tab.ASMTextField:SetLabel1("Show when timer reaches:")
		tab.ASMTextField:SetLabel2("")
		tab.ASMTextField:SetTextField("")
		tab.ASMTextField:Show()
		tab.ASMColorPicker:Hide()
		tab.ASMColorPickerButton:SetVisible(false)
	end
	
	tab.ShowModeActiveRadioButton = CCD.UI.RadioButton:Create(tabName .. ".ShowModeActiveRadioButton", tab, "Active")
	tab.ShowModeActiveRadioButton:SetPoint("TOPLEFT", tab.ShowModeMissingRadioButton, "TOPRIGHT")
	function tab.ShowModeActiveRadioButton.Event:RadioButtonSelect()
		RBT:SetGlobalShowMode(AlertShowMode.Active)
		tab.ASMTextField:SetLabel1("ONLY show when timer reaches:")
		tab.ASMTextField:SetLabel2("")
		tab.ASMTextField:SetTextField("")
		tab.ASMTextField:Show()
		tab.ASMColorPicker:Hide()
		tab.ASMColorPickerButton:SetVisible(false)
	end
	
	-- Alert Show Mode Radio Button Controller
	tab.ShowModeRadioButtonController = Library.LibSimpleWidgets.RadioButtonGroup(tabName .. ".AlertAlwaysShowRadioButtonController")
	tab.ShowModeRadioButtonController:AddRadioButton(tab.ShowModeNeverRadioButton)
	tab.ShowModeRadioButtonController:AddRadioButton(tab.ShowModeAlwaysRadioButton)
	tab.ShowModeRadioButtonController:AddRadioButton(tab.ShowModeMissingRadioButton)
	tab.ShowModeRadioButtonController:AddRadioButton(tab.ShowModeActiveRadioButton)
	
	-- ASM Timer Value Text Field
	tab.ASMTextField = CCD.UI.CreateTextField(tabName .. ".ASMTextField", tab, "", "")
	tab.ASMTextField:SetPoint("TOPLEFT", tab.ShowModeNeverRadioButton, "BOTTOMLEFT")
	tab.ASMTextField:Hide()
	tab.ASMTextField:SetTextFieldChangedEvent(function ()
		local value = tab.ASMTextField.TextField:GetConvertedText()

		if value then
			RBT:SetGlobalASMTimerValue(value)
		end
	end, tab.ASMTextField:GetName() .. ".OnTextFieldChanged*")
	
	-- Always Show Mode Color Picker
	tab.ASMColorPicker = CCD.UI.CreateColorPicker(tabName .. ".ASMColorPicker", tab, "")
	tab.ASMColorPicker:SetPoint("TOPLEFT", tab.ASMTextField, "BOTTOMLEFT", 0, 5)
	tab.ASMColorPicker:Hide()

	-- Set Color Button
	tab.ASMColorPickerButton = UI.CreateFrame("RiftButton", tabName .. ".ASMColorPickerButton", tab)
	tab.ASMColorPickerButton:SetText("Set Color")
	tab.ASMColorPickerButton:SetPoint("TOPLEFT", tab.ASMColorPicker, "TOPRIGHT", 0, 0)
	tab.ASMColorPickerButton:SetVisible(false)
	function tab.ASMColorPickerButton.Event:LeftPress()
		 RBT:SetGlobalASMAlwaysTimerColor(tab.ASMColorPicker:GetRed(), tab.ASMColorPicker:GetGreen(), tab.ASMColorPicker:GetBlue(), tab.ASMColorPicker:GetAlpha())
	end
	
	-- Global Enable Combat Show Radio Button
	tab.AlertEnableCombatShowRadioButton = CCD.UI.RadioButton:Create(tabName .. ".AlertEnableCombatShowRadioButton", tab, "Only show alerts while in combat")
	tab.AlertEnableCombatShowRadioButton:SetPoint("TOPLEFT", tab.ASMColorPicker, "BOTTOMLEFT", 0, 5)
	function tab.AlertEnableCombatShowRadioButton.Event:RadioButtonSelect()
		for _, frame in ipairs(RBT.Frames) do
			frame:SetCombatShow(true)
		end
	end
	
	-- Global Disable Combat Show Radio Button
	tab.AlertDisableCombatShowRadioButton = CCD.UI.RadioButton:Create(tabName .. ".AlertDisableCombatShowRadioButton", tab, "Show alerts regardless of my combat status")
	tab.AlertDisableCombatShowRadioButton:SetPoint("TOPLEFT", tab.AlertEnableCombatShowRadioButton, "TOPRIGHT")
	function tab.AlertDisableCombatShowRadioButton.Event:RadioButtonSelect()
		for _, frame in ipairs(RBT.Frames) do
			frame:SetCombatShow(false)
		end
	end
	
	-- Combat Show Radio Button Controller
	tab.AlertCombatShowRadioButtonController = Library.LibSimpleWidgets.RadioButtonGroup("AlertCombatShowRadioButtonController")
	tab.AlertCombatShowRadioButtonController:AddRadioButton(tab.AlertEnableCombatShowRadioButton)
	tab.AlertCombatShowRadioButtonController:AddRadioButton(tab.AlertDisableCombatShowRadioButton)
				
	-- Frame Scale Slider
	tab.FrameScaleSlider = CCD.UI.CreateSlider(tabName .. ".FrameScaleSlider", tab)
	tab.FrameScaleSlider:SetText("Icon Size")
	tab.FrameScaleSlider:SetRange(CCD.UI.AlertFrame.IconSizeRange.Min, CCD.UI.AlertFrame.IconSizeRange.Max, 48)
	tab.FrameScaleSlider.Label:SetWidth(110)
	tab.FrameScaleSlider:SetPoint("TOPLEFT", tab.AlertEnableCombatShowRadioButton, "BOTTOMLEFT", 0, 10)
	tab.FrameScaleSlider:SetVisible(true)

	-- Frame Scale Button
	tab.FrameScaleButton = UI.CreateFrame("RiftButton", tabName .. ".FrameScaleButton", tab)
	tab.FrameScaleButton:SetText("Set All")
	tab.FrameScaleButton:SetPoint("TOPLEFT", tab.FrameScaleSlider, "TOPRIGHT", 5, -7)
	tab.FrameScaleButton:SetVisible(true)
	function tab.FrameScaleButton.Event:LeftPress()
		RBT:SetGlobalIconScale(tab.FrameScaleSlider:GetPosition())
	end

	-- Timer Font Scale
	tab.TimerFontScaleSlider = CCD.UI.CreateSlider(tabName .. ".TimerFontScaleSlider", tab)
	tab.TimerFontScaleSlider:SetText("Font Size:")
	tab.TimerFontScaleSlider:SetRange(CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max, 24)
	tab.TimerFontScaleSlider.Label:SetWidth(110)
	tab.TimerFontScaleSlider:SetPoint("TOPLEFT", tab.FrameScaleSlider, "BOTTOMLEFT")
	tab.TimerFontScaleSlider:SetVisible(true)

	-- Timer Font Scale Button
	tab.TimerFontScaleButton = UI.CreateFrame("RiftButton", tabName .. ".TimerFontScaleButton", tab)
	tab.TimerFontScaleButton:SetText("Set All")
	tab.TimerFontScaleButton:SetPoint("TOPLEFT", tab.TimerFontScaleSlider, "TOPRIGHT", 5, -7)
	tab.TimerFontScaleButton:SetVisible(true)
	function tab.TimerFontScaleButton.Event:LeftPress()
		RBT:SetGlobalTimerScale(tab.TimerFontScaleSlider:GetPosition())
	end

	-- Global Frame Spacing
	tab.FrameSpacingSlider = CCD.UI.CreateSlider(tabName .. ".FrameSpacingSlider", tab)
	tab.FrameSpacingSlider:SetText("Frame Spacing:")
	tab.FrameSpacingSlider:SetRange(0, 50, 10)
	tab.FrameSpacingSlider.Label:SetWidth(110)
	tab.FrameSpacingSlider:SetPoint("TOPLEFT", tab.TimerFontScaleSlider, "BOTTOMLEFT")
	tab.FrameSpacingSlider:SetVisible(true)
	function tab.FrameSpacingSlider.Slider.Event:SliderChange()
		for _, frame in ipairs(RBT.Frames) do
			frame.NextLinkSpace = self:GetPosition()
		end
		RBT.MessageFrame.NextLinkSpace = self:GetPosition()
	end

	-- Dropdown
	tab.FrameSpacingDropdown = UI.CreateFrame("SimpleSelect", tabName .. ".FrameSpacingDropdown", tab)
	local items = {"Horizontal", "Vertical",}
	local values = {"TOPRIGHT", "BOTTOMLEFT",}
	tab.FrameSpacingDropdown:SetItems(items, values)
	tab.FrameSpacingDropdown:SetSelectedIndex(1)
	tab.FrameSpacingDropdown:ResizeToFit()
	tab.FrameSpacingDropdown:SetPoint("TOPLEFT", tab.FrameSpacingSlider, "TOPRIGHT", 10, 0)
	tab.FrameSpacingDropdown:SetVisible(true)
	function tab.FrameSpacingDropdown.Event:ItemSelect(Item, Value, Index)
		for _, frame in ipairs(RBT.Frames) do
			frame.NextLinkPoint = Value
		end
		RBT.MessageFrame.NextLinkPoint = Value
	end

	-- Link Frames Button
	tab.LinkFramesButton = UI.CreateFrame("RiftButton", tabName .. ".LinkFramesButton", tab)
	tab.LinkFramesButton:SetText("Link Frames")
	tab.LinkFramesButton:SetPoint("TOPLEFT", tab.FrameSpacingDropdown, "TOPRIGHT", 5, -7)
	tab.LinkFramesButton:SetVisible(true)
	function tab.LinkFramesButton.Event:LeftPress()
		RBT:SetOrderByX()
		RBT:LinkFrames(tab.FrameSpacingDropdown:GetSelectedValue(), tab.FrameSpacingSlider:GetPosition())
	end

	-- Unlink Frames Button
	tab.UnlinkFramesButton = UI.CreateFrame("RiftButton", tabName .. "UnlinkFramesButton", tab)
	tab.UnlinkFramesButton:SetText("Unlink Frames")
	tab.UnlinkFramesButton:SetPoint("TOPLEFT", tab.LinkFramesButton, "TOPRIGHT")
	tab.UnlinkFramesButton:SetVisible(true)
	function tab.UnlinkFramesButton.Event:LeftPress()
		RBT:UnlinkFrames()
	end

	-- Alert Alpha Slider
	tab.GlobalAlertAlphaSlider = CCD.UI.CreateSlider(tabName .. ".GlobalAlertAlphaSlider", tab)
	tab.GlobalAlertAlphaSlider.Label:SetText("Opacity:")
	tab.GlobalAlertAlphaSlider.Label:SetWidth(110)
	tab.GlobalAlertAlphaSlider:SetRange(0, 100, 100)
	tab.GlobalAlertAlphaSlider:SetPoint("TOPLEFT", tab.FrameSpacingSlider, "BOTTOMLEFT", 0, 0)
	function tab.GlobalAlertAlphaSlider.Slider.Event:SliderChange()
		local pos = tab.GlobalAlertAlphaSlider.Slider:GetPosition()
		RBT:SetGlobalAlertAlpha(pos)
	end
	
	-- GSV Text Field
	tab.GSVTextField = CCD.UI.CreateTextField(tabName .. ".GSVTextField", tab, "Only show in group of", "or more")
	tab.GSVTextField.TextField:SetValue("valueType", "number")
	tab.GSVTextField:SetPoint("TOPLEFT", tab.GlobalAlertAlphaSlider, "BOTTOMLEFT", 0, 10)
	tab.GSVTextField:SetVisible(true)
	tab.GSVTextField:SetTextFieldChangedEvent(function ()
		local value = tab.GSVTextField.TextField:GetConvertedText()

		if value then
			for _, frame in ipairs(RBT.Frames) do
				frame:SetGroupShowValue(value)
			end
			RBT.MessageFrame:SetGroupShowValue(value)
		end
	end, tab.GSVTextField:GetName() .. ".OnTextFieldChanged")

		
	-- Color Picker
	local colorPicker = CCD.UI.CreateColorPicker("RBTFrameTimerTextColorPicker", tab, "Default Colour:")
	colorPicker:SetPoint("TOPLEFT", tab.GSVTextField, "BOTTOMLEFT", 0, 10)

	-- Set Color Button
	local colorButton = UI.CreateFrame("RiftButton", tabName .. "_colorButton", tab)
	colorButton:SetText("Set Colour")
	colorButton:SetPoint("TOPLEFT", colorPicker, "TOPRIGHT", 0, 0)
	colorButton:SetVisible(true)
	function colorButton.Event:LeftPress()
		RBT:SetAllTimerTextColors(colorPicker:GetRed(), colorPicker:GetGreen(), colorPicker:GetBlue(), colorPicker:GetAlpha())
	end
	
	tab.TooltipViewModeLabel = CCD.UI.CreateText(tabName .. ".TooltipViewModeLabel", tab)
	tab.TooltipViewModeLabel:SetText("Tooltip View Mode", 15, Color.Green())
	tab.TooltipViewModeLabel:SetPoint("TOPLEFT", colorPicker, "BOTTOMLEFT", 0, 5)
	
	tab.TVMOffRadioButton = CCD.UI.RadioButton:Create(tabName .. ".TVMOffRadioButton", tab, "Off")
	tab.TVMOffRadioButton:SetPoint("TOPLEFT", tab.TooltipViewModeLabel, "BOTTOMLEFT", 0, 0)
	function tab.TVMOffRadioButton.Event:RadioButtonSelect()
		RBT.TooltipViewMode = TooltipViewMode.Off
	end
	
	tab.TVMAlwaysRadioButton = CCD.UI.RadioButton:Create(tabName .. ".TVMAlwaysRadioButton", tab, "Always")
	tab.TVMAlwaysRadioButton:SetPoint("TOPLEFT", tab.TVMOffRadioButton, "TOPRIGHT", 0, 0)
	function tab.TVMAlwaysRadioButton.Event:RadioButtonSelect()
		RBT.TooltipViewMode = TooltipViewMode.Always
	end
	
	tab.TVMCombatOnlyRadioButton = CCD.UI.RadioButton:Create(tabName .. ".TVMCombatOnlyRadioButton", tab, "Combat Only")
	tab.TVMCombatOnlyRadioButton:SetPoint("TOPLEFT", tab.TVMAlwaysRadioButton, "TOPRIGHT", 0, 0)
	function tab.TVMCombatOnlyRadioButton.Event:RadioButtonSelect()
		RBT.TooltipViewMode = TooltipViewMode.CombatOnly
	end
	
	tab.TVMPassiveOnlyRadioButton = CCD.UI.RadioButton:Create(tabName .. ".TVMPassiveOnlyRadioButton", tab, "Outside of Combat Only")
	tab.TVMPassiveOnlyRadioButton:SetPoint("TOPLEFT", tab.TVMCombatOnlyRadioButton, "TOPRIGHT", 0, 0)
	function tab.TVMPassiveOnlyRadioButton.Event:RadioButtonSelect()
		RBT.TooltipViewMode = TooltipViewMode.PassiveOnly
	end
	
	tab.TVMRadioButtonController = Library.LibSimpleWidgets.RadioButtonGroup("TVMRadioButtonController")
	tab.TVMRadioButtonController:AddRadioButton(tab.TVMOffRadioButton)
	tab.TVMRadioButtonController:AddRadioButton(tab.TVMAlwaysRadioButton)
	tab.TVMRadioButtonController:AddRadioButton(tab.TVMCombatOnlyRadioButton)
	tab.TVMRadioButtonController:AddRadioButton(tab.TVMPassiveOnlyRadioButton)
	tab.TVMRadioButtonController:SetSelectedIndex(RBT.TooltipViewMode, true)
	
	--[[ Raid Buff Checkboxes
	tab.Group1Checkbox = CCD.UI.CreateCheckbox(tabName .. ".Group1Checkbox", tab, RBT.Frames[1].IsMobile)
	tab.Group1Checkbox:SetText(RBT.Frames[1].GroupName, 18)
	tab.Group1Checkbox:SetPoint("TOPLEFT", colorPicker, "BOTTOMLEFT", 0, 10)
	function tab.Group1Checkbox.Event:CheckboxChange()
		RBT.Frames[1].IsMobile = self:GetChecked()
	end
	--]]
	return self.Tab
end

function RBT:CheckRoles()
	--[[
	This function is called:
	1.	CCD.OnGroupJoin() - Whenever someone joins my group
	2.	CCD.OnRoleCastCompleted_Task() - When a group member finishes their role change cast
	3.	CCD.OnEnterFullAvailability() - When a group member enters full availability
	4.	CCD.OnRoleChange() - When a group member's role flag changes
	5.	RBT:OnEnterPartialAvailability() - When an ally enters partial availability
	6.	RBT:OnEnterNoneAvailability() - When an ally enters no availability
	7.	RBT:OnGroupLeave() - When an ally leaves group
	8.	RBT:OnOffline() - When an ally goes offline
	--]]
	self.SupportPresent.Bard = false
	self.SupportPresent.Oracle = false
	self.SupportPresent.Beastmaster = false

	for _, ally in pairs(Allies.ValuesById) do
		if ally.Availability == "full" then
			if ally.Role == "support" then
				if ally.Class == "rogue" then self.SupportPresent.Bard = true
				elseif ally.Class == "cleric" then self.SupportPresent.Oracle = true
				elseif ally.Class == "warrior" then self.SupportPresent.Beastmaster = true end
			elseif ally.Role == "heal" then
				if ally.Class == "mage" then
					RBT:AddLEProvider(ally) -- not entirely accurate since it is possible for a mage with a heal flag to not have LE
				end
			end
			if ally.IsArchon then self.SupportPresent.Archon = true end
			-- Archon is not flagged here because it cannot be determined from the Role Flag system
		end
	end

	RBT:UpdateAllFrames()
end
function RBT:IsClassSupportPresent(SupportClass)
	if SupportClass == "BARD" then
		return self.SupportPresent.Bard
	elseif SupportClass == "ORACLE" then
		return self.SupportPresent.Oracle
	elseif SupportClass == "ARCHON" then
		return self.SupportPresent.Archon
	elseif SupportClass == "BEASTMASTER" then
		return self.SupportPresent.Beastmaster
	elseif SupportClass == "CHLOROMANCER" then
		return self.SupportPresent.Chloromancer
	elseif SupportClass == "SELF" then
		return true
	end
end

-- Archon
function RBT:AddArchon(Ally)
	Ally.IsArchon = true
	
	if not RBT.SupportPresent.Archon then
		RBT.SupportPresent.Archon = true
		RBT:UpdateAllFrames() -- this can be replaced with a function that only updates archon frames
		--RBT:UpdateMessageFrame()
	end
end
function RBT:RemoveArchon(Ally)
	Ally.IsArchon = false
	
	local found = false
	for _, ally in pairs(Allies.ValuesById) do		
		if ally.IsArchon then
			found = true
		end
	end
	
	if not found then
		RBT.SupportPresent.Archon = false
		RBT:UpdateAllFrames() -- this can be replaced with a function that only updates archon frames
		--RBT:UpdateMessageFrame()
	end	
end
function RBT:IsArchonBuff(BuffType)
	for buffType, values in pairs(self.Buffs) do
		local class = values[2]
		if class == "ARCHON" and buffType == BuffType then
			return true
		end
	end

	return false
end

-- Living Energy
function RBT:AddLEProvider(Ally)
	Ally.CanProvideLivingEnergy = true
	
	if not RBT.SupportPresent.Chloromancer then
		RBT.SupportPresent.Chloromancer = true
		RBT:UpdateAllFrames() -- this can be replaced with a function that only updates the Living Enegy frame
		--RBT:UpdateMessageFrame()
	end
end
function RBT:RemoveLEProvider(Ally)
	Ally.CanProvideLivingEnergy = false
	
	local found = false
	for _, ally in pairs(Allies.ValuesById) do		
		if Ally.CanProvideLivingEnergy then
			found = true
		end
	end
	
	if not found then
		RBT.SupportPresent.Chloromancer = false
		RBT:UpdateAllFrames() -- this can be replaced with a function that only updates the Living Enegy frame
		--RBT:UpdateMessageFrame()
	end	
end

function RBT:IsLivingEnergy(Buff)
	if Buff.Type == self.LivingEnergy then
		return true
	else
		return false
	end
end

function RBT:SetOrderByX()
	local ordered = {}

	for i, frame in ipairs(self.Frames) do
		table.insert(ordered, {i, frame:GetLeft(), frame:GetTop()})
	end

	table.sort(ordered, Compare)

	self.GroupOrder = {}
	
	for i, values in ipairs(ordered) do
		table.insert(self.GroupOrder, values[1])
	end
end
function RBT:SetMyEditModeTextures()
	local resource = CCD.Me:GetResource()

	if resource == "MANA" then
		RBT.EditModeTextures[19] = "medium_potion_spherical_3.dds"
		RBT.EditModeTextures[21] = "fatestone_02_blue.dds"
	else	 -- for any classes other than mage and cleric
		RBT.EditModeTextures[19] = "medium_potion_spherical_1.dds"
		RBT.EditModeTextures[21] = "fatestone_03_violet.dds"
	end
end

--[[
Archons are detected by:
1. Archon Buffs - Any buff (not debuff) tracked by the Buff Tracker that is flagged as ARCHON
2. Vitality of Stone Ability (Self) - If the player has the VoS ability (not buff)
3. CCD Messenger - Any ally that broadcasts an "Archon" message
--]]