local TOC, CCD = ...

-- Globals
CCD.UI = CCD.UI or {}
CCD.Modules = CCD.Modules or {}
CCD.Modules.RCT = CCD.Modules.RCT or {}
CCD.Modules.RCT.CooldownDataCollection = {} -- used in CCD.RaidCooldown
CCD.Modules.RCT.Role = {
	Damage = 1,
	Heal = 2,
	Tank = 3,
}
CCD.MainWindow = CCD.MainWindow or UI.CreateFrame("SimpleWindow", "CCDMainWindow", CCD.UI.Context)
CCD.Allies = CCD.Allies or {}
CCD.Debug = CCD.Debug or {}
CCD.RCDStatus = CCD.RCDStatus or {}

-- Aliases
local RCT = CCD.Modules.RCT
local Colors = CCD.UI.Colors
local Allies = CCD.Allies
local Debug = CCD.Debug
local Status = CCD.RCDStatus
local Color = CCD.UI.Color

-- Locals
local EditModeTextures = {
	"humbling_blow_01.dds", -- FP
	"bard_orchestra_of_the_planes.dds", -- Orchestra
	"verse_of_joy.dds", -- VoJ
	"spiritoftree1a.dds", -- Wild Growth
	"beastmaster_enraged_companion_01_a.dds", -- Enrage
	"tempest_identified_weakness_b.dds", -- Focused Fire
	"defiler_unstable_transformation_a.dds", -- Unstable Transformation
	"castigation3.dds", -- Shield Defense
	"beastmaster_shared_bonds_b.dds", -- Symbiosis
	"paragon_alacricity.dds", -- Fiery will
	"blazeofglory1a.dds", -- Command to Attack
	"blessing_of_the_flames.dds", -- Flame of Life
	"ardent_doctrine_01_a.dds", -- Healer's Covenant
	"ward3.dds", -- Spiritual Conflagration
	"sentinel_wrathful_exuberance.dds", -- Biofeedback
}
local BattleWearyId = "BFD1360782FA647DF" -- Icon: shaman-glory_of_the_chosen.dds
local BattleWearyCheckTimestamp
local BuffFlag = 1
local AbilityFlag = 2
local NextHealAlertIndex = 1
local HealCooldownAlertAnchorTexture = "Images\\Heal Anchor.png"
local DamageCooldownAlertAnchorTexture = "Images\\Damage Anchor.png"

function string.startswith(String, Start1, Start2, Start3)
   if string.sub(String,1,string.len(Start1))==Start1 then return true
   elseif Start2 and string.sub(String,1,string.len(Start2))==Start2 then return true
   elseif Start3 and string.sub(String,1,string.len(Start3))==Start3 then return true
   else return false end
end

-- Classes
local CooldownData = {}
function CooldownData:Create(Name, Type, Cooldown, Duration, Group, GreyImage, AbilityId, Role, Class) -- also saves the created object to the <insert> table
	local rcdData = {}
	setmetatable(rcdData, self)
	self.__index = self

	rcdData.Name = Name
	rcdData.Type = Type
	rcdData.Cooldown = Cooldown
	rcdData.Duration = Duration
	rcdData.Group = Group
	rcdData.GreyImage = GreyImage
	rcdData.AbilityId = AbilityId
	rcdData.Role = Role or 0
	rcdData.Enabled = true
	rcdData.Class = Class

	RCT.CooldownDataCollection[rcdData.Type] = rcdData
	if rcdData.AbilityId then RCT.CooldownDataCollection[rcdData.AbilityId] = rcdData end

	return rcdData
end

-- All Cooldowns
RCT.AllCooldownsByType = {}			-- Key: Buff Type;	Value: Table of RaidCooldowns indexed by Caster
RCT.AllCooldownsByGroup = {}		-- Key: Buff Group;	Value: List of RaidCooldowns
RCT.AllCooldownsByCaster = {}		-- Key: Ally;		Value: Table of RaidCooldowns indexed by Buff Type

-- Damage Cooldown Lists
RCT.DamageCooldownsByType = {}		-- Key: Buff Type;	Value: Table of RaidCooldowns indexed by Caster
RCT.DamageCooldownsByGroup = {}		-- Key: Buff Group;	Value: List of RaidCooldowns
RCT.DamageCooldownsByCaster = {}	-- Key: Ally;		Value: Table of RaidCooldowns indexed by Buff Type

-- Heal Cooldown Lists
RCT.HealCooldownsByType = {}		-- Key: Buff Type;	Value: Table of RaidCooldowns indexed by Caster
RCT.HealCooldownsByGroup = {}		-- Key: Buff Group;	Value: List of RaidCooldowns
RCT.HealCooldownsByCaster = {}		-- Key: Ally;		Value: Table of RaidCooldowns indexed by Buff Type
RCT.HealCooldowns = {}				-- Key: Index;		Value: List of CCD.RaidCooldown

function RCT.Initialize()
	RCT:PopulateCooldownDataCollection()
	RCT.Id = "RCT"
	RCT.On = true
	RCT.ShowInertTimers = true
	RCT.UseGreyscaleForInertCooldowns = true
	RCT.GroupNames = {
		"Flaring Power",
		"Orchestra of the Planes",
		"Verse of Joy",
		"Wild Growth",
		"Enrage",
		"Focused Fire",
		"Unstable Transformation",
		"Shield Defense",
		"Symbiosis",
		"Fiery Will",
		"Command to Attack",
		"Flame of Life",
		"Healer's Covenant",
		"Spiritual Conflagration",
		"Biofeedback",
	}
	RCT.StatusText = { -- stores the text associated with each cooldown Status; used for CPT as well
		"Active",
		"Ready",
		"Soon",
		"Down",
	}
	RCT.SoonReadyStartTime = 10
	RCT.CooldownTimerUpdateFrequency = 0.1 -- seconds between each update; used for bar and alerts
	RCT.ShowBars = true
	RCT.ShowBarsInCombatOnly = false
	RCT.DamageBarGSV = 0
	RCT.StatusColors = {
		Colors.Green,
		Colors.Yellow,
		Colors.Turquoise,
		Colors.Red,
	}
	RCT.GlobalCooldownBarOpacity = 1
	RCT.CooldownBarBackgroundOpacity = 1
	RCT.ShowAlerts = true
	RCT.AbilityTimerTags = {}
	RCT.FrameHeaderHeight = 0
	--
	RCT.HealAlertMenu = RCT:CreateHealAlertMenu()
	--

	RCT.DamageAlertsByGroup = {}
	for _, rcdData in pairs(RCT.CooldownDataCollection) do
		if rcdData.Role == RCT.Role.Damage and not RCT.DamageAlertsByGroup[rcdData.Group] then
			local frame = RCT:CreateDamageAlert(rcdData.Group)
			RCT.DamageAlertsByGroup[rcdData.Group] = frame
		end
	end

	RCT.DamageBars = {}
	RCT.DamageBarLength = 500
	RCT.DamageBarHeight = 24
	RCT.DamageBarFontSize = 18
	RCT.DamageBarLengthRange = { Min = 125, Max = 500}	-- static
	RCT.CooldownBarHeightRange = { Min = 12, Max = 48} -- static
	RCT.CooldownBarFontSizeRange = { Min = 12, Max = 48} -- static
	RCT.NextCooldownBarIndex = 1

	-- Damage Bar Container
	RCT.DamageBarContainer = RCT:CreateDamageBarContainer()
	RCT.DamageBarContainer:Show() -- linkable frames are hidden on creation
		
	-- Damage Bar 1 - Created early for Edit Mode
	RCT:CreateBar()	
		
	RCT.HealAlerts = {}
	RCT.HealAlertOptions = {
		Enabled = true,
		CombatOnly = false,
		GSV = 0,
		IconSize = 48,
		HeaderTextLengthLocked = false,
		HeaderTextFontSizeLocked = false,
		TimerTextFontSizeLocked = false,
		FooterTextFontSizeLocked = false,
		HeaderTextLength = 5,
		HeaderTextFontSize = 18,
		TimerTextFontSize = 24,
		FooterTextFontSize = 18,
		Alpha = 1,
	}
	RCT:CreateHealAlertAnchor()

	RCT.HealBars = {}
	RCT.HealBarManager = {}
	RCT.HealBarManager.Properties = {
		X = 0,
		Y = 0,
		Enabled = true,
		CombatOnly = false,
		GSV = 0,
		Length = 500,
		Height = 24,
		FontSize = 18,
		Opacity = 1,
		BackgroundOpacity = 1,
		Linked,
		LinkPoint,
		AnchorId,
		LinkSpace,
	}
	RCT.HBM = RCT.HealBarManager
	function RCT.HealBarManager:SetEnabled(Bool)
		if Bool == nil then Bool = not self.Properties.Enabled end
		self.Properties.Enabled = Bool
		RCT.HealBarContainer:UpdateForeground()
		RCT.HealBarContainer.IsMobile = self.Properties.Enabled
	end
	function RCT.HealBarManager:HideAllBars()
		for _, bar in ipairs(RCT.HealBars) do
			bar:Hide()
		end
	end
	function RCT.HBM:SetCombatOnly(Bool)
		if Bool == nil then Bool = not self.Properties.CombatOnly end
		self.Properties.CombatOnly = Bool
		RCT.HealBarContainer:UpdateForeground()
	end
	function RCT.HBM:SetGSV(Number)
		self.Properties.GSV = Number
		RCT:OrganizeHealBars()
	end
	function RCT.HBM:SetLength(Number)
		Number = Number or self.Properties.Length
		self.Properties.Length = CCD.Math.MinMax(Number, RCT.DamageBarLengthRange.Min, RCT.DamageBarLengthRange.Max)

		for _, bar in ipairs(RCT.HealBars) do
			bar:SetWidth(self.Properties.Length)
		end
		
		RCT.HealBarContainer:Resize()
	end
	function RCT.HBM:SetHeight(Number)
		Number = Number or self.Properties.Height
		self.Properties.Height = CCD.Math.MinMax(Number, RCT.CooldownBarHeightRange.Min, RCT.CooldownBarHeightRange.Max)

		for _, bar in ipairs(RCT.HealBars) do
			bar:SetHeight(self.Properties.Height)
		end
		
		RCT.HealBarContainer:Resize()
	end
	function RCT.HBM:SetFontSize(Number)
		Number = Number or self.Properties.FontSize
		self.Properties.FontSize = CCD.Math.MinMax(Number, RCT.CooldownBarFontSizeRange.Min, RCT.CooldownBarFontSizeRange.Max)

		for _, bar in ipairs(RCT.HealBars) do
			bar:SetFontSize(self.Properties.FontSize)
		end
	end
	function RCT.HBM:SetOpacity(Number)
		if Number then self.Properties.Opacity = CCD.Math.ConvertOpacity(Number) end

		for _, bar in ipairs(RCT.HealBars) do
			bar.Filling:SetAlpha(self.Properties.Opacity)
			bar.Icon:SetAlpha(self.Properties.Opacity)
		end
	end
	function RCT.HBM:SetBackgroundOpacity(Number)
		if Number then self.Properties.BackgroundOpacity = CCD.Math.ConvertOpacity(Number) end

		for _, bar in ipairs(RCT.HealBars) do
			bar:SetBackgroundAlpha(self.Properties.BackgroundOpacity)
		end
	end
	function RCT.HBM:Start()
		RCT.HBM:SetLength()
		RCT.HBM:SetHeight()
		RCT.HBM:SetFontSize()
		RCT.HBM:SetOpacity()
		RCT.HBM:SetBackgroundOpacity()
	end

	RCT.HBM.Properties.Length = 500
	RCT.HBM.Properties.Height = 24
	RCT.HBM.Properties.FontSize = 18
	RCT.HealBarLengthRange = { Min = 125, Max = 500}	-- static
	RCT.HealBarHeightRange = { Min = 12, Max = 48} -- static
	RCT.HealBarFontSizeRange = { Min = 12, Max = 48} -- static
	RCT.NextHealBarIndex = 1
	RCT.HBM.Properties.Opacity = 1
	RCT.HBM.Properties.BackgroundOpacity = 1

	-- Heal Bar Container
	RCT.HealBarContainer = RCT:CreateHealBarContainer()
	RCT.HealBarContainer:Show() -- linkable frames are hidden on creation
	
	-- Heal Bar 1
	RCT:CreateHealBar()

	RCT.ContextMenu = RCT.CreateDamageAlertMenu()
	RCT.CooldownBarContextMenu = RCT.CreateDamageBarMenu()
	RCT.HealBarMenu = RCT.CreateHealBarMenu()
	RCT.HealCooldownSelectFrame = RCT:CreateHealCooldownSelectFrame()
	RCT.HealCooldownSelectFrame:Hide()
end
function RCT.OnFirstRun()
	RCT.ResetFramePositions()
	RCT.Start()
end
function RCT.Load(Settings)
	if Settings then
		if Settings.On ~= nil then RCT.On = Settings.On end
		if Settings.UseGreyscaleForInertCooldowns ~= nil then RCT.UseGreyscaleForInertCooldowns = Settings.UseGreyscaleForInertCooldowns end
		if Settings.ShowInertTimers ~= nil then RCT.ShowInertTimers = Settings.ShowInertTimers	end
		RCT.SoonReadyStartTime = Settings.SoonReadyStartTime or RCT.SoonReadyStartTime
		if Settings.ShowBars ~= nil then RCT.ShowBars = Settings.ShowBars end
		if Settings.ShowBarsInCombatOnly ~= nil then RCT.ShowBarsInCombatOnly = Settings.ShowBarsInCombatOnly end
		RCT.DamageBarGSV = Settings.DamageBarGSV or Settings.CooldownBarGroupShowValue or RCT.DamageBarGSV -- variable name change in 3.4.2
		RCT.GlobalCooldownBarOpacity = Settings.GlobalCooldownBarOpacity or RCT.GlobalCooldownBarOpacity -- added in 3.2.4
		RCT.CooldownBarBackgroundOpacity = Settings.CooldownBarBackgroundOpacity or RCT.CooldownBarBackgroundOpacity -- removed in 3.2.4; readded in 3.3.4
		if Settings.ShowAlerts ~= nil then RCT.ShowAlerts = Settings.ShowAlerts end
		-- pre v3.1.5 settings
		if Settings.FrameDirection then RCT.FrameDirection = Settings.FrameDirection end
		if Settings.FrameSpacing then RCT.FrameSpacing = Settings.FrameSpacing end
		--
		if Settings.Frames then
			for i, frame in pairs(RCT.DamageAlertsByGroup) do
				if Settings.Frames[i] then
					frame.X = Settings.Frames[i].X or frame.X
					frame.Y = Settings.Frames[i].Y or frame.Y

					frame.HeaderTextFontSize = Settings.Frames[i].HeaderTextFontSize or frame.HeaderTextFontSize
					frame.HeaderTextLength = Settings.HeaderLength or Settings.Frames[i].HeaderTextLength or frame.HeaderTextLength -- Pre 3.2.4: Header length was global
					-- texture dimensions are set by frame.IconSize
					frame.TimerTextFontSize = Settings.Frames[i].TimerFontSize or Settings.Frames[i].TimerTextFontSize or frame.TimerTextFontSize -- Pre 3.2.4: Settings variable was named TimerFontSize
					frame.FooterTextFontSize = Settings.Frames[i].FooterTextFontSize or frame.FooterTextFontSize
					frame.IconSize = Settings.Frames[i].Scale or frame.IconSize
					if frame.IconSize <= 10 then frame.IconSize = 39 + ((frame.IconSize - 1) * 11) end

					if Settings.Frames[i].TimerTextFontSizeLocked ~= nil then frame.TimerTextFontSizeLocked = Settings.Frames[i].TimerTextFontSizeLocked end

					if Settings.Frames[i].IsTracked ~= nil then frame.IsTracked = Settings.Frames[i].IsTracked end
					if Settings.Frames[i].CombatShow ~= nil then frame.CombatShow = Settings.Frames[i].CombatShow end
					frame.GroupShowValue = Settings.Frames[i].GroupShowValue or frame.GroupShowValue
					if Settings.Frames[i].ShowDamageBar == nil then
						if Settings.Frames[i].ShowCooldownBars ~= nil then frame.ShowDamageBar = Settings.Frames[i].ShowCooldownBars end -- Pre 3.4.2
					else
						frame.ShowDamageBar = Settings.Frames[i].ShowDamageBar
					end
					frame.Alpha = Settings.Frames[i].Alpha or frame.Alpha

					if Settings.Frames[i].Linked ~= nil then frame.Linked = Settings.Frames[i].Linked end
					frame.LinkPoint = Settings.Frames[i].LinkPoint or frame.LinkPoint
					frame.AnchorId = Settings.Frames[i].AnchorId or frame.AnchorId
					frame.LinkSpace = Settings.Frames[i].LinkSpace or frame.LinkSpace

					if Settings.Frames[i].HeaderOn ~= nil then frame.HeaderOn = Settings.Frames[i].HeaderOn end
					if Settings.Frames[i].FooterOn ~= nil then frame.FooterOn = Settings.Frames[i].FooterOn end
				end
			end
		end
		--
		if Settings.StatusText then
			for i, text in ipairs(Settings.StatusText) do
				RCT.StatusText[i] = text
			end
		end
		--
		RCT.DamageBarLength = Settings.DamageBarLength or Settings.CooldownBarLength or Settings.CooldownBarWidth or RCT.DamageBarLength -- pre 3.2.1 var was referenced "Width" rather than length; pre 3.4.2 was CooldownBarLength
		RCT.DamageBarHeight = Settings.DamageBarHeight or Settings.CooldownBarHeight or RCT.DamageBarHeight -- pre 3.4.2 was CooldownBarHeight
		RCT.DamageBarFontSize = Settings.DamageBarFontSize or Settings.CooldownBarFontSize or RCT.DamageBarFontSize -- pre 3.4.2 was CooldownBarFontSize
		if RCT.DamageBarLength < RCT.DamageBarLengthRange.Min then RCT.DamageBarLength = RCT.DamageBarLengthRange end -- pre v3.2.1 allowed length of less than 250; 3.3.4+ min length = 125
		
		-- Damage Bar Container
		local dbcSettings
		if Settings.CooldownBars and Settings.CooldownBars[1] then
			-- pre 3.4.0
			dbcSettings = Settings.CooldownBars[1]
		elseif Settings.DamageBarContainer then
			dbcSettings = Settings.DamageBarContainer
		end
		if dbcSettings then
			RCT.DamageBarContainer.X = dbcSettings.X or RCT.DamageBarContainer.X
			RCT.DamageBarContainer.Y = dbcSettings.Y or RCT.DamageBarContainer.Y
			if dbcSettings.Linked ~= nil then RCT.DamageBarContainer.Linked = dbcSettings.Linked end
			RCT.DamageBarContainer.LinkPoint = dbcSettings.LinkPoint or RCT.DamageBarContainer.LinkPoint
			RCT.DamageBarContainer.AnchorId = dbcSettings.AnchorId or RCT.DamageBarContainer.AnchorId
			RCT.DamageBarContainer.LinkSpace = dbcSettings.LinkSpace or RCT.DamageBarContainer.LinkSpace		
		else
			RCT.ResetCooldownBarContainerPositions()
		end

		-- Heal Alerts
		if Settings.HealAlert then
			RCT.HealAlertAnchor.X = Settings.HealAlert.X or RCT.HealAlertAnchor.X
			RCT.HealAlertAnchor.Y = Settings.HealAlert.Y or RCT.HealAlertAnchor.y
			--
			if Settings.HealAlert.Linked then RCT.HealAlertAnchor.Linked = Settings.HealAlert.Linked end
			RCT.HealAlertAnchor.LinkPoint = Settings.HealAlert.LinkPoint or RCT.HealAlertAnchor.LinkPoint
			RCT.HealAlertAnchor.AnchorId = Settings.HealAlert.AnchorId or RCT.HealAlertAnchor.AnchorId
			RCT.HealAlertAnchor.LinkSpace = Settings.HealAlert.LinkSpace or RCT.HealAlertAnchor.LinkSpace
			--
			if Settings.HealAlert.Options then
				for key, value in pairs(Settings.HealAlert.Options) do
					RCT.HealAlertOptions[key] = value
				end
			end
		end

		-- Heal Bar
		if Settings.HealBar then
			for key, value in pairs(Settings.HealBar) do
				RCT.HBM.Properties[key] = value
			end
		end

		if Settings.CooldownDataCollection then
			for key, data in pairs(RCT.CooldownDataCollection) do
				if Settings.CooldownDataCollection[key] then
					data.Enabled = Settings.CooldownDataCollection[key].Enabled
					if RCT.HealCooldownSelectFrame.CheckboxesByBuffType[key] then
						RCT.HealCooldownSelectFrame.CheckboxesByBuffType[key]:SetChecked(data.Enabled)
					end
				end
			end
		end
		
		RCT.Start()
	else
		RCT.OnFirstRun()
	end
end
function RCT.Start()
	for i, frame in pairs(RCT.DamageAlertsByGroup) do
		frame:Initialize()
		frame:SetTracked()
		
		if frame.Linked then
			if frame.AnchorId == "NONE" then
				-- this means we're loading pre-3.1.5 settings
				frame:Link(RCT.DamageAlertsByGroup[i-1], RCT.FrameDirection, RCT.FrameSpacing)
			else
				-- load normally
				frame:Link(CCD.UI.LinkableFramesById[frame.AnchorId], frame.LinkPoint, frame.LinkSpace)
			end
		else
			frame.EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", frame.X, frame.Y)
		end
	end

	local frame = RCT.DamageBarContainer
	if frame.Linked then
		frame:Link(CCD.UI.LinkableFramesById[frame.AnchorId], frame.LinkPoint, frame.LinkSpace)
	else
		frame.EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", frame.X, frame.Y)
	end

	RCT:SetBarLength()
	RCT:SetBarHeight()
	RCT:SetBarFontSize()
	RCT:UpdateBarForegrounds()
	RCT:SetGlobalCooldownBarOpacity() -- added in 3.2.4
	RCT:SetBarBackgroundOpacity()
	frame.IsMobile = RCT.ShowBars
	
	RCT.HealAlertAnchor.SuperFunctions:Initialize()
	if RCT.HealAlertAnchor.Linked then
		RCT.HealAlertAnchor:Link(CCD.UI.LinkableFramesById[RCT.HealAlertAnchor.AnchorId], RCT.HealAlertAnchor.LinkPoint, RCT.HealAlertAnchor.LinkSpace)
	else
		RCT.HealAlertAnchor.EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", RCT.HealAlertAnchor.X, RCT.HealAlertAnchor.Y)
	end
	RCT.HealAlertAnchor:UpdateForeground()
	RCT.HealAlertAnchor.IsMobile = RCT.HealAlertOptions.Enabled

	RCT.HealBarContainer:UpdateForeground()
	RCT.HealBarContainer.IsMobile = RCT.HBM.Properties.Enabled
	if RCT.HBM.Properties.Linked then
		RCT.HealBarContainer:Link(CCD.UI.LinkableFramesById[RCT.HBM.Properties.AnchorId], RCT.HBM.Properties.LinkPoint, RCT.HBM.Properties.LinkSpace)
	else
		RCT.HealBarContainer.EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", RCT.HBM.Properties.X, RCT.HBM.Properties.Y)
	end
	RCT.HBM:Start()

	RCT.Set(RCT.On)
end
function RCT.CreateSaveTable()
	local saveTable = {
		On = RCT.On,
		ShowInertTimers = RCT.ShowInertTimers,
		UseGreyscaleForInertCooldowns = RCT.UseGreyscaleForInertCooldowns,
		SoonReadyStartTime = RCT.SoonReadyStartTime,
		Frames = {},
		StatusText = {},		
		ShowBars = RCT.ShowBars,
		ShowAlerts = RCT.ShowAlerts,
		HealAlert = {},
	}

	for i, frame in pairs(RCT.DamageAlertsByGroup) do
		saveTable.Frames[i] = {}
		saveTable.Frames[i].Linked = frame.Linked
		saveTable.Frames[i].LinkPoint = frame.LinkPoint
		saveTable.Frames[i].AnchorId = frame.AnchorId
		saveTable.Frames[i].LinkSpace = frame.LinkSpace

		saveTable.Frames[i].X = frame.X
		saveTable.Frames[i].Y = frame.Y

		saveTable.Frames[i].HeaderTextFontSize = frame.HeaderTextFontSize
		saveTable.Frames[i].HeaderTextLength = frame.HeaderTextLength
		-- texture dimensions are set by frame.IconSize
		saveTable.Frames[i].TimerTextFontSize = frame.TimerTextFontSize -- Pre 3.2.4: Settings variable was named TimerFontSize
		saveTable.Frames[i].FooterTextFontSize = frame.FooterTextFontSize
		saveTable.Frames[i].Scale = frame.IconSize

		saveTable.Frames[i].TimerTextFontSizeLocked = frame.TimerTextFontSizeLocked

		saveTable.Frames[i].IsTracked = frame.IsTracked
		saveTable.Frames[i].CombatShow = frame.CombatShow
		saveTable.Frames[i].GroupShowValue = frame.GroupShowValue
		saveTable.Frames[i].ShowDamageBar = frame.ShowDamageBar
		saveTable.Frames[i].Alpha = frame.Alpha

		saveTable.Frames[i].HeaderOn = frame.HeaderOn
		saveTable.Frames[i].FooterOn = frame.FooterOn
		
		saveTable.Frames[i].Group = frame.Group
	end

	for i, text in ipairs(RCT.StatusText) do
		saveTable.StatusText[i] = text
	end

	saveTable.DamageBarLength = RCT.DamageBarLength
	saveTable.DamageBarHeight = RCT.DamageBarHeight
	saveTable.DamageBarFontSize = RCT.DamageBarFontSize
	saveTable.ShowBars = RCT.ShowBars
	saveTable.ShowBarsInCombatOnly = RCT.ShowBarsInCombatOnly
	saveTable.DamageBarGSV = RCT.DamageBarGSV
	saveTable.GlobalCooldownBarOpacity = RCT.GlobalCooldownBarOpacity -- added in 3.2.4
	saveTable.CooldownBarBackgroundOpacity = RCT.CooldownBarBackgroundOpacity -- removed in 3.2.4; readded in 3.3.4

	saveTable.DamageBarContainer = {
		X = RCT.DamageBarContainer.X,
		Y = RCT.DamageBarContainer.Y,
		Linked = RCT.DamageBarContainer.Linked,
		LinkPoint = RCT.DamageBarContainer.LinkPoint,
		AnchorId = RCT.DamageBarContainer.AnchorId,
		LinkSpace = RCT.DamageBarContainer.LinkSpace,
	}

	-- Heal Alerts
	saveTable.HealAlert.X = RCT.HealAlertAnchor:GetLeft()
	saveTable.HealAlert.Y = RCT.HealAlertAnchor:GetTop()
	--
	saveTable.HealAlert.Linked = RCT.HealAlertAnchor.Linked
	saveTable.HealAlert.LinkPoint = RCT.HealAlertAnchor.LinkPoint
	saveTable.HealAlert.AnchorId = RCT.HealAlertAnchor.AnchorId
	saveTable.HealAlert.LinkSpace = RCT.HealAlertAnchor.LinkSpace
	--
	saveTable.HealAlert.Options = {}
	for key, value in pairs(RCT.HealAlertOptions) do
		saveTable.HealAlert.Options[key] = value
	end
	--

	-- Heal Bar
	RCT.HBM.Properties.X = RCT.HealBarContainer:GetLeft()
	RCT.HBM.Properties.Y = RCT.HealBarContainer:GetTop()
	RCT.HBM.Properties.Linked = RCT.HealBarContainer.Linked
	RCT.HBM.Properties.LinkPoint = RCT.HealBarContainer.LinkPoint
	RCT.HBM.Properties.AnchorId = RCT.HealBarContainer.AnchorId
	RCT.HBM.Properties.LinkSpace = RCT.HealBarContainer.LinkSpace
	saveTable.HealBar = {}
	for key, value in pairs(RCT.HBM.Properties) do
		saveTable.HealBar[key] = value
	end

	saveTable.CooldownDataCollection = {}
	for key, data in pairs(RCT.CooldownDataCollection) do
		saveTable.CooldownDataCollection[key] = {}
		saveTable.CooldownDataCollection[key].Enabled = data.Enabled
	end
	
	return saveTable
end

function RCT:PopulateCooldownDataCollection()
	-- Damage
	CooldownData:Create("FlaringPower",			"B2F55F12AF5DA3E23", 300,	30, 1,	"flaring.png",		"A67D9C8E34456AC35",	CCD.Modules.RCT.Role.Damage)
	CooldownData:Create("OrchestraOfThePlanes",	"BFD9F4FF8303ACEE6", 120,	15, 2,	"orchestra.png",	"AFAA9C085A3AD3A8E",	CCD.Modules.RCT.Role.Damage)
	CooldownData:Create("DefendTheFallen",		"B39FB71DBD14135BE", 120,	15, 2,	"dtf.png",			"A2BBE0D93CAE15935",	CCD.Modules.RCT.Role.Damage)
	CooldownData:Create("LavaField",			"B14A2E6D509F79153", 60,	11, 2,	"lava.png",			"A2303677658E3BEFD",	CCD.Modules.RCT.Role.Damage)
	CooldownData:Create("VerseOfJoy",			"B6264D731057EDA4F", 120,	15, 3,	"voj.png",			"A2DAD149799991550",	CCD.Modules.RCT.Role.Damage)
	CooldownData:Create("ResuscitatingChant",	"B3BD1851CE8D9DE52", 120,	15, 3,	"reschant.png",		"A3B0C08974A8812D2",	CCD.Modules.RCT.Role.Damage)
	CooldownData:Create("WildGrowth",			"B087BE1DF34A67269", 58.5,	12, 4,	"wild.png",			"A1228F93DC7326968",	CCD.Modules.RCT.Role.Damage)
	CooldownData:Create("PowerCore",			"B0118235BB73BE55B", 60,	15, 4,	"power.png",		"AFBF79829B2DD89A0",	CCD.Modules.RCT.Role.Damage)
	CooldownData:Create("Enrage",				"B0B046904062527A2", 60,	10, 5,	"enrage.png",		"AFFA7EFCDAA9506A3",	CCD.Modules.RCT.Role.Damage)
	CooldownData:Create("CommandToAttack",		"BFA7A188CC92D648B", 300,	30, 11, "cta.png",			"AFAEE513B677B63D2",	CCD.Modules.RCT.Role.Damage)
	-- Heal
	CooldownData:Create("Fiery Will", "B6317AF35D5658FAD", 60, 15,					20, "fiery_will_grey.png", "A2D8B55747FE03229", CCD.Modules.RCT.Role.Heal, "CLERIC")
	CooldownData:Create("Flame of Life", "B1BC33DBE7CFA0B28", 90, 6,				21, "flame_of_life_grey.png", "AFABF9CC466A13DEE", CCD.Modules.RCT.Role.Heal, "CLERIC")
	CooldownData:Create("Life's Rapture", "B7CCBCE2EF6633EA1", 90, 6,				21, nil, "A483A8B14922C2865", CCD.Modules.RCT.Role.Heal, "WARRIOR")
	CooldownData:Create("Gathering of Flames", "BFE6320F8868060AF", 60, 30,			22, nil, "AFC6A7C6553CFAF07", CCD.Modules.RCT.Role.Heal, "CLERIC")
	CooldownData:Create("Rite of the Forge", "B1412EB3418417D27", 120, 15,			23, nil, "A0C1F4BB4C0AF1CD3", CCD.Modules.RCT.Role.Heal, "CLERIC")
	CooldownData:Create("Spiritual Conflagration", "B380B25D249D2F653", 60, 5,		24, nil, "A5D83EB134C61A309", CCD.Modules.RCT.Role.Heal, "CLERIC")
	CooldownData:Create("Healer's Covenant", "B677C02F4965465D5", 60, 10,			25, nil, "A10BC68E643725C47", CCD.Modules.RCT.Role.Heal, "CLERIC")
	CooldownData:Create("Unstable Transformation", "BFD76644415EA643D", 60, 10,		26, "unstable transformation.png", "AFBEDD6F07A8364D7", CCD.Modules.RCT.Role.Heal, "CLERIC")
	CooldownData:Create("Symbiosis (Non-Synthesis)", "B19EEE7BA9C218028", 45, 10,	27, "symbiosis_grey.png", "A4EF38CC08FA3FB34", CCD.Modules.RCT.Role.Heal, "MAGE") -- Non-Synthesis
	CooldownData:Create("Symbiosis (Synthesis)", "B07B4AD401C182BB0", 45, 10,		27, "symbiosis_grey.png", "A4EF38CC08FA3FB34", CCD.Modules.RCT.Role.Heal, "MAGE") -- Synthesis
	CooldownData:Create("Natural Splendor (60s)", "B70400B13CA99AD9C", 60, 8,				28, nil, "A40C424984174CB64", CCD.Modules.RCT.Role.Heal, "MAGE")
	CooldownData:Create("Biofeedback", "B15D743C8FA734A24", 60, 5,					29, nil, "A43C6AFFF6D66A7D6", CCD.Modules.RCT.Role.Heal, "ROGUE")
	CooldownData:Create("Life Insurance", "B23926F1847222504", 60, 10,				30, nil, "A52D58EF2B55774DF", CCD.Modules.RCT.Role.Heal, "ROGUE")
	CooldownData:Create("Intensive Care", "B7AA77FB8ECEB27E8", 120, 10,				31, nil, "A7B839DE92DA3A304", CCD.Modules.RCT.Role.Heal, "ROGUE")
	--CooldownData:Create("Apothecary's Will (Buff)", "B4EEFD27725DA9B48", 60, 5,		32, nil, "A6AA0FF8A38A544E1", CCD.Modules.RCT.Role.Heal)
	--CooldownData:Create("Apothecary's Will (Debuff)", "BFEAED9BE1650AD17", 60, 5,	32, nil, "A6AA0FF8A38A544E1", CCD.Modules.RCT.Role.Heal)
	CooldownData:Create("Redress", "B221B43BB7A12BAB1", 60, 8,						33, nil, "A26C92F13E56F1D87", CCD.Modules.RCT.Role.Heal, "WARRIOR")
	CooldownData:Create("Mass Rescue", "B1602847D3E3FFFA7", 60, 15,					34, nil, "A229C371DBE60CB22", CCD.Modules.RCT.Role.Heal, "WARRIOR")
	CooldownData:Create("Medical Facility", "B338DB29C84E08104", 60, 6,				35, nil, "A781808662A146214", CCD.Modules.RCT.Role.Heal, "WARRIOR")
	CooldownData:Create("Command to Stand Ground (5m)", "BFA61745B438D5210", 300, 30,	36, nil, "AFE2FD850D13CEABE", CCD.Modules.RCT.Role.Heal, "WARRIOR")
	CooldownData:Create("Command to Recover (5m)", "BFEB4CDB54A4B5899", 300, 30,		37, nil, "AFD00EEA28FF099C9", CCD.Modules.RCT.Role.Heal, "WARRIOR")
	--CooldownData:Create("Shield Defense", "BFAC0BA623B24AA33", 60, 10, 8, "shield defense.png", "AFA1A89D25BE4F842", CCD.Modules.RCT.Role.Heal) -- needs to be changed back to Tank
	--CooldownData:Create("monsoon", nil, 60, 15, 18, nil, "A0983900D47E93A3A", CCD.Modules.RCT.Role.Heal)
	--[[ Test
	CooldownData:Create("HealingSpray", "B2117281B48810943", 45, 15, 1, nil) -- test; remove on release
	CooldownData:Create("HealingFlood", "B54CE31C6FDF59F25", 45, 15, 5, nil) -- test; remove on release
	CooldownData:Create("SmybolOfTheTorch", "B495934659ECF7164", 60, 30, 3, nil) -- test; remove on release
	CooldownData:Create("Resurgence", "BFB4BF1F76B80B033", 10, 6, 2, nil) -- test; remove on release
	CooldownData:Create("MotifOfBravery", "B75CF79A3A7B75E32", 30, 20, 1, nil) -- test; remove on release
	CooldownData:Create("MotifOfTenacity", "B60F47F310B4793F6", 30, 20, 2, nil) -- test; remove on release
	CooldownData:Create("MotifOfRegeneration", "B0FA41E09E7A89F58", 30, 20, 6, nil) -- test; remove on release]]

	-- Initialise CooldownsByType and CooldownsByGroup Lists; CooldownsByCaster Lists are initialised OnNewAlly()
	for buffType, data in pairs(RCT.CooldownDataCollection) do
		RCT.AllCooldownsByType[buffType] = {}
		RCT.AllCooldownsByGroup[data.Group] = RCT.AllCooldownsByGroup[data.Group] or {}

		RCT.DamageCooldownsByType[buffType] = {}
		RCT.DamageCooldownsByGroup[data.Group] = RCT.DamageCooldownsByGroup[data.Group] or {}

		RCT.HealCooldownsByType[buffType] = {}
		RCT.HealCooldownsByGroup[data.Group] = RCT.HealCooldownsByGroup[data.Group] or {}
	end
end
function RCT.ResetFramePositions()
	local x = 99
	for i, frame in pairs(RCT.DamageAlertsByGroup) do
		x = x + 1
		RCT.DamageAlertsByGroup[i].EditBolt:ClearAll()
		RCT.DamageAlertsByGroup[i].EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", x, 400)
	end
	RCT.LinkFrames("TOPRIGHT", 5)
	
	RCT.HealAlertAnchor:Unlink()
	RCT.HealAlertAnchor.EditBolt:ClearAll()
	RCT.HealAlertAnchor.EditBolt:SetPoint("TOPLEFT", UIParent, "TOPLEFT", RCT.DamageAlertsByGroup[1]:GetLeft(), RCT.DamageAlertsByGroup[1]:GetBottom() + 50)
	
	RCT.ResetCooldownBarContainerPositions()
end
function RCT.ResetCooldownBarContainerPositions()
	RCT.DamageBarContainer:Unlink()
	RCT.DamageBarContainer:Link(RCT.DamageAlertsByGroup[1], "BOTTOMLEFT", 5)
	
	RCT.HealBarContainer:Unlink()
	RCT.HealBarContainer:Link(RCT.HealAlertAnchor, "BOTTOMLEFT", 5)
end
function RCT.Set(NewState)
	RCT.On = NewState

	if RCT.On then
		RCT:UpdateAllDamageAlerts()
		
		for _, bar in ipairs(RCT.DamageBars) do
			if RCT:ShouldShowBars() then
				bar:Update()
			end
		end
	elseif not CCD.UI.EditModeOn then
		for _, frame in pairs(RCT.DamageAlertsByGroup) do
			frame:SetVisible(false)
		end
		RCT:HideAllBars()
	end

end

function RCT:CreateTab(TabWidth)
	-- Base Frame
	RCT.Tab = UI.CreateFrame("Mask", "RaidCooldownTrackerTab", CCD.MainWindow)
	RCT.Tab:SetBackgroundColor(Colors.Black.R, Colors.Black.G, Colors.Black.B)
	local tab = RCT.Tab
	local description = tab:GetName()

	-- On Check Box
	tab.OnCheckbox = UI.CreateFrame("SimpleCheckbox", description .. "_OnCheckbox", tab)
	tab.OnCheckbox:SetFontSize(15)
	tab.OnCheckbox:SetText("Enable Cooldown Tracker")
	tab.OnCheckbox:SetChecked(RCT.On)
	tab.OnCheckbox:SetPoint("TOPLEFT", tab, "TOPLEFT", 10, 10)
	tab.OnCheckbox:SetVisible(true)
	function tab.OnCheckbox.Event:CheckboxChange()
		RCT.Set(self:GetChecked())
		RCT.ContextMenu.OnCheckbox:SetChecked(self:GetChecked())
	end

	-- Show Damage Alerts Check Box
	tab.ShowAlertsCheckbox = CCD.UI.CreateCheckbox(description .. ".ShowAlertsCheckbox", tab, RCT.ShowAlerts)
	tab.ShowAlertsCheckbox:SetText("Show Damage Alerts", 15, Colors.White)
	tab.ShowAlertsCheckbox:SetPoint("TOPLEFT", tab.OnCheckbox, "TOPRIGHT", 5, 0)
	function tab.ShowAlertsCheckbox.Event:CheckboxChange()
		RCT:SetShowAlerts(self:GetChecked(), self)
	end

	-- Reset Frames Button
	tab.ResetButton = UI.CreateFrame("RiftButton", description .. "_tab.ResetButton", tab)
	tab.ResetButton:SetText("Reset Frames")
	tab.ResetButton:SetPoint("TOPRIGHT", tab, "TOPRIGHT", 0, 2)
	tab.ResetButton:SetVisible(true)
	function tab.ResetButton.Event:LeftPress()
		RCT.ResetFramePositions()
	end

	-- Show Down Timers Checkbox
	tab.InertTimersCheckbox = UI.CreateFrame("SimpleCheckbox", description .. "_InertTimersCheckbox", tab)
	tab.InertTimersCheckbox:SetFontSize(15)
	tab.InertTimersCheckbox:SetText("Show Alerts for DOWN cooldowns")
	tab.InertTimersCheckbox:SetChecked(RCT.ShowInertTimers)
	tab.InertTimersCheckbox:SetPoint("TOPLEFT", tab.OnCheckbox, "BOTTOMLEFT", 0, 10)
	tab.InertTimersCheckbox:SetVisible(true)
	function tab.InertTimersCheckbox.Event:CheckboxChange()
		local isChecked = self:GetChecked()
		RCT.ShowInertTimers = isChecked
		RCT.ContextMenu.ShowInertTimersCheckbox:SetChecked(isChecked)
	end

	-- Use Greyscale Checkbox
	tab.GreyscaleCheckbox = UI.CreateFrame("SimpleCheckbox", description .. "_GreyscaleCheckbox", tab)
	tab.GreyscaleCheckbox:SetFontSize(15)
	tab.GreyscaleCheckbox:SetText("Use Greyscale for DOWN cooldowns")
	tab.GreyscaleCheckbox:SetChecked(RCT.UseGreyscaleForInertCooldowns)
	tab.GreyscaleCheckbox:SetPoint("TOPLEFT", tab.InertTimersCheckbox, "BOTTOMLEFT", 0, 0)
	tab.GreyscaleCheckbox:SetVisible(true)
	function tab.GreyscaleCheckbox.Event:CheckboxChange()
		local isChecked = self:GetChecked()
		RCT.UseGreyscaleForInertCooldowns = isChecked
		RCT.ContextMenu.UseGreyscaleCheckbox:SetChecked(isChecked)
	end

	local sliderXOffset = 115

	-- Frame Scale

	-- Label
	local frameScaleLabel = UI.CreateFrame("Text", description .. "_frameScaleLabel", tab)
	frameScaleLabel:SetText("Icon Size:")
	frameScaleLabel:SetEffectGlow({
		colorA = 0.8, colorB = 0, colorG = 0, colorR = 0,
		offsetX = 0, offsetY = 0,
		blurX = 1, blurY = 1,
		knockout = false,
		replace = false,
		strength = 3
	})
	frameScaleLabel:SetFontSize(15)
	frameScaleLabel:SetFontColor(1, 1, 1) -- white
	frameScaleLabel:SetPoint("TOPLEFT", tab.GreyscaleCheckbox, "BOTTOMLEFT", 0, 5)
	frameScaleLabel:SetVisible(true)

	-- Slider
	local frameScaleSlider = UI.CreateFrame("SimpleSlider", description .. "_frameScaleSlider", tab)
	frameScaleSlider:SetRange(CCD.UI.AlertFrame.IconSizeRange.Min, CCD.UI.AlertFrame.IconSizeRange.Max)
	frameScaleSlider:SetPosition(48)
	frameScaleSlider:SetPoint("TOPLEFT", frameScaleLabel, "TOPLEFT", sliderXOffset, 0)
	frameScaleSlider:SetVisible(true)

	-- Button
	local frameScaleButton = UI.CreateFrame("RiftButton", description .. "_frameScaleButton", tab)
	frameScaleButton:SetText("Set All")
	frameScaleButton:SetPoint("TOPLEFT", frameScaleSlider, "TOPRIGHT", 5, -7)
	frameScaleButton:SetVisible(true)
	function frameScaleButton.Event:LeftPress()
		RCT:SetAllIconSizes(frameScaleSlider:GetPosition())
	end

	----------------------

	-- Timer Scale

	-- Label
	local timerScaleLabel = UI.CreateFrame("Text", description .. "_timerScaleLabel", tab)
	timerScaleLabel:SetText("Timer Font:")
	timerScaleLabel:SetEffectGlow({
		colorA = 0.8, colorB = 0, colorG = 0, colorR = 0,
		offsetX = 0, offsetY = 0,
		blurX = 1, blurY = 1,
		knockout = false,
		replace = false,
		strength = 3
	})
	timerScaleLabel:SetFontSize(15)
	timerScaleLabel:SetFontColor(1, 1, 1) -- white
	timerScaleLabel:SetPoint("TOPLEFT", frameScaleLabel, "BOTTOMLEFT", 0, 0)
	timerScaleLabel:SetVisible(true)

	-- Slider
	local timerScaleSlider = UI.CreateFrame("SimpleSlider", description .. "_timerScaleSlider", tab)
	timerScaleSlider:SetRange(CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max)
	timerScaleSlider:SetPosition(24)
	timerScaleSlider:SetPoint("TOPLEFT", timerScaleLabel, "TOPLEFT", sliderXOffset, 0)
	timerScaleSlider:SetVisible(true)

	-- Button
	local timerScaleButton = UI.CreateFrame("RiftButton", description .. "_timerScaleButton", tab)
	timerScaleButton:SetText("Set All")
	timerScaleButton:SetPoint("TOPLEFT", timerScaleSlider, "TOPRIGHT", 5, -7)
	timerScaleButton:SetVisible(true)
	function timerScaleButton.Event:LeftPress()
		RCT:SetAllTimerTextSizes(timerScaleSlider:GetPosition())
	end

	----------------------

	-- Header Length

	-- Label
	local headerLengthLabel = UI.CreateFrame("Text", description .. "_headerLengthLabel", tab)
	headerLengthLabel:SetText("Name Length:")
	headerLengthLabel:SetEffectGlow({
		colorA = 0.8, colorB = 0, colorG = 0, colorR = 0,
		offsetX = 0, offsetY = 0,
		blurX = 1, blurY = 1,
		knockout = false,
		replace = false,
		strength = 3
	})
	headerLengthLabel:SetFontSize(15)
	headerLengthLabel:SetFontColor(1, 1, 1) -- white
	headerLengthLabel:SetPoint("TOPLEFT", timerScaleLabel, "BOTTOMLEFT", 0, 0)
	headerLengthLabel:SetVisible(true)

	-- Slider
	local headerLengthSlider = UI.CreateFrame("SimpleSlider", description .. "_headerLengthSlider", tab)
	headerLengthSlider:SetRange(3, 19)
	headerLengthSlider:SetPosition(5)
	headerLengthSlider:SetPoint("TOPLEFT", headerLengthLabel, "TOPLEFT", sliderXOffset, 0)
	headerLengthSlider:SetVisible(true)
	function headerLengthSlider.Event:SliderChange()
		local length = self:GetPosition()
		RCT.SetAllHeaderTextLengths(length)
	end

	----------------------

	-- Spacing Label
	tab.SpacingLabel = UI.CreateFrame("Text", description .. "_SpacingLabel", tab)
	tab.SpacingLabel:SetText("Frame Spacing:")
	tab.SpacingLabel:SetEffectGlow({
		colorA = 0.8, colorB = 0, colorG = 0, colorR = 0,
		offsetX = 0, offsetY = 0,
		blurX = 1, blurY = 1,
		knockout = false,
		replace = false,
		strength = 3
	})
	tab.SpacingLabel:SetFontSize(15)
	tab.SpacingLabel:SetFontColor(1, 1, 1) -- white
	tab.SpacingLabel:SetPoint("TOPLEFT", headerLengthLabel, "BOTTOMLEFT", 0, 0)
	tab.SpacingLabel:SetVisible(true)

	-- Spacing Width Slider
	tab.SpacingWidthSlider = UI.CreateFrame("SimpleSlider", description .. "_SpacingWidthSlider", tab)
	tab.SpacingWidthSlider:SetEditable(true)
	tab.SpacingWidthSlider:SetRange(0, 50)
	tab.SpacingWidthSlider:SetPosition(5)
	tab.SpacingWidthSlider:SetPoint("TOPLEFT", tab.SpacingLabel, "TOPLEFT", sliderXOffset, 0)
	tab.SpacingWidthSlider:SetVisible(true)

	-- Spacing Direction Dropdown Menu
	tab.SpacingDirectionDropdown = UI.CreateFrame("SimpleSelect", description .. "_SpacingDirectionDropdown", tab)
	local items = {"Horizontal", "Vertical",}
	local values = {"TOPRIGHT", "BOTTOMLEFT",}
	tab.SpacingDirectionDropdown:SetItems(items, values)
	tab.SpacingDirectionDropdown:SetSelectedIndex(1)
	tab.SpacingDirectionDropdown:ResizeToFit()
	tab.SpacingDirectionDropdown:SetPoint("TOPLEFT", tab.SpacingWidthSlider, "TOPRIGHT", 5, 0)
	tab.SpacingDirectionDropdown:SetVisible(true)

	-- Link Cooldowns Button
	tab.LinkButton = UI.CreateFrame("RiftButton", description .. "_LinkButton", tab)
	tab.LinkButton:SetText("Link Cooldowns")
	tab.LinkButton:SetPoint("TOPLEFT", tab.SpacingDirectionDropdown, "TOPRIGHT", 5, -7)
	tab.LinkButton:SetVisible(true)
	function tab.LinkButton.Event:LeftPress()
		RCT.LinkFrames(tab.SpacingDirectionDropdown:GetSelectedValue(), tab.SpacingWidthSlider:GetPosition())
	end

	-- Alert Header Font Size Slider
	tab.HeaderTextFontSizeSlider = CCD.UI.CreateSlider(description .. ".HeaderTextFontSizeSlider", tab)
	tab.HeaderTextFontSizeSlider.Label:SetText("Header Font:")
	tab.HeaderTextFontSizeSlider.Label:SetWidth(sliderXOffset)
	tab.HeaderTextFontSizeSlider:SetRange(CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max, 18)
	tab.HeaderTextFontSizeSlider:SetPoint("TOPLEFT", tab.SpacingLabel, "BOTTOMLEFT", 0, 0)
	function tab.HeaderTextFontSizeSlider.Slider.Event:SliderChange()
		local pos = tab.HeaderTextFontSizeSlider.Slider:GetPosition()
		RCT:SetAllHeaderTextFontSizes(pos)
	end

	tab.ShowHeadersButton = UI.CreateFrame("RiftButton", description .. ".ShowHeadersButton", tab)
	tab.ShowHeadersButton:SetText("Show Headers")
	tab.ShowHeadersButton:SetPoint("TOPLEFT", tab.HeaderTextFontSizeSlider, "TOPRIGHT", 5, -7)
	tab.ShowHeadersButton:SetVisible(true)
	function tab.ShowHeadersButton.Event:LeftPress()
		if self.HideHeaders then
			RCT:SetAllHeaderVisibilities(false)
			self:SetText("Show Headers")
			self.HideHeaders = false
		else
			RCT:SetAllHeaderVisibilities(true)
			self:SetText("Hide Headers")
			self.HideHeaders = true
		end
	end
	
	-- Alert Footer Font Size Slider
	tab.FooterTextFontSizeSlider = CCD.UI.CreateSlider(description .. ".FooterTextFontSizeSlider", tab)
	tab.FooterTextFontSizeSlider.Label:SetText("Footer Font:")
	tab.FooterTextFontSizeSlider.Label:SetWidth(sliderXOffset)
	tab.FooterTextFontSizeSlider:SetRange(CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max, 18)
	tab.FooterTextFontSizeSlider:SetPoint("TOPLEFT", tab.HeaderTextFontSizeSlider, "BOTTOMLEFT", 0, 0)
	function tab.FooterTextFontSizeSlider.Slider.Event:SliderChange()
		local pos = tab.FooterTextFontSizeSlider.Slider:GetPosition()
		RCT:SetAllFooterTextFontSizes(pos)
	end

	tab.ShowFootersButton = UI.CreateFrame("RiftButton", description .. ".ShowFootersButton", tab)
	tab.ShowFootersButton:SetText("Show Footers")
	tab.ShowFootersButton:SetPoint("TOPLEFT", tab.FooterTextFontSizeSlider, "TOPRIGHT", 5, -7)
	tab.ShowFootersButton:SetVisible(true)
	function tab.ShowFootersButton.Event:LeftPress()
		if self.HideFooters then
			RCT:SetAllFooterVisibilities(false)
			self:SetText("Show Footers")
			self.HideFooters = false
		else
			RCT:SetAllFooterVisibilities(true)
			self:SetText("Hide Footers")
			self.HideFooters = true
		end
	end
	
	-- Alert Alpha Slider
	tab.GlobalAlertAlphaSlider = CCD.UI.CreateSlider(description .. ".GlobalAlertAlphaSlider", tab)
	tab.GlobalAlertAlphaSlider.Label:SetText("Opacity:")
	tab.GlobalAlertAlphaSlider.Label:SetWidth(sliderXOffset)
	tab.GlobalAlertAlphaSlider:SetRange(0, 100, 100)
	tab.GlobalAlertAlphaSlider:SetPoint("TOPLEFT", tab.FooterTextFontSizeSlider, "BOTTOMLEFT", 0, 0)
	function tab.GlobalAlertAlphaSlider.Slider.Event:SliderChange()
		local pos = tab.GlobalAlertAlphaSlider.Slider:GetPosition()
		RCT:SetAllAlphasFromSlider(pos)
	end

	-- Status Text
	-- Label
	tab.StatusTextLabel = UI.CreateFrame("Text", description .. "_StatusTextLabel", tab)
	tab.StatusTextLabel:SetText("Cooldown Footers:")
	tab.StatusTextLabel:SetFontSize(15)
	tab.StatusTextLabel:SetPoint("TOPLEFT", tab.GlobalAlertAlphaSlider, "BOTTOMLEFT", 0, 5)
	local anchor = tab.StatusTextLabel
	tab.StatusTextFields = {}
	for i, text in ipairs(RCT.StatusText) do
		-- Text Fields
		tab.StatusTextFields[i] = EnKai.uiCreateFrame("nkTextField", description .. "_" .. text .. "_nkTextField", tab)
		tab.StatusTextFields[i]:SetText(text)
		tab.StatusTextFields[i]:SetPoint("TOPLEFT", anchor, "TOPRIGHT", 5, 0)
		local textField = tab.StatusTextFields[i]
		function textField:OnTextFieldChanged()
			RCT.StatusText[i] = textField:GetText()
		end
		Command.Event.Attach(EnKai.events[tab.StatusTextFields[i]:GetName()].TextfieldChanged, textField.OnTextFieldChanged, textField:GetName() .. "_nkTextfieldChanged")
		anchor = tab.StatusTextFields[i]
	end

	-- Footer Note
	tab.RCTFooterNote = CCD.UI.CreateText(tab:GetName() .. ".RCTFooterNote", tab)
	tab.RCTFooterNote:SetText("*Press enter after typing to save all text fields.", 15, Colors.Yellow)
	tab.RCTFooterNote:SetPoint("TOPLEFT", tab.StatusTextFields[4], "TOPRIGHT", 2, 0)

	-- SoonStartTime Label
	tab.SoonStartTimeLabel = UI.CreateFrame("Text", description .. "_SoonStartTimeLabel", tab)
	tab.SoonStartTimeLabel:SetText("Flag cooldowns as being close to ready at:")
	tab.SoonStartTimeLabel:SetFontSize(15)
	tab.SoonStartTimeLabel:SetPoint("TOPLEFT", tab.StatusTextLabel, "BOTTOMLEFT", 0, 5)

	-- SoonStartTime TextField
	tab.SoonStartTimeTextField = EnKai.uiCreateFrame("nkTextField", description .. "_SoonStartTimeTextField", tab)
	tab.SoonStartTimeTextField:SetValue("valueType", "number")
	tab.SoonStartTimeTextField:SetText(RCT.SoonReadyStartTime)
	tab.SoonStartTimeTextField:SetPoint("TOPLEFT", tab.SoonStartTimeLabel, "TOPRIGHT", 5, 0)
	function tab.SoonStartTimeTextField:OnTextFieldChanged()
		local value = tab.SoonStartTimeTextField:GetConvertedText()

		if value then
			RCT.SoonReadyStartTime = tab.SoonStartTimeTextField:GetConvertedText()
			if CCD.UI.EditModeOn then
				for _, frame in pairs(RCT.DamageAlertsByGroup) do
					frame.TimerText:SetText(tostring(RCT.SoonReadyStartTime))
				end
			end
		else
			tab.SoonStartTimeTextField:SetText(RCT.SoonReadyStartTime)
		end
	end
	Command.Event.Attach(EnKai.events[tab.SoonStartTimeTextField:GetName()].TextfieldChanged, tab.SoonStartTimeTextField.OnTextFieldChanged, tab.SoonStartTimeTextField:GetName() .. "_nkTextfieldChanged")

	-- SoonStartTime Label2
	tab.SoonStartTimeLabel2 = UI.CreateFrame("Text", description .. "_SoonStartTimeLabel2", tab)
	tab.SoonStartTimeLabel2:SetText("seconds")
	tab.SoonStartTimeLabel2:SetFontSize(15)
	tab.SoonStartTimeLabel2:SetPoint("TOPLEFT", tab.SoonStartTimeTextField, "TOPRIGHT", 2, 0)

	-- Cooldown Bar Options
	-- Show Check Box
	tab.ShowCooldownBarsCheckbox = CCD.UI.CreateCheckbox(description .. ".ShowCooldownBarsCheckbox", tab, true)
	tab.ShowCooldownBarsCheckbox:SetText("Show Cooldown Bars:", 15, Colors.White)
	tab.ShowCooldownBarsCheckbox:SetPoint("TOPLEFT", tab.SoonStartTimeLabel, "BOTTOMLEFT", 0, 10)
	function tab.ShowCooldownBarsCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT:SetBarVisibility(checked, self)
		RCT.HealBarManager:SetEnabled(checked)
	end

	-- Combat Only Check Box
	tab.ShowCooldownBarsInCombatCheckbox = CCD.UI.CreateCheckbox(description .. ".ShowCooldownBarsInCombatCheckbox", tab, true)
	tab.ShowCooldownBarsInCombatCheckbox:SetText("In Combat Only,", 15, Colors.White)
	tab.ShowCooldownBarsInCombatCheckbox:SetPoint("TOPLEFT", tab.ShowCooldownBarsCheckbox, "TOPRIGHT")
	function tab.ShowCooldownBarsInCombatCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT:SetBarVisibilityInCombat(checked, self)
		RCT.HealBarManager:SetCombatOnly(checked)
	end

	-- GSV Text Field
	tab.CooldownBarsGSVTextField = CCD.UI.CreateTextField(description .. ".CooldownBarsGSVTextField", tab, "In groups of", " or more")
	tab.CooldownBarsGSVTextField.TextField:SetWidth(40)
	tab.CooldownBarsGSVTextField:SetPoint("TOPLEFT", tab.ShowCooldownBarsInCombatCheckbox, "TOPRIGHT")
	tab.CooldownBarsGSVTextField:SetTextFieldChangedEvent(function ()
		local value = tab.CooldownBarsGSVTextField:GetConvertedText()

		if value then
			RCT:SetAllBarGroupShowValues(value, tab.CooldownBarsGSVTextField.TextField)
			RCT.HealBarManager:SetGSV(value)
		end
	end, tab.CooldownBarsGSVTextField:GetName() .. ".OnTextFieldChanged")

	-- Bar Length Slider
	tab.CooldownBarLengthSlider = CCD.UI.CreateSlider(description .. ".CooldownBarLengthSlider", tab)
	tab.CooldownBarLengthSlider.Label:SetText("Bar Length:")
	tab.CooldownBarLengthSlider:SetRange(RCT.DamageBarLengthRange.Min, RCT.DamageBarLengthRange.Max, RCT.DamageBarLength)
	tab.CooldownBarLengthSlider:SetPoint("TOPLEFT", tab.ShowCooldownBarsCheckbox, "BOTTOMLEFT", 0, 5)
	function tab.CooldownBarLengthSlider.Slider.Event:SliderChange()
		local pos = tab.CooldownBarLengthSlider.Slider:GetPosition()
		RCT:SetBarLength(pos, self)
		RCT.HealBarManager:SetLength(pos)
	end

	-- Bar Height Slider
	tab.CooldownBarHeightSlider = CCD.UI.CreateSlider(description .. ".CooldownBarHeightSlider", tab)
	tab.CooldownBarHeightSlider.Label:SetText("Bar Height:")
	tab.CooldownBarHeightSlider:SetRange(RCT.CooldownBarHeightRange.Min, RCT.CooldownBarHeightRange.Max, RCT.DamageBarHeight)
	tab.CooldownBarHeightSlider:SetPoint("TOPLEFT", tab.CooldownBarLengthSlider, "BOTTOMLEFT", 0, 0)
	function tab.CooldownBarHeightSlider.Slider.Event:SliderChange()
		local pos = tab.CooldownBarHeightSlider.Slider:GetPosition()
		RCT:SetBarHeight(pos, self)
		RCT.HealBarManager:SetHeight(pos)
	end

	-- Bar FontSize Slider
	tab.CooldownBarFontSizeSlider = CCD.UI.CreateSlider(description .. ".CooldownBarFontSizeSlider", tab)
	tab.CooldownBarFontSizeSlider.Label:SetText("Bar Font Size:")
	tab.CooldownBarFontSizeSlider:SetRange(RCT.CooldownBarFontSizeRange.Min, RCT.CooldownBarFontSizeRange.Max, RCT.DamageBarFontSize)
	tab.CooldownBarFontSizeSlider:SetPoint("TOPLEFT", tab.CooldownBarHeightSlider, "BOTTOMLEFT", 0, 0)
	function tab.CooldownBarFontSizeSlider.Slider.Event:SliderChange()
		local pos = tab.CooldownBarFontSizeSlider.Slider:GetPosition()
		RCT:SetBarFontSize(pos, self)
		RCT.HealBarManager:SetFontSize(pos)
	end

	-- Cooldown Bar Opacity Slider
	tab.CooldownBarOpacitySlider = CCD.UI.CreateSlider(description .. ".CooldownBarOpacitySlider", tab)
	tab.CooldownBarOpacitySlider.Label:SetText("Bar Opacity:")
	tab.CooldownBarOpacitySlider:SetRange(0, 100, CCD.Math.Round(RCT.GlobalCooldownBarOpacity * 100))
	tab.CooldownBarOpacitySlider:SetPoint("TOPLEFT", tab.CooldownBarFontSizeSlider, "BOTTOMLEFT", 0, 0)
	function tab.CooldownBarOpacitySlider.Slider.Event:SliderChange()
		local pos = tab.CooldownBarOpacitySlider.Slider:GetPosition()
		RCT:SetGlobalCooldownBarOpacity(pos)
		RCT.HealBarManager:SetOpacity(pos)
	end

	--[[ Bar Background Opacity Slider -- removed in 3.2.4
	tab.CooldownBarBackgroundOpacitySlider = CCD.UI.CreateSlider(description .. ".CooldownBarBackgroundOpacitySlider", tab)
	tab.CooldownBarBackgroundOpacitySlider.Label:SetText("Bar Background Opacity:")
	tab.CooldownBarBackgroundOpacitySlider:SetRange(0, 100, math.floor(RCT.CooldownBarBackgroundOpacity * 100))
	tab.CooldownBarBackgroundOpacitySlider:SetPoint("TOPLEFT", tab.CooldownBarOpacitySlider, "BOTTOMLEFT", 0, 0)
	function tab.CooldownBarBackgroundOpacitySlider.Slider.Event:SliderChange()
		local sliderPosition = tab.CooldownBarBackgroundOpacitySlider.Slider:GetPosition()
		RCT:SetBarBackgroundOpacity(sliderPosition)
	end--]]

	-- How RCT Works Label
	tab.RCTGuideLabel = CCD.UI.CreateText(tab:GetName() .. ".RCTGuideLabel", tab)
	tab.RCTGuideLabel:SetText("How the Cooldown Tracker Works", 18, CCD.UI.Color.Green())
	tab.RCTGuideLabel:SetPoint("TOPLEFT", tab.CooldownBarOpacitySlider, "BOTTOMLEFT", 0, 8)
	tab.RCTGuideLabel:SetVisible(true)

	local instructions = {}
	table.insert(instructions, "1. Cooldowns are grouped as follows: (1) Flaring Power, (2) Orchestra of the Planes, Defend the Fallen and Lava Field, (3) Verse of Joy and Resuscitating Chant, (4) Wild Growth and Power Core, (5) Enrage and (6) Command to Attack.")
	table.insert(instructions, "2. Each group (not cooldown) has its own frame.")
	table.insert(instructions, "3. Cooldowns are only tracked after the first cast. A player's cooldowns will be removed whenever that player changes spec.")
	table.insert(instructions, "4. Cooldowns effectively have 3 states: Active, Ready and Down. The cooldown closest to Active will always be shown first.")
	table.insert(instructions, "5. There is no option to have cooldowns of the same group in separate frames.")

	local lastLineText = tab.RCTGuideLabel
	tab.RCTGuideLines = {}
	for i, text in ipairs(instructions) do
		tab.RCTGuideLines[i] = UI.CreateFrame("Text", description .. "_RCTGuideLines#" .. i, tab)
		tab.RCTGuideLines[i]:SetText(text)
		tab.RCTGuideLines[i]:SetWordwrap(true)
		tab.RCTGuideLines[i]:SetWidth(TabWidth - 5) -- Tabs are only assigned a width after being added to Tab View
		tab.RCTGuideLines[i]:SetEffectGlow({
			colorA = 0.8, colorB = 0, colorG = 0, colorR = 0,
			offsetX = 0, offsetY = 0,
			blurX = 1, blurY = 1,
			knockout = false,
			replace = false,
			strength = 3
		})
		tab.RCTGuideLines[i]:SetFontSize(15)
		tab.RCTGuideLines[i]:SetFontColor(1, 1, 1) -- white
		--if i == 1 then
			tab.RCTGuideLines[i]:SetPoint("TOPLEFT", lastLineText, "BOTTOMLEFT")
		--else
		--	tab.RCTGuideLines[i]:SetPoint("TOPLEFT", lastLineText, "TOPLEFT", 0, lasttab.RCTGuideLines[i]:GetHeight())
		--end
		tab.RCTGuideLines[i]:SetVisible(true)

		lastLineText = tab.RCTGuideLines[i]
	end

	return RCT.Tab
end

function RCT:GetAllCooldownInfo()
	local myCooldowns = RCT.AllCooldownsByCaster[CCD.Me]
	local cooldownRemsByAbilityId = {}

	for _, rcd in pairs(myCooldowns) do
		if rcd.AbilityId then
			local details = Inspect.Ability.New.Detail(rcd.AbilityId)
			cooldownRemsByAbilityId[rcd.AbilityId] = rcd:GetGreyTime() or 0
		else
			CCD.Modules.ErrorManager:AddError("RCT:GetAllCooldownInfo()", string.format("%s of type %s has no ability ID.", rcd.Name, rcd.Type))
		end
	end

	return cooldownRemsByAbilityId
end
function RCT:AddCooldown(Cooldown)
	-- All
	RCT.AllCooldownsByType[Cooldown.BuffType][Cooldown.Caster.Id] = Cooldown
	table.insert(RCT.AllCooldownsByGroup[Cooldown.Group], Cooldown)
	RCT.AllCooldownsByCaster[Cooldown.Caster][Cooldown.BuffType] = Cooldown

	-- Damage
	if RCT.CooldownDataCollection[Cooldown.BuffType].Role == RCT.Role.Damage then
		RCT.DamageCooldownsByType[Cooldown.BuffType][Cooldown.Caster.Id] = Cooldown
		table.insert(RCT.DamageCooldownsByGroup[Cooldown.Group], Cooldown)
		RCT.DamageCooldownsByCaster[Cooldown.Caster][Cooldown.BuffType] = Cooldown		
	elseif RCT.CooldownDataCollection[Cooldown.BuffType].Role == RCT.Role.Heal then
		RCT.HealCooldownsByType[Cooldown.BuffType][Cooldown.Caster.Id] = Cooldown
		table.insert(RCT.HealCooldownsByGroup[Cooldown.Group], Cooldown)
		RCT.HealCooldownsByCaster[Cooldown.Caster][Cooldown.BuffType] = Cooldown
		table.insert(RCT.HealCooldowns, Cooldown)
	end

	Cooldown.Role = RCT.CooldownDataCollection[Cooldown.BuffType].Role
end

-- Cooldown Removal
function RCT:RemoveAllCooldowns(Ally)
	RCT:RemoveCooldownsFromBigList(Ally)
	RCT:RemoveDamageCooldowns(Ally)
	RCT:RemoveHealCooldowns(Ally)
	Debug:Record("Removed cooldowns from %s", Ally.Name)
	RCT:SetupAlerts()
	RCT:OrganizeDamageBars()
	RCT:OrganizeHealBars()
end
function RCT:RemoveCooldownsFromBigList(Ally) -- Removes cooldowns from RCT.AllCooldowns-ByType/ByGroup/ByCaster
	for buffType, cooldownsByAllyId in pairs(self.AllCooldownsByType) do
		if cooldownsByAllyId[Ally.Id] then
			--print(string.format("Removed %s's %s", Ally.Name, cooldownsByAllyId[Ally.Id].Buff.Name))
			cooldownsByAllyId[Ally.Id] = nil
		end
	end

	for _, rcds in ipairs(self.AllCooldownsByGroup) do
		for i = #rcds, 1, -1 do -- iterate reverse table indexed numerically
			local rcd = rcds[i]
			--print(string.format("Checking %s's %s", rcd.Caster.Name, rcd.Name))
			if rcd.Caster == Ally then
				--print(string.format("Setting %s's %s to nil", rcd.Caster.Name, rcd.Name))
				table.remove(rcds, i) -- using rcds[i] = nil caused table to empty
				if rcd.Tag then Library.LibCron.delete(rcd.Tag) end
				if rcd.Bar then rcd.Bar = nil end
			end
		end
	end

	for buffType, rcd in pairs(self.AllCooldownsByCaster[Ally]) do
		self.AllCooldownsByCaster[Ally][buffType] = nil
	end
end
function RCT:RemoveDamageCooldowns(Ally)
	for buffType, cooldownsByAllyId in pairs(self.DamageCooldownsByType) do
		if cooldownsByAllyId[Ally.Id] then
			--print(string.format("Removed %s's %s", Ally.Name, cooldownsByAllyId[Ally.Id].Buff.Name))
			cooldownsByAllyId[Ally.Id] = nil
		end
	end

	for _, rcds in pairs(self.DamageCooldownsByGroup) do
		for i = #rcds, 1, -1 do -- iterate reverse table indexed numerically
			local rcd = rcds[i]
			--print(string.format("Checking %s's %s", rcd.Caster.Name, rcd.Name))
			if rcd.Caster == Ally then
				--print(string.format("Setting %s's %s to nil", rcd.Caster.Name, rcd.Name))
				table.remove(rcds, i) -- using rcds[i] = nil caused table to empty
				if rcd.Tag then Library.LibCron.delete(rcd.Tag) end
				if rcd.Bar then rcd.Bar = nil end
			end
		end
	end

	for buffType, rcd in pairs(self.DamageCooldownsByCaster[Ally]) do
		self.DamageCooldownsByCaster[Ally][buffType] = nil
	end

	for _, frame in pairs(RCT.DamageAlertsByGroup) do
		if frame.RCD then
			if frame.RCD.Caster == Ally then
				frame.RCD = nil
				frame:Hide()
			end
		end
	end

	--Debug:Record("Removed cooldowns from: %s", Ally.Name)
	--print(string.format("Removed cooldowns from: %s", Ally.Name))
end
function RCT:RemoveHealCooldowns(Ally)
	for buffType, cooldownsByAllyId in pairs(self.HealCooldownsByType) do
		if cooldownsByAllyId[Ally.Id] then
			--print(string.format("Removed %s's %s", Ally.Name, cooldownsByAllyId[Ally.Id].Buff.Name))
			cooldownsByAllyId[Ally.Id] = nil
			RCT.AllCooldownsByType[buffType][Ally.Id] = nil
		end
	end

	for _, rcds in pairs(self.HealCooldownsByGroup) do
		for i = #rcds, 1, -1 do -- iterate reverse table indexed numerically
			local rcd = rcds[i]
			--print(string.format("Checking %s's %s", rcd.Caster.Name, rcd.Name))
			if rcd.Caster == Ally then
				--print(string.format("Setting %s's %s to nil", rcd.Caster.Name, rcd.Name))
				table.remove(rcds, i) -- using rcds[i] = nil caused table to empty
				if rcd.Tag then Library.LibCron.delete(rcd.Tag) end
				if rcd.Bar then rcd.Bar = nil end
			end
		end
	end

	for buffType, rcd in pairs(self.HealCooldownsByCaster[Ally]) do
		self.HealCooldownsByCaster[Ally][buffType] = nil
	end

	for _, rcds in ipairs(self.HealCooldowns) do
		for i = #rcds, 1, -1 do -- iterate reverse table indexed numerically
			local rcd = rcds[i]
			--print(string.format("Checking %s's %s", rcd.Caster.Name, rcd.Name))
			if rcd.Caster == Ally then
				--print(string.format("Setting %s's %s to nil", rcd.Caster.Name, rcd.Name))
				table.remove(rcds, i) -- using rcds[i] = nil caused table to empty
				if rcd.Tag then Library.LibCron.delete(rcd.Tag) end
				if rcd.Bar then rcd.Bar = nil end
			end
		end
	end

	for _, frame in ipairs(RCT.HealAlerts) do
		if frame.RCD then
			if frame.RCD.Caster == Ally then
				frame.RCD = nil
				frame:Hide()
			end
		end
	end

	--Debug:Record("Removed cooldowns from: %s", Ally.Name)
	--print(string.format("Removed cooldowns from: %s", Ally.Name))
end
function RCT:RemoveAllHealCooldowns()
	for _, ally in pairs(CCD.Allies.ValuesById) do
		RCT:RemoveHealCooldowns(ally)
	end
end

function RCT:DisableCooldown(BuffType)
	local data = RCT.CooldownDataCollection[BuffType]
	
	if data then
		data.Enabled = false
		RCT.CooldownDataCollection[data.AbilityId].Enabled = false
		RCT:OrganizeHealAlerts()
		RCT:OrganizeHealBars()
	end
end
function RCT:EnableCooldown(BuffType)
	local data = RCT.CooldownDataCollection[BuffType]
	
	if data then
		data.Enabled = true
		RCT.CooldownDataCollection[data.AbilityId].Enabled = true
		RCT:OrganizeHealAlerts()
		RCT:OrganizeHealBars()
	end
end

-- Global Alert
function RCT:SetAllIconSizes(Scale)
	for _, frame in pairs(RCT.DamageAlertsByGroup) do
		frame:SetIconSize(Scale)
	end
	
	RCT.HealAlertAnchor.SuperFunctions:SetIconSize(Scale)
end
function RCT:SetAllTimerTextSizes(Size)
	for _, frame in pairs(RCT.DamageAlertsByGroup) do
		frame:SetTimerTextFontSize(Size)
	end
	
	 RCT.HealAlertAnchor.SuperFunctions:SetTimerTextFontSize(Size)
end
function RCT.SetAllHeaderTextLengths(Length)
	for _, frame in pairs(RCT.DamageAlertsByGroup) do
		frame:SetHeaderTextLength(Length)
	end
	
	RCT.HealAlertAnchor.SuperFunctions:SetHeaderTextLength(Length)
end
function RCT:SetAllHeaderTextFontSizes(Size)
	for _, frame in pairs(RCT.DamageAlertsByGroup) do
		frame:SetHeaderTextFontSize(Size)
	end
	
	RCT.HealAlertAnchor.SuperFunctions:SetHeaderTextFontSize(Size)
end
function RCT:SetAllFooterTextFontSizes(Size)
	for _, frame in pairs(RCT.DamageAlertsByGroup) do
		frame:SetFooterTextFontSize(Size)
	end
	
	RCT.HealAlertAnchor.SuperFunctions:SetFooterTextFontSize(Size)
end
function RCT:SetAllAlphasFromSlider(Position)
	for _, frame in pairs(RCT.DamageAlertsByGroup) do
		frame:SetAlphaFromSlider(Position)
	end
	
	RCT.HealAlertAnchor.SuperFunctions:SetAlphaFromSlider(Position)
end
function RCT:SetAllHeaderVisibilities(IsVisible)
	for _, frame in pairs(RCT.DamageAlertsByGroup) do
		frame:SetHeaderVisibility(IsVisible)
	end
	
	RCT.HealAlertAnchor.SuperFunctions:SetHeaderVisibility(IsVisible)
end
function RCT:SetAllFooterVisibilities(IsVisible)
	for _, frame in pairs(RCT.DamageAlertsByGroup) do
		frame:SetFooterVisibility(IsVisible)
	end
	
	RCT.HealAlertAnchor.SuperFunctions:SetFooterVisibility(IsVisible)
end

-- Damage Alerts
function RCT:CreateDamageAlert(Group)
	local id = string.format("%s.Frame%d", RCT.Id, Group) -- Examples RCT.Frame1, RCT.Frame5
	local frame = CCD.UI.AlertFrame.Create(id, CCD.UI.Context)

	--frame:SetBackgroundColor(Colors.Black)
	frame.Group = Group
	-- default settings
	frame.IsTracked = true
	frame.ShowDamageBar = true
	-- Locks
	frame.TimerTextFontSizeLocked = false
	--
	function frame:OnGroupShowValueChanged()
		frame:Update()
	end
	function frame:Update()
		-- Mobility
		if RCT.On and RCT.ShowAlerts and frame.IsTracked then
			frame.IsMobile = true
		else
			frame.IsMobile = false
		end

		if RCT:ShouldShowAlert(self) and self.RCD then
			local RCD = self.RCD

			if RCD.Status == CCD.RCDStatus.Ready then
				self:SetFontColor(RCT.StatusColors[RCD.Status])
				self.Texture:SetTexture("Rift", RCD.Icon)
				self.TimerText:SetVisible(false)
				self:SetHeaderText(RCD.Caster.Name)
				self.Footer.Text:SetText(RCT.StatusText[RCD.Status])
				self:SetVisible(true)
			end
		elseif not CCD.UI.EditModeOn then
			self:Hide()
		end
	end
	function frame:SetTracked(Bool)
		if Bool ~= nil then	self.IsTracked = Bool end
		if RCT.On then
			self.IsMobile = self.IsTracked
		end
		self:UpdateForeground()
	end
	function frame:SetCombatShow(Status)
		self.CombatShow = Status
		self:UpdateForeground()
		self:Update()
	end
	function frame:UpdateForeground()
		if self.IsTracked then
			if self.CombatShow then
				self:SetForegroundColor(CCD.UI.Colors.LightRedTint)
			else
				self:SetForegroundColor(CCD.UI.Colors.ForegroundTint)
			end
		else
			self:SetForegroundColor(CCD.UI.Colors.ForegroundBlackTint)
		end
	end
	function frame:SetShowCooldownBars(Value)
		assert(type(Value) == "boolean", "Parameter 2 (Value) must be boolean!")
		self.ShowDamageBar = Value
	end

	function frame:OnIconSizeChanged()
		if RCT.ContextMenu.Group == frame.Group then
			RCT.ContextMenu.IconSizeSlider:SetPosition(self.IconSize, true)
		end
	end
	function frame:OnTimerTextFontSizeChanged()
		if RCT.ContextMenu.Group == frame.Group then
			RCT.ContextMenu.TimerTextFontSizeSlider:SetPosition(self.TimerTextFontSize, true)
		end
	end

	-- Edit Bolt
	frame.EditBolt:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	function frame.EditBolt.Event:RightClick()
		if RCT.ContextMenu.Group == frame.Group then
			RCT.ContextMenu:Hide()
			RCT.ContextMenu.Group = -1
		else
			RCT.ContextMenu.Group = frame.Group
			RCT.ContextMenu:Show(frame)
		end
	end
	
	return frame
end
function RCT.CreateDamageAlertMenu()
	-- Context Menu
	local id = RCT.Id .. ".ContextMenu"
	local frame = CCD.UI.CreateContextMenu(RCT.Id .. ".ContextMenu", 325, 643)

	-- Enable Alert Checkbox
	frame.EnableAlertCheckbox = frame:AddCheckbox(id .. ".EnableAlertCheckbox", "Enable Alert", false)
	function frame.EnableAlertCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.DamageAlertsByGroup[frame.Group]:SetTracked(checked)
	end--]]

	-- Show Cooldown Bars... Checkbox
	frame.ShowCooldownBarsCheckbox = frame:AddCheckbox(id .. ".ShowCooldownBarsCheckbox", "Enable Cooldown Bars for this group", false)
	function frame.ShowCooldownBarsCheckbox.Event:CheckboxChange()
		RCT.DamageAlertsByGroup[frame.Group]:SetShowCooldownBars(self:GetChecked())
	end

	-- Combat Show Checkbox
	frame.CombatShowCheckbox = frame:AddCheckbox(id .. ".CombatShowCheckbox", "Only show while in combat")
	function frame.CombatShowCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.DamageAlertsByGroup[frame.Group]:SetCombatShow(checked)
	end

	-- GSV Text Field
	frame.GSVTextField = frame:AddTextField(frame:GetName() .. ".GSVTextField", "Only show in groups of", "or more", true)
	function frame.GSVTextField.OnTextFieldChanged()
		local selectedFrame = RCT.DamageAlertsByGroup[frame.Group]
		local value = frame.GSVTextField:GetConvertedText()

		if value then
			selectedFrame:SetGroupShowValue(value)
		end
	end
	frame.GSVTextField:SetTextFieldChangedEvent(frame.GSVTextField.OnTextFieldChanged, frame.GSVTextField:GetName() .. ".OnTextFieldChanged")

	-- Opacity Slider
	frame.AlphaSlider = frame:AddSlider(id .. ".AlphaSlider", "Opacity:", 0, 100)
	function frame.AlphaSlider.Slider.Event:SliderChange()
		local sliderPosition = self:GetPosition()
		RCT.DamageAlertsByGroup[frame.Group]:SetAlphaFromSlider(sliderPosition)
	end

	-- Icon Size Slider
	frame.IconSizeSlider = frame:AddSlider(id .. ".IconSizeSlider", "Icon Size:", CCD.UI.AlertFrame.IconSizeRange.Min, CCD.UI.AlertFrame.IconSizeRange.Max, false)
	function frame.IconSizeSlider.Slider.Event:SliderChange()
		RCT.DamageAlertsByGroup[frame.Group]:SetIconSize(self:GetPosition())
	end

	-- Timer Font Size Slider
	frame.TimerTextFontSizeSlider = frame:AddSlider(id .. ".TimerTextFontSizeSlider", "Timer Font Size:", CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max, false)
	function frame.TimerTextFontSizeSlider.Slider.Event:SliderChange()
		local size = self:GetPosition()
		RCT.DamageAlertsByGroup[frame.Group]:SetTimerTextFontSize(size)
	end

	-- Timer Font Size Locked Check Box
	frame.TimerTextFontSizeLockedCheckbox = frame:AddCheckbox(id .. ".TimerTextFontSizeLockedCheckbox", "Lock Timer Font Size", true)
	function frame.TimerTextFontSizeLockedCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.DamageAlertsByGroup[frame.Group].TimerTextFontSizeLocked = checked
	end

	-- Show Header Check Box
	frame.ShowHeaderCheckbox = frame:AddCheckbox(id .. ".ShowHeaderCheckbox", "Show Header", false)
	function frame.ShowHeaderCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.DamageAlertsByGroup[frame.Group]:SetHeaderVisibility(checked)
		frame.HeaderTextLengthSlider:SetVisible(checked)
		frame.HeaderTextFontSizeSlider:SetVisible(checked)
		frame.HeaderTextFontSizeSlider.Line:SetVisible(checked)

		if checked then
			frame.ShowFooterCheckbox:AnchorTo("ORIGINAL")
		else
			frame.ShowFooterCheckbox:AnchorTo(frame.ShowHeaderCheckbox, 0)
		end
	end

	-- Header Text Length Slider
	frame.HeaderTextLengthSlider = frame:AddSlider(id .. ".HeaderTextLengthSlider", "Header Length:", 3, 19)
	function frame.HeaderTextLengthSlider.Slider.Event:SliderChange()
		RCT.DamageAlertsByGroup[frame.Group]:SetHeaderTextLength(self:GetPosition())
	end

	-- Header Font Size Slider
	frame.HeaderTextFontSizeSlider = frame:AddSlider(id .. ".HeaderTextFontSizeSlider", "Header Font Size:", CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max, true)
	function frame.HeaderTextFontSizeSlider.Slider.Event:SliderChange()
		local size = self:GetPosition()
		RCT.DamageAlertsByGroup[frame.Group]:SetHeaderTextFontSize(size)
	end

	-- Show Footer Check Box
	frame.ShowFooterCheckbox = frame:AddCheckbox(id .. ".ShowFooterCheckbox", "Show Footer", false)
	function frame.ShowFooterCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.DamageAlertsByGroup[frame.Group]:SetFooterVisibility(checked)

		frame.FooterTextFontSizeSlider:SetVisible(checked)

		if checked then
			frame.FooterTextFontSizeSlider.Line:AnchorTo("ORIGINAL")
		else
			frame.FooterTextFontSizeSlider.Line:AnchorTo(frame.ShowFooterCheckbox, 0)
		end
	end

	-- Footer Font Size Slider
	frame.FooterTextFontSizeSlider = frame:AddSlider(id .. ".FooterTextFontSizeSlider", "Footer Font Size:", CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max, true)
	function frame.FooterTextFontSizeSlider.Slider.Event:SliderChange()
		local size = self:GetPosition()
		RCT.DamageAlertsByGroup[frame.Group]:SetFooterTextFontSize(size)
	end


	-- Global Header
	frame.GlobalOptionsLabel = frame:AddHeader(id .. ".GlobalOptionsLabel", "Global Options")

	-- On/Off Checkbox
	frame.OnCheckbox = frame:AddCheckbox(id .. ".OnCheckbox", "Enable Cooldown Tracker", false)
	function frame.OnCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.Set(checked)
		RCT.Tab.OnCheckbox:SetChecked(RCT.On)
	end

	frame.ShowAlertsCheckbox = frame:AddCheckbox(id .. ".ShowAlertsCheckbox", "Enable Damage Alerts", false)
	function frame.ShowAlertsCheckbox.Event:CheckboxChange()
		RCT:SetShowAlerts(self:GetChecked(), self)
	end

	-- Show Down Timers Checkbox
	frame.ShowInertTimersCheckbox = frame:AddCheckbox(id .. "ShowInertTimersCheckbox", "Show Alerts for DOWN cooldowns", false)
	function frame.ShowInertTimersCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.ShowInertTimers = checked
		RCT.Tab.InertTimersCheckbox:SetChecked(RCT.ShowInertTimers)
	end

	-- Use Greyscale Checkbox
	frame.UseGreyscaleCheckbox = frame:AddCheckbox(id .. ".UseGreyscaleCheckbox", "Use Greyscale for DOWN cooldowns", true)
	function frame.UseGreyscaleCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.UseGreyscaleForInertCooldowns = checked
		RCT.Tab.GreyscaleCheckbox:SetChecked(RCT.UseGreyscaleForInertCooldowns)
	end

	frame.FooterText = frame:AddFooterNote(id .. ".FooterText", "Click here for more options")
	function frame.FooterText.Event:LeftClick()
		CCD.MainWindow:Show()
		CCD.MainWindow.TabView:SetActiveTab(1)
	end

	-- Load Function
	function frame:Load()
		self.EnableAlertCheckbox:SetChecked(RCT.DamageAlertsByGroup[self.Group].IsTracked)
		self.CombatShowCheckbox:SetChecked(RCT.DamageAlertsByGroup[self.Group].CombatShow)
		self.GSVTextField:SetTextField(RCT.DamageAlertsByGroup[self.Group].GroupShowValue)
		self.HeaderText:SetText(RCT.GroupNames[self.Group])
		self.OnCheckbox:SetChecked(RCT.On)
		self.ShowAlertsCheckbox:SetChecked(RCT.ShowAlerts)
		self.ShowInertTimersCheckbox:SetChecked(RCT.ShowInertTimers)
		self.UseGreyscaleCheckbox:SetChecked(RCT.UseGreyscaleForInertCooldowns)
		local rctFrame = RCT.DamageAlertsByGroup[self.Group]
		self.ShowCooldownBarsCheckbox:SetChecked(rctFrame.ShowDamageBar)

		self.IconSizeSlider:SetPosition(rctFrame.IconSize, true)
		self.HeaderTextFontSizeSlider:SetPosition(rctFrame.HeaderTextFontSize, true)
		self.HeaderTextLengthSlider:SetPosition(rctFrame.HeaderTextLength, true)
		self.TimerTextFontSizeSlider:SetPosition(rctFrame.TimerTextFontSize, true)
		self.FooterTextFontSizeSlider:SetPosition(rctFrame.FooterTextFontSize, true)
		self.AlphaSlider:SetPosition(CCD.Math.Round(rctFrame.Alpha * 100), true)

		self.ShowHeaderCheckbox:SetChecked(rctFrame.HeaderOn)
		self.ShowFooterCheckbox:SetChecked(rctFrame.FooterOn)
		self.TimerTextFontSizeLockedCheckbox:SetChecked(rctFrame.TimerTextFontSizeLocked)
	end

	return frame
end
function RCT:UpdateAllDamageAlerts()
	for _, frame in pairs(RCT.DamageAlertsByGroup) do
		frame:Update()
	end
end

function RCT.LinkFrames(Point, Spacing)
	local frames = {}

	for _, frame in pairs(RCT.DamageAlertsByGroup) do
		table.insert(frames, frame)
	end

	if Point == "TOPRIGHT" then
		table.sort(frames, function(a, b) return a.X < b.X end)
	elseif Point == "BOTTOMLEFT" then
		table.sort(frames, function(a, b) return a.Y < b.Y end)
	else return end

	for i, frame in ipairs(frames) do
		frame:Unlink()

		if frames[i-1] then
			frame:Link(frames[i-1], Point, Spacing)
		end
	end
end

function RCT:SetShowAlerts(Value, Source)
	assert(type(Value) == "boolean", "Parameter 2 (Value) must be boolean!")
	RCT.ShowAlerts = Value
	RCT:UpdateAllDamageAlerts()

	-- update check boxes
	local checkboxes = { RCT.Tab.ShowAlertsCheckbox, RCT.ContextMenu.ShowAlertsCheckbox }
	for _, checkbox in ipairs(checkboxes) do
		if not Source or Source ~= checkbox then
			checkbox:SetChecked(RCT.ShowAlerts)
		end
	end
end
function RCT:ShouldShowAlert(Frame)
	if		CCD.UI.EditModeOn then return false
	elseif	not RCT.On then return false
	elseif	not RCT.ShowAlerts then return false
	elseif	not Frame.IsTracked then return false
	elseif	Frame.CombatShow and not CCD.Me.InCombat then return false
	elseif	Frame.GroupShowValue >= 2 and CCD.Allies.Count < Frame.GroupShowValue then return false
	elseif	Frame.RCD and not RCT.ShowInertTimers and Frame.RCD.Status == Status.Inert then return false
	elseif	CCD.DisableInWarfronts and CCD.Me.ZoneType == "warfront" then return false
	else return true end
end
function RCT:SetupAlerts()
	for group, cooldownList in pairs(RCT.DamageCooldownsByGroup) do
		local frame = RCT.DamageAlertsByGroup[group]

		for _, rcd in ipairs(cooldownList) do
			rcd:UpdateStatus()

			-- determine if this rcd should be shown
			if frame.RCD == nil then -- nil suggests that either a cd has just changed from ACTIVE or no cds of this group have been used
				frame.RCD = rcd
				--print(string.format("%s is being set as frame.RCD because frame.RCD was nil", rcd.Name))
			elseif rcd.Status < frame.RCD.Status then
				frame.RCD = rcd
				--print(string.format("%s is being set as frame.RCD because its status is lower than current", rcd.Name))
			elseif rcd.Status == frame.RCD.Status then
				if rcd.Status == CCD.RCDStatus.Active then
					if rcd:GetActiveTime(timestamp) > frame.RCD:GetActiveTime(timestamp) then
						frame.RCD = rcd
						--print(string.format("%s is being set as frame.RCD because both cds are Active and its time is longer", rcd.Name))
					end
				elseif rcd.Status ~= CCD.RCDStatus.Ready then
					if rcd:GetGreyTime(timestamp) < frame.RCD:GetGreyTime(timestamp) then
						frame.RCD = rcd
						--print(string.format("%s is being set as frame.RCD because both cds are NOT ready and its time is closer to ready", rcd.Name))
					end
				elseif rcd.PriorityIndex < frame.RCD.PriorityIndex then
					frame.RCD = rcd
					--print(string.format("%s is being set as frame.RCD because its priority index is lower", rcd.Name))
				end
			end
		end
	end
end
--
function RCT.DamageCooldownTimer_Task(RCD)
	RCD:UpdateStatus()
	local alert = RCT.DamageAlertsByGroup[RCD.Group]
	if RCD.LastStatus == Status.Active and RCD.Status ~= Status.Active then
		-- fixes active alerts fading and being displayed as down even for a fraction of a second even when another cooldown is ready
		RCT:SetupAlerts() 
		if alert.RCD and alert.RCD ~= RCD then
			alert:Update()
		end
	end
	
	local statusColor = RCT.StatusColors[RCD.Status]

	if RCD.Bar then
		if RCT:ShouldShowBars() then
			if RCD.Status == Status.Active then
				RCD.ActiveUpdateValue = RCT.DamageBarLength / RCD.Duration * RCT.CooldownTimerUpdateFrequency -- amount of Bar that is removed per update
				RCD.CurrentFill = (RCT.DamageBarLength * (RCD.Duration - RCD:GetTimeElapsed())) / RCD.Duration
				RCD.Bar:SetPercentage(RCD.CurrentFill / RCD.Bar:GetWidth() * 100)
				RCD.Bar:SetTimerText(tostring(math.floor(RCD:GetActiveTime()))) -- 3.2.3: changed from ceil to floor
			elseif RCD.Status == Status.Inert or RCD.Status == Status.Soon then
				RCD.InertFill = RCT.DamageBarLength / RCD.Cooldown * RCD:GetTimeElapsed()
				RCD.Bar:SetPercentage(RCD.InertFill / RCD.Bar:GetWidth() * 100)
				RCD.Bar:SetTimerText(tostring(math.floor(RCD:GetGreyTime()))) -- 3.2.3: changed from ceil to floor
			elseif RCD.Status == Status.Ready then
				RCD.Bar:SetPercentage(100)
				RCD.Bar:SetTimerText("Ready")
			end
			RCD.Bar:SetColor(statusColor, 0.6)
			RCD.Bar:SetTitle(RCD.Caster.Name)
			RCD.Bar:SetTexture("Rift", RCD.Icon)
			RCD.Bar:Show()
		elseif not CCD.UI.EditModeOn then
			RCD.Bar:Hide()
		end
	end

	if alert.RCD and alert.RCD == RCD then
		if RCT:ShouldShowAlert(alert) then
			if RCD.Status == Status.Active then
				alert.Texture:SetTexture("Rift", RCD.Icon)
				alert.TimerText:SetVisible(true)
			elseif RCD.Status == Status.Inert or RCD.Status == Status.Soon then
				if RCT.UseGreyscaleForInertCooldowns and RCT.CooldownDataCollection[RCD.BuffType].GreyImage then
					alert.Texture:SetTexture("CCD", "Images\\" .. RCT.CooldownDataCollection[RCD.BuffType].GreyImage)
				else
					alert.Texture:SetTexture("Rift", RCD.Icon)
				end
				alert.TimerText:SetVisible(true)
			elseif RCD.Status == Status.Ready then
				alert.Texture:SetTexture("Rift", RCD.Icon)
				alert.TimerText:SetVisible(false)
			end

			alert:SetFontColor(statusColor)
			alert:SetHeaderText(RCD.Caster.Name)
			alert.TimerText:SetText(tostring(math.floor(RCD:GetDisplayTime()))) -- 3.2.3: changed from ceil to floor
			alert.Footer.Text:SetText(RCT.StatusText[RCD.Status])
			alert:SetVisible(true)
		elseif not CCD.UI.EditModeOn then
			alert:Hide()
		end
	end

	if RCD.Status == Status.Ready then -- timer deletion
		Library.LibCron.delete(RCD.Tag)
		RCD.Tag = nil
		--print(string.format("%s's %s timer has been deleted.", RCD.Caster.Name, RCD.Name))
	end
end

-- Damage Bar
function RCT:CreateDamageBarContainer()
	local container = CCD.UI.CreateLinkableFrame("RCT.DamageBarContainer", CCD.UI.Context)
	--Library.LibSimpleWidgets.SetBorder("plain", container, 2, 0, 0, 0, 1, "tblr") -- thin black border	
	
	function container:Resize()
		if CCD.UI.EditModeOn then
			self:SetSize(RCT.DamageBarLength, RCT.DamageBarHeight)
		else
			self:SetSize(RCT.DamageBarLength, RCT.DamageBarHeight * RCT:GetActiveDamageBarCount())
		end
	end
	--
	function container.Event:WheelForward()
		if not CCD.UI.EditModeOn then return end
		RCT:SetBarLength(RCT.DamageBarLength + 25, self)
	end
	function container.Event:WheelBack()
		if not CCD.UI.EditModeOn then return end
		RCT:SetBarLength(RCT.DamageBarLength - 25, self)
	end
	function container.EditBolt.Event:RightClick()
		RCT.CooldownBarContextMenu:Show(self)
	end
	
	return container
end
function RCT:CreateBar()
	local i = RCT.NextCooldownBarIndex
	local id = string.format("%s.CooldownBar%d", RCT.Id, i) -- Examples RCT.CooldownBar1, RCT.CooldownBar5

	local bar = CCD.UI.CreateColorBar(id, RCT.DamageBarContainer, Colors.Green)
	table.insert(RCT.DamageBars, bar)
	bar:SetSize(RCT.DamageBarLength, RCT.DamageBarHeight)
	bar:SetFontSize(RCT.DamageBarFontSize)
	if RCT.DamageBars[i-1] then bar:SetPoint("TOPLEFT", RCT.DamageBars[i-1], "BOTTOMLEFT", 0, 0)
	else bar:SetPoint("TOPLEFT", RCT.DamageBarContainer, "TOPLEFT") end
	bar:Hide()
	bar.Filling:SetAlpha(RCT.GlobalCooldownBarOpacity)
	bar.Icon:SetAlpha(RCT.GlobalCooldownBarOpacity)
	bar:SetBackgroundAlpha(RCT.CooldownBarBackgroundOpacity)

	function bar:Update()
		--if CCD.UI.EditModeOn or not RCT.ShowBars then return end
		if bar.RCD and bar.RCD.Status == Status.Ready then
			bar:SetColor(Colors.Yellow, 0.6)
			bar:SetTitle(bar.RCD.Caster.Name)
			bar:SetTexture("Rift", bar.RCD.Icon)
			bar:SetPercentage(100)
			bar:SetTimerText("Ready")
			if RCT.DamageAlertsByGroup[bar.RCD.Group].ShowDamageBar then
				bar:Show()
			else
				bar:Hide()
			end
		else
			bar:Hide()
		end
	end

	RCT.NextCooldownBarIndex = RCT.NextCooldownBarIndex + 1
	return bar
end
function RCT.CreateDamageBarMenu()
	-- Context Menu
	local id = RCT.Id .. "CooldownBar.ContextMenu"
	local frame = CCD.UI.CreateContextMenu(RCT.Id, 310, 425)

	-- Show Cooldown Bars Checkbox
	frame.ShowBarsCheckbox = frame:AddCheckbox(id .. ".ShowBarsCheckbox", "Enable Damage Bars")
	function frame.ShowBarsCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT:SetBarVisibility(checked, self)
	end

	-- Combat Show Checkbox
	frame.ShowBarsInCombatOnlyCheckbox = frame:AddCheckbox(id .. ".ShowBarsInCombatOnlyCheckbox", "Only show while in combat")
	function frame.ShowBarsInCombatOnlyCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT:SetBarVisibilityInCombat(checked, self)
	end

	-- GSV Text Field
	frame.GSVTextField = frame:AddTextField(frame:GetName() .. ".GSVTextField", "Only show in groups of", "or more", true)
	function frame.GSVTextField.OnTextFieldChanged()
		local value = frame.GSVTextField:GetConvertedText()

		if value then
			RCT:SetAllBarGroupShowValues(value, frame.GSVTextField)
		end
	end
	frame.GSVTextField:SetTextFieldChangedEvent(frame.GSVTextField.OnTextFieldChanged, frame.GSVTextField:GetName() .. ".OnTextFieldChanged")

	-- Length Slider
	frame.FrameSizeSlider = frame:AddSlider(id .. ".FrameSizeSlider", "Length:", RCT.DamageBarLengthRange.Min, RCT.DamageBarLengthRange.Max)
	function frame.FrameSizeSlider.Slider.Event:SliderChange()
		local length = self:GetPosition()
		RCT:SetBarLength(length, self)
	end

	-- Bar Height Slider
	frame.BarHeightSlider = frame:AddSlider(id .. ".BarHeightSlider", "Height:", RCT.CooldownBarHeightRange.Min, RCT.CooldownBarHeightRange.Max)
	function frame.BarHeightSlider.Slider.Event:SliderChange()
		local length = self:GetPosition()
		RCT:SetBarHeight(length, self)
	end

	-- Bar Font Size Slider
	frame.BarFontSizeSlider = frame:AddSlider(id .. ".BarFontSizeSlider", "Font Size:", RCT.CooldownBarFontSizeRange.Min, RCT.CooldownBarFontSizeRange.Max)
	function frame.BarFontSizeSlider.Slider.Event:SliderChange()
		local length = self:GetPosition()
		RCT:SetBarFontSize(length, self)
	end

	-- Cooldown Bar Opacity Slider
	frame.BarOpacitySlider = frame:AddSlider(id .. ".OpacitySlider", "Opacity:", 0, 100)
	function frame.BarOpacitySlider.Slider.Event:SliderChange()
		local pos = self:GetPosition()
		RCT:SetGlobalCooldownBarOpacity(pos)
	end	--]]

	-- Cooldown Bar Background Opacity Slider  -- removed in 3.2.4; readded in 3.3.4
	frame.BarBackgroundOpacitySlider = frame:AddSlider(id .. ".OpacitySlider", "Background Opacity:", 0, 100, true)
	function frame.BarBackgroundOpacitySlider.Slider.Event:SliderChange()
		local pos = self:GetPosition()
		RCT:SetBarBackgroundOpacity(pos, self)
	end	 -- removed in 3.2.4; readded in 3.3.4]]

	-- Global Header
	frame.GlobalOptionsLabel = frame:AddHeader(id .. ".GlobalOptionsLabel", "Global Options")

	-- On/Off Checkbox
	frame.OnCheckbox = frame:AddCheckbox(id .. ".OnCheckbox", "Enable Cooldown Tracker")
	function frame.OnCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.Set(checked)
		RCT.Tab.OnCheckbox:SetChecked(RCT.On)
	end

	-- Reset Bars button
	frame.ResetBarsButton = frame:AddLabel(id .. ".ResetBarsButton", "Reset All Bar Positions", true)
	function frame.ResetBarsButton.Event:LeftClick()
		RCT.ResetCooldownBarContainerPositions()
	end

	frame.FooterText = frame:AddFooterNote(id .. ".FooterText", "Click here for more options")
	function frame.FooterText.Event:LeftClick()
		CCD.MainWindow:Show()
		CCD.MainWindow.TabView:SetActiveTab(1)
	end

	-- Load Function
	function frame:Load()
		self.HeaderText:SetText("Damage Bar Anchor")

		frame.ShowBarsCheckbox:SetChecked(RCT.ShowBars)
		frame.ShowBarsInCombatOnlyCheckbox:SetChecked(RCT.ShowBarsInCombatOnly)
		frame.GSVTextField:SetTextField(RCT.DamageBarGSV)

		self.FrameSizeSlider:SetPosition(RCT.DamageBarLength, true)
		self.BarHeightSlider:SetPosition(RCT.DamageBarHeight, true)
		self.BarFontSizeSlider:SetPosition(RCT.DamageBarFontSize, true)
		self.BarOpacitySlider:SetPosition(CCD.Math.Round(RCT.GlobalCooldownBarOpacity * 100), true)
		self.BarBackgroundOpacitySlider:SetPosition(CCD.Math.Round(RCT.CooldownBarBackgroundOpacity * 100), true) -- Removed in 3.2.4; Readded in 3.3.4
		
		-- Global
		self.OnCheckbox:SetChecked(RCT.On)		
	end

	return frame
end
--
function RCT:SetBarVisibility(Value)
	if Value == true or Value == false then
		RCT.ShowBars = Value
		RCT.DamageBarContainer.IsMobile = RCT.ShowBars
	else
		return
	end

	for _, bar in ipairs(RCT.DamageBars) do
		if RCT:ShouldShowBars() then
			bar:Update()
		--else
			--bar:Hide()
		end
	end

	RCT:UpdateBarForegrounds()
end
function RCT:SetBarVisibilityInCombat(Value)
	if Value == true or Value == false then
		RCT.ShowBarsInCombatOnly = Value
	else
		return
	end

	for _, bar in ipairs(RCT.DamageBars) do
		if RCT:ShouldShowBars() then
			bar:Update()
		--else
			--bar:Hide()
		end
	end

	RCT:UpdateBarForegrounds()
end
function RCT:SetAllBarGroupShowValues(Value)
	if Value then
		-- not using LinkableFrames's SetGroupShowValue()
		RCT.DamageBarGSV = Value
		if not CCD.UI.EditModeOn then
			RCT:UpdateAllBars()
		end
	end
end
function RCT:SetBarLength(Length, Source)
	RCT.DamageBarLength = Length or RCT.DamageBarLength
	RCT.DamageBarLength = CCD.Math.MinMax(RCT.DamageBarLength, RCT.DamageBarLengthRange.Min, RCT.DamageBarLengthRange.Max)
	RCT.DamageBarContainer:Resize()
	
	for _, bar in ipairs(RCT.DamageBars) do
		bar:SetWidth(RCT.DamageBarLength)
		bar:SetPercentage(50)
	end

	--update sliders
	if not RCT.Tab then return end -- Main Window may not be initialized yet
	local sliders = {RCT.Tab.CooldownBarLengthSlider, RCT.CooldownBarContextMenu.FrameSizeSlider}
	for _, slider in ipairs(sliders) do
		if not Source or Source ~= slider then
			slider:SetPosition(RCT.DamageBarLength, true)
		end
	end
end
function RCT:SetBarHeight(Height, Source)
	RCT.DamageBarHeight = Height or RCT.DamageBarHeight
	RCT.DamageBarContainer:Resize()
	
	for _, bar in ipairs(RCT.DamageBars) do
		bar:SetHeight(RCT.DamageBarHeight)
	end

	--update sliders
	if not RCT.Tab then return end -- Main Window may not be initialized yet
	local sliders = {RCT.Tab.CooldownBarHeightSlider, RCT.CooldownBarContextMenu.BarHeightSlider}
	for _, slider in ipairs(sliders) do
		if not Source or Source ~= slider then
			slider:SetPosition(RCT.DamageBarHeight, true)
		end
	end
end
function RCT:SetBarFontSize(FontSize, Source)
	RCT.DamageBarFontSize = FontSize or RCT.DamageBarFontSize

	for _, bar in ipairs(RCT.DamageBars) do
		bar:SetFontSize(RCT.DamageBarFontSize)
	end

	--update sliders
	if not RCT.Tab then return end -- Main Window may not be initialized yet
	local sliders = {RCT.Tab.CooldownBarFontSizeSlider, RCT.CooldownBarContextMenu.BarFontSizeSlider}
	for _, slider in ipairs(sliders) do
		if not Source or Source ~= slider then
			slider:SetPosition(RCT.DamageBarFontSize, true)
		end
	end--]]
end
function RCT:SetBarBackgroundOpacity(Alpha, Source)
	local adjustedValue = Alpha -- Value may need to be rounded (down)

	if adjustedValue and adjustedValue >= 1 then adjustedValue = adjustedValue / 100 end
	RCT.CooldownBarBackgroundOpacity = adjustedValue or RCT.CooldownBarBackgroundOpacity

	for _, bar in ipairs(RCT.DamageBars) do
		bar:SetBackgroundAlpha(RCT.CooldownBarBackgroundOpacity)
	end

	--update sliders
	if not RCT.Tab then return end -- Main Window may not be initialized yet
	local sliders = {RCT.CooldownBarContextMenu.OpacitySlider}
	for _, slider in ipairs(sliders) do
		if not Source or Source ~= slider then
			slider:SetPosition(CCD.Math.Round(RCT.CooldownBarBackgroundOpacity * 100), true)
		end
	end
end--]]
function RCT:SetGlobalCooldownBarOpacity(Value) -- sets the opacity for ALL cooldown bars
	local adjustedValue = Value -- Value may need to be rounded (down)

	if adjustedValue and adjustedValue >= 1 then adjustedValue = adjustedValue / 100 end

	RCT.GlobalCooldownBarOpacity = adjustedValue or RCT.GlobalCooldownBarOpacity

	for _, cooldownBar in ipairs(RCT.DamageBars) do
		cooldownBar.Filling:SetAlpha(RCT.GlobalCooldownBarOpacity)
		cooldownBar.Icon:SetAlpha(RCT.GlobalCooldownBarOpacity)
	end

	if CCD.ActiveContextMenu == RCT.CooldownBarContextMenu then
		local opacityAsSliderValue = CCD.Math.Round(RCT.GlobalCooldownBarOpacity * 100)
		RCT.Tab.CooldownBarOpacitySlider:SetPosition(opacityAsSliderValue, true)
		RCT.CooldownBarContextMenu.BarOpacitySlider:SetPosition(opacityAsSliderValue, true)
	end
end

function RCT:GetNextBar()
	for _, bar in ipairs(RCT.DamageBars) do
		if bar.RCD == nil then return bar end
	end
	
	local bar = RCT:CreateBar()
	
	return bar
end
function RCT:AssignCooldownToBar(RCD, Bar)
	Bar.RCD = RCD
	RCD.Bar = Bar
	Bar:SetTexture("Rift", RCD.Icon)

	if RCD.Status == Status.Ready then
		Bar:SetColor(Colors.Yellow, 0.6)
		Bar:SetTitle(RCD.Caster.Name)
		Bar:SetPercentage(100)
		Bar:SetTimerText("Ready")
	end
	
	if RCT:ShouldShowBars() then
		Bar:Show()
	end
end
function RCT:UnassignAllCooldownBars()
	for _, bar in ipairs(RCT.DamageBars) do
		if bar.RCD then bar.RCD.Bar = nil end -- hopefully fixes the issue with multiple cds using the same bar
		bar.RCD = nil
		bar:Hide()
	end
end
function RCT:OrganizeDamageBars()
	if CCD.UI.EditModeOn then return end
	
	self:UnassignAllCooldownBars()
	local lastBarUsed
	
	for _, rcdList in pairs(self.DamageCooldownsByGroup) do
		for _, rcd in ipairs(rcdList) do
			local alert = RCT.DamageAlertsByGroup[rcd.Group]

			if alert and alert.ShowDamageBar then
				local bar = self:GetNextBar()
				if bar then
					self:AssignCooldownToBar(rcd, bar)
					lastBarUsed = bar
				end
			end
		end
	end
	
	RCT.DamageBarContainer:Resize()
end
function RCT:ShouldShowBars()
	if		CCD.UI.EditModeOn then return false
	elseif	not RCT.On then return false
	elseif	not RCT.ShowBars then return false
	elseif	RCT.ShowBarsInCombatOnly and not CCD.Me.InCombat then return false
	elseif 	RCT.DamageBarGSV >= 2 and CCD.Allies.Count < RCT.DamageBarGSV then return false
	elseif	CCD.DisableInWarfronts and CCD.Me.ZoneType == "warfront" then return false
	else return true end
end
function RCT:HideAllBars()
	for _, bar in ipairs(RCT.DamageBars) do
		bar:Hide()
	end
end
function RCT:UpdateAllBars()
	for _, bar in ipairs(RCT.DamageBars) do
		bar:Update()
	end
end
function RCT:UpdateBarForegrounds()
	local color

	if RCT.ShowBars then
		if RCT.ShowBarsInCombatOnly then
			color = CCD.UI.Colors.LightRedTint
		else
			color = CCD.UI.Colors.ForegroundTint
		end
	else
		color = CCD.UI.Colors.ForegroundBlackTint
	end

	RCT.DamageBarContainer:SetForegroundColor(color)	
end
function RCT:GetActiveDamageBarCount()
	local count = 0
	
	for _, damageBar in ipairs(RCT.DamageBars) do
		if damageBar.RCD then count = count + 1 end
	end
	
	return count
end
--

-- Heal Alerts
function RCT:CreateHealAlertAnchor()
	RCT.HealAlertAnchor = RCT:CreateHealAlert(1)
	RCT.HealAlertAnchor.IsMobile = true
	RCT.HealAlertAnchor.NeverShow = false
	RCT.HealAlertAnchor:Hide()

	local anchor = RCT.HealAlertAnchor
	anchor.SuperFunctions = {}
	--[[	SuperFunctions are used to first set the anchor's settings (which validates the data),
			then the global setting,
			and then the settings for all other alerts]]--

	function anchor.SuperFunctions:SetIconSize(Value, Silent, DontScale)
		anchor:SetIconSize(Value, Silent, DontScale)
		RCT.HealAlertOptions.IconSize = anchor.IconSize
		
		for i = 2, #RCT.HealAlerts, 1 do -- iterate all frames after anchor
			RCT.HealAlerts[i]:SetIconSize(RCT.HealAlertOptions.IconSize, Silent, DontScale)
		end
	end
	function anchor.SuperFunctions:SetTimerTextFontSize(Value)
		anchor:SetTimerTextFontSize(Value)
		RCT.HealAlertOptions.TimerTextFontSize = anchor.TimerTextFontSize
		
		for i = 2, #RCT.HealAlerts, 1 do -- iterate all frames after anchor
			RCT.HealAlerts[i]:SetTimerTextFontSize(Value)
		end
	end
	function anchor.SuperFunctions:SetHeaderVisibility(Checked)
		anchor:SetHeaderVisibility(Checked)
		RCT.HealAlertOptions.HeaderOn = anchor.HeaderOn
		
		for i = 2, #RCT.HealAlerts, 1 do -- iterate all frames after anchor
			RCT.HealAlerts[i]:SetHeaderVisibility(Checked)
		end
	end
	function anchor.SuperFunctions:SetFooterVisibility(Checked)
		anchor:SetFooterVisibility(Checked)
		RCT.HealAlertOptions.FooterOn = anchor.FooterOn
		
		for i = 2, #RCT.HealAlerts, 1 do -- iterate all frames after anchor
			RCT.HealAlerts[i]:SetFooterVisibility(Checked)
		end
	end
	function anchor.SuperFunctions:SetTimerTextFontSizeLock(Checked) -- There is no base function for this but to keep consistency a super fxn was used
		anchor.TimerTextFontSizeLocked = Checked
		RCT.HealAlertOptions.TimerTextFontSizeLocked = anchor.TimerTextFontSizeLocked
		
		for i = 2, #RCT.HealAlerts, 1 do -- iterate all frames after anchor
			RCT.HealAlerts[i].TimerTextFontSizeLocked = Checked
		end
	end
	function anchor.SuperFunctions:SetHeaderTextLength(Length)
		anchor:SetHeaderTextLength(Length)
		RCT.HealAlertOptions.HeaderTextLength = anchor.HeaderTextLength
				
		for i = 2, #RCT.HealAlerts, 1 do -- iterate all frames after anchor
			RCT.HealAlerts[i]:SetHeaderTextLength(RCT.HealAlertOptions.HeaderTextLength)
		end
	end
	function anchor.SuperFunctions:SetHeaderTextFontSize(Size)
		anchor:SetHeaderTextFontSize(Size)
		RCT.HealAlertOptions.HeaderTextFontSize = anchor.HeaderTextFontSize
				
		for i = 2, #RCT.HealAlerts, 1 do -- iterate all frames after anchor
			RCT.HealAlerts[i]:SetHeaderTextFontSize(RCT.HealAlertOptions.HeaderTextFontSize)
		end
	end
	function anchor.SuperFunctions:SetFooterTextFontSize(Size)
		anchor:SetFooterTextFontSize(Size)
		RCT.HealAlertOptions.FooterTextFontSize = anchor.FooterTextFontSize
				
		for i = 2, #RCT.HealAlerts, 1 do -- iterate all frames after anchor
			RCT.HealAlerts[i]:SetFooterTextFontSize(RCT.HealAlertOptions.FooterTextFontSize)
		end
	end
	function anchor.SuperFunctions:SetAlphaFromSlider(Alpha)
		anchor:SetAlphaFromSlider(Alpha)
		RCT.HealAlertOptions.Alpha = anchor.Alpha
				
		for i = 2, #RCT.HealAlerts, 1 do -- iterate all frames after anchor
			RCT.HealAlerts[i].Alpha = RCT.HealAlertOptions.Alpha
			RCT.HealAlerts[i]:UpdateAlpha()
		end
	end
	function anchor.SuperFunctions:UpdateAlpha() -- for consistency
		anchor:UpdateAlpha()
				
		for i = 2, #RCT.HealAlerts, 1 do -- iterate all frames after anchor
			RCT.HealAlerts[i]:UpdateAlpha()
		end
	end
	function anchor.SuperFunctions:Initialize()
		for k, v in pairs(RCT.HealAlertOptions) do
			anchor[k] = v
		end
		self:SetIconSize(nil, true, true)
		self:SetTimerTextFontSize()	
		self:UpdateAlpha()
		self:SetHeaderVisibility()
		--self:SetHeaderText()
		self:SetHeaderTextLength()
		self:SetHeaderTextFontSize()
		self:SetFooterVisibility()
		--self:SetFooterText()
		self:SetFooterTextFontSize()		
	end
	
	function RCT.HealAlertAnchor:OnIconSizeChanged()
		RCT.HealAlertMenu.FrameSizeSlider:SetPosition(self.IconSize, true)
	end
	function RCT.HealAlertAnchor:OnTimerTextFontSizeChanged()
		RCT.HealAlertMenu.TimerTextFontSizeSlider:SetPosition(self.TimerTextFontSize, true)
	end
	function RCT.HealAlertAnchor.Event:WheelForward()
		if not CCD.UI.EditModeOn then return end
		anchor.SuperFunctions:SetIconSize(self.IconSize + 1)
	end
	function RCT.HealAlertAnchor.Event:WheelBack()
		if not CCD.UI.EditModeOn then return end
		anchor.SuperFunctions:SetIconSize(self.IconSize - 1)
	end--]]

end
function RCT:CreateHealAlert()
	local id = string.format("%s.HealAlertFrame%d", RCT.Id, NextHealAlertIndex)
	local frame = CCD.UI.AlertFrame.Create(id, CCD.UI.Context)

	frame.IsMobile = false -- only anchor will be shown
	frame:SetIconSize(RCT.HealAlertOptions.IconSize, true, true)
	frame:SetTimerTextFontSize(RCT.HealAlertOptions.TimerTextFontSize)
	frame:SetHeaderTextLength(RCT.HealAlertOptions.HeaderTextLength)
	frame:SetHeaderTextFontSize(RCT.HealAlertOptions.HeaderTextFontSize)
	frame:SetTimerTextFontSize(RCT.HealAlertOptions.TimerTextFontSize)
	frame:SetFooterTextFontSize(RCT.HealAlertOptions.FooterTextFontSize)
	frame:SetAlpha(RCT.HealAlertOptions.Alpha)
	frame:SetHeaderVisibility(RCT.HealAlertOptions.HeaderOn)
	frame:SetFooterVisibility(RCT.HealAlertOptions.FooterOn)	

	function frame:Update()
		if not CCD.UI.EditModeOn then
			frame:Hide()
		end
	end
	function frame:UpdateForeground()
		if RCT.HealAlertOptions.Enabled then
			if RCT.HealAlertOptions.CombatOnly then
				self:SetForegroundColor(CCD.UI.Colors.LightRedTint)
			else
				self:SetForegroundColor(CCD.UI.Colors.ForegroundTint)
			end
		else
			self:SetForegroundColor(CCD.UI.Colors.ForegroundBlackTint)
		end
	end

	function frame.EditBolt.Event:RightClick()
		RCT.HealAlertMenu:Show(frame)
	end

	table.insert(RCT.HealAlerts, frame)
	NextHealAlertIndex = NextHealAlertIndex + 1
	return frame
end
function RCT:CreateHealCooldownSelectFrame()
	local container = CCD.UI.CreateFrame("RCT.HealCooldownSelectFrame", CCD.UI.Context)
	container:SetBackgroundColor(Color.Black(0.75))
	container:SetLayer(3)
	container:Hide()
	
	-- Close Button
	container.CloseButtonTexture = UI.CreateFrame("Texture", "RCT.HealCooldownSelectFrame.CloseButtonTexture", container)
	container.CloseButtonTexture:SetTexture("CCD", "Images\\ContextMenuCloseButton.png")
	container.CloseButtonTexture:SetWidth(28)
	container.CloseButtonTexture:SetHeight(28)
	container.CloseButtonTexture:SetPoint("TOPRIGHT", container, "TOPRIGHT", 0, 0)
	container.CloseButtonTexture:SetVisible(true)
	function container.CloseButtonTexture.Event:MouseIn()
		container.CloseButtonTexture:SetTexture("CCD", "Images\\ContextMenuCloseButtonHighlight.png")
	end
	function container.CloseButtonTexture.Event:MouseOut()
		container.CloseButtonTexture:SetTexture("CCD", "Images\\ContextMenuCloseButton.png")
	end
	function container.CloseButtonTexture.Event:LeftClick()
		container:Hide()
	end

	local anchor = nil
	local height = 0
	local width = 0
	container.Checkboxes = {}
	container.CheckboxesByBuffType = {}
	for key, data in pairs(RCT.CooldownDataCollection) do		
		if data.Role == RCT.Role.Heal and string.startswith(key, "B") then		
			local checkbox = CCD.UI.CreateCheckbox("RCT.HealCooldownSelectFrame.Checkbox" .. data.Name, container, data.Enabled)
			table.insert(container.Checkboxes, checkbox)
			container.CheckboxesByBuffType[key] = checkbox
			checkbox.BuffName = data.Name
			checkbox:SetText(data.Name, 18)			
			if data.Class == "CLERIC" then checkbox:SetFontColor(Color.Green())
			elseif data.Class == "MAGE" then checkbox:SetFontColor(Color:Create(0.75, 0.5, 1, 0))
			elseif data.Class == "WARRIOR" then checkbox:SetFontColor(Color.Red())
			elseif data.Class == "ROGUE" then checkbox:SetFontColor(Color.Yellow())
			else checkbox:SetFontColor(Color.White()) end -- primalist?
			function checkbox.Event:CheckboxChange()
				if self:GetChecked() then
					RCT:EnableCooldown(data.Type)
				else
					RCT:DisableCooldown(data.Type)
				end
			end
			
			height = height + checkbox:GetHeight()
			width = math.max(width, checkbox:GetWidth())
		end
	end
	
	table.sort(container.Checkboxes, function(a, b) return a.BuffName < b.BuffName end)
	
	for _, checkbox in ipairs(container.Checkboxes) do
		if anchor then checkbox:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT")
		else checkbox:SetPoint("TOPLEFT", container, "TOPLEFT") end
		anchor = checkbox	
		--print(checkbox.BuffName)
	end
	container:SetHeight(height + 1)
	container:SetWidth(width + 1)
	
	return container
end
--
function RCT:OrganizeHealAlerts()
	RCT:UnassignAllHealAlerts()

	local healCds = {}

	for _, v in ipairs(RCT.HealCooldowns) do
		v:UpdateStatus()

		if RCT.CooldownDataCollection[v.BuffType].Enabled and v.Status == Status.Active then
			table.insert(healCds, v)
		end
	end

	table.sort(healCds, function(a, b) return a.LastCast < b.LastCast end)

	for _, cooldown in pairs(healCds) do
		if cooldown.Status == Status.Active then
			RCT:LinkHealCooldownToAlert(cooldown)
		end
	end
	
	for _, alert in ipairs(RCT.HealAlerts) do
		if not alert.RCD then alert:HideX() end
	end
end
function RCT:LinkHealCooldownToAlert(Cooldown)
	for _, alert in ipairs(RCT.HealAlerts) do
		if not alert.RCD then
			alert.RCD = Cooldown
			Cooldown.Alert = alert
			return
		end
	end

	--print(RCT.HealAlerts[1].RCD.Name .. " already has an alert")
	--print("Making new alert for " .. Cooldown.Name)
	-- create new alert
	local alert = RCT:CreateHealAlert()
	alert:Link(RCT.HealAlerts[NextHealAlertIndex-2], "TOPRIGHT", 5)
	alert.RCD = Cooldown
	Cooldown.Alert = alert
end
function RCT:UnassignAllHealAlerts()
	for _, alert in ipairs(RCT.HealAlerts) do
		if alert.RCD then alert.RCD.Alert = nil end
		alert.RCD = nil
		--alert:HideX()
	end
end
function RCT:UpdateAllHealAlerts()
	for _, alert in pairs(RCT.HealAlerts) do
		alert:Update()
	end
end
function RCT:ShouldShowHealAlert(Frame)
	if		CCD.UI.EditModeOn then return false
	elseif	not RCT.On then return false
	elseif	not RCT.HealAlertOptions.Enabled then return false
	elseif	RCT.HealAlertOptions.CombatOnly and not CCD.Me.InCombat then return false
	elseif	RCT.HealAlertOptions.GSV >= 2 and CCD.Allies.Count < RCT.HealAlertOptions.GSV then return false
	elseif	Frame.RCD and not RCT.ShowInertTimers and Frame.RCD.Status == Status.Inert then return false
	elseif	CCD.DisableInWarfronts and CCD.Me.ZoneType == "warfront" then return false
	else return true end
end
function RCT:HideAllHealAlerts()
	for _, alert in ipairs(RCT.HealAlerts) do
		alert:Hide()
	end
end
function RCT:EnableHealAlerts()
	if RCT.HealAlertOptions.Enabled then return end

	--

	RCT.HealAlertOptions.Enabled = true
	RCT.HealAlertAnchor:UpdateForeground()
	RCT.HealAlertAnchor.IsMobile = true
end
function RCT:DisableHealAlerts()
	if not RCT.HealAlertOptions.Enabled then return end

	RCT:HideAllHealAlerts()

	RCT.HealAlertOptions.Enabled = false
	RCT.HealAlertAnchor:UpdateForeground()
	RCT.HealAlertAnchor.IsMobile = false
end
function RCT:SetHealAlertOption(Key, Value)
	if RCT.HealAlertOptions[Key] == nil then return end

	RCT.HealAlertOptions[Key] = Value

	if Key == "CombatOnly" then
		RCT.HealAlertAnchor:UpdateForeground()
	end
end
--
function RCT.CreateHealAlertMenu()
	-- Context Menu
	local id = RCT.Id .. ".HealAlertMenu"
	local frame = CCD.UI.CreateContextMenu(id, 315, 540, "Healing CD Alert Anchor")
	function frame:OnNewElementHighlighted(Element)
		if Element ~= self.EnableAlertCheckbox then
			RCT.HealCooldownSelectFrame:Hide()
		end
	end
	function frame:OnClosed()
		RCT.HealCooldownSelectFrame:Hide()
	end
	function frame:OnLostFocus()
		RCT.HealCooldownSelectFrame:Hide()
	end
	
	-- Enable Alert Checkbox
	frame.EnableAlertCheckbox = frame:AddCheckbox(id .. ".EnableAlertCheckbox", "Enable Healing Cooldown Alerts", false, nil, true)
	function frame.EnableAlertCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		if checked then
			RCT:EnableHealAlerts()
		else
			RCT:DisableHealAlerts()
		end
	end
	--
	local baseMouseIn = frame.EnableAlertCheckbox.Event.MouseIn
	function frame.EnableAlertCheckbox.Event:MouseIn()
		--self:SetBackgroundColor(Color.Grey())
		baseMouseIn(self)
		RCT.HealCooldownSelectFrame:SetPoint("TOPLEFT", frame, "TOPRIGHT", 10, 0)
		RCT.HealCooldownSelectFrame:Show()
	end--]]
	
	-- Combat Show Checkbox
	frame.CombatShowCheckbox = frame:AddCheckbox(id .. ".CombatShowCheckbox", "Only show while in combat")
	function frame.CombatShowCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT:SetHealAlertOption("CombatOnly", checked)
	end

	-- GSV Text Field
	frame.GSVTextField = frame:AddTextField(frame:GetName() .. ".GSVTextField", "Only show in groups of", "or more", true)
	function frame.GSVTextField.OnTextFieldChanged()
		local value = frame.GSVTextField:GetConvertedText()

		if value then
			RCT.HealAlertOptions.GSV = value
		end
	end
	frame.GSVTextField:SetTextFieldChangedEvent(frame.GSVTextField.OnTextFieldChanged, frame.GSVTextField:GetName() .. ".OnTextFieldChanged")

	-- Opacity Slider
	frame.AlphaSlider = frame:AddSlider(id .. ".AlphaSlider", "Opacity:", 0, 100)
	function frame.AlphaSlider.Slider.Event:SliderChange()
		local sliderPosition = self:GetPosition()
		RCT.HealAlertAnchor.SuperFunctions:SetAlphaFromSlider(sliderPosition)
	end

	-- Icon Size Slider
	frame.FrameSizeSlider = frame:AddSlider(id .. ".FrameSizeSlider", "Icon Size:", CCD.UI.AlertFrame.IconSizeRange.Min, CCD.UI.AlertFrame.IconSizeRange.Max, false)
	function frame.FrameSizeSlider.Slider.Event:SliderChange()
		RCT.HealAlertAnchor.SuperFunctions:SetIconSize(self:GetPosition())
		frame.HeaderTextLengthSlider:SetPosition(RCT.HealAlertOptions.HeaderTextLength, true)
		frame.HeaderTextFontSizeSlider:SetPosition(RCT.HealAlertOptions.HeaderTextFontSize, true)
		--frame.TimerTextFontSizeSlider:SetPosition(RCT.HealAlertOptions.TimerTextFontSize, true)
		frame.FooterTextFontSizeSlider:SetPosition(RCT.HealAlertOptions.FooterTextFontSize, true)
	end

	-- Timer Font Size Slider
	frame.TimerTextFontSizeSlider = frame:AddSlider(id .. ".TimerTextFontSizeSlider", "Timer Font Size:", CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max, false)
	function frame.TimerTextFontSizeSlider.Slider.Event:SliderChange()
		local size = self:GetPosition()
		RCT.HealAlertAnchor.SuperFunctions:SetTimerTextFontSize(size)
	end

	-- Timer Font Size Locked Check Box
	frame.TimerTextFontSizeLockedCheckbox = frame:AddCheckbox(id .. ".TimerTextFontSizeLockedCheckbox", "Lock Timer Font Size", true)
	function frame.TimerTextFontSizeLockedCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.HealAlertAnchor.SuperFunctions:SetTimerTextFontSizeLock(checked)
	end

	-- Show Header Check Box
	frame.ShowHeaderCheckbox = frame:AddCheckbox(id .. ".ShowHeaderCheckbox", "Show Header", false)
	function frame.ShowHeaderCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.HealAlertAnchor.SuperFunctions:SetHeaderVisibility(checked)
		frame.HeaderTextLengthSlider:SetVisible(checked)
		frame.HeaderTextFontSizeSlider:SetVisible(checked)
		frame.HeaderTextFontSizeSlider.Line:SetVisible(checked)

		if checked then
			frame.ShowFooterCheckbox:AnchorTo("ORIGINAL")
		else
			frame.ShowFooterCheckbox:AnchorTo(frame.ShowHeaderCheckbox, 0)
		end
	end

	-- Name Length Slider
	frame.HeaderTextLengthSlider = frame:AddSlider(id .. ".HeaderTextLengthSlider", "Header Length:", 3, 19)
	function frame.HeaderTextLengthSlider.Slider.Event:SliderChange()
		RCT.HealAlertAnchor.SuperFunctions:SetHeaderTextLength(self:GetPosition())
	end

	-- Header Font Size Slider
	frame.HeaderTextFontSizeSlider = frame:AddSlider(id .. ".HeaderTextFontSizeSlider", "Header Font Size:", CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max, true)
	function frame.HeaderTextFontSizeSlider.Slider.Event:SliderChange()
		local size = self:GetPosition()
		RCT.HealAlertAnchor.SuperFunctions:SetHeaderTextFontSize(size)
	end

	-- Show Footer Check Box
	frame.ShowFooterCheckbox = frame:AddCheckbox(id .. ".ShowFooterCheckbox", "Show Footer")
	function frame.ShowFooterCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.HealAlertAnchor.SuperFunctions:SetFooterVisibility(checked)

		frame.FooterTextFontSizeSlider:SetVisible(checked)

		if checked then
			frame.FooterTextFontSizeSlider.Line:AnchorTo("ORIGINAL")
		else
			frame.FooterTextFontSizeSlider.Line:AnchorTo(frame.ShowFooterCheckbox, 0)
		end
	end

	-- Footer Font Size Slider
	frame.FooterTextFontSizeSlider = frame:AddSlider(id .. ".FooterTextFontSizeSlider", "Footer Font Size:", CCD.UI.AlertFrame.TextFontSizeRange.Min, CCD.UI.AlertFrame.TextFontSizeRange.Max, true)
	function frame.FooterTextFontSizeSlider.Slider.Event:SliderChange()
		local size = self:GetPosition()
		RCT.HealAlertAnchor.SuperFunctions:SetFooterTextFontSize(size)
	end


	-- Global Header
	frame.GlobalOptionsLabel = frame:AddHeader(id .. ".GlobalOptionsLabel", "Global Options")

	-- On/Off Checkbox
	frame.OnCheckbox = frame:AddCheckbox(id .. ".OnCheckbox", "Enable Cooldown Tracker", true)
	function frame.OnCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.Set(checked)
		RCT.Tab.OnCheckbox:SetChecked(RCT.On)
	end

	frame.FooterText = frame:AddFooterNote(id .. ".FooterText", "Click here for more options")
	function frame.FooterText.Event:LeftClick()
		CCD.MainWindow:Show()
		CCD.MainWindow.TabView:SetActiveTab(1)
	end

	-- Load Function
	function frame:Load()
		self.EnableAlertCheckbox:SetChecked(RCT.HealAlertOptions.Enabled)
		self.CombatShowCheckbox:SetChecked(RCT.HealAlertOptions.CombatOnly)
		self.GSVTextField:SetTextField(RCT.HealAlertOptions.GSV)

		self.FrameSizeSlider:SetPosition(RCT.HealAlertOptions.IconSize, true)
		self.ShowHeaderCheckbox:SetChecked(RCT.HealAlertAnchor.HeaderOn)
		self.ShowFooterCheckbox:SetChecked(RCT.HealAlertAnchor.FooterOn)
		self.TimerTextFontSizeLockedCheckbox:SetChecked(RCT.HealAlertOptions.TimerTextFontSizeLocked)

		self.HeaderTextLengthSlider:SetPosition(RCT.HealAlertOptions.HeaderTextLength, true)
		self.HeaderTextFontSizeSlider:SetPosition(RCT.HealAlertOptions.HeaderTextFontSize, true)
		self.TimerTextFontSizeSlider:SetPosition(RCT.HealAlertOptions.TimerTextFontSize, true)
		self.FooterTextFontSizeSlider:SetPosition(RCT.HealAlertOptions.FooterTextFontSize, true)
		self.AlphaSlider:SetPosition(CCD.Math.Round(RCT.HealAlertOptions.Alpha * 100), true)

		self.OnCheckbox:SetChecked(RCT.On)--]]
	end

	return frame
end

-- Heal Bars
function RCT:CreateHealBarContainer()
	local container = CCD.UI.CreateLinkableFrame("RCT.HealBarContainer", CCD.UI.Context)
	--Library.LibSimpleWidgets.SetBorder("plain", container.Foreground, 5, 1, 1, 1, 1, "tblr") -- thin black border	
	
	function container:Resize()
		if CCD.UI.EditModeOn then
			self:SetSize(RCT.HBM.Properties.Length, RCT.HBM.Properties.Height)
		else
			self:SetSize(RCT.HBM.Properties.Length, RCT.HBM.Properties.Height * RCT:GetActiveHealBarCount())
		end
	end
	--
	function container.Event:WheelForward()
		if not CCD.UI.EditModeOn then return end
		RCT.HBM:SetLength(RCT.HBM.Properties.Length + 25, self)
	end
	function container.Event:WheelBack()
		if not CCD.UI.EditModeOn then return end
		RCT.HBM:SetLength(RCT.HBM.Properties.Length - 25, self)
	end	
	function container.EditBolt.Event:RightClick()
		RCT.HealBarMenu:Show(self)
	end
	function container:UpdateForeground()
		if RCT.HBM.Properties.Enabled then
			if RCT.HBM.Properties.CombatOnly then
				self:SetForegroundColor(CCD.UI.Colors.LightRedTint)
			else
				self:SetForegroundColor(CCD.UI.Colors.ForegroundTint)
			end
		else
			self:SetForegroundColor(CCD.UI.Colors.ForegroundBlackTint)
		end
	end
	
	return container
end
function RCT:CreateHealBar()
	local i = RCT.NextHealBarIndex
	local id = string.format("%s.HealBar%d", RCT.Id, i) -- Examples RCT.HealBar1, RCT.HealBar5

	local bar = CCD.UI.CreateColorBar(id, RCT.HealBarContainer, Colors.Green)
	bar:SetSize(RCT.HBM.Properties.Length, RCT.HBM.Properties.Height)
	bar:SetFontSize(RCT.HBM.Properties.FontSize)
	if RCT.HealBars[i-1] then bar:SetPoint("TOPLEFT", RCT.HealBars[i-1], "BOTTOMLEFT", 0, 0)
	else bar:SetPoint("TOPLEFT", RCT.HealBarContainer, "TOPLEFT") end
	
	bar.Filling:SetAlpha(RCT.HBM.Properties.Opacity)
	bar.Icon:SetAlpha(RCT.HBM.Properties.Opacity)
	bar:SetBackgroundAlpha(RCT.HBM.Properties.BackgroundOpacity)

	function bar:Update()
		if bar.RCD and bar.RCD.Status == Status.Ready then
			bar:SetColor(Colors.Yellow, 0.6)
			bar:SetTitle(bar.RCD.Caster.Name)
			bar:SetTexture("Rift", bar.RCD.Icon)
			bar:SetPercentage(100)
			bar:SetTimerText("Ready")
			if RCT.HealAlertsByGroup[bar.RCD.Group].ShowHealBars then
				bar:Show()
			else
				bar:Hide()
			end
		else
			bar:Hide()
		end
	end

	table.insert(RCT.HealBars, bar)
	RCT.NextHealBarIndex = RCT.NextHealBarIndex + 1
	return bar
end
function RCT:GetNextHealBar()
	for i, bar in ipairs(RCT.HealBars) do
		if bar.RCD == nil then return bar end
	end

	local bar = RCT:CreateHealBar()
	return bar
end
function RCT:OrganizeHealBars()
	if CCD.UI.EditModeOn then return end
	
	if not RCT:ShouldShowHealBars() then
		RCT.HBM:HideAllBars()
		return
	end

	RCT:UnassignAllHealBars()

	for _, cooldownList in pairs(RCT.HealCooldownsByGroup) do
		for _, rcd in ipairs(cooldownList) do
			if RCT.CooldownDataCollection[rcd.BuffType].Enabled then
				local bar = RCT:GetNextHealBar()
				if bar then RCT:AssignHealCooldownToBar(rcd, bar) end
			end
		end
	end

	RCT.HealBarContainer:Resize()
end
function RCT:AssignHealCooldownToBar(RCD, Bar)
	Bar.RCD = RCD
	RCD.Bar = Bar

	if RCT:ShouldShowHealBars() then
		Bar:SetTitle(RCD.Caster.Name)
		Bar:SetTexture("Rift", RCD.Icon)

		if RCD.Status == Status.Ready then
			Bar:SetColor(Colors.Yellow, 0.6)
			Bar:SetTitle(RCD.Caster.Name)
			Bar:SetPercentage(100)
			Bar:SetTimerText("Ready")
		end

		Bar:Show()
	end
end
function RCT:UnassignAllHealBars()
	for _, bar in ipairs(RCT.HealBars) do
		if bar.RCD then bar.RCD.Bar = nil end -- hopefully fixes the issue with multiple cds using the same bar
		bar.RCD = nil
		bar:Hide()
	end
end
function RCT:GetHealBarLength()
	return RCT.HealBars[1]:GetWidth()
end
function RCT.CreateHealBarMenu()
	-- Context Menu
	local id = RCT.Id .. "HealBarMenu"
	local frame = CCD.UI.CreateContextMenu(id, 310, 425, "Heal Bar Anchor")

	-- Show Cooldown Bars Checkbox
	frame.ShowBarsCheckbox = frame:AddCheckbox(id .. ".ShowBarsCheckbox", "Enable Heal Bars")
	function frame.ShowBarsCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.HealBarManager:SetEnabled(checked)
	end

	-- Combat Show Checkbox
	frame.ShowBarsInCombatOnlyCheckbox = frame:AddCheckbox(id .. ".ShowBarsInCombatOnlyCheckbox", "Only show while in combat")
	function frame.ShowBarsInCombatOnlyCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.HBM:SetCombatOnly(checked)
	end

	-- GSV Text Field
	frame.GSVTextField = frame:AddTextField(frame:GetName() .. ".GSVTextField", "Only show in groups of", "or more", true)
	function frame.GSVTextField.OnTextFieldChanged()
		local value = frame.GSVTextField:GetConvertedText()

		if value then
			RCT.HBM:SetGSV(value)
		end
	end
	frame.GSVTextField:SetTextFieldChangedEvent(frame.GSVTextField.OnTextFieldChanged, frame.GSVTextField:GetName() .. ".OnTextFieldChanged")

	-- Length Slider
	frame.FrameSizeSlider = frame:AddSlider(id .. ".FrameSizeSlider", "Length:", RCT.DamageBarLengthRange.Min, RCT.DamageBarLengthRange.Max)
	function frame.FrameSizeSlider.Slider.Event:SliderChange()
		local length = self:GetPosition()
		RCT.HBM:SetLength(length)
	end

	-- Bar Height Slider
	frame.BarHeightSlider = frame:AddSlider(id .. ".BarHeightSlider", "Height:", RCT.CooldownBarHeightRange.Min, RCT.CooldownBarHeightRange.Max)
	function frame.BarHeightSlider.Slider.Event:SliderChange()
		local length = self:GetPosition()
		RCT.HBM:SetHeight(length)
	end

	-- Bar Font Size Slider
	frame.BarFontSizeSlider = frame:AddSlider(id .. ".BarFontSizeSlider", "Font Size:", RCT.CooldownBarFontSizeRange.Min, RCT.CooldownBarFontSizeRange.Max)
	function frame.BarFontSizeSlider.Slider.Event:SliderChange()
		local length = self:GetPosition()
		RCT.HBM:SetFontSize(length)
	end

	-- Cooldown Bar Opacity Slider
	frame.BarOpacitySlider = frame:AddSlider(id .. ".OpacitySlider", "Opacity:", 0, 100)
	function frame.BarOpacitySlider.Slider.Event:SliderChange()
		local pos = self:GetPosition()
		RCT.HBM:SetOpacity(pos)
	end

	-- Cooldown Bar Background Opacity Slider
	frame.BarBackgroundOpacitySlider = frame:AddSlider(id .. ".OpacitySlider", "Background Opacity:", 0, 100, true)
	function frame.BarBackgroundOpacitySlider.Slider.Event:SliderChange()
		local pos = self:GetPosition()
		RCT.HBM:SetBackgroundOpacity(pos)
	end

	-- Global Header
	frame.GlobalOptionsLabel = frame:AddHeader(id .. ".GlobalOptionsLabel", "Global Options")

	-- On/Off Checkbox
	frame.OnCheckbox = frame:AddCheckbox(id .. ".OnCheckbox", "Enable Cooldown Tracker")
	function frame.OnCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		RCT.Set(checked)
		RCT.Tab.OnCheckbox:SetChecked(RCT.On)
	end

	-- Reset Bars button
	frame.ResetBarsButton = frame:AddLabel(id .. ".ResetBarsButton", "Reset All Bar Positions", true)
	function frame.ResetBarsButton.Event:LeftClick()
		RCT.ResetCooldownBarContainerPositions()
	end

	frame.FooterText = frame:AddFooterNote(id .. ".FooterText", "Click here for more options")
	function frame.FooterText.Event:LeftClick()
		CCD.MainWindow:Show()
		CCD.MainWindow.TabView:SetActiveTab(1)
	end

	-- Load Function
	function frame:Load()
		frame.ShowBarsCheckbox:SetChecked(RCT.HBM.Properties.Enabled)
		frame.ShowBarsInCombatOnlyCheckbox:SetChecked(RCT.HBM.Properties.CombatOnly)
		frame.GSVTextField:SetTextField(RCT.HBM.Properties.GSV)
		frame.FrameSizeSlider:SetPosition(RCT.HBM.Properties.Length, true)
		frame.BarHeightSlider:SetPosition(RCT.HBM.Properties.Height, true)
		frame.BarFontSizeSlider:SetPosition(RCT.HBM.Properties.FontSize, true)
		frame.BarOpacitySlider:SetPosition(CCD.Math.Round(RCT.HBM.Properties.Opacity * 100), true)
		frame.BarBackgroundOpacitySlider:SetPosition(CCD.Math.Round(RCT.HBM.Properties.BackgroundOpacity * 100), true)

		-- global
		self.OnCheckbox:SetChecked(RCT.On)
	end

	return frame
end
function RCT:ShouldShowHealBars()
	if		CCD.UI.EditModeOn then return false
	elseif	not RCT.On then return false
	elseif	not RCT.HBM.Properties.Enabled then return false
	elseif	RCT.HBM.Properties.CombatOnly and not CCD.Me.InCombat then return false
	elseif  RCT.HBM.Properties.GSV >= 2 and CCD.Allies.Count < RCT.HBM.Properties.GSV then return false
	elseif	CCD.DisableInWarfronts and CCD.Me.ZoneType == "warfront" then return false
	else return true end
end
function RCT:GetActiveHealBarCount()
	local count = 0
	
	for _, healBar in ipairs(self.HealBars) do
		if healBar.RCD then count = count + 1 end
	end
	
	return count
end

function RCT:ResetFlaring(Cooldown)
	for _, rcd in ipairs(RCT.DamageCooldownsByGroup[1]) do
		rcd.Cooldown = Cooldown
	end

	for _, rcd in ipairs(RCT.DamageCooldownsByGroup[11]) do
		rcd.Cooldown = Cooldown
	end
end

-- Event Handlers
function RCT:OnEditModeBegin()
	-- update damage alerts
	for _, alert in pairs(RCT.DamageAlertsByGroup) do
		alert:SetHeaderText("1234567890")
		alert.Texture:SetTexture("Rift", EditModeTextures[alert.Group])
		alert.TimerText:SetText(string.format("%d", RCT.SoonReadyStartTime))
		alert.TimerText:SetVisible(true)
		alert.Footer.Text:SetText("Edit")
		alert:SetFontColor(Colors.White)
		alert:UpdateForeground()
	end

	-- Damage Bar Container
	for i = 2, #RCT.DamageBars, 1 do
		RCT.DamageBars[i]:Hide()
	end
	RCT.DamageBarContainer:Resize()
	RCT.DamageBars[1]:Show()
	RCT.DamageBars[1]:SetPercentage(50)
	if not RCT.DamageBars[1].RCD then
		RCT.DamageBars[1]:SetColor(Colors.Yellow, 0.6)
		RCT.DamageBars[1]:SetTitle(CCD.Me.Name)
		RCT.DamageBars[1]:SetTexture("CCD", DamageCooldownAlertAnchorTexture)
		RCT.DamageBars[1]:SetTimerText("Edit")
	end
	
	-- update heal alert anchor
	RCT.HealAlerts[1].Texture:SetTexture("CCD", HealCooldownAlertAnchorTexture)
	RCT.HealAlerts[1].Texture:SetVisible(true)
	RCT.HealAlerts[1].TimerText:SetText("99")
	RCT.HealAlertAnchor.Footer.Text:SetText("Edit") -- not sure why this is set directly; same with above
	RCT.HealAlertAnchor:SetFontColor(Color.White())

	-- Heal Bar Container
	for i = 2, #RCT.HealBars, 1 do
		RCT.HealBars[i]:Hide()
	end
	RCT.HealBarContainer:Resize()
	RCT.HealBars[1]:Show()
	RCT.HealBars[1]:SetPercentage(50)
	if not RCT.HealBars[1].RCD then
		RCT.HealBars[1]:SetColor(Colors.Yellow, 0.6)
		RCT.HealBars[1]:SetTitle(CCD.Me.Name)
		RCT.HealBars[1]:SetTexture("CCD", HealCooldownAlertAnchorTexture)
		RCT.HealBars[1]:SetTimerText("Edit")
	end
end
function RCT:OnEditModeEnd()
	RCT:UpdateAllDamageAlerts()
	RCT:UpdateAllHealAlerts()
	RCT:OrganizeDamageBars()
	RCT:OrganizeHealBars()
	RCT.HealCooldownSelectFrame:Hide()
end

function RCT:OnNewAlly(Ally)
	RCT.AllCooldownsByCaster[Ally] = {}
	RCT.DamageCooldownsByCaster[Ally] = {}
	RCT.HealCooldownsByCaster[Ally] = {}
end
function RCT:OnBuffAdd(UnitId, BuffId, BuffType)
	-- cancel Wild Growth charge draining buff
	if BuffType == "B087BE1DF34A67269" and UnitId == CCD.Me.Id and CCD.AutoCancelWildGrowthOn then
		Command.Buff.Cancel(BuffId)
	end

	-- track battle weary
	if BuffType == BattleWearyId then
		local ally = CCD.Allies:Get(UnitId)

		if ally then
			battleWeary = CCD.Buff:Create(ally.Id, BuffId)
			ally:AddBuff(battleWeary)
		end
	end

	RCT:Read(BuffFlag, UnitId, BuffId, BuffType)
end
function RCT:Read(Flag, UnitId, Arg1, Arg2) -- Flag: 1 = Buff, 2 = Ability; Arg1 = BuffId or CooldownData; Arg2 = BuffType or AbilityCooldownRemaining
	local buffType, abilityId, cooldownName, cooldownIcon, cooldownStartTime, cooldownRemaining
	local buffId, buff, target -- Buff Flag
	local rcdData, abilityDetails -- Ability Flag
	local caster

	if Flag == BuffFlag then
		buffType = Arg2

		if RCT.CooldownDataCollection[buffType] then
			buffId = Arg1

			target = CCD.Allies:Get(UnitId)
			if target then buff = target.Buffs[buffId] end -- check if this buff ID has already been processed
			buff = buff or CCD.Buff:Create(UnitId, buffId)

			-- unknown bug
			if buff.Type ~= buffType or not RCT.CooldownDataCollection[buff.Type] then
				CCD.Modules.ErrorManager:AddError("RCT Buff Type", string.format("%s was matched to %s but when inspected had a .type of %s.", buffType, buff.Name, buff.Type))
				print(string.format("%s has a .Type of %s vs given %s. Please screenshot and report this.",buff.Name, buff.Type, buffType))
				CCD.Debug:Record("%s has a .Type of %s vs given %s", buff.Name, buff.Type, buffType)
				return
			end

			cooldownName = buff.Name
			cooldownIcon = buff.Icon
			cooldownStartTime = buff.StartTime

			caster = CCD.Allies:Get(buff.CasterId)
			if caster == CCD.Me then return end -- don't process buffs from Me (only do abilities)
		end

		if not buffId then return end -- not a buff we care about
	elseif Flag == AbilityFlag then		
		rcdData = Arg1
		buffType = rcdData.Type
		abilityId = rcdData.AbilityId
		abilityDetails = Inspect.Ability.New.Detail(abilityId)
		-- Check if Player's cast was broken
		if UnitId == CCD.Me.Id and abilityDetails and not abilityDetails.currentCooldownRemaining then return end

		cooldownName = abilityDetails.name
		cooldownIcon = abilityDetails.icon
		cooldownStartTime = Inspect.Time.Frame() -- this will change to ability start time either from self or msging system
		cooldownRemaining = Arg2 or abilityDetails.currentCooldownRemaining
		caster = CCD.Allies:Get(UnitId)
	end

	if not caster then return end -- this happens in CQ sometimes (because of lag?)
	
	local casterCooldowns = RCT.AllCooldownsByCaster[caster]
	local rcd = casterCooldowns[buffType]

	-- check if new cooldown
	if not rcd then
		-- check for Symbiosis because it has 2 buff types but should only have 1 alert and bar
		if buffType == "B19EEE7BA9C218028" then
			rcd = casterCooldowns["B07B4AD401C182BB0"]
		elseif buffType == "B07B4AD401C182BB0" then
			rcd = casterCooldowns["B19EEE7BA9C218028"]
		end

		-- if still nil then...
		if not rcd then
			rcd = CCD.RaidCooldown:Create(caster, buffType, abilityId, cooldownName, cooldownIcon)
			RCT:AddCooldown(rcd)
		end
	end

	if not rcd:IsReady() then Debug:Record("%s was not Ready when it was used.", cooldownName) end -- this check may be removed if CCD is not as accurate as I think

	--print(rcd.Name .. "has status of " .. rcd.Status .. " and flag of " .. Flag)
	--T = T or rcd
	--if rcd ~= T then print("diff") end	
	if rcd.Status ~= Status.Active or (buffType == "B7FC56FE0D2639B3C" and rcd:IsReady()) then
		if rcd.Group == 1 or rcd.Group == 11 then
			cooldownRemaining = RCT.CooldownDataCollection[rcd.BuffType].Cooldown
		end
		rcd:JustUsed(cooldownStartTime, cooldownRemaining)
		RCT:OnCooldownUsed(rcd)
		if caster.Id == CCD.Me.Id and Flag == AbilityFlag then
			CCD.Messaging.BroadcastAbilityBegin(abilityId, cooldownRemaining)
		end
	end
end

function RCT:OnCooldownUsed(Cooldown)
	if Cooldown.Role == RCT.Role.Damage then
		RCT:OnDamageCooldownUsed(Cooldown)
	elseif Cooldown.Role == RCT.Role.Heal then
		RCT:OnHealCooldownUsed(Cooldown)
	end
end
function RCT:OnDamageCooldownUsed(RCD)
	-- Alert
	local alert = RCT.DamageAlertsByGroup[RCD.Group]
	if not alert then -- to be removed; only handle damage cds atm
		alert = RCT.HealAlerts[1]
		alert.Header.Text:SetFontColor(Colors.Green)
		alert:SetHeaderText(RCD.Caster.Name)
		alert.TimerText:SetFontColor(Colors.Green)
		alert.TimerText:SetText(string.format("%d", RCD.Duration - 1))
		alert.TimerText:SetVisible(true)
		alert.Footer.Text:SetText(RCT.StatusText[RCD.Status])
		alert.Footer.Text:SetFontColor(Colors.Green)
		alert.Texture:SetTexture("Rift", RCD.Icon)
		alert:SetVisible(true)
		RCD.Alert = alert
		local repetitions = math.huge -- ensures timer runs until it is deleted
		local preventDeletion = false
		RCD.Tag = Library.LibCron.new(TOC.identifier, RCT.CooldownTimerUpdateFrequency, repetitions, preventDeletion, RCT.ActiveHealCooldown_TimerTask, RCD)
		return
	end
	if not alert.RCD or alert.RCD.Status ~= Status.Active or RCD:GetActiveTime() > alert.RCD:GetActiveTime() then
		alert.RCD = RCD
		if RCT:ShouldShowAlert(alert) then
			alert.Header.Text:SetFontColor(Colors.Green)
			alert:SetHeaderText(RCD.Caster.Name)
			alert.TimerText:SetFontColor(Colors.Green)
			alert.TimerText:SetText(string.format("%d", RCD.Duration - 1))
			alert.TimerText:SetVisible(true)
			alert.Footer.Text:SetText(RCT.StatusText[RCD.Status])
			alert.Footer.Text:SetFontColor(Colors.Green)
			alert.Texture:SetTexture("Rift", RCD.Icon)
			alert:SetVisible(true)
		end
	end

	-- Bar
	RCT:OrganizeDamageBars()

	local barLength = RCT.DamageBarLength
	RCD.ActiveUpdateValue = barLength / RCD.Duration * RCT.CooldownTimerUpdateFrequency -- amount of Bar that is removed per update
	RCD.CurrentFill = barLength - RCD.ActiveUpdateValue
	RCD.InertUpdateValue = barLength / RCD.Cooldown * RCT.CooldownTimerUpdateFrequency -- amount of Bar that is filled per update
	RCD.InertFill = 0

	if RCT:ShouldShowBars() and RCD.Bar then
		RCD.Bar:SetColor(Colors.Green, 0.5)
		RCD.Bar:SetPercentage(RCD.CurrentFill / barLength * 100)
		if alert.ShowDamageBar then RCD.Bar:Show() end
	end

	-- timer
	if RCD.Tag then
		--print("deleting cuz active: " .. RCD.Tag)
		Library.LibCron.delete(RCD.Tag)
		RCD.Tag = nil
	end

	local repetitions = math.huge -- ensures timer runs until it is deleted
	local preventDeletion = false
	RCD.Tag = Library.LibCron.new(TOC.identifier, RCT.CooldownTimerUpdateFrequency, repetitions, preventDeletion, RCT.DamageCooldownTimer_Task, RCD)
end
function RCT:OnHealCooldownUsed(Cooldown)
	RCT:OrganizeHealAlerts()
	local alert = Cooldown.Alert
	if alert then
		if RCT:ShouldShowHealAlert(alert) then
			alert.Header.Text:SetFontColor(Colors.Green)
			alert:SetHeaderText(Cooldown.Caster.Name)
			alert.TimerText:SetFontColor(Colors.Green)
			alert.TimerText:SetText(string.format("%d", Cooldown.Duration - 1))
			alert.TimerText:SetVisible(true)
			alert.Footer.Text:SetText(RCT.StatusText[Cooldown.Status])
			alert.Footer.Text:SetFontColor(Colors.Green)
			alert.Texture:SetTexture("Rift", Cooldown.Icon)
			alert:SetVisible(true)
		end
	end
	
	local repetitions = math.huge -- ensures timer runs until it is deleted
	local preventDeletion = false
	Cooldown.Tag = Library.LibCron.new(TOC.identifier, RCT.CooldownTimerUpdateFrequency, repetitions, preventDeletion, RCT.ActiveHealCooldown_TimerTask, Cooldown)
	
	RCT:OrganizeHealBars()
end

function RCT:OnAbilityAdded(AbilityId, AbilityCooldownRemaining, AllyId)
	local ally = CCD.Allies:Get(AllyId)
	if not ally then return end
	local CooldownData = RCT.CooldownDataCollection[AbilityId]
	if not CooldownData then return end

	local rcd = RCT.AllCooldownsByCaster[ally][CooldownData.Type]

	if rcd then -- update cd rather than add a new one
		if ally == CCD.Me then
			--CCD.Messaging.BroadcastAbilityNew(AbilityId, rcd:GetGreyTime())	-- inaccurate time
		else
			rcd.LastCast = Inspect.Time.Frame() - rcd.Cooldown + (AbilityCooldownRemaining or 0)
			--print("last cast on " .. rcd.Name .. " = " .. rcd.LastCast)
			--print("current time: " .. Inspect.Time.Frame())
		end
	else
		local abilityDetails = Inspect.Ability.New.Detail(CooldownData.AbilityId)
		if not abilityDetails then
			CCD.Modules.ErrorManager:AddError("RCT:OnAbilityAdded()", string.format("Inspection of %s returned nil.", CooldownData.AbilityId))
			-- probably occurs when a toon receives cooldown messages while being partially available
			return
		end
			
		abilityDetails.currentCooldownRemaining = AbilityCooldownRemaining or abilityDetails.currentCooldownRemaining
		local rcd = CCD.RaidCooldown:Create(ally, CooldownData.Type, CooldownData.AbilityId, abilityDetails.name, abilityDetails.icon)

		if abilityDetails.currentCooldownRemaining then
			rcd.LastCast = Inspect.Time.Frame() - abilityDetails.cooldown + abilityDetails.currentCooldownRemaining
		end

		RCT:AddCooldown(rcd)
		RCT:OrganizeDamageBars()
		RCT:OrganizeHealBars()

		if rcd.LastCast then
			local repetitions = math.huge -- ensures timer runs until it is deleted
			local preventDeletion = false

			if CooldownData.Role == RCT.Role.Damage then
				rcd.Tag = Library.LibCron.new(TOC.identifier, RCT.CooldownTimerUpdateFrequency, repetitions, preventDeletion, RCT.DamageCooldownTimer_Task, rcd)
			elseif CooldownData.Role == RCT.Role.Heal then
				rcd.Tag = Library.LibCron.new(TOC.identifier, RCT.CooldownTimerUpdateFrequency, repetitions, preventDeletion, RCT.ActiveHealCooldown_TimerTask, rcd)
			end
		end

		if ally == CCD.Me then
			 CCD.Messaging.BroadcastSingleCooldown(AbilityId, abilityDetails.currentCooldownRemaining)
		end
	end
end
function RCT:OnAbilityNewCooldownBegin(UnitId, AbilityId, CooldownValue)
	local rcdData = RCT.CooldownDataCollection[AbilityId]

	if rcdData then
		local details = Inspect.Ability.New.Detail(AbilityId)
		CooldownValue = tonumber(CooldownValue)

		if UnitId == CCD.Me.Id then
			if CooldownValue and CCD.Math.Round(CooldownValue) ~= CCD.Math.Round(details.cooldown) then	return end -- makes sure that the real ability cooldown was triggered not just the GCD

			if details.castingTime and not details.channeled then
				-- if the ability has a cast time then execute a function when the cast SHOULD finish that determines whether the cast was successful or not
				-- channeled abilities have a .castingTime value equivalent to their channel time so if player ever get an ability that has a cast time and then channels
				-- this check will need to be reevaluated. Examples of such abilities: Zilas' Beam, Salvarola's Beams, Seathe's Triferno
				RCT.AbilityTimerTags[AbilityId] = Library.LibCron.new(TOC.identifier, details.castingTime, 0, false, RCT.Read, RCT, AbilityFlag, UnitId, rcdData, nil)
			else				
				RCT:Read(AbilityFlag, UnitId, rcdData, nil)
			end
			-- CooldownValue is not passed to RCT.Read for Player because we can poll for ability details to get a more accurate reading
			-- Users notified of cooldowns via the Messaging system have to rely on the CooldownValue passed
		else
			RCT:Read(AbilityFlag, UnitId, rcdData, CooldownValue)
		end
	end
end
function RCT:OnAbilityNewCooldownEnd(UnitId, AbilityId, CooldownValue)
	local rcdData = RCT.CooldownDataCollection[AbilityId]

	if rcdData then
		if RCT.AbilityTimerTags[AbilityId] then Library.LibCron.delete(RCT.AbilityTimerTags[AbilityId]) end
	end
end
function RCT:OnCombatStatusChanged()
	RCT:UpdateAllDamageAlerts()

	if RCT.ShowBarsInCombatOnly and not CCD.Me.InCombat then
		RCT:HideAllBars()
	elseif RCT:ShouldShowBars() then
		RCT:UpdateAllBars()
	end

	RCT:OrganizeHealBars()
end
function RCT:OnAllyCountChanged()
	RCT:UpdateAllDamageAlerts()

	if RCT:ShouldShowBars() then
		RCT:UpdateAllBars()
	else
		RCT:HideAllBars()
	end

	RCT:OrganizeHealBars() -- may be removed if too much flickering
end
function RCT:OnUpdate()
	RCT:SetupAlerts()
	RCT:UpdateAllDamageAlerts() -- Removed in 3.2.4; Readded in 3.2.6

	-- battle weary debuff check (every 5 seconds)
	local timestamp = Inspect.Time.Frame()
	if not BattleWearyCheckTimestamp or timestamp >= BattleWearyCheckTimestamp + 5 then
		BattleWearyCheckTimestamp = timestamp
		local debuffCount = 0

		for k, ally in pairs(CCD.Allies.ValuesById) do
			if ally.BuffsByType[BattleWearyId] then
				debuffCount = debuffCount + 1
			end
		end

		-- check if less than half the raid has the debuff
		if debuffCount < CCD.Math.Round(CCD.Allies.Count / 2) then
			RCT:ResetFlaring(120)
		else
			RCT:ResetFlaring(300)
		end
	end
end
function RCT:OnRoleCastCompleted(Ally)
	RCT:RemoveAllCooldowns(Ally)
end
function RCT:OnRoleFlagChanged(Ally)
	if Ally.Role == false then -- empty role
		RCT:RemoveAllCooldowns(Ally)
	end
end

function RCT.ActiveHealCooldown_TimerTask(RCD)
	RCD:UpdateStatus()

	local statusColor = RCT.StatusColors[RCD.Status]
	local alert = RCD.Alert

	if RCD.Bar then
		if RCT:ShouldShowHealBars() then
			local barLength = RCT:GetHealBarLength()

			if RCD.Status == Status.Active then
				RCD.ActiveUpdateValue = barLength / RCD.Duration * RCT.CooldownTimerUpdateFrequency -- amount of Bar that is removed per update
				RCD.CurrentFill = (barLength * (RCD.Duration - RCD:GetTimeElapsed())) / RCD.Duration
				RCD.Bar:SetPercentage(RCD.CurrentFill / RCD.Bar:GetWidth() * 100)
				RCD.Bar:SetTimerText(tostring(math.floor(RCD:GetActiveTime()))) -- 3.2.3: changed from ceil to floor
			elseif RCD.Status == Status.Inert or RCD.Status == Status.Soon then
				RCD.InertFill = barLength / RCD.Cooldown * RCD:GetTimeElapsed()
				RCD.Bar:SetPercentage(RCD.InertFill / RCD.Bar:GetWidth() * 100)
				RCD.Bar:SetTimerText(tostring(math.floor(RCD:GetGreyTime()))) -- 3.2.3: changed from ceil to floor
			elseif RCD.Status == Status.Ready then
				RCD.Bar:SetPercentage(100)
				RCD.Bar:SetTimerText("Ready")
			end
			RCD.Bar:SetColor(statusColor, 0.6)
			RCD.Bar:SetTitle(RCD.Caster.Name)
			RCD.Bar:SetTexture("Rift", RCD.Icon)
			RCD.Bar:Show()
		elseif not CCD.UI.EditModeOn then
			RCD.Bar:Hide()
		end
	end

	if RCD.Alert then
		if RCT:ShouldShowHealAlert(alert) then
			if RCD.Status == Status.Active then
				alert.Texture:SetTexture("Rift", RCD.Icon)
				alert.TimerText:SetVisible(true)
				alert:SetFontColor(statusColor)
				alert:SetHeaderText(RCD.Caster.Name)
				alert.TimerText:SetText(tostring(math.floor(RCD:GetDisplayTime())))
				alert.Footer.Text:SetText(RCT.StatusText[RCD.Status])
				alert:SetVisible(true)
			elseif RCD.LastStatus == Status.Active and RCD.Status ~= Status.Active then
				RCD.Alert = nil
				alert.RCD = nil
				alert:HideX()
				RCT:OrganizeHealAlerts() -- added to reorganize alerts when one fades
			else
				alert:Hide()
			end
		elseif not CCD.UI.EditModeOn then
			alert:Hide()
		end
	end

	if RCD.Status == Status.Ready then -- timer deletion
		Library.LibCron.delete(RCD.Tag)
		RCD.Tag = nil
		--print(string.format("%s's %s timer has been deleted.", RCD.Caster.Name, RCD.Name))
	end
end
