local TOC, CCD = ...

-- globals
CCD.RCDStatus = CCD.RCDStatus or {}
CCD.RCDStatus.Active = 1
CCD.RCDStatus.Ready = 2
CCD.RCDStatus.Soon = 3
CCD.RCDStatus.Inert = 4
CCD.UI = CCD.UI or {}
CCD.UI.Colors = CCD.UI.Colors or {}

-- locals
local PriorityIndex = 1
local Locals = {}
local Colors = CCD.UI.Colors

-- Aliases
local Status = CCD.RCDStatus
local RCT = CCD.Modules.RCT

CCD.ColorBars = {}

CCD.RaidCooldown = {
	Caster,
	BuffType,
	Cooldown,		-- max cooldown in seconds i.e. Wild Growth would have value of 58.5
	Duration,		-- max duration of the buff/debuff
	Group,			-- buff stack group
	LastCast,		-- time stamp of when ability was last cast
	Frame,			-- base frame
	PriorityIndex,	-- determines how to sort cooldowns that have the same status; lower index = higher priority
	Status,			-- 1 = Active, 2 = Ready, 3 = Soon Ready, 4 = Inert
	HasReset,
}
function CCD.RaidCooldown:Create(Caster, BuffType, AbilityId, Name, Icon)
	local rcd = {}
	setmetatable(rcd, self)
	self.__index = self

	rcd.Caster = Caster
	rcd.BuffType = BuffType
	rcd.AbilityId = AbilityId
	rcd.Name = Name
	rcd.Icon = Icon
	rcd.Cooldown = CCD.Modules.RCT.CooldownDataCollection[rcd.BuffType].Cooldown
	rcd.Duration = CCD.Modules.RCT.CooldownDataCollection[rcd.BuffType].Duration
	rcd.Group = CCD.Modules.RCT.CooldownDataCollection[rcd.BuffType].Group or -1
	rcd.PriorityIndex = PriorityIndex
	rcd.Status = Status.Ready
	PriorityIndex = PriorityIndex + 1
	rcd.HasReset = false
	rcd._IsReady = false
	
	return rcd
end
function CCD.RaidCooldown:JustUsed(StartTime, Cooldown)
	self.StartTime = StartTime or Inspect.Time.Frame()
	self.Cooldown = Cooldown or self.Cooldown
	self.LastCast = self.StartTime -- some buffs don't have a details.begin member e.g. OotP
	self.Status = Status.Active
	self:SetReady(false)
end
function CCD.RaidCooldown:GetGreyTime(Timestamp)
	if not self.LastCast then
		return Inspect.Ability.New.Detail(self.AbilityId).currentCooldownRemaining or 0
	end
	
	if Timestamp == nil then
		Timestamp = Inspect.Time.Frame()
	end

	local timeElapsed = Timestamp - self.LastCast + 0 	-- seconds passed since cooldown was last used; changed from +1 to +0 in 3.2.3
	local timeLeft = self.Cooldown - timeElapsed			-- seconds left on cooldown

	return timeLeft
end
function CCD.RaidCooldown:GetActiveTime(Timestamp)
	if Timestamp == nil then
		Timestamp = Inspect.Time.Frame()
	end
	
	local timeElapsed = Timestamp - self.LastCast + 0	-- seconds passed since cooldown was last used; changed from +1 to +0 in 3.2.3
	local activeTime = self.Duration - timeElapsed		-- seconds left on cooldown

	return activeTime
end
function CCD.RaidCooldown:GetDisplayTime()
	local displayTime
	if self.Status == Status.Active then displayTime =  self:GetActiveTime()
	elseif self.Status == Status.Soon or self.Status == Status.Inert then displayTime = self:GetGreyTime()
	else displayTime = 0 end
	
	--[[if displayTime > 18000 then displayTime = "5h"
	elseif displayTime > 14400 then displayTime = "4h"
	elseif displayTime > 10800 then displayTime = "3h"
	elseif displayTime > 7200 then displayTime = "2h"
	elseif displayTime > 3600 then displayTime = "1h"
	elseif displayTime > 60 then displayTime = math.floor(displayTime/60) .. "m" end--]]
	
	return displayTime
end
function CCD.RaidCooldown:SetReady(Value)
	self._IsReady = Value
end
function CCD.RaidCooldown:IsReady()
	return self._IsReady
end
function CCD.RaidCooldown:UpdateStatus(Timestamp)
	self.LastStatus = self.Status -- not sure why this is here instead of at the top
	
	-- if cd was never cast then it must be ready
	if not self.LastCast then
		self.Status = Status.Ready
		return
	end
	
	Timestamp = Timestamp or Inspect.Time.Frame()	
	
	-- check if active
	if self:GetActiveTime(Timestamp) > 0 then
		self.Status = Status.Active
		return
	end
	
	-- if cd is ready AND not active then set to ready
	if self:IsReady() then
		self.Status = Status.Ready
		return
	end
	
	-- set status based on time since LastCast
	local ttr = self:GetGreyTime(Timestamp) -- time til ready
	if ttr <= 0 then
		self:SetReady(true)
		self.Status = Status.Ready		
	elseif ttr <= RCT.SoonReadyStartTime then
		self.Status = Status.Soon
	else
		self.Status = Status.Inert
	end
end
function CCD.RaidCooldown:GetTimeElapsed(Timestamp)
	Timestamp = Timestamp or Inspect.Time.Frame()
	return Timestamp - self.LastCast
end