local _, CCD = ...
CCD.Modules.RBT = CCD.Modules.RBT or {}
local RBT = CCD.Modules.RBT
CCD.Modules.RCT = CCD.Modules.RCT or {}
local RCT = CCD.Modules.RCT
CCD.Buff = CCD.Buff or {}
local Buff = CCD.Buff

local Target = CCD.Target
function Target:Create(UnitId)
	local details = Inspect.Unit.Detail(UnitId)
	if not details then return nil end -- don't know why this happens; reported on Johlen by Kabloom
	
	local target = {}
	setmetatable(target, self)
	self.__index = self
	
	target.Id = details.id
	target.IsHostile = (details.relation ~= "friendly")
	target.RaidDebuffs = {}
	target.RaidDebuffsById = {}

	return target
end
function Target:CreateFromDetails(Details)
	local target = {}
	setmetatable(target, self)
	self.__index = self

	local details = Details
	target.Id = details.id
	target.IsHostile = (details.relation ~= "friendly")
	target.RaidDebuffs = {}
	target.RaidDebuffsById = {}

	return target
end
function Target:ReadBuffs(IdsByType)
	for buffId, buffType in pairs(IdsByType) do
		local buff = Buff:Create(self.Id, buffId)
		--print(string.format("%s = %s = %s", buff.Name, buff.Type, buff.Icon))
		if RBT.Debuffs[buff.Type] then 		-- only read Raid Debuffs			
			self.RaidDebuffs[buff.Type] = buff
			self.RaidDebuffsById[buff.Id] = buff
			RBT:SetBuffStatus(buff, true)
		else
			--RCT:OnBuffAdd(self, buffId, buffType, nil)
		end
	end
end
function Target:ReadAllBuffs()
	local buffIds = Inspect.Buff.List(self.Id)

	for buffId in pairs(buffIds) do
		local buff = Buff:Create(self.Id, buffId)

		if RBT.Debuffs[buff.Type] then
			self.RaidDebuffs[buff.Type] = buff
			self.RaidDebuffsById[buff.Id] = buff
		end
	end
end
function Target:RemoveBuffs(BuffIds)
	for buffId in pairs(BuffIds) do
		local buff = self.RaidDebuffsById[buffId]

		if buff then
			self.RaidDebuffs[buff.Type] = nil
			self.RaidDebuffsById[buff.Id] = nil
			RBT:SetBuffStatus(buff, false)
			buff:Delete()
		end
	end
end
function Target:DeleteBuffs()
	for _, buff in pairs(self.RaidDebuffs) do
		buff:Delete()
	end
end
