local TOC, CCD = ...

-- Globals
CCD.UI = CCD.UI or {}
CCD.UI.Color = CCD.UI.Color or {}
CCD.UI.AlertFrame = CCD.UI.AlertFrame or {}
CCD.UI.AlertFrame.IconSizeRange = {
	Min = 12,
	Max = 200,
}
CCD.UI.AlertFrame.TextFontSizeRange = {
	Min = CCD.Math.Round(CCD.UI.AlertFrame.IconSizeRange.Min / 2),
	Max = CCD.Math.Round(CCD.UI.AlertFrame.IconSizeRange.Max / 2),
}

-- Aliases
local Color = CCD.UI.Color

-- Functions
function CCD.UI.AlertFrame.Create(Name, Parent)
	local frame = CCD.UI.CreateLinkableFrame(Name, Parent)
	--frame:SetBackgroundColor(Color.Black())
	--frame:Show()

	-- Header - For centering header text
	frame.Header = CCD.UI.CreateFrame(Name .. ".Header", frame, Color.Black())
	frame.Header:SetPoint("TOPLEFT", frame, "TOPLEFT")
	
	-- Header Text
	frame.Header.Text = CCD.UI.CreateText(frame.Header:GetName() .. ".Text", frame) -- parent is base frame rather than header because of opacity
	frame.Header.Text:SetEffectGlow(CCD.UI.FrameTextEffectGlow)
	frame.Header.Text:SetLayer(frame.Header:GetLayer() + 1)
	frame.Header.Text:SetPoint("CENTER", frame.Header, "CENTER")
	--frame.Header.Text:SetBackgroundColor(Color.Black())
	
	-- Texture i.e. Icon
	frame.Texture = CCD.UI.CreateTexture(frame:GetName() .. ".Texture", frame)
	frame.Texture:SetPoint("TOPLEFT", frame.Header, "BOTTOMLEFT")
	
	-- Timer
	frame.TimerText = CCD.UI.CreateText(frame:GetName() .. ".TimerText", frame)
	frame.TimerText:SetEffectGlow(CCD.UI.FrameTextEffectGlow)
	frame.TimerText:SetLayer(frame.Texture:GetLayer() + 1)
	frame.TimerText:SetPoint("CENTER", frame.Texture, "CENTER")
	frame.TimerText:SetText("100")
	frame.TimerText:Show()
	
	-- Footer
	frame.Footer = CCD.UI.CreateFrame(frame:GetName() .. ".Footer", frame, Color.Black())
	frame.Footer:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT")
	
	-- Footer Text
	frame.Footer.Text = CCD.UI.CreateText(frame.Footer:GetName() .. ".Text", frame)
	frame.Footer.Text:SetEffectGlow(CCD.UI.FrameTextEffectGlow)
	frame.Footer.Text:SetLayer(frame.Footer:GetLayer() + 1)
	frame.Footer.Text:SetPoint("CENTER", frame.Footer, "CENTER")
	--frame.Footer.Text:SetBackgroundColor(Color.Black())
	
	-- Edit Bolt
	frame.EditBolt:SetPoint("CENTER", UIParent, "CENTER", 0, 0)
	
	frame.IsMobile = true
	frame.IconSize = 48 -- IconSize is actually just Icon Size
	frame.HeaderText = "1234567890"
	frame.HeaderTextLength = 5
	frame.HeaderTextFontSize = 18
	frame.TimerTextFontSize = frame.IconSize / 2
	frame.FooterTextLength = 5
	frame.FooterText = "abcdefghijk"
	frame.FooterTextFontSize = 18
	frame.Alpha = 1
	frame.HeaderOn = true
	frame.FooterOn = true	
	frame.TimerTextFontSizeLocked = false
	
	frame.OnIconSizeChanged = nil
	frame.OnTimerTextFontSizeChanged = nil
		
	function frame.Event:WheelForward()
		if not CCD.UI.EditModeOn then return end
		self:SetIconSize(self.IconSize + 1)
	end
	function frame.Event:WheelBack()
		if not CCD.UI.EditModeOn then return end
		self:SetIconSize(self.IconSize - 1)
	end
	--
	function frame:UpdateHeight()
		local height = frame.IconSize
		if frame.HeaderOn then height = height + frame.Header:GetHeight() end
		if frame.FooterOn then height = height + frame.Footer:GetHeight() end
		frame:SetHeight(height)
	end
	function frame:UpdateWidth()		
		frame.Header:SetWidth(frame.IconSize)
		--frame.Header.Text:SetWidth(frame.TextureWidth)
		frame.Footer:SetWidth(frame.IconSize)
		--frame.Footer.Text:SetWidth(frame.TextureWidth)
		frame:SetWidth(frame.IconSize)
	end
	function frame:UpdateSize()
		frame:UpdateWidth()
		frame:UpdateHeight()
	end
	function frame:ShowHeader()
		frame.HeaderOn = true
		frame.Header:Show()
		frame.Header.Text:Show()
		frame.Texture:SetPoint("TOPLEFT", frame.Header, "BOTTOMLEFT")
		frame:UpdateHeight()
	end
	function frame:HideHeader()
		frame.HeaderOn = false
		frame.Header:Hide()
		frame.Header.Text:Hide()
		frame.Texture:SetPoint("TOPLEFT", frame, "TOPLEFT")
		frame:UpdateHeight()
	end
	function frame:SetHeaderVisibility(IsVisible)
		if IsVisible == nil then IsVisible = frame.HeaderOn end
		if IsVisible then
			frame:ShowHeader()
		else
			frame:HideHeader()
		end
	end
	function frame:ShowFooter()
		frame.FooterOn = true
		frame.Footer:Show()
		frame.Footer.Text:Show()
		frame:UpdateHeight()
	end
	function frame:HideFooter()
		frame.FooterOn = false
		frame.Footer:Hide()
		frame.Footer.Text:Hide()
		frame:UpdateHeight()
	end
	function frame:SetFooterVisibility(IsVisible)
		if IsVisible == nil then IsVisible = frame.FooterOn end
		if IsVisible then
			frame:ShowFooter()
		else
			frame:HideFooter()
		end
	end
	--
	function frame:SetFontColor(Color)
		frame.Header.Text:SetFontColor(Color)
		frame.TimerText:SetFontColor(Color)
		frame.Footer.Text:SetFontColor(Color)
	end
	function frame:SetHeaderText(Text)
		frame.HeaderText = Text or frame.HeaderText
		frame.Header.Text:SetText(string.sub(frame.HeaderText, 1, frame.HeaderTextLength))
	end
	function frame:SetHeaderTextLength(Length)
		frame.HeaderTextLength = Length or frame.HeaderTextLength
		frame.HeaderTextLength = math.min(frame.HeaderTextLength, 19)
		frame.Header.Text:SetText(string.sub(frame.HeaderText, 1, frame.HeaderTextLength))
	end
	function frame:SetFooterText(Text)
		frame.FooterText = Text or frame.FooterText
		frame.Footer.Text:SetText(string.sub(frame.FooterText, 1, frame.FooterTextLength))
	end
	function frame:SetFooterTextLength(Length)
		frame.FooterTextLength = Length or frame.FooterTextLength
		frame.FooterTextLength = math.min(frame.FooterTextLength, 19)
		frame.Footer.Text:SetText(string.sub(frame.FooterText, 1, frame.FooterTextLength))
	end
	--
	function frame:SetHeaderTextFontSize(Size)
		frame.HeaderTextFontSize = Size or frame.HeaderTextFontSize
		frame.Header.Text:SetFontSize(frame.HeaderTextFontSize)
		frame.Header:SetHeight(frame.Header.Text:GetHeight())
		frame:UpdateHeight()
	end
	function frame:SetFooterTextFontSize(Size)
		frame.FooterTextFontSize = Size or frame.FooterTextFontSize
		frame.Footer.Text:SetFontSize(frame.FooterTextFontSize)
		frame.Footer:SetHeight(frame.Footer.Text:GetHeight())
		frame:UpdateHeight()
	end
	function frame:SetIconSize(IconSize, Silent, DontScale)
		frame.IconSize = CCD.Math.MinMax(IconSize or frame.IconSize, CCD.UI.AlertFrame.IconSizeRange.Min, CCD.UI.AlertFrame.IconSizeRange.Max)
		frame.Texture:SetSize(frame.IconSize)
		frame:UpdateSize()
		
		if not Silent and frame.OnIconSizeChanged then frame:OnIconSizeChanged() end
		if not DontScale and not frame.TimerTextFontSizeLocked then	frame:SetTimerTextFontSize(CCD.Math.Round(frame.IconSize * 0.5)) end		
	end
	function frame:SetTimerTextFontSize(Size)
		frame.TimerTextFontSize = Size or frame.TimerTextFontSize
		if frame.OnTimerTextFontSizeChanged then frame:OnTimerTextFontSizeChanged() end
		frame.TimerText:SetFontSize(frame.TimerTextFontSize)	
	end
	--	
	function frame:SetAlphaFromSlider(SliderPosition)
		if not SliderPosition then return end
		frame.Alpha = SliderPosition / 100		
		frame:UpdateAlpha()
	end
	function frame:UpdateAlpha()
		--frame:SetBackgroundAlpha(frame.Alpha)
		frame.Texture:SetAlpha(frame.Alpha)
		frame.Header:SetBackgroundAlpha(frame.Alpha)
		frame.Footer:SetBackgroundAlpha(frame.Alpha)
		-- text if unaffected
	end
	function frame:SetAlpha(Value)
		frame.Alpha = Value or frame.Alpha
		frame:UpdateAlpha()
	end
	--
	function frame:Initialize()
		self:SetIconSize(nil, true, true)
		self:SetTimerTextFontSize()	
		self:UpdateAlpha()
		self:SetHeaderVisibility()
		self:SetHeaderText()
		self:SetHeaderTextLength()
		self:SetHeaderTextFontSize()
		self:SetFooterVisibility()
		self:SetFooterText()
		self:SetFooterTextFontSize()		
	end
	
	frame:Initialize()
	return frame
end

