local TOC, CCD = ...
CCD.UI = CCD.UI or {}

function CCD.UI.CreateColorPicker(Name, Parent, Text, ShowAlphaSlider)
	local container = CCD.UI.CreateFrame(Name, Parent)
	--Library.LibSimpleWidgets.SetBorder("plain", container, 1, 1, 1, 1, 1, "tblr")
	container.OnColorChange = nil -- treated as an event
	local OnSliderChange

	local label = UI.CreateFrame("Text", Name .. "_container_label", container)
	--label:SetWidth(self:GetWidth())
	label:SetFontSize(18)
	label:SetText(Text)
	label:SetPoint("TOPLEFT", container, "TOPLEFT")
	label:SetVisible(true)

	-- colour bar
	local colorBar = UI.CreateFrame("Frame", Name, container)
	--colorBar:SetWidth(container:GetWidth() - label:GetWidth() - 4)
	colorBar:SetHeight(30)
	colorBar:SetBackgroundColor(1, 1, 1, 1)
	colorBar:SetPoint("TOPLEFT", label, "TOPRIGHT", 4, 0)
	colorBar:SetPoint("TOPRIGHT", container, "TOPRIGHT")
	colorBar:SetVisible(true)

	-- Red Slider
	container.RedSlider = CCD.UI.CreateSlider(Name .. ".RedSlider", container)
	container.RedSlider:SetText("Red", 15, CCD.UI.Colors.White)
	container.RedSlider:SetRange(0, 100, 100)
	container.RedSlider:SetPoint("TOPLEFT", label, "TOPLEFT", 0, colorBar:GetHeight() + 5)

	container.GreenSlider = CCD.UI.CreateSlider(Name .. "greenSlider", container)
	container.GreenSlider:SetText("Green", 15, CCD.UI.Colors.White)
	container.GreenSlider:SetRange(0, 100, 100)
	container.GreenSlider:SetPoint("TOPLEFT", container.RedSlider, "BOTTOMLEFT")

	container.BlueSlider = CCD.UI.CreateSlider(Name .. "blueSlider", container, "Blue", 0, 100, 100, 50)
	container.BlueSlider:SetText("Blue", 15, CCD.UI.Colors.White)
	container.BlueSlider:SetRange(0, 100, 100)
	container.BlueSlider:SetPoint("TOPLEFT", container.GreenSlider, "BOTTOMLEFT")
	
	-- set slider label widths so that they are all parallel
	local width = math.max(container.RedSlider.Label:GetWidth(), container.GreenSlider.Label:GetWidth(), container.BlueSlider.Label:GetWidth()) + 17
	container.RedSlider.Label:SetWidth(width)
	container.GreenSlider.Label:SetWidth(width)
	container.BlueSlider.Label:SetWidth(width)

	local frameHeight = colorBar:GetHeight() + container.RedSlider:GetHeight() * 3 + 10

	if ShowAlphaSlider then
		container.AlphaSlider = CCD.UI.CreateSlider(Name .. ".AlphaSlider", container)
		container.AlphaSlider:SetText("Opacity", 15, CCD.UI.Colors.White)
		container.AlphaSlider.Label:SetWidth(width)
		container.AlphaSlider:SetRange(0, 100, 100)
		container.AlphaSlider:SetPoint("TOPLEFT", container.BlueSlider, "BOTTOMLEFT")
		function container.AlphaSlider.Slider.Event:SliderChange()
			OnSliderChange()
		end
		function container:GetAlpha()
			return container.AlphaSlider:GetPosition() / 100
		end
		frameHeight = frameHeight + container.AlphaSlider.Label:GetHeight()
	end	

	OnSliderChange = function ()
		local r = container.RedSlider:GetPosition() / 100
		local g = container.GreenSlider:GetPosition() / 100
		local b = container.BlueSlider:GetPosition() / 100
		local a = 1
		if ShowAlphaSlider then
			a = container.AlphaSlider:GetPosition() / 100
		end

		colorBar:SetBackgroundColor(r, g, b, a)

		if container.OnColorChange then
			container.OnColorChange(r, g, b, a)
		end
	end
	function container.RedSlider.Slider.Event:SliderChange()
		OnSliderChange()
	end
	function container.GreenSlider.Slider.Event:SliderChange()
		OnSliderChange()
	end
	function container.BlueSlider.Slider.Event:SliderChange()
		OnSliderChange()
	end

	function container:SetColor(R, G, B, A)
		if not G then
			local color = R
			R = color.R
			G = color.G
			B = color.B
			A = color.A or 1 -- provides backward compatibility with 3.0.0 where fonts didn't save their alpha value
		end
		local sliderR = math.floor(R * 100)
		local sliderG = math.floor(G * 100)
		local sliderB = math.floor(B * 100)
		local sliderA = math.floor(A * 100)

		colorBar:SetBackgroundColor(R, G, B, A)
		container.RedSlider.Slider:SetPosition(sliderR, true)
		container.GreenSlider:SetPosition(sliderG, true)
		container.BlueSlider:SetPosition(sliderB, true)

		if ShowAlphaSlider then
			container.AlphaSlider:SetPosition(sliderA, true)
		end
	end
	function container:GetRed()
		return container.RedSlider:GetPosition() / 100
	end
	function container:GetGreen()
		return container.GreenSlider:GetPosition() / 100
	end
	function container:GetBlue()
		return container.BlueSlider:GetPosition() / 100
	end

	container:SetWidth(container.RedSlider:GetWidth())
	container:SetHeight(frameHeight)

	return container
end
