local TOC, CCD = ...
CCD.UI = CCD.UI or {}
CCD.UI.Context = CCD.UI.Context or UI.CreateContext("CCDContext")
CCD.UI.ContextMenu = {}
CCD.UI.ContextMenu.Saved = {X = 0, Y = 0}

function CCD.UI.CreateContextMenu(Name1, Width, Height, Title, Immobile)
	local description = Name1

	-- Base Frame
	local frame = UI.CreateFrame("Frame", description, CCD.UI.Context)
	frame.Load = nil -- this stores a function that is called before frame:Show(); use to initialize frame before showing it
	frame.Children = {}
	frame:SetWidth(Width)
	frame:SetHeight(Height)
	frame:SetBackgroundColor(0, 0, 0, 0.85) -- heavy black tint
	frame:SetLayer(2)
	Library.LibSimpleWidgets.SetBorder("rounded", frame)
	frame:SetVisible(false)
	--
	function frame.Event:WheelForward()
		if not CCD.UI.EditModeOn then return end
		frame:SetHeight(frame:GetHeight() + 1)
		print(frame:GetHeight())
	end
	function frame.Event:WheelBack()
		if not CCD.UI.EditModeOn then return end
		frame:SetHeight(frame:GetHeight() - 1)
		print(frame:GetHeight())
	end--]]

	
	frame.OnNewElementHighlighted = nil
	frame.HighlightedElement = nil
	frame.OnClosed = nil
	frame.OnLostFocus = nil
	
	function frame:Show(Anchor)
		assert(type(Anchor) == "table", "Parameter 1 (Anchor) must be a valid frame!")
		
		if self.Load then self:Load() end
		
		if Immobile then
			self:ClearPoint("BOTTOMLEFT")
			self:ClearPoint("TOPLEFT")
			self:ClearPoint("BOTTOMRIGHT")
			self:ClearPoint("TOPRIGHT")		
			
			if Anchor:GetBottom() > UIParent:GetHeight() / 2 then
				if Anchor:GetLeft() > UIParent:GetWidth() / 2 then
					self:SetPoint("BOTTOMRIGHT", Anchor, "TOPLEFT", -20, 0)
				else
					self:SetPoint("BOTTOMLEFT", Anchor, "TOPLEFT", 0, -15)
				end
			else
				if Anchor:GetLeft() > UIParent:GetWidth() / 2 then
					self:SetPoint("TOPRIGHT", Anchor, "BOTTOMLEFT", -5, 0)
				else
					self:SetPoint("TOPLEFT", Anchor, "BOTTOMLEFT", 0, 5)
				end
			end
		else
			local anchorMiddle = Anchor:GetTop() + Anchor:GetHeight() / 2
			local menuHalf = self:GetHeight() / 2
			local futureBottom = anchorMiddle + menuHalf
			local futureTop = anchorMiddle - menuHalf
			local bottomOver = math.max(futureBottom - UIParent:GetHeight(), 0)
			local topOver = math.max(0 - futureTop, 0)
			
			--print("futureBottom = " .. futureBottom)
			--print("UIParent height = " .. UIParent:GetHeight())
			--print("bottomOver = " .. bottomOver)
			--print("topOver = " .. topOver)
			--[[
			--]]

			self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", CCD.UI.ContextMenu.Saved.X, CCD.UI.ContextMenu.Saved.Y)
		end
		
		self:SetVisible(true)
		CCD.UI.SetActiveContextMenu(self)
	end
	function frame:ShowAtAnchor(Anchor)
		assert(type(Anchor) == "table", "Parameter 1 (Anchor) must be a valid frame!")
		
		if self.Load then self:Load() end
		
		self:ClearPoint("CENTERLEFT")
		self:ClearPoint("CENTERRIGHT")		
		self:ClearPoint("TOPLEFT") -- for manual positioning
		
		local anchorMiddle = Anchor:GetTop() + Anchor:GetHeight() / 2
		local menuHalf = self:GetHeight() / 2
		local futureBottom = anchorMiddle + menuHalf
		local futureTop = anchorMiddle - menuHalf
		local bottomOver = math.max(futureBottom - UIParent:GetHeight(), 0)
		local topOver = math.max(0 - futureTop, 0)
		
		--print("futureBottom = " .. futureBottom)
		--print("UIParent height = " .. UIParent:GetHeight())
		--print("bottomOver = " .. bottomOver)
		--print("topOver = " .. topOver)
		--[[
		if Anchor:GetLeft() > UIParent:GetWidth() / 2 then
			--print("top if")
			if bottomOver > 0 then
				self:SetPoint("CENTERRIGHT", Anchor, "CENTERLEFT", -15, -bottomOver)
			else
				self:SetPoint("CENTERRIGHT", Anchor, "CENTERLEFT", -15, topOver)
			end
		else
			--print("bottom if")
			if bottomOver > 0 then
				self:SetPoint("CENTERLEFT", Anchor, "CENTERRIGHT", 0, -bottomOver)
			else
				self:SetPoint("CENTERLEFT", Anchor, "CENTERRIGHT", 0, topOver)
			end
		end--]]

		self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", CCD.UI.ContextMenu.Saved.X, CCD.UI.ContextMenu.Saved.Y)
		self:SetVisible(true)
		CCD.UI.SetActiveContextMenu(self)
	end
	
	function frame:Hide()
		self:SetVisible(false)
		CCD.UI.SetActiveContextMenu(nil)
	end
	function frame:Toggle(Anchor)
		assert(type(Anchor) == "table", "Parameter 1 (Anchor) must be a valid frame!")
		
		if self:GetVisible() then
			self:Hide()
		else
			self:Show(Anchor)
		end
	end
	if not Immobile then
		function frame.Event:LeftDown()
			self.MouseDown = true
			mouseData = Inspect.Mouse()
			self.MyStartX = self:GetLeft()
			self.MyStartY = self:GetTop()
			self.StartX = mouseData.x - self.MyStartX
			self.StartY = mouseData.y - self.MyStartY
		end
		function frame.Event:MouseMove(mouseX, mouseY)
			if self.MouseDown then
				self:ClearPoint("CENTERLEFT")
				self:ClearPoint("CENTERRIGHT")	
				self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", (mouseX - self.StartX), (mouseY - self.StartY))
			end
		end
		function frame.Event:LeftUp()
			if self.MouseDown then
				CCD.UI.ContextMenu.Saved.X = self:GetLeft()
				CCD.UI.ContextMenu.Saved.Y = self:GetTop()
			end
			self.MouseDown = false
		end
	end
	local verticalSpacing
	
	function frame:AddMenuLine(Child)
		local line = UI.CreateFrame("Texture", Child:GetName() .. "_line", frame)
		Child.Line = line
		
		line:SetWidth(frame:GetWidth())
		line:SetHeight(2)
		line:SetTexture("CCD", "Images\\menu line.png")
		line:SetLayer(5)
		line:SetPoint("TOPLEFT", Child, "BOTTOMLEFT", 0, 5)
		line:SetVisible(true)
		function line:AnchorTo(Frame)
			if Frame == "ORIGINAL" then
				self:SetPoint("TOPLEFT", Child, "BOTTOMLEFT", 0, 5)
			else
				line:SetPoint("TOPLEFT", Frame, "BOTTOMLEFT", 0, 5)
			end
		end
		
		frame.LastChild = line
		verticalSpacing = 10
		return line
	end
	function frame:AddCheckbox(Name, Text, DrawLine, Anchor, IncludeArrow)
		local checkbox = CCD.UI.CreateCheckbox(self:GetName() .. "_" .. Name, frame)
		local anchor = Anchor or self.LastChild
		
		checkbox:SetChecked(true)
		checkbox:SetFontSize(18)
		checkbox:SetText(Text)
		checkbox:SetWidth(frame:GetWidth())
		checkbox:SetPoint("TOPLEFT", Anchor or self.LastChild, "BOTTOMLEFT", 0, verticalSpacing)
		checkbox:SetVisible(true)
		function checkbox:AnchorTo(Frame, Spacing)
			if Frame == "ORIGINAL" then
				self:SetPoint("TOPLEFT", anchor, "BOTTOMLEFT", 0, verticalSpacing)
			else
				self:SetPoint("TOPLEFT", Frame, "BOTTOMLEFT", 0, Spacing or verticalSpacing)
			end
		end		
		--[[function checkbox.Event:MouseIn()
			if frame.HighlightedElement ~= self then
				frame.HighlightedElement = self
				if frame.OnNewElementHighlighted then frame:OnNewElementHighlighted() end
			end
		end--]]
		function checkbox.Event:MouseIn()
			self:SetBackgroundColor(0.25, 0.25, 0.25, 1) -- heavy grey
			
			if frame.HighlightedElement ~= self then
				frame.HighlightedElement = self
				if frame.OnNewElementHighlighted then frame:OnNewElementHighlighted(self) end
			end
		end
		function checkbox.Event:MouseOut()
			self:SetBackgroundColor(0, 0, 0, 0) -- transparent
		end
		
		if IncludeArrow then
			checkbox.Arrow = CCD.UI.CreateTexture(self:GetName() .. ".Arrow", checkbox)
			checkbox.Arrow:SetTexture("CCD", "Images\\menu_right_arrow.png")
			checkbox.Arrow:SetSize(checkbox:GetHeight())			
			checkbox.Arrow:SetPoint("TOPRIGHT", checkbox, "TOPRIGHT")
			checkbox.Arrow:SetVisible(true)
		end
		
		self.LastChild = checkbox
		verticalSpacing = 0

		if DrawLine then
			self:AddMenuLine(checkbox)
		end

		return checkbox
	end
	function frame:AddSlider(Name, Text, MinValue, MaxValue, DrawLine)
		local slider = CCD.UI.CreateSlider(Name, frame, Text)
		slider:SetRange(MinValue, MaxValue)
		slider.Label:SetFontSize(18)
		slider.Slider:SetWidth(slider.Slider:GetWidth() - slider.Label:GetWidth() + 45)
		slider:SetPoint("TOPLEFT", self.LastChild, "BOTTOMLEFT", 0, verticalSpacing)
		slider:SetVisible(true)
		slider.BaseFunctions = {
			SetVisible = slider.SetVisible
		}
		function slider.Event:MouseIn()
			self:SetBackgroundColor(0.25, 0.25, 0.25, 1) -- heavy grey
			
			if frame.HighlightedElement ~= self then
				frame.HighlightedElement = self
				if frame.OnNewElementHighlighted then frame:OnNewElementHighlighted(self) end
			end
		end
		function slider.Event:MouseOut()
			self:SetBackgroundColor(0, 0, 0, 0) -- transparent
		end
		
		self.LastChild = slider
		verticalSpacing = 0

		if DrawLine then
			self:AddMenuLine(self.LastChild)
		end

		return slider
	end
	function frame:AddTextField(Name, Text, Text2, DrawLine, Anchor)
		--[[ Label
		local label = CCD.UI.CreateText(Name .. ".Label", frame)
		label:SetFontSize(18)
		label:SetText(Text)
		label:SetPoint("TOPLEFT", Anchor or frame.LastChild, "BOTTOMLEFT", 0, verticalSpacing)
		label:SetVisible(true)

		-- textField
		local textField = EnKai.uiCreateFrame("nkTextField", Name .. ".TextField", frame)
		textField:SetWidth(40)
		textField:SetValue("valueType", "number")
		textField:SetPoint("TOPLEFT", label, "TOPRIGHT", 2, 2)
		textField:SetVisible(true)
		function textField:SetTextFieldChangedEvent(OnTextFieldChanged)
			Command.Event.Attach(EnKai.events[textField:GetName()].TextfieldChanged, OnTextFieldChanged, self:GetName() .. "_TextfieldChanged")
		end
		
		if Text2 then
			-- Timer Show Value Label #2
			local label2 = UI.CreateFrame("Text", self:GetName() .. "_" .. Name .. "Label2", frame)
			label2:SetFontSize(18)
			label2:SetText(Text2)
			label2:SetPoint("TOPLEFT", textField, "TOPRIGHT", 2, 0)
			label2:SetVisible(true)
		end--]]
		
		local textField = CCD.UI.CreateTextField(Name, frame, Text, Text2)
		textField:SetFontSize(18)
		textField:SetTextFieldWidth(40)
		textField:SetPoint("TOPLEFT", Anchor or frame.LastChild, "BOTTOMLEFT", 0, verticalSpacing)
		function textField.Event:MouseIn()			
			if frame.HighlightedElement ~= self then
				frame.HighlightedElement = self
				if frame.OnNewElementHighlighted then frame:OnNewElementHighlighted(self) end
			end
		end
		
		self.LastChild = textField
		verticalSpacing = 0

		if DrawLine then
			self:AddMenuLine(textField)
		end

		return textField
	end
	function frame:AddLabel(Name, Text, DrawLine)
		local label = UI.CreateFrame("Text", self:GetName() .. "_" .. Name, frame)
		label:SetWidth(self:GetWidth())
		label:SetFontSize(18)
		label:SetText(Text)
		label:SetPoint("TOPLEFT", self.LastChild, "BOTTOMLEFT", 0, verticalSpacing)
		label:SetVisible(true)
		function label.Event:MouseIn()
			self:SetBackgroundColor(0, 0, 1, 0.5) -- medium blue tint
			if frame.HighlightedElement ~= self then
				frame.HighlightedElement = self
				if frame.OnNewElementHighlighted then frame:OnNewElementHighlighted(self) end
			end
		end
		function label.Event:MouseOut()
			self:SetBackgroundColor(0, 0, 0, 0) -- transparent
		end

		self.LastChild = label
		verticalSpacing = 0

		if DrawLine then
			self:AddMenuLine(label)
		end

		return label
	end
	function frame:AddHeader(Name, Text)
		local label = UI.CreateFrame("Text", self:GetName() .. "_" .. Name, frame)
		label:SetWidth(self:GetWidth())
		label:SetFontSize(18)
		label:SetText(Text)
		label:SetPoint("TOPLEFT", self.LastChild, "BOTTOMLEFT", 0, 20)
		label:SetVisible(true)

		self.LastChild = label
		self:AddHeaderLine(label)

		return label
	end
	function frame:AddHeaderLine(Frame)
		local newLine = UI.CreateFrame("Texture", Frame:GetName() .. "_newLine", Frame:GetParent())
		table.insert(self.Children, newLine)
		newLine:SetWidth(Frame:GetParent():GetWidth())
		newLine:SetHeight(2)
		newLine:SetTexture("CCD", "Images\\header line.png")
		newLine:SetLayer(5)
		newLine:SetPoint("TOPLEFT", Frame, "BOTTOMLEFT", 0, 5)
		newLine:SetVisible(true)
		self.LastChild = newLine
		verticalSpacing = 5
		return newLine
	end
	function frame:AddFooterNote(Name, Text)
		local container = CCD.UI.CreateFrame(Name .. "_container", frame)
		container:SetWidth(frame:GetWidth())
		container:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT")
		container:SetVisible(true)
		function container.Event:MouseIn()
			self:SetBackgroundColor(0, 0, 1, 0.5) -- medium blue tint
		end
		function container.Event:MouseOut()
			self:SetBackgroundColor(0, 0, 0, 0) -- transparent
		end
		
		container.Label = CCD.UI.CreateText(container:GetName() .. ".Label", container)
		container.Label:SetText(Text, 18, CCD.UI.Colors.Green)
		container.Label:SetWordwrap(true)
		container.Label:SetPoint("CENTER", container, "CENTER", 0, 0)
		container.Label:SetVisible(true)
		
		container:SetHeight(container.Label:GetHeight())
		
		return container
	end
	function frame:AddColourPicker(Name, Text, ShowAlphaSlider, DrawLine)
		local container = CCD.UI.CreateColorPicker(Name, self, Text, ShowAlphaSlider)
		container:SetPoint("TOPLEFT", self.LastChild, "BOTTOMLEFT", 0, verticalSpacing)
	
		self.LastChild = container
		verticalSpacing = 0

		if DrawLine then self:AddMenuLine(self.LastChild) end

		return container
	end
	function frame:AddRadioButton(Name, Text, ToTheRight, DrawLine)
		local radioButton = CCD.UI.RadioButton:Create(self:GetName() .. "_" .. Name, frame, Text)
		function radioButton.Event:MouseIn()			
			if frame.HighlightedElement ~= self then
				frame.HighlightedElement = self
				if frame.OnNewElementHighlighted then frame:OnNewElementHighlighted(self) end
			end
		end
		--radioButton:SetFontSize(18)
		--radioButton:SetText(Text)
		if ToTheRight then
			radioButton:SetPoint("TOPLEFT", self.LastChild, "TOPRIGHT", 0, 0)
		else
			radioButton:SetPoint("TOPLEFT", self.LastChild, "BOTTOMLEFT", 0, verticalSpacing)
		end
		radioButton:SetVisible(true)
		self.LastChild = radioButton
		verticalSpacing = 0

		if DrawLine then
			self:AddMenuLine(radioButton)
		end

		return radioButton
	end
	
	-- Header Frame
	frame.Header = UI.CreateFrame("Frame", description .. "_Header", frame)
	table.insert(frame.Children, frame.Header)
	frame.Header:SetWidth(frame:GetWidth() - 2)
	frame.Header:SetBackgroundColor(0/255, 30/255, 205/255, 0) -- blue but transparent
	frame.Header:SetPoint("TOPLEFT", frame, "TOPLEFT", 1, 1)
	frame.Header:SetVisible(true)
	--Library.LibSimpleWidgets.SetBorder("plain", frame.Header, 1, 1, 1, 1, 1, "tblr") -- thin white border

	-- Header Text
	frame.HeaderText = CCD.UI.CreateText(description .. "_HeaderText", frame.Header)
	frame.HeaderText:SetText(Title or frame.HeaderText:GetName(), 22, CCD.UI.Colors.Orange)
	frame.HeaderText:SetEffectGlow(CCD.UI.DefaultTextEffectGlow)
	frame.HeaderText:SetPoint("TOPCENTER", frame.Header, "TOPCENTER", 0, 0)
	frame.HeaderText:SetVisible(true)
	local headerTextHeight = frame.HeaderText:GetHeight()

	-- Close Button
	frame.CloseButtonTexture = UI.CreateFrame("Texture", description .. "_CloseButtonTexture", frame.Header)
	frame.CloseButtonTexture:SetTexture("CCD", "Images\\ContextMenuCloseButton.png")
	frame.CloseButtonTexture:SetWidth(headerTextHeight)
	frame.CloseButtonTexture:SetHeight(headerTextHeight)
	frame.CloseButtonTexture:SetPoint("TOPRIGHT", frame.Header, "TOPRIGHT", 0, 0)
	frame.CloseButtonTexture:SetVisible(true)
	function frame.CloseButtonTexture.Event:MouseIn()
		frame.CloseButtonTexture:SetTexture("CCD", "Images\\ContextMenuCloseButtonHighlight.png")
	end
	function frame.CloseButtonTexture.Event:MouseOut()
		frame.CloseButtonTexture:SetTexture("CCD", "Images\\ContextMenuCloseButton.png")
	end
	function frame.CloseButtonTexture.Event:LeftClick()
		frame:Hide()
		if frame.OnClosed then frame:OnClosed() end
	end

	-- Set Header Height
	frame.Header:SetHeight(headerTextHeight)

	-- Header Line
	frame.HeaderLine = UI.CreateFrame("Texture", description .. "_HeaderLine", frame)
	table.insert(frame.Children, frame.HeaderLine)
	frame.LastChild = frame.HeaderLine
	frame.HeaderLine:SetWidth(frame:GetWidth())
	frame.HeaderLine:SetHeight(2)
	frame.HeaderLine:SetTexture("CCD", "Images\\header line.png")
	frame.HeaderLine:SetPoint("TOPLEFT", frame.Header, "BOTTOMLEFT", 0, 5)
	frame.HeaderLine:SetVisible(true)
	verticalSpacing = 5

	return frame
end

function CCD.UI.SetActiveContextMenu(Frame)
	if CCD.ActiveContextMenu and CCD.ActiveContextMenu ~= Frame then
		CCD.ActiveContextMenu:SetVisible(false)
		
		if CCD.ActiveContextMenu.Group then
			CCD.ActiveContextMenu.Group = -1
		end
		
		if CCD.ActiveContextMenu.OnLostFocus then CCD.ActiveContextMenu:OnLostFocus() end
	end

	CCD.ActiveContextMenu = Frame
end

function CCD.UI.ContextMenu:CreateSaveTable()
	return self.Saved
end
function CCD.UI.ContextMenu:Load(SaveTable)
	if not SaveTable then return end
	for k, v in pairs(SaveTable) do
		self.Saved[k] = v
	end
end
