local TOC, CCD = ...

-- Globals
CCD.UI = CCD.UI or {}
CCD.UI.LinkableFramesById = {}
CCD.UI.Colors = CCD.UI.Colors or {}
CCD.UI.EditModeOn = false
CCD.UI.Context = CCD.UI.Context or UI.CreateContext("CCDContext")

-- Locals
local EditBoltLockedTexture = "Images\\Gadget Locked.png"

CCD.UI.EditModeWindow = CCD.UI.CreateFrame("EditModeWindow", CCD.UI.Context, CCD.UI.Color.Black(), 180, 90)
local EditModeWindow = CCD.UI.EditModeWindow
--EditModeWindow.Frame:SetBackgroundColor(0, 0, 0, 0.5) -- medium black tint
EditModeWindow.Saved = { FrameSpacing = 0}
-- point is set initially in SetEditMode
EditModeWindow:SetLayer(5)
Library.LibSimpleWidgets.SetBorder("rounded", EditModeWindow)
--EditModeWindow.Frame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", self.X, self.Y)
EditModeWindow:SetVisible(false)
function EditModeWindow.Event:LeftDown()
	self.MouseDown = true
	mouseData = Inspect.Mouse()
	self.MyStartX = self:GetLeft()
	self.MyStartY = self:GetTop()
	self.StartX = mouseData.x - self.MyStartX
	self.StartY = mouseData.y - self.MyStartY
end
function EditModeWindow.Event:MouseMove(mouseX, mouseY)
	if self.MouseDown then
		self:ClearPoint("TOPLEFT")
		self:ClearPoint("CENTER")
		self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", (mouseX - self.StartX), (mouseY - self.StartY))
	end
end
function EditModeWindow.Event:LeftUp()
	if self.MouseDown then
		self.MouseDown = false
	end
end
function EditModeWindow.Event:Move()
	EditModeWindow.Saved.X = self:GetLeft()
	EditModeWindow.Saved.Y = self:GetTop()
end
--
-- Close Button
EditModeWindow.CloseButtonTexture = UI.CreateFrame("Texture", "EditModeWindow.CloseButtonTexture", EditModeWindow)
EditModeWindow.CloseButtonTexture:SetTexture("CCD", "Images\\ContextMenuCloseButton.png")
EditModeWindow.CloseButtonTexture:SetWidth(24)
EditModeWindow.CloseButtonTexture:SetHeight(24)
EditModeWindow.CloseButtonTexture:SetPoint("TOPRIGHT", EditModeWindow, "TOPRIGHT", 0, 0)
EditModeWindow.CloseButtonTexture:SetVisible(true)
function EditModeWindow.CloseButtonTexture.Event:MouseIn()
	EditModeWindow.CloseButtonTexture:SetTexture("CCD", "Images\\ContextMenuCloseButtonHighlight.png")
end
function EditModeWindow.CloseButtonTexture.Event:MouseOut()
	EditModeWindow.CloseButtonTexture:SetTexture("CCD", "Images\\ContextMenuCloseButton.png")
end
function EditModeWindow.CloseButtonTexture.Event:LeftClick()
	EditModeWindow:Hide()
end

--
EditModeWindow.Button = UI.CreateFrame("RiftButton", "EditModeWindow.Button", EditModeWindow)
EditModeWindow.Button:SetText("Show Disabled")
EditModeWindow.Button:SetPoint("TOPCENTER", EditModeWindow, "TOPCENTER")
EditModeWindow.Button:SetVisible(true)
function EditModeWindow.Button.Event:LeftPress()
	-- only recognize click if frame has not been moved
	if not EditModeWindow.HasMoved then
		for _, frame in pairs(CCD.UI.LinkableFramesById) do
			if not frame.IsMobile and not frame.NeverShow then		
				frame:ShowEditBolt()
				frame:ShowForeground()
				frame:Show()
			end
		end
		
		--EditModeWindow:Hide()
	end
	
	EditModeWindow.HasMoved = false
end
function EditModeWindow.Button.Event:LeftDown()
	EditModeWindow.MouseDown = true
	mouseData = Inspect.Mouse()
	EditModeWindow.MyStartX = EditModeWindow:GetLeft()
	EditModeWindow.MyStartY = EditModeWindow:GetTop()
	EditModeWindow.StartX = mouseData.x - EditModeWindow.MyStartX
	EditModeWindow.StartY = mouseData.y - EditModeWindow.MyStartY
end
function EditModeWindow.Button.Event:MouseMove(mouseX, mouseY)
	if EditModeWindow.MouseDown then
		EditModeWindow:ClearPoint("TOPLEFT")
		EditModeWindow:ClearPoint("CENTER")
		EditModeWindow:SetPoint("TOPLEFT", UIParent, "TOPLEFT", (mouseX - EditModeWindow.StartX), (mouseY - EditModeWindow.StartY))
		EditModeWindow.HasMoved = true
	end
end
function EditModeWindow.Button.Event:LeftUp()
	if EditModeWindow.MouseDown then
		EditModeWindow.MouseDown = false
	end
	--self.DownTime = false
end
function EditModeWindow.Button.Event:Move()
	EditModeWindow.X = EditModeWindow:GetLeft()
	EditModeWindow.Y = EditModeWindow:GetTop()
end
--
EditModeWindow.LinkSpaceLabel = CCD.UI.CreateText("EditModeWindow.LinkSpaceLabel", EditModeWindow)
EditModeWindow.LinkSpaceLabel:SetText("Frame Spacing")
EditModeWindow.LinkSpaceLabel:SetPoint("TOPCENTER", EditModeWindow.Button, "BOTTOMCENTER", 0, 5)
--
EditModeWindow.LinkSpaceSlider = CCD.UI.CreateSlider("EditModeWindow.LinkSpaceSlider", EditModeWindow)
EditModeWindow.LinkSpaceSlider:SetRange(0, 100, EditModeWindow.Saved.FrameSpacing)
EditModeWindow.LinkSpaceSlider.Slider:SetWidth(175)
EditModeWindow.LinkSpaceSlider:SetPoint("TOPCENTER", EditModeWindow.LinkSpaceLabel, "BOTTOMCENTER")
function EditModeWindow.LinkSpaceSlider.Slider.Event:SliderChange()
	CCD.UI:SetAllLinkSpace(self:GetPosition())
	EditModeWindow.Saved.FrameSpacing = self:GetPosition()
end
function EditModeWindow:CreateSaveTable()
	return EditModeWindow.Saved
end
function EditModeWindow:Load(SaveTable)
	if not SaveTable then return end
	for k, v in pairs(SaveTable) do
		EditModeWindow.Saved[k] = v
	end
	if EditModeWindow.Saved.X and EditModeWindow.Saved.Y then EditModeWindow:SetPoint("TOPLEFT", UIParent, "TOPLEFT", EditModeWindow.Saved.X, EditModeWindow.Saved.Y) end
	EditModeWindow.LinkSpaceSlider:SetPosition(EditModeWindow.Saved.FrameSpacing)
end
-- Aliases
local LinkableFrames = CCD.UI.LinkableFramesById
local Colors = CCD.UI.Colors

-- Private Functions
local function ShowAvailableLinkPoints(ActiveFrame)
	for _, frame in pairs(LinkableFrames) do
		if frame ~= ActiveFrame then
			frame.TopRightSquare:SetVisible(true)
			frame.BottomLeftSquare:SetVisible(true)
		end
	end
end
local function HideAvailableLinkPoints(ActiveFrame)
	for _, frame in pairs(LinkableFrames) do
		if frame ~= ActiveFrame then
			frame.TopRightSquare:SetVisible(false)
			frame.BottomLeftSquare:SetVisible(false)
		end
	end
end

-- Public Functions
function CCD.UI.CreateLinkableFrame(Id, Parent)
	if LinkableFrames[Id] then error(string.format("A frame with the ID %s already exists.", Id)) end
	
	local frame = CCD.UI.CreateFrame(Id, Parent)

	LinkableFrames[Id] = frame -- all created frames are automatically added to the table
	-- 	
	frame:SetMouseMasking("limited")
	frame:SetVisible(false)
	--
	frame.Id = Id
	frame.X = 0
	frame.Y = 0
	frame.CombatShow = false
	frame.GroupShowValue = 10 -- 2+ are valid values; anything less than 2 means this value is irrelevant
	frame.IsMobile = true -- if true, frame will be shown in edit mode
	frame.Alpha = 1 -- this value is initialized in the :Load() function of most modules
	frame.NeverShow = false
	--
	frame.Linked = false
	frame.AnchorId = "NONE"
	frame.LinkPoint = "NONE"
	frame.LinkSpace = "NONE"
	frame.NextLinkPoint = "TOPRIGHT"
	frame.NextLinkSpace = 0
	--
	function frame.Event:Move()
		self.X = self:GetLeft()
		self.Y = self:GetTop()
	end
	function frame:IsInside(Anchor)
		return self.X <= Anchor:GetRight() and frame.X >= Anchor:GetLeft() -- x
		and frame.Y >= Anchor:GetTop() and frame.Y <= Anchor:GetBottom() -- y
	end
	function frame:GetInsidePoint(Anchor)
		local bottomLeftSqRight = Anchor:GetLeft() + 20
		local bottomRightSqTop = Anchor:GetBottom() - 20
		
		if self.X <= bottomLeftSqRight and frame.Y >= bottomRightSqTop then
			return "BOTTOMLEFT"
		else
			return "TOPRIGHT"
		end
	end
	function frame:Link(Anchor, LinkPoint, LinkSpace)
		if not Anchor then
			Anchor = LinkableFrames[self.AnchorId]
			if not Anchor then
				self:Unlink(true)
				return
			end
			LinkPoint = self.LinkPoint
			LinkSpace = self.LinkSpace
		elseif not LinkPoint then
			LinkPoint = frame:GetInsidePoint(Anchor)
			LinkSpace = frame.NextLinkSpace
		end

		local xOffset, yOffset = 0, 0

		if LinkPoint == "TOPRIGHT" then
			xOffset = LinkSpace
		elseif LinkPoint == "BOTTOMLEFT" then
			yOffset = LinkSpace		
		else
			print(string.format("Error: LinkableFrame:Link() was attempted with an invalid LinkPoint of %s for ID %s.", LinkPoint or "nil", frame.Id))
			return
		end

		local anchorX = Anchor:GetLeft()
		local anchorY = Anchor:GetTop()

		frame.EditBolt:SetTexture("CCD", EditBoltLockedTexture)
		frame.EditBolt:ClearAll()
		frame.EditBolt:SetPoint("CENTER", Anchor, LinkPoint, xOffset, yOffset)

		frame.Linked = true
		frame.AnchorId = Anchor.Id
		frame.LinkPoint = LinkPoint
		frame.LinkSpace = LinkSpace
	end
	function frame:Unlink(SetPoint)
		frame.Linked = false
		frame.AnchorId = "NONE"
		frame.LinkPoint = "NONE"
		frame.LinkSpace = "NONE"
		frame.EditBolt:SetTexture("CCD", "Images\\Gadget.png")

		if SetPoint then
			local x = frame.X
			local y = frame.Y
			frame.EditBolt:ClearAll()
			frame.EditBolt:SetPoint("CENTER", UIParent, "TOPLEFT", x, y)
		end
	end
	-- functions
	function frame:Show()
		frame:SetVisible(true)
	end
	function frame:Hide() -- Does not work if CCD.UI.EditModeOn is true
		if not CCD.UI.EditModeOn then
			frame:SetVisible(false)
		end
	end
	function frame:ShowEditBolt()
		frame.EditBolt:SetVisible(true)
	end
	function frame:HideEditBolt()
		frame.EditBolt:SetVisible(false)
	end
	function frame:ShowForeground()
		frame.Foreground:SetVisible(true)
	end
	function frame:HideForeground()
		frame.Foreground:SetVisible(false)
	end
	function frame:SetForegroundColor(Color)
		self.Foreground:SetBackgroundColor(Color.R, Color.G, Color.B, Color.A)
	end
	function frame:ShowX(Condition)
		local visible = true
		
		if Condition ~= nil and not Condition then visible = false end
		if frame.CombatShow and not CCD.Me.InCombat then visible = false end
		if frame.GroupShowValue >= 2 and CCD.Allies.Count < frame.GroupShowValue then visible = false end
		if CCD.DisableInWarfronts and CCD.Me.ZoneType == "warfront" then visible = false end
		
		if visible then
			frame:Show()			
		else
			frame:Hide()
		end
		
		return visible
	end
	function frame:HideX(Condition) -- Hides the frame if all conditions are met; otherwise, does nothing
		local hiding = true
		
		if Condition ~= nil and not Condition then hiding = false end
		if CCD.UI.EditModeOn then hiding = false end -- redundant but necessary for returning false
		
		if hiding then
			frame:Hide()
		end
		
		return hiding
	end
	function frame:SetGroupShowValue(Value)
		frame.GroupShowValue = Value
		
		if frame.OnGroupShowValueChanged then
			frame:OnGroupShowValueChanged()
		end
	end
	--[[function frame:SetMobilitity(Value)
		self.IsMobile = Value
	end
	function frame:SetAlphaFromSliderPosition(Position)
		if Position and Position >= 1 then Position = Position / 100 end
		frame:SetAlpha(Position)
	end
	function frame:GetAlphaAsSliderPosition()
		return CCD.Math.Round(frame:GetAlpha() * 100)
	end--]]
	
	-- Edit Bolt
	frame.EditBolt = CCD.UI.CreateTexture(frame.Id .. ".EditBolt", CCD.UI.Context, "CCD", "Images\\Gadget.png")
	frame.EditBolt:SetSize(24, 24)
	frame.EditBolt:SetLayer(1) -- find how high this needs to be
	frame.EditBolt:SetVisible(false) -- will only be shown in edit mode
	-- Event Handlers
	function frame.EditBolt.Event:LeftDown()
		if not CCD.UI.EditModeOn then return end
		CCD.UI.SetActiveContextMenu(nil)
		self.MouseDown = true
		local mouseData = Inspect.Mouse()
		self.MyStartX = self:GetLeft()
		self.MyStartY = self:GetTop()
		self.StartX = mouseData.x - self.MyStartX
		self.StartY = mouseData.y - self.MyStartY
		
		ShowAvailableLinkPoints(self)
	end
	function frame.EditBolt.Event:MouseMove(MouseX, MouseY)
		if not CCD.UI.EditModeOn then return end
		if not self.MouseDown then return end
		self:ClearAll()
		self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", (MouseX - self.StartX), (MouseY - self.StartY))
	end
	function frame.EditBolt.Event:LeftUp()
		if not CCD.UI.EditModeOn then return end
		if not self.MouseDown then return end
		self.MouseDown = false
		frame:Unlink(true) -- added true in 3.2.5 to hopefully solve issue with left clicking bolt not always unlinking
		HideAvailableLinkPoints(self) -- done early so we can just return after linking
		-- drag-and-drop linking
		for _, otherFrame in pairs(LinkableFrames) do
			if otherFrame ~= frame and frame:IsInside(otherFrame) then
				frame:Link(otherFrame)
				return
			end
		end
	end
	-- RightClick will be set to open a context menu in the module's Initialize()

	-- attach frame to edit bolt
	frame:SetPoint("TOPLEFT", frame.EditBolt, "CENTER")
	
	-- Linking Squares
	frame.TopRightSquare = CCD.UI.CreateFrame(frame.Id .. ".TopRightSquare", frame)
	frame.TopRightSquare:SetWidth(15)
	frame.TopRightSquare:SetHeight(15)
	frame.TopRightSquare:SetLayer(2)
	frame.TopRightSquare:SetPoint("TOPRIGHT", frame, "TOPRIGHT", 1, -1)
	frame.TopRightSquare:SetVisible(false)
	Library.LibSimpleWidgets.SetBorder("plain", frame.TopRightSquare, 3, Colors.Green.R, Colors.Green.G, Colors.Green.B, Colors.Green.A, "tr")
	--
	frame.BottomLeftSquare = CCD.UI.CreateFrame(frame.Id .. ".BottomLeftSquare", frame)
	frame.BottomLeftSquare:SetWidth(15)
	frame.BottomLeftSquare:SetHeight(15)
	frame.TopRightSquare:SetLayer(2)
	frame.BottomLeftSquare:SetPoint("BOTTOMLEFT", frame, "BOTTOMLEFT", -1, 1)
	frame.BottomLeftSquare:SetVisible(false)
	Library.LibSimpleWidgets.SetBorder("plain", frame.BottomLeftSquare, 3, Colors.Green.R, Colors.Green.G, Colors.Green.B, Colors.Green.A, "bl")

	-- Foreground
	frame.Foreground = CCD.UI.CreateFrame(frame.Id .. ".Foreground", frame, Colors.ForegroundTint)
	frame.Foreground:SetLayer(5)
	frame.Foreground:SetPoint("TOPLEFT", frame, "TOPLEFT")
	frame.Foreground:SetVisible(false)
	function frame.Foreground:SetSize(W, H)
		self:SetWidth(W)
		self:SetHeight(H)
	end
		
	-- events
	function frame.Event:Size()
		self.Foreground:SetSize(self:GetWidth(), self:GetHeight())
		
		if self.OnSizeChanged then
			self:OnSizeChanged()
		end
	end
	
	return frame
end
function CCD.UI.SetEditMode(Switch)
	if Switch == nil then -- toggle
		CCD.UI.EditModeOn = not CCD.UI.EditModeOn
	elseif Switch == CCD.UI.EditModeOn then
		return
	else
		CCD.UI.EditModeOn = Switch
	end
	
	if CCD.UI.EditModeOn then
		if not EditModeWindow.Saved.X then
			EditModeWindow:SetPoint("CENTER", UIParent, "CENTER")
		end
		EditModeWindow:SetVisible(true)
		if CCD.Modules.RCT.OnEditModeBegin then CCD.Modules.RCT:OnEditModeBegin() end
		if CCD.Modules.RBT.OnEditModeBegin then CCD.Modules.RCT:OnEditModeBegin() end
		if CCD.Modules.CPT.OnEditModeBegin then CCD.Modules.RCT:OnEditModeBegin() end
	end

	for _, frame in pairs(LinkableFrames) do
		if CCD.UI.EditModeOn then
			if frame.OnEditModeBegin then
				frame:OnEditModeBegin()
			end
			
			if frame.IsMobile then
				frame:ShowEditBolt()
				frame:ShowForeground()
				frame:Show()
			else
				--frame:Show()
				frame:SetVisible(false) -- can't use :Hide() because it checks for EditModeOn
			end
		else
			if frame.OnEditModeEnd then
				frame:OnEditModeEnd()
			end
		
			frame:HideEditBolt()
			frame:HideForeground()
		end
	end

	if not CCD.UI.EditModeOn then
		EditModeWindow:SetVisible(false)
		CCD.UI.SetActiveContextMenu(nil)
		if CCD.Modules.RBT.OnEditModeEnd then CCD.Modules.RBT:OnEditModeEnd() end
		if CCD.Modules.RCT.OnEditModeEnd then CCD.Modules.RCT:OnEditModeEnd() end
	end
end

function CCD.UI:SetAllLinkSpace(Value)
	for _, frame in pairs(CCD.UI.LinkableFramesById) do
		frame.NextLinkSpace = Value
	end
end

