local TOC, CCD = ...
CCD.UI = CCD.UI or {}
CCD.UI.Context = CCD.UI.Context or UI.CreateContext("CCDContext")
local MyUIContext = CCD.UI.Context
CCD.MainWindow = CCD.MainWindow or UI.CreateFrame("SimpleWindow", "CCDMainWindow", MyUIContext)
local MainWindow = CCD.MainWindow
CCD.Importer = CCD.Importer or {}
local Importer = CCD.Importer
CCD.UI.MenuButton = CCD.UI.MenuButton or {}
local MenuButton = CCD.UI.MenuButton

MenuButton.BaseFrame = nil
MenuButton.X = 0
MenuButton.Y = 0
MenuButton.ContextMenu = nil
MenuButton.PositionLocked = false
MenuButton.SizeLocked = false
MenuButton.DefaultSize = 24
MenuButton.CurrentSize = MenuButton.DefaultSize
MenuButton.Hidden = false

function MenuButton:SetSize(Value)
	MenuButton.CurrentSize = Value or MenuButton.CurrentSize
	MenuButton.CurrentSize = CCD.Math.MinMax(MenuButton.CurrentSize, 20, 60)
	MenuButton.BaseFrame:SetWidth(MenuButton.CurrentSize)
	MenuButton.BaseFrame:SetHeight(MenuButton.CurrentSize)		
	
	MenuButton.ContextMenu.ResizeButtonSlider:SetPosition(MenuButton.CurrentSize, true)
end
function MenuButton:Create() -- Occurs AFTER :Load()
	local description = "Frame_Button"	
	
	if CCD.IsFirstTime then
		MenuButton.DefaultSize = MenuButton.DefaultSize * 2
	end

	MenuButton.BaseFrame = UI.CreateFrame("Frame", description, MyUIContext)

	MenuButton.BaseFrame:SetWidth(MenuButton.CurrentSize or MenuButton.DefaultSize)
	MenuButton.BaseFrame:SetHeight(MenuButton.CurrentSize or MenuButton.DefaultSize)
	if (CCD.IsFirstTime) then
		MenuButton.BaseFrame:SetPoint("CENTER", UIParent, "CENTER", 0 , 0)
	else
		MenuButton.BaseFrame:SetPoint("TOPLEFT", UIParent, "TOPLEFT", self.X , self.Y)
	end
	MenuButton.BaseFrame:SetMouseMasking("limited")
	MenuButton.BaseFrame:SetLayer(1)
	MenuButton.BaseFrame:SetVisible(not MenuButton.Hidden)

	MenuButton.BaseFrame.Texture = UI.CreateFrame("Texture", description .. "_Texure", MenuButton.BaseFrame)
	MenuButton.BaseFrame.Texture:SetWidth(MenuButton.CurrentSize or MenuButton.DefaultSize)
	MenuButton.BaseFrame.Texture:SetHeight(MenuButton.CurrentSize or MenuButton.DefaultSize)
	MenuButton.BaseFrame.Texture:SetPoint("CENTER", MenuButton.BaseFrame, "CENTER")
	MenuButton.BaseFrame.Texture:SetLayer(0)
	MenuButton.BaseFrame.Texture:SetTexture("CCD", "Images\\Button1.png")
	MenuButton.BaseFrame.Texture:SetVisible(true)

	MenuButton.BaseFrame.Tooltip = UI.CreateFrame("Frame", description .. "_Tooltip", MenuButton.BaseFrame)
	MenuButton.BaseFrame.Tooltip:SetWidth(150)
	MenuButton.BaseFrame.Tooltip:SetHeight(12)
	MenuButton.BaseFrame.Tooltip:SetBackgroundColor(0, 0, 0, 0.85) -- heavy black tint
	MenuButton.BaseFrame.Tooltip:SetVisible(false)

	MenuButton.BaseFrame.Tooltip.Text = UI.CreateFrame("Text", description .. "_Tooltip_Text", MenuButton.BaseFrame.Tooltip)
	MenuButton.BaseFrame.Tooltip.Text:SetText("Right Click for menu")
	MenuButton.BaseFrame.Tooltip.Text:SetFontSize(15)
	MenuButton.BaseFrame.Tooltip.Text:SetPoint("TOPLEFT", MenuButton.BaseFrame.Tooltip, "TOPLEFT")
	MenuButton.BaseFrame.Tooltip.Text:SetVisible(true)

	MenuButton.BaseFrame.Tooltip:SetWidth(MenuButton.BaseFrame.Tooltip.Text:GetWidth())
	MenuButton.BaseFrame.Tooltip:SetHeight(MenuButton.BaseFrame.Tooltip.Text:GetHeight())

	function MenuButton.BaseFrame.Event:LeftDown()
		if MenuButton.PositionLocked then
			CCD.UI.SetEditMode()
			if CCD.UI.EditModeOn then
				MenuButton.BaseFrame.Texture:SetTexture("CCD", "Images\\Button2.png")
			else
				MenuButton.BaseFrame.Texture:SetTexture("CCD", "Images\\Button3.png")
			end
			return
		end
		MenuButton.ContextMenu:Hide()
		self.MouseDown = true
		mouseData = Inspect.Mouse()
		self.MyStartX = self:GetLeft()
		self.MyStartY = self:GetTop()
		self.StartX = mouseData.x - self.MyStartX
		self.StartY = mouseData.y - self.MyStartY
		self.HasMoved = false
	end
	function MenuButton.BaseFrame.Event:MouseMove(mouseX, mouseY)
		if self.MouseDown then
			self:ClearPoint("CENTER")
			self:SetPoint("TOPLEFT", UIParent, "TOPLEFT", (mouseX - self.StartX), (mouseY - self.StartY))
			self.HasMoved = true
		end
	end
	function MenuButton.BaseFrame.Event:LeftUp()
		if self.MouseDown then
			self.MouseDown = false
			self:SetWidth(MenuButton.CurrentSize or MenuButton.DefaultSize)
			self:SetHeight(MenuButton.CurrentSize or MenuButton.DefaultSize)
			MenuButton.X = self:GetLeft() -- these two lines
			MenuButton.Y = self:GetTop() -- may need to be changed
		end
	end
	function MenuButton.BaseFrame.Event:RightClick()
		MenuButton.ContextMenu:Toggle(MenuButton.BaseFrame)
	end

	function MenuButton.BaseFrame.Event:LeftClick()
		if MenuButton.PositionLocked then return end
		if self.HasMoved then return end
		CCD.UI.SetEditMode()
		if CCD.UI.EditModeOn then
			MenuButton.BaseFrame.Texture:SetTexture("CCD", "Images\\Button2.png")
		else
			MenuButton.BaseFrame.Texture:SetTexture("CCD", "Images\\Button3.png")
		end
	end
	function MenuButton.BaseFrame.Event:MiddleClick()
		CCD.Modules.RBT.ContextMenu:Hide()
		MainWindow:ToggleVisibility()
	end
	function MenuButton.BaseFrame.Event:MouseIn()
		self.Texture:SetTexture("CCD", "Images\\Button2.png")
		self.Tooltip:ClearPoint("BOTTOMLEFT")
		self.Tooltip:ClearPoint("TOPLEFT")
		self.Tooltip:ClearPoint("BOTTOMRIGHT")
		self.Tooltip:ClearPoint("TOPRIGHT")
		if self:GetBottom() > UIParent:GetHeight() / 2 then
			if self:GetLeft() > UIParent:GetWidth() / 2 then
				self.Tooltip:SetPoint("BOTTOMRIGHT", self, "TOPLEFT", -5, 0)
			else
				self.Tooltip:SetPoint("BOTTOMLEFT", self, "TOPLEFT", 0, -5)
			end
		else
			if self:GetLeft() > UIParent:GetWidth() / 2 then
				self.Tooltip:SetPoint("TOPRIGHT", self, "BOTTOMLEFT", -5, 0)
			else
				self.Tooltip:SetPoint("TOPLEFT", self, "BOTTOMLEFT", 0, 5)
			end
		end

		self.Tooltip:SetVisible(true)
	end
	function MenuButton.BaseFrame.Event:MouseOut()
		if not CCD.UI.EditModeOn then
			MenuButton.BaseFrame.Texture:SetTexture("CCD", "Images\\Button1.png")
		end
		self.Tooltip:SetVisible(false)
	end
	function MenuButton.BaseFrame.Event:Move()
		MenuButton.X = self:GetLeft()
		MenuButton.Y = self:GetTop()
	end

	function MenuButton.BaseFrame.Event:Size()
		local size = MenuButton.BaseFrame:GetWidth()
		MenuButton.BaseFrame.Texture:SetWidth(size)
		MenuButton.BaseFrame.Texture:SetHeight(size)
	end
	function MenuButton.BaseFrame.Event:WheelForward()
		if MenuButton.SizeLocked then return end
		MenuButton:SetSize(MenuButton.CurrentSize + 2)
	end
	function MenuButton.BaseFrame.Event:WheelBack()
		if MenuButton.SizeLocked then return end
		MenuButton:SetSize(MenuButton.CurrentSize - 2)
	end

	-- Context Menu
	local menuId = MenuButton.BaseFrame:GetName() .. ".ContextMenu"
	self.ContextMenu = CCD.UI.CreateContextMenu(MenuButton.BaseFrame:GetName() .. ".ContextMenu", 300, 300, "CCD Menu", true)
	
	self.ContextMenu.MainWindowLabel = self.ContextMenu:AddLabel("MainWindowLabel", "Main Window", true)
	function self.ContextMenu.MainWindowLabel.Event:LeftClick()
		MenuButton.ContextMenu:Hide()
		MainWindow:Show()
	end
	
	self.ContextMenu.ImporterLabel = self.ContextMenu:AddLabel("ImporterLabel", "Importer", true)
	function self.ContextMenu.ImporterLabel.Event:LeftClick()
		MenuButton.ContextMenu:Hide()
		if not Inspect.System.Secure() then
			Importer.Frame:SetVisible(true)
		else
			print("Can't open Importer while in combat.")
		end
	end

	self.ContextMenu.ResetFramePositionsLabel = self.ContextMenu:AddLabel("ResetFramePositionsLabel", "Reset Frame Positions", true)
	function self.ContextMenu.ResetFramePositionsLabel.Event:LeftClick()
		MenuButton.ContextMenu:Hide()
		CCD.Modules.RCT.ResetFramePositions()
		CCD.Modules.RBT:ResetFramePositions()
		CCD.Modules.CPT:ResetFramePositions()
		CCD.Modules.CPUT:ResetFramePositions()
	end

	self.ContextMenu.ResizeButtonSlider = self.ContextMenu:AddSlider(menuId .. ".ResizeButtonSlider", "Button Size:", 20, 60, true)
	self.ContextMenu.ResizeButtonSlider:SetPosition(CCD.Math.MinMax(MenuButton.CurrentSize or MenuButton.DefaultSize, 20, 60))
	function self.ContextMenu.ResizeButtonSlider.Slider.Event:SliderChange()
		local size = self:GetPosition()
		MenuButton:SetSize(size)
	end
	
	self.ContextMenu.MenuButtonPositionLockedCheckbox = self.ContextMenu:AddCheckbox(menuId .. ".MenuButtonPositionLockedCheckbox", "Lock Position", false)
	self.ContextMenu.MenuButtonPositionLockedCheckbox:SetChecked(MenuButton.PositionLocked)
	function self.ContextMenu.MenuButtonPositionLockedCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		MenuButton.PositionLocked = checked
	end
	
	self.ContextMenu.MenuButtonSizeLockedCheckbox = self.ContextMenu:AddCheckbox(menuId .. ".MenuButtonSizeLockedCheckbox", "Lock Size", false)
	self.ContextMenu.MenuButtonSizeLockedCheckbox:SetChecked(MenuButton.SizeLocked)
	function self.ContextMenu.MenuButtonSizeLockedCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		MenuButton.SizeLocked = checked
	end

	self.ContextMenu.MenuButtonHiddenCheckbox = self.ContextMenu:AddCheckbox(menuId .. ".MenuButtonHiddenCheckbox", "Hidden", false)
	self.ContextMenu.MenuButtonHiddenCheckbox:SetChecked(MenuButton.Hidden)
	function self.ContextMenu.MenuButtonHiddenCheckbox.Event:CheckboxChange()
		local checked = self:GetChecked()
		MenuButton.Hidden = checked
		MenuButton.BaseFrame:SetVisible(not MenuButton.Hidden)
	end

	end
function MenuButton:ResetFramePositions()
	self.BaseFrame:ClearAll()
	self.BaseFrame:SetPoint("TOPLEFT", UIParent, "CENTER", -100, -400)
end
function MenuButton:Flashing_Timer()
	if self.IsBlue then
		MenuButton.BaseFrame.Texture:SetTexture("CCD", "Images\\Button4.png")
		self.IsBlue = false
	else
		MenuButton.BaseFrame.Texture:SetTexture("CCD", "Images\\Button1.png")
		self.IsBlue = true
	end
end
function MenuButton:StartFlashing()
	self.IsBlue = true
	Library.LibCron.new(TOC.identifier, 0.5, 10, false, self.Flashing_Timer, self)
end
function MenuButton:Locate()
	self:Show()
	MenuButton.BaseFrame:SetWidth((MenuButton.CurrentSize or MenuButton.DefaultSize) * 1.5)
	MenuButton.BaseFrame:SetHeight((MenuButton.CurrentSize or MenuButton.DefaultSize) * 1.5)
	self:StartFlashing()
end
function MenuButton:CreateSaveTable()
	return {
		X = MenuButton.X,
		Y = MenuButton.Y,
		Size = MenuButton.CurrentSize,
		PositionLocked = MenuButton.PositionLocked,
		SizeLocked = MenuButton.SizeLocked,
		Hidden = MenuButton.Hidden,
	}
end
function MenuButton:Load(Settings)
	if not Settings then return end
	MenuButton.X = Settings.X or MenuButton.X
	MenuButton.Y = Settings.Y or MenuButton.Y
	MenuButton.CurrentSize = Settings.Size or MenuButton.CurrentSize
	MenuButton.PositionLocked = Settings.PositionLocked or MenuButton.PositionLocked
	MenuButton.SizeLocked = Settings.SizeLocked or MenuButton.SizeLocked
	if Settings.Hidden ~= nil then MenuButton.Hidden = Settings.Hidden end
end
function MenuButton:Show()
	CCD.UI.MenuButton.BaseFrame:SetVisible(true)
	CCD.UI.MenuButton.ContextMenu.MenuButtonHiddenCheckbox:SetChecked(false)
end